(* 
 *
 * Copyright (c) 2003-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



open Cil
open Commondefs
open Facts

module H = Hashtbl

module StringSet = Set.Make(struct type t = string let compare = compare end)                
 

type hashOfGlobals = (string, global) H.t
type hashOfGlobalLocations = (string, location) H.t
type hashOfStringToInt = (string, int) H.t
type hashOfIntToString = (int, string) H.t
type hashOfStringToString = (string, string) H.t
type hashOfIntToInt = (int, int) H.t
type hashOfIntToIntList = (int, int list) H.t
type hashOfIntPairToInt = (int * int, int) H.t
type hashOfCallGraph = ((string * string), bool) H.t

exception ImDone

(* Begin PLEASE_ELIM_THIS_AS_IT_IS_DUPLICATED *)

(* tells us whether we're processing the first or second file *)
type stage_t = ProcessFile1Stage | ProcessFile2Stage
let stage : stage_t ref = ref ProcessFile1Stage

(* filter on globals; only type definitions and functions will pass *)
let filterOutGlobalDeclarations (g : Cil.global) : bool = 
  match g with
      GCompTagDecl _ | GEnumTagDecl _ | GVarDecl _ | GAsm _ | GPragma _ | GText _ -> false
    | _ -> true 

(* Iter on globals that pass the given filter *)
let iterGlobalsFiltered (file : Cil.file) 
                        (filterFun : Cil.global -> bool) 
                        (iter : Cil.global -> unit) = 
   Cil.iterGlobals 
      file 
      (fun g -> if filterFun g then iter g)

(* End PLEASE_ELIM_THIS_AS_IT_IS_DUPLICATED *)

(* standardised results, meant to be retruned to clients *)
type semdiffResult = {
   sr_vars        : (string, globalVarFact) H.t;
   sr_functions   : (string, functionFact) H.t;
   sr_typedefs    : (string, typedefFact) H.t;
   sr_comps  : (string, compFact) H.t;
};;

let globalsFile1: hashOfGlobals = H.create 65535
let globalsFile2: hashOfGlobals = H.create 65535

let globalLocations: hashOfGlobalLocations = H.create 65535

let globalMatchingStatus: hashOfStringToInt = H.create 65535

let statusStore1: hashOfStringToInt = H.create 65535
let statusStore2: hashOfStringToInt = H.create 65535

let typeEnvironment1: hashOfStringToInt = H.create 65535
let typeEnvironment2: hashOfStringToInt = H.create 65535

let typeEnvironmentReverseMapping1: hashOfIntToString = H.create 65535
let typeEnvironmentReverseMapping2: hashOfIntToString = H.create 65535

let typeConstraints: hashOfIntToIntList = H.create 65535
let unsatisfiedTypeConstraints: hashOfIntPairToInt = H.create 127
let satisfiedTypeConstraints: hashOfIntPairToInt = H.create 65535

let globalVarCensus : (string, globalVarFact) H.t = H.create 4099
let typedefCensus : (string, typedefFact) H.t = H.create 4099
let functionCensus : (string, functionFact) H.t = H.create 4099
let compCensus : (string, compFact) H.t = H.create 4099 
let enumCensus : (string, enumFact) H.t = H.create 127

let functionCallGraph1 : hashOfCallGraph = H.create 4099
let functionCallGraph2 : hashOfCallGraph = H.create 4099

let ignoreIt : string = "IGNORED"

let globalStatistics   = new Statistics.aggregateStatistics "Global Definitions"
let functionStatistics = new Statistics.aggregateStatistics "Functions"
let compStatistics     = new Statistics.aggregateStatistics "Structs/Unions"
let typedefStatistics  = new Statistics.aggregateStatistics "Typedefs"
let varStatistics      = new Statistics.aggregateStatistics "Global Variables"
let enumStatistics     = new Statistics.aggregateStatistics "Enums"

let funargAddedStatistics     = new Statistics.firstClassStatistics "Function arguments added"
let funargDeletedStatistics   = new Statistics.firstClassStatistics "Function arguments deleted"
let funargTypeChangedStatistics = new Statistics.firstClassStatistics "Function arguments type changes"
let compfieldsAddedStatistics     = new Statistics.firstClassStatistics "Struct/Union fields added"
let compfieldsDeletedStatistics   = new Statistics.firstClassStatistics "Struct/Union fields deleted"
let compfieldsTypeChangeStatistics   = new Statistics.firstClassStatistics "Struct/Union fields type changed"

let enumvalsAddedStatistics     = new Statistics.firstClassStatistics "Enum vals added"
let enumvalsDeletedStatistics   = new Statistics.firstClassStatistics "Enum vals deleted"
let enumvalsChangedStatistics   = new Statistics.firstClassStatistics "Enum vals exp changed"

let typedefPartialMap = new Partialmap.partialMap "typedefs"
let compPartialMap = new Partialmap.partialMap "comps"
let gvarPartialMap = new Partialmap.partialMap "global vars"
let enumPartialMap = new Partialmap.partialMap "enums"
let functionPartialMap = new Partialmap.partialMap "functions"

let localCount : int ref = ref 0
let freshLocalCount =  fun () -> incr localCount; !localCount

let pass : int ref = ref 1 
let nextPass  = fun () -> incr pass 

let direction : int ref = ref 0 
let flipDirection =  fun () -> direction := (1 - !direction)

(******************************** Constants ******************)
let sNOT_CHECKED_YET : int = 1
let sMATCH : int = 2
let sNO_MATCH : int = 3

let addCallEdge h caller callee = 
  Commonutil.hashtbl_add_once h (caller,callee) true


let getLocationOfString (s : string) : location = 
  try
    let l = H.find globalLocations s in l
  with Not_found ->          (* Printf.printf "### just looked up %s\n" s;*) locUnknown  


let argMatch (s1 : string) (s2 : string) : bool =
(* needed when matching arguments, e.g. "count___4" matches "count___5"  *)
(*  and "count" matches "count___0"  *)
   let all_numbers s = 
      let res = ref true in
      for i = 0 to (String.length s) - 1 do
         if (s.[i] < '0' || s.[i] > '9') then res:=false 
      done;
      !res && (s <> "")
   in
   let explode_around_underscores s = 
      (* tokenize the string around "___"; 
         if that's not possible, return [s,"",""] 
      *)
      try
         let last_underscore = String.rindex s '_' in
         let first_underscore = last_underscore - 2 in
         let last_char = String.length s in
         let part1 = String.sub s 0 first_underscore 
         and part2 = String.sub s first_underscore 3 
         and part3 = String.sub s (last_underscore + 1) (last_char - last_underscore - 1) in
         if (part2 = "___")  && (all_numbers part3)
         then [part1, part2, part3] 
         else [s,"",""]    
          
     with _ -> 
         [s,"",""]
   in 
      (* see the scenarios at the top of this function *)
      match (explode_around_underscores s1),(explode_around_underscores s2) with 
            [s1',"",""],    [s2',"___",_] -> s1' = s2'
         |  [s1',"___",_], [s2',"",""]    -> s1' = s2'
         |  [s1',"___",_], [s2',"___",_]   -> s1' = s2'
         |  _, _ -> false

let argMatch2 (s1 : string) (s2 : string) : bool = 
 startsWith s1 s2 || startsWith s2 s1 || argMatch s1 s2

let anonMatch (s1 : string) (s2 : string) : bool = 
(* needed when matching CIL-generated things
"struct anonstruct_irq_127" "struct anonstruct_irq_135" *)
  let isTrueAnon s = 
    (Cil.startsWith "union __anonunion" s) ||
    (Cil.startsWith "struct __anonstruct" s) 
  in
  if not (isTrueAnon s1) || not (isTrueAnon s2)  
  then 
    false
  else
  begin
    let rec firstDifference (index : int) : int = 
      begin
        try 
          let crt1 = s1.[index] in 
          let crt2 = s2.[index] in
            if crt1 <> crt2 
            then 
              index
            else (* crt1 = crt2 *)
              firstDifference (index+1)
        with Invalid_argument _ -> (index - 1)
      end    
    in

    let rec allNumbers (s : string) (index : int): bool = 
      try let si = s.[index] in
        if (si <= '9' && si >= '0') (* || (si <= 'f' && si >= 'a')  *)|| (si = '\"')
        then allNumbers s (index + 1) 
        else false
      with Invalid_argument _ -> true
    in 

    let rec allNumbersOrUndesrcores (s : string) (index : int): bool = 
      try let si = s.[index] in
        if (si <= '9' && si >= '0') (* || (si <= 'f' && si >= 'a') *) || (si = '_')  || (si = '\"')
        then allNumbersOrUndesrcores s (index + 1) 
        else false
      with Invalid_argument _ -> true
    in 

    begin
        let fd = firstDifference 0 in 
        let res = allNumbers s1 fd && allNumbers s2 fd
        in 
          if res 
          then begin (*Printf.printf "anonMatch: SUCCESS \"%s\" \"%s\"\n" s1 s2;*) true end
          else 
(*               let l1 = (getLocationOfString s1) and l2 = (getLocationOfString s2) in *)
                if (*FIXME l1.file = l2.file && l1.line = l2.line && l1 <> locUnknown &&*) allNumbersOrUndesrcores s1 fd && allNumbersOrUndesrcores s2 fd
                then (*Printf.printf "anonMatch: FAIL    \"%s\" \"%s\" but location would have saved the day\n" s1 s2*) true
                else false
    end  
end


let getTypeVariable (t : typ) (typeEnvironment : hashOfStringToInt) (typeEnvironmentRevMap : hashOfIntToString)   = 
  (* let nada = Printf.printf " get type variable\n" in *)
  let tn = (Commoncilutil.prettyPrintType t)
  in
    try 
      H.find typeEnvironment tn
    with Not_found -> 
      begin 
        let tv = freshLocalCount () in
          (*Printf.printf "Adding type variable %s -> %d\n" tn tv;*)
          H.add typeEnvironment tn tv;
          H.add typeEnvironmentRevMap tv tn;
          tv
      end

let addRenamingConstraint (t1 : typ) (t2 : typ) : unit = 
   let aux tn1 tn2 pm =
      if !direction = 0 
      then pm#addConstraint (tn1,tn2)
      else pm#addConstraint (tn2,tn1)
   in
      match t1,t2 with
          TNamed (ti1, _), TNamed (ti2, _) -> 
             aux ti1.tname ti2.tname typedefPartialMap

        | TComp (c1, _), TComp (c2, _) -> 
             aux c1.cname c2.cname compPartialMap

        | TEnum (e1, _), TEnum (e2, _) -> 
             aux e1.ename e2.ename enumPartialMap

        | _,_ -> ()
;;

let rec addTypeConstraint (t1 : typ) (t2 : typ) : unit = 
  let tE1, tER1, tE2, tER2 = 
    if !direction = 0 then
      typeEnvironment1, typeEnvironmentReverseMapping1, typeEnvironment2, typeEnvironmentReverseMapping2 
    else
      typeEnvironment2, typeEnvironmentReverseMapping2, typeEnvironment1, typeEnvironmentReverseMapping1
  in
  let doAddIt (t1 : typ) (t2 : typ) : unit =
     let tv1 = getTypeVariable t1 tE1 tER1 
     and tv2 = getTypeVariable t2 tE2 tER2 in
     begin
        try 
           let clist = H.find typeConstraints tv1 in 
           begin
              if not (List.mem tv2 clist) 
              then
                begin
                  H.remove typeConstraints tv1;
                  H.add typeConstraints tv1 (tv2::clist);
                end
           end
        with Not_found -> 
          begin 
            (*Printf.printf "Adding type constraint (%d,%d)\n" tv1 tv2;*)
            H.add typeConstraints tv1 [tv2]
          end
      end
  in
    begin
      match t1,t2 with
          TNamed (ty1, _), TNamed (ty2, _) -> 
             addRenamingConstraint t1 t2;
             addTypeConstraint ty1.ttype ty2.ttype

        | TPtr (ty1, _), TPtr (ty2, _) -> 
             addTypeConstraint ty1 ty2

        | TArray (ty1, _, _), TArray (ty2, _, _) -> 
             addTypeConstraint ty1 ty2

        | TComp (c1, _), TComp (c2, _) -> 
             addRenamingConstraint t1 t2;
             doAddIt t1 t2

        | TEnum (e1, _), TEnum (e2, _) -> 
             addRenamingConstraint t1 t2

        | _,_ -> ()
    end

type eitem = (string * exp * location)
    
let  checkEnumIsomorphism (e1 : enuminfo) (e2 : enuminfo) : int =
  let valsDeletedList = ref [] 
  and valsAddedList = ref [] 
  and valsChangedList = ref [] 
  and ch : enumFact = { 
     e_name = e1.ename; 
     e_change = ValChanged;
     e_valsChange = []}
  in
  let e1name = "enum \"" ^ e1.ename ^ "\"" in
  let getName ((s,e,l) : eitem) = s in
  let getExp  ((s,e,l) : eitem) = e in

  let sameLength (ee1 : enuminfo) (ee2 : enuminfo) : bool = 
     List.length ee1.eitems =  List.length ee2.eitems 
  in
  let checkPresenceInOther (op : char) (g2l : eitem list) (g1 : eitem) : int =
    let sameName (f : eitem) (g : eitem) : bool =  (getName f = getName g) in
    let findSameName (f : eitem) (fl :  eitem list) : eitem = 
      List.find (sameName f) fl 
    in
    let   checkEitemExp (f1 : eitem) (f2 : eitem) : int = 
      let summy =  (* Printf.printf "checkEitemExp %s %s %s\n" e1.ename (getName f1) (getName f2); *)
        if op = '+' then
          begin
            if (getExp f1) <> (getExp f2) then 
              begin
                valsChangedList := (getName f1)::!valsChangedList;
                sNO_MATCH
              end
            else
              sMATCH
          end
        else
          sMATCH
      in summy
    in
      
    let found = 
      try    
         ignore (checkEitemExp g1 (findSameName g1 g2l));
         sMATCH
      with Not_found ->
        begin
          begin
            if op = '+' 
            then valsAddedList := (getName g1)::!valsAddedList
            else valsDeletedList := (getName g1)::!valsDeletedList 
          end;
          sNO_MATCH
        end
    in   found
  in
    
  let checkListOneWay (op : char) (f1 : eitem list) (f2 : eitem list) : bool =
    let resultslist = List.map (checkPresenceInOther op f2) f1 in
    let finalval = List.for_all (fun x -> if x = sNO_MATCH then false else true)  resultslist
    in   finalval
  in 

  let sameLen = sameLength e1 e2 in
  let all3 = 
    (* let nichts = compStatistics#incr "total" in *)
    (* Printf.printf "ENUM: e1 has %d items and e2 has %d\n" (List.length e1.eitems) (List.length e2.eitems); *)
    let homomorphismOldNew =  checkListOneWay '-' e1.eitems  e2.eitems in
    let homomorphismNewOld =  checkListOneWay '+' e2.eitems  e1.eitems in
      sameLen && homomorphismOldNew && homomorphismNewOld
  in

  let () = 
    if (!pass = 2) then
      begin
        let 
          fdel = List.length !valsDeletedList and
           fadd = List.length !valsAddedList and
          fvalchange  = List.length !valsChangedList in
          begin
            if not sameLen then
              begin
                (*Printf.printf "function \"%s\": argument count changed \n" name;*)

                enumStatistics#incr "item count changes";
              end;
            if fdel > 0 then
              begin
                (*Printf.printf "%s: %d vals deleted: " c1name fdel;
                List.iter (fun s -> Printf.printf "%s " s) !valsDeletedList;
                Printf.printf "\n";*)

                List.iter 
                  (fun s -> ch.e_valsChange <- {v_name = s; v_change = Deleted }::ch.e_valsChange) 
                            !valsDeletedList;
                enumvalsDeletedStatistics#add e1name fdel ((getLocationOfString e1name).file);
                enumStatistics#incrN "items deleted" fdel;
              end;
            if fadd > 0 then
              begin
                (*Printf.printf "%s: %d vals added: " c1name fadd;
                List.iter (fun s -> Printf.printf "%s " s) !valsAddedList;
                Printf.printf "\n"; *)

                List.iter 
                  (fun s -> ch.e_valsChange <- {v_name = s; v_change = Added }::ch.e_valsChange) 
                            !valsAddedList;


                enumvalsAddedStatistics#add e1name fadd ((getLocationOfString e1name).file);
                enumStatistics#incrN "items added" fadd;
              end;
            if fvalchange > 0 then
              begin

                List.iter 
                  (fun n -> ch.e_valsChange <- 
                     {v_name = n; v_change = ValChanged}::ch.e_valsChange) 
                            !valsChangedList;


                enumvalsChangedStatistics#add e1name fvalchange ((getLocationOfString e1name).file);
                enumStatistics#incrN "val exp changes" fvalchange;
              end;
          end;
          if fdel > 0 || fadd > 0 || fvalchange > 0
          then 
            begin 
              ch.e_change <- OtherChange "change";
              H.add enumCensus e1.ename ch; 
            end;
      end

  in 
    if all3 
    then sMATCH 
    else sNO_MATCH


let checkVarinfoIsomorphism (v1 : Cil.varinfo) (v2 : Cil.varinfo) : bool = v1.vname = v2.vname;; 
let checkFieldinfoIsomorphism (f1 : Cil.fieldinfo) (f2 : Cil.fieldinfo) : bool = f1.fname = f2.fname;;

let  rec checkTypeIsomorphism (t1 : typ) (t2 : typ) : int = (*Printf.printf "checkTypeIsomorphism...";*)
  if !pass = 1 
  then
    ((match t1,t2 with
        TComp (c1, _), TComp (c2, _) -> addTypeConstraint t1 t2
          (* checkme : should't be ty1/ty2 ??? *)
      | TNamed (ty1, _), TNamed (ty2, _) -> addTypeConstraint t1 t2
      | TFun (_, _, _, _),  TFun (_, _, _, _) -> 
          begin 
            ignore(checkFunctionIsomorphism "" t1 t2)
          end
      | _,_ -> ());
    sMATCH)
  else
     begin match t1,t2 with
          TVoid _, TVoid _ -> sMATCH
        | TInt (ik1, _), TInt (ik2, _)  -> if ik1 = ik2 then sMATCH else sNO_MATCH
        | TFloat (fk1, _), TFloat (fk2, _) -> if fk1 = fk2 then sMATCH else sNO_MATCH
        | TNamed (ty1, _), TNamed (ty2, _) -> checkTypeIsomorphism ty1.ttype ty2.ttype
        | TPtr (ty1, _), TPtr (ty2, _) -> checkTypeIsomorphism  ty1 ty2
        | TArray (ty1, eo1, _), TArray (ty2, eo2, _) -> 
             if checkTypeIsomorphism ty1 ty2 = sMATCH 
             then 
                begin 
                   if checkExpOptionIsomorphism eo1 eo2 
                   then sMATCH
                   else sNO_MATCH
                end
             else 
                sNO_MATCH

        | TComp (c1, _), TComp (c2, _) -> checkCompinfoIsomorphismShallow c1 c2
        | TEnum (_, _), TEnum (_, _) -> sMATCH
        | TFun (_, _, _, _),  TFun (_, _, _, _) -> checkFunctionIsomorphism "" t1 t2
        | TBuiltin_va_list _, TBuiltin_va_list _ -> sMATCH
        | _,_ -> sNO_MATCH
    end

and 

   checkExpOptionIsomorphism (eo1:Cil.exp option) (eo2:Cil.exp option) : bool = 
   begin
      match eo1,eo2 with

        None, None -> true

      | Some e1, Some e2 -> checkExpIsomorphism e1 e2

      | None, Some _ 
      | Some _, None -> false

   end

   

and
  (************************** TYPEDEF stuff starts here ********************************************)   


  checkTypedefIsomorphism (t1 : typeinfo) (t2 : typeinfo) : int =

  if !pass = 1 then
     begin
        (*Printf.printf "Adding constraint based on global var \"%s\"\n" v1.vname;*)
        ignore (checkTypeIsomorphism t1.ttype t2.ttype);
        sMATCH
     end
  else
    let sameType = checkTypeIsomorphism t1.ttype t2.ttype in
    begin
      (* typedefStatistics#incr "total"; *)
      if sameType = sNO_MATCH then
        begin
          (*
            Printf.printf "typedef \"%s\": changed base type " t1.tname;
            Printf.printf "from %s " (Commoncilutil.prettyPrintType t1.ttype);
            Printf.printf "to %s\n" (Commoncilutil.prettyPrintType t2.ttype);*)

          let ch : typedefFact = 
            { t_name = t1.tname;
              t_change = TypeChanged((Commoncilutil.prettyPrintType t1.ttype), (Commoncilutil.prettyPrintType t2.ttype))} 
          in
            H.add typedefCensus t1.tname ch; 
            typedefStatistics#incr "base type changes"
        end
          (*else 
            Printf.printf "var \"%s\": did not change type\n" v1.vname*)
    end;
    sameType

and 

  (************************** OFFSET stuff starts here ********************************************)   
   checkOffsetIsomorphism (o1: Cil.offset)(o2: Cil.offset) : bool = match o1,o2 with 
   
           NoOffset, NoOffset -> true

         | Field(fi1,o11),Field(fi2,o22) -> 
              (checkFieldinfoIsomorphism fi1 fi2) && (checkOffsetIsomorphism o11 o22)

         | Index(e1,o11),Index(e2,o22) -> 
              (checkExpIsomorphism e1 e2) && (checkOffsetIsomorphism o11 o22)

         | _,_ -> false

and 
  (************************** LHOST stuff starts here ********************************************)   
   checkLhostIsomorphism (lh1 : Cil.lhost) (lh2 : Cil.lhost) : bool = match lh1,lh2 with
           Var(vi1),Var(vi2) -> checkVarinfoIsomorphism vi1 vi2
         | Mem(e1),Mem(e2) -> checkExpIsomorphism e1 e2
         | _,_ -> false

and 
  (************************** LVAL stuff starts here ********************************************)   
   checkLvalIsomorphism (lv1 : Cil.lval) (lv2 : Cil.lval) : bool = 

   let (lh1,of1), (lh2,of2) = lv1,lv2 in
   (checkLhostIsomorphism lh1 lh2) && (checkOffsetIsomorphism of1 of2)

and 
  (************************** CONST stuff starts here ********************************************)   

   checkConstIsomorphism (c1 : Cil.constant) (c2 : Cil.constant) : bool = match c1,c2 with

        CInt64(i1,ik1,_),CInt64(i2,ik2,_) -> (i1 = i2) && (ik1 = ik2) 

      | CStr(s1),CStr(s2) -> s1 = s2

      | CWStr(l1),CWStr(l2) -> 
           begin 
              try
                 List.for_all2 (fun x y -> x = y) l1 l2
              with _ -> 
                 false
           end

     | CChr(c1),CChr(c2) -> c1 = c2

     | CReal(f1,fk1,Some s1),CReal(f2,fk2,Some s2) -> (f1 = f2) && (fk1 = fk2)

     | _,_ -> false

and
  (************************** EXP stuff starts here ********************************************)   
 checkExpIsomorphism (e1 : Cil.exp)  (e2 : Cil.exp) : bool = 
   let e1 =  Cil.constFold true e1 in
   let e2 =  Cil.constFold true e2 in
   match e1,e2 with
        Const(c1), Const(c2) -> checkConstIsomorphism c1 c2
      | Lval(lv1),Lval(lv2) -> checkLvalIsomorphism lv1 lv2
      | SizeOf(t1),SizeOf(t2) -> (checkTypeIsomorphism t1 t2) = sMATCH
      | SizeOfE(e1),SizeOfE(e2) -> checkExpIsomorphism e1 e2
      | SizeOfStr(s1),SizeOfStr(s2) -> s1 = s2
      | AlignOf(t1),AlignOf(t2) -> (checkTypeIsomorphism t1 t2) = sMATCH
      | AlignOfE(e1),AlignOfE(e2) -> checkExpIsomorphism e1 e2
      | UnOp(op1,e1,t1),UnOp(op2,e2,t2) -> 
           op1 = op2 
        && checkExpIsomorphism e1 e2
        && (checkTypeIsomorphism t1 t2 = sMATCH)
          
      | BinOp(op1,e1,e1',t1),BinOp(op2,e2,e2',t2) -> 
           op1 = op2 
        && checkExpIsomorphism e1 e2
        && checkExpIsomorphism e1' e2'
        && (checkTypeIsomorphism t1 t2) = sMATCH 

      | CastE(t1,e1),CastE(t2,e2) -> 
           checkExpIsomorphism e1 e2
        && (checkTypeIsomorphism t1 t2) = sMATCH

      | AddrOf(lv1),AddrOf(lv2) -> checkLvalIsomorphism lv1 lv2
      | StartOf(lv1),StartOf(lv2) -> checkLvalIsomorphism lv1 lv2
      | _,_ -> false

and
  (************************** init stuff starts here ********************************************)   

   checkInitIsomorphism (i1 : Cil.init) (i2 : Cil.init) : bool = match i1,i2 with
        SingleInit e1, SingleInit e2 -> checkExpIsomorphism e1 e2
      | CompoundInit (t1,l1), CompoundInit (t2,l2) -> checkCompoundInitIsomorphism (t1,l1) (t2,l2)
      | _ -> false

and
  (************************** CompoundInit stuff starts here ********************************************)   
   checkCompoundInitIsomorphism ((t1,c1) : typ * (offset * init) list) ((t2, c2) : typ * (offset * init) list) : bool = 
   let checkCompoundInitIsomorphism_inner (off1,i1) (off2,i2) = 
         checkOffsetIsomorphism off1 off2
      && checkInitIsomorphism i1 i2

   in
   let inner_res = 
      begin try
         List.for_all2 checkCompoundInitIsomorphism_inner c1 c2
      with _ ->
         false
      end
   in
        (checkTypeIsomorphism t1 t2) = sMATCH
     && inner_res

and
  (************************** GVAR stuff starts here ********************************************)   
  checkGVarIsomorphism 
     ((v1, ii1) : (varinfo * initinfo)) 
     ((v2, ii2) : (varinfo * initinfo)) : int =

  if !pass = 1
  then
    begin
      (*Printf.printf "Adding constraint based on global var \"%s\"\n" v1.vname;*)
(*       Printf.printf "tttttttt Will add (%s,%s) constraint\n"  *)
(*         (Commoncilutil.prettyPrintType v1.vtype) (Commoncilutil.prettyPrintType v2.vtype) ; *)
      addTypeConstraint v1.vtype v2.vtype;
      sMATCH
    end
  else
     let sameType = (checkTypeIsomorphism v1.vtype v2.vtype) in
        if sameType = sNO_MATCH then
           begin
              (*Printf.printf "Global var \"%s\": changed type " v1.vname;
                Printf.printf "from %s " (Commoncilutil.prettyPrintType v1.vtype);
                Printf.printf "to %s\n" (Commoncilutil.prettyPrintType v2.vtype);*)

              let ch : globalVarFact = 
                 { g_name = v1.vname;
                   g_change = TypeChanged((Commoncilutil.prettyPrintType v1.vtype), (Commoncilutil.prettyPrintType v2.vtype))} 
              in
              H.add globalVarCensus v1.vname ch; 
              varStatistics#incr "var type changes";
              sameType
           end
        else 
           (* Var type didn't change. How about the value ? *)
           let res_bool = 
              begin
                 match ii1, ii2 with
                      {init = Some i1}, {init = Some i2} -> checkInitIsomorphism i1 i2
                    | _ -> true
              end
           in
              if res_bool 
              then 
                 sMATCH 
              else
                 begin
                    let ch : globalVarFact = 
                       { g_name = v1.vname;
                         g_change = ValChanged} 
                    in
                       H.add globalVarCensus v1.vname ch; 
                       varStatistics#incr "static initializer changes";
                       sNO_MATCH
                 end

and
  (************************** FUN stuff starts here ********************************************)   
  nameTypeListMatch (l1 : (string * typ) list) 
                    (l2 : (string * typ) list) : bool = 
     if List.length l1 <>  List.length l2 
     then 
        false
     else
        not (List.exists2 (fun (n1,t1) (n2,t2) -> checkTypeIsomorphism t1 t2 = sNO_MATCH) l1 l2)
and  

  checkFunctionIsomorphism (name : string) (t1 : typ) (t2 : typ) : int =
  let 
    argsDeletedList = ref [] and
    argsAddedList = ref [] and
    argsTypeChangedList = ref [] and
    ch : functionFact = { f_name = name; 
                          f_change = NoChange; 
                          f_returnValChange = NoChange;
                          f_argumentsChange = []}
  in

  let typeConstraintList = ref [] in 
  
  let addLocalTypeConstraint (t1 : typ) (t2 : typ) : unit = 
    begin
       if !pass = 1
       then typeConstraintList := (t1,t2)::!typeConstraintList 
    end 
  in 
  
  let flushLocalConstraints = fun () ->
    begin
      (*Printf.printf "flushLocalConstraints (%d constraints)\n" (List.length !typeConstraintList);*)
       if !pass = 1
       then List.iter (fun p -> let (x,y) = p in addTypeConstraint x y) !typeConstraintList
    end
  in

  let  sameLength 
    (al1 : (string * typ * attributes) list) 
    (al2 : (string * typ * attributes) list) : bool = 

    let old_len = List.length al1 and new_len = List.length al2 in
    old_len = new_len 


  in 

  let sameName 
    (a1 : string * typ * attributes) 
    (a2 : string * typ * attributes) : bool = 
    let haveSameName = 
      begin
        
        match a1,a2 with
            (n1,_,_), (n2,_,_) -> (*Printf.printf "<%s><%s>\n" n1 n2;*)(n1 = n2) || (argMatch n1 n2) 
      end
    in haveSameName
  in    

  let findSameName 
    (a : (string * typ * attributes)) 
    (al : (string * typ * attributes) list) :  (string * typ * attributes) = 
    List.find (sameName a) al 
  in

  let checkPresenceInOther (op : char)  (a2l : (string * typ * attributes) list) (a : (string * typ * attributes)) : bool =
    let   checkArgType (a1 : (string * typ * attributes)) (a2 : (string * typ * attributes)) : bool =

      let summy =  
        if op = '+' then
          begin
            let (n1, t1, _), (n2, t2, _) = a1, a2 in
               if !pass = 1
               then
                begin 
                  addLocalTypeConstraint t1 t2;
                  true
                end
              else
              if (checkTypeIsomorphism t1 t2) = sNO_MATCH then 
                begin
                  argsTypeChangedList := (n1,t1,t2)::!argsTypeChangedList;
                  false
                end
              else
                begin
                  (*Printf.printf "  function: arg <%s> did not change type \n" n1;*)
                  true
                end
          end
        else
           begin 
            true 
          end
      in summy
    in
    let found =
      begin
        try    checkArgType a (findSameName a a2l)
        with Not_found -> 
          let n1 = 
            match a with
                (nm,_,_) -> nm
          in
            if op = '+' 
            then argsAddedList := n1::!argsAddedList
            else argsDeletedList := n1::!argsDeletedList;
            false
      end
    in found
  in

  let isNamelessArgument (a:string * typ * attributes) : bool = 
    match a with 
        ("",_,_) -> true
      | _ -> false
  in

  let isPointerToFunction (al : (string * typ * attributes) list) : bool = 
    List.exists isNamelessArgument al
  in

  let sideBySideMatch (al1 : (string * typ * attributes) list) 
                       (al2 : (string * typ * attributes) list) : bool = 
     let filter3to2 = List.map ( fun (s,t,_) -> (s,t)) in
     nameTypeListMatch (filter3to2 al1) (filter3to2 al2)  
  in

  let checkListOneWay (op : char)
                      (al1 : (string * typ * attributes) list) 
                      (al2 : (string * typ * attributes) list) : bool = 
     if isPointerToFunction al1 || isPointerToFunction al2
     then sideBySideMatch al1 al2 
     else List.for_all (checkPresenceInOther op al2) al1 
  in 

  let dummy = 
    begin
      (* functionStatistics#incr "total"; *)
      match t1,t2 with
          TFun (ret1, alr1, va1, _),  TFun (ret2, alr2, va2, _) ->

            let al1 = argsToList alr1 in
            let al2 = argsToList alr2 in
            let sameRet =               
               if !pass = 1
               then
                begin 
                  addLocalTypeConstraint t1 t2;
                  true;
                end
              else
                (checkTypeIsomorphism ret1 ret2) = sMATCH 
            in
            let homomorphismOldNew =  (checkListOneWay '-' al1 al2) in
            let homomorphismNewOld =  (checkListOneWay '+' al2 al1) in
            let () = 
               if !pass = 2 then
                begin
                  let 
                    adel = List.length !argsDeletedList and
                    aadd = List.length !argsAddedList and
                    atc  = List.length !argsTypeChangedList in
                    begin

                      if not sameRet then
                        begin
                          (*Printf.printf "function \"%s\": return type changed " name;
                          Printf.printf "from %s " (Commoncilutil.prettyPrintType ret1);
                          Printf.printf "to %s\n" (Commoncilutil.prettyPrintType ret2);*)
                          ch.f_returnValChange <- TypeChanged (Commoncilutil.prettyPrintType ret1, Commoncilutil.prettyPrintType ret2);

                          functionStatistics#incr "return types changes";
                        end;

                      if not (sameLength al1 al2) then
                        begin
                          (*Printf.printf "function \"%s\": argument count changed \n" name;*)
                          (*functionStatistics#incr "argument count changes"; *)
                        end;

                      if adel > 0 then
                        begin
                          (*Printf.printf "function \"%s\": %d arguments deleted: " name adel;
                          List.iter (fun s -> Printf.printf "%s " s) !argsDeletedList;
                          Printf.printf "\n";*)
                          List.iter 
                            (fun s -> ch.f_argumentsChange <- {v_name = s; v_change = Deleted }::ch.f_argumentsChange) 
                            !argsDeletedList;

                          funargDeletedStatistics#add name adel ((getLocationOfString ("function \"" ^ name ^ "\"")).file);
                        end;

                      if aadd > 0 then
                        begin
                          (*Printf.printf "function \"%s\": %d arguments added: " name aadd;
                          List.iter (fun s -> Printf.printf "%s " s) !argsAddedList;
                          Printf.printf "\n";*)
                          List.iter 
                            (fun s -> ch.f_argumentsChange <- {v_name = s; v_change = Added }::ch.f_argumentsChange) 
                            !argsAddedList;

                          funargAddedStatistics#add name aadd ((getLocationOfString ("function \"" ^ name ^ "\"")).file);
                        end;

                      if atc > 0 && name <> "" then
                        begin
                          (*Printf.printf "function \"%s\": %d arguments changed type: \n" name atc;

                          List.iter (
                            fun (n,t1,t2) -> Printf.printf "\t argument %s changed type from %s to %s \n" 
                              n 
                              (Commoncilutil.prettyPrintType t1) 
                              (Commoncilutil.prettyPrintType t2)) 
                            !argsTypeChangedList;*)
                          List.iter 
                            (fun (n,t1,t2) -> ch.f_argumentsChange <- 
                               {v_name = n; v_change = TypeChanged((Commoncilutil.prettyPrintType t1),(Commoncilutil.prettyPrintType t2))}::ch.f_argumentsChange) 
                            !argsTypeChangedList;

                          functionStatistics#incrN "arguments type changes" atc;
                          funargTypeChangedStatistics#add name atc ((getLocationOfString ("function \"" ^ name ^ "\"")).file);
                        end;

                      if adel > 0 || aadd > 0 || atc > 0 || (not sameRet)
                      then 
                        begin 
                          ch.f_change <- TypeChanged ((Commoncilutil.prettyPrintType t1),(Commoncilutil.prettyPrintType t2)) ;
                          H.add functionCensus name ch;
                          functionStatistics#incr "function signature changes";
                        end
                    end
                end
            in
            if  sameRet && (sameLength al1 al2) && homomorphismOldNew && homomorphismNewOld
            then 
              begin
                 if !pass = 1
                 then
                  flushLocalConstraints ();
                sMATCH
              end
                
            else sNO_MATCH

        | _,_ -> sNO_MATCH
    end
  in dummy
       
       
  and
  (************************** COMP stuff starts here ********************************************)
  checkCompinfoIsomorphismShallow (c1 : compinfo) (c2 : compinfo) : int =
   begin
      let n1 = Commoncilutil.prettyPrintType (TComp(c1,[]))
      and n2 = Commoncilutil.prettyPrintType (TComp(c2,[])) in
 
      if n1 = n2
      then sMATCH
      else 
         if anonMatch n1 n2
         then sMATCH
         else
           let tv1 = H.find typeEnvironment1 n1
           and tv2 = H.find typeEnvironment2 n2 in  
           if H.mem satisfiedTypeConstraints (tv1,tv2)
           then sMATCH
           else sNO_MATCH
   end

   and

  (* should we add type constraints upon field matching as well ??  *)

  checkCompinfoIsomorphismPass1 (c1 : compinfo) (c2 : compinfo) : int =
  begin
    ignore(addTypeConstraint (TComp (c1, [])), (TComp (c2, [])));
    sMATCH
  end


  and

  checkCompinfoIsomorphism (c1 : compinfo) (c2 : compinfo) : int =
  let 
    fieldsDeletedList = ref [] and
    fieldsAddedList = ref [] and
    fieldsTypeChangedList = ref [] and
    ch : compFact = { c_name = c1.cname; 
                           c_isStruct = if c1.cstruct then true else false; 
                           c_change = NoChange;
                           c_fieldsChange = []}

  in
  let compinfo2name (c : compinfo) : string =
    if c.cstruct 
    then ("struct \"" ^ c.cname ^ "\"")
    else ("union \""  ^ c.cname ^ "\"")  in 

  let c1name = compinfo2name c1 in

  let sameLength (c1 : compinfo) (c2 : compinfo) : bool = 
    if c1.cstruct <> c2.cstruct then
      begin
        Printf.printf "%s changed, struct <-> union \n" (c1.cname);
        false
      end
    else
      begin
        let old_len = List.length c1.cfields in
        let new_len = List.length c2.cfields in
          old_len = new_len
      end
  in    
  let checkPresenceInOther (op : char) (f2l : fieldinfo list) (f1 : fieldinfo) : int =
    let sameName (f : fieldinfo) (g : fieldinfo) : bool =  f.fname = g.fname in
    let findSameName (f : fieldinfo) (fl : fieldinfo list) : fieldinfo = 
      List.find (sameName f) fl in
    let   checkFieldinfoType (f2 : fieldinfo) (f1 : fieldinfo) : int = 
      let summy =  (*Printf.printf "checkFieldinfoType %s %s\n"  f1.fname f2.fname;*)
        if op = '+' then
          begin
            (*Printf.printf "+branch\n";*)
            let uf1 = unrollTypeDeep f1.ftype in 
            let uf2 = unrollTypeDeep f2.ftype in 
            let cr = checkTypeIsomorphism uf1 uf2 in
              if cr = sNO_MATCH then 
                begin
                  fieldsTypeChangedList := (f1.fname,f1.ftype,f2.ftype)::!fieldsTypeChangedList;
                end;
              cr
          end
        else
           begin 
            (*Printf.printf "-branch\n";*)
            sMATCH
          end
      in summy
    in
      
    let found = 
      try    
          ignore(checkFieldinfoType f1 (findSameName f1 f2l));
          sMATCH
      with Not_found ->
        begin
          if op = '+' 
          then fieldsAddedList := f1.fname::!fieldsAddedList
          else fieldsDeletedList := f1.fname::!fieldsDeletedList;
          sNO_MATCH
        end
    in   found
  in
    
  let checkListOneWay (op : char) (f1 : fieldinfo list) (f2 : fieldinfo list) : bool =
    let resultslist = List.map (checkPresenceInOther op f2) f1 in
    let finalval = List.for_all (fun x -> if x = sNO_MATCH then false else true)  resultslist
    in   finalval
  in 

  let sameLen = sameLength c1 c2 in
  let all3 = 
    (* let nichts = compStatistics#incr "total" in *)
    let homomorphismOldNew =  checkListOneWay '-' c1.cfields  c2.cfields in
    let homomorphismNewOld =  checkListOneWay '+' c2.cfields  c1.cfields in
      sameLen && homomorphismOldNew && homomorphismNewOld
  in

  let () = 
     if !pass = 2 (*&& (not (Cil.startsWith "union \"__anon" c1name)) &&  (not (Cil.startsWith "struct \"__anon" c1name)) *)
     then
      begin
        let 
          fdel = List.length !fieldsDeletedList and
           fadd = List.length !fieldsAddedList and
          ftc  = List.length !fieldsTypeChangedList in
          begin
            if not sameLen then
              begin
                (*Printf.printf "function \"%s\": argument count changed \n" name;*)
                compStatistics#incr "field count changes";
              end;
            if fdel > 0 then
              begin
                (*Printf.printf "%s: %d fields deleted: " c1name fdel;
                List.iter (fun s -> Printf.printf "%s " s) !fieldsDeletedList;
                Printf.printf "\n";*)
                List.iter 
                  (fun s -> ch.c_fieldsChange <- {v_name = s; v_change = Deleted }::ch.c_fieldsChange) 
                            !fieldsDeletedList;
                compfieldsDeletedStatistics#add c1name fdel ((getLocationOfString c1name).file);
                compStatistics#incrN "fields deleted" fdel;
              end;
            if fadd > 0 then
              begin
                (*Printf.printf "%s: %d fields added: " c1name fadd;
                List.iter (fun s -> Printf.printf "%s " s) !fieldsAddedList;
                Printf.printf "\n"; *)
                List.iter 
                  (fun s -> ch.c_fieldsChange <- {v_name = s; v_change = Added }::ch.c_fieldsChange) 
                            !fieldsAddedList;


                compfieldsAddedStatistics#add c1name fadd ((getLocationOfString c1name).file);
                compStatistics#incrN "fields added" fadd;
              end;
            if ftc > 0 then
              begin
                (*Printf.printf "%s: %d fields changed type: \n" c1name ftc;

                List.iter (
                  fun (n,t1,t2) -> Printf.printf "\t field %s changed type from %s to %s \n" 
                    n 
                    (Commoncilutil.prettyPrintType t1) 
                    (Commoncilutil.prettyPrintType t2)) 
                  !fieldsTypeChangedList; *)

                List.iter 
                  (fun (n,t1,t2) -> ch.c_fieldsChange <- 
                     {v_name = n; v_change = TypeChanged((Commoncilutil.prettyPrintType t1),(Commoncilutil.prettyPrintType t2))}::ch.c_fieldsChange) 
                            !fieldsTypeChangedList;


                compfieldsTypeChangeStatistics#add c1name ftc ((getLocationOfString c1name).file);
                compStatistics#incrN "field type changes" ftc;
              end;
            if fdel > 0 || fadd > 0 || ftc > 0
            then 
              begin 
                ch.c_change <- OtherChange "change";
                compStatistics#incr "comp type changes";

                      H.add compCensus c1name ch
              end
          end
      end

  in 
    if all3 
    then sMATCH 
    else sNO_MATCH

and

(************************* CHECK BODY ISOMORPHISM *******************************************)

  checkBodyIsomorphism (fname : string) (f1 : fundec) (f2 : fundec) : bool = 

  let f1map: hashOfStringToInt = H.create 127 in
  let f2map: hashOfStringToInt = H.create 127 in

  (* let nichts = Printf.printf "Checking function %s...\n" fname in *)
  let endResult = ref true in
  let someNamesChanged = ref false in
  let nameConstraintList = ref [] in 
  let typeConstraintList = ref [] in 
  
  let addLocalTypeConstraint (t1 : typ) (t2 : typ) : unit = 
    begin
      if (!pass = 1)
      then typeConstraintList := (t1,t2)::!typeConstraintList 
    end 
  in 
  
  let flushLocalConstraints = fun () ->
    begin
      (*Printf.printf "flushLocalConstraints (%d constraints)\n" (List.length !typeConstraintList);*)
      if (!pass = 1)
      then List.iter (fun p -> let (x,y) = p in addTypeConstraint x y) !typeConstraintList
    end
  in
  
  let setResultFalse  = fun () -> endResult := false;() in 
  let setSomeNamesChanged  = fun () -> someNamesChanged := true;() in 

  let addVar (hr  :  hashOfStringToInt) (v : varinfo ) : unit =
    try
      let r = H.find hr v.vname in 
        Printf.printf "WARNING: %s already mapped (to %d)!\n" v.vname r
    with
      Not_found -> 
            let fl = freshLocalCount () in
            H.add hr v.vname fl
              (* Printf.printf "Mapping (\"%s\",%d) added\n" v.vname fl *)
  in

  let addLocalVarConstraint (s1 : string) (s2 : string) : unit = 
    begin
      (* Printf.printf "Adding constraint [%s,%s]\n" s1 s2; *)
      begin
        if (not (argMatch2 s1  s2)) then setSomeNamesChanged ();
        try 
          let i1 = H.find f1map s1 in
            begin
              try
                let i2 = H.find f2map s2 in
                  begin
                  if not (List.mem (i1,i2) !nameConstraintList) 
                  then nameConstraintList := (i1,i2)::!nameConstraintList
                end
              with Not_found -> (* Printf.printf "Cannot find \"%s\" in File2 function hash ?!\n" s2 *) ()
            end
        with Not_found -> (* Printf.printf "Cannot find \"%s\" in File1 function hash ?!\n" s1 *) ()
      end
    end
  in

  let addGlobalVarConstraint (s1 : string) (s2 : string) : unit = 
     begin
        (* In a nutshell, if the Lvals are function names then introduce renaming constraints 
           between functions, else if they're global vars then introduce renaming constraints 
           between global vars *)
        let fun_ish1 =  Commoncilutil.getGlobalNameLight_gfun s1 
        and fun_ish2 =  Commoncilutil.getGlobalNameLight_gfun s2 
        and gvar_ish1 = Commoncilutil.getGlobalNameLight_gvar s1
        and gvar_ish2 = Commoncilutil.getGlobalNameLight_gvar s2
        in
           if (H.mem globalsFile1 fun_ish1) && (H.mem globalsFile2 fun_ish2)
           then functionPartialMap#addConstraint (s1,s2)
           else 
              if (H.mem globalsFile1 gvar_ish1) && (H.mem globalsFile2 gvar_ish2)
              then gvarPartialMap#addConstraint (s1,s2)
              else () 
     end
  in

  let addVarinfoConstraint (v1 : Cil.varinfo) (v2: Cil.varinfo) : bool = 

     match v1.vglob, v2.vglob with
          true, true -> (addGlobalVarConstraint v1.vname v2.vname; true)
        | false, false -> (addLocalVarConstraint v1.vname v2.vname; true)
        | _ -> false
  in

  let rec analyzeExpressions (exp1 : exp) (exp2 : exp) : unit = 
    match exp1,exp2 with
        Const c1, Const c2 -> if c1 <> c2 then  setResultFalse (); ()
      | Lval l1, Lval l2 -> analyzeLvals l1 l2
      | SizeOf t1, SizeOf t2 -> if (checkTypeIsomorphism t1 t2) = sNO_MATCH then setResultFalse (); ()
      | SizeOfE e1, SizeOfE e2 -> analyzeExpressions e1 e2
      | SizeOfStr s1, SizeOfStr s2 -> if s1 <> s2 then setResultFalse (); ()
      | AlignOf t1, AlignOf t2 -> if (checkTypeIsomorphism t1 t2) = sNO_MATCH then setResultFalse (); ()
      | AlignOfE e1, AlignOfE e2 -> analyzeExpressions e1 e2
              
      | UnOp (op1,e1,_), UnOp (op2,e2,_) -> 
          begin
            if op1 <> op2 
            then setResultFalse ()
            else analyzeExpressions e1 e2
          end
      | BinOp (op1,e1,e1',_), BinOp (op2,e2,e2',_) ->
          begin
            if op1 <> op2 
            then setResultFalse ()
            else 
              begin 
                analyzeExpressions e1 e2;
                analyzeExpressions e1' e2'
              end
          end
      | CastE (t1,e1), CastE (t2,e2) ->
          if (checkTypeIsomorphism t1 t2) = sNO_MATCH
          then setResultFalse ()
          else analyzeExpressions e1 e2

      | AddrOf lval1, AddrOf lval2 ->   analyzeLvals lval1 lval2
      | StartOf lval1, StartOf lval2 ->  analyzeLvals lval1 lval2
      | _,_ -> setResultFalse ()

  and analyzeLvals (lv1 : lval) (lv2 : lval) : unit = 
    match lv1, lv2 with
        (*   must also check offset !    *)
        (Var v1, o1), (Var v2, o2) -> 
          analyzeOffsets o1 o2;
          begin
            (* if ASTs are off sync, make no sense to keep adding constraints*)
            if (!pass = 1) then
                begin
                  if (!endResult) then
                    if not (addVarinfoConstraint v1 v2)
                    then setResultFalse ()
                      (*Printf.printf "Function %s adding Lval constraint %s %s\n" fname (Commoncilutil.prettyPrintType v1.vtype) (Commoncilutil.prettyPrintType v2.vtype);*)
                    else addLocalTypeConstraint v1.vtype v2.vtype
                end
              else (* pass = 2 *)
                if (not !endResult) || (v1.vname = v2.vname) 
                then ()
                else
                  begin (* possible renaming *)
                    let dealWithRenaming isFun =
                      let gv1 =
                        if isFun
                        then Commoncilutil.getGlobalNameLight_gfun v1.vname
                        else Commoncilutil.getGlobalNameLight_gvar v1.vname
                      and gv2 = 
                        if isFun
                        then Commoncilutil.getGlobalNameLight_gfun v2.vname
                        else Commoncilutil.getGlobalNameLight_gvar v2.vname
(*                     in *)
(*                     let () = Printf.printf "analyzeLvals %s %s %b %b %b %b\n" gv1 gv2 (H.mem globalsFile1 gv1) (H.mem globalsFile2 gv1) *)
(*                       (H.mem globalsFile1 gv2) (H.mem globalsFile2 gv2) *)
                    in
                      if ((H.mem globalsFile1 gv1) && not (H.mem globalsFile2 gv1) && (H.mem globalsFile2 gv2) && not (H.mem globalsFile1 gv2))
                        || ((H.mem globalsFile1 gv2) && not (H.mem globalsFile2 gv2) && (H.mem globalsFile2 gv1) && not (H.mem globalsFile1 gv1))
                      then
                        begin
                          let renamingsList =
                            if isFun
                            then gvarPartialMap#getRenamings ()
                            else functionPartialMap#getRenamings ()
                          in
                            if (List.mem (gv1,gv2) renamingsList)  ||
                              (List.mem (gv2,gv1 )renamingsList)
                            then ()
                            else
                              begin
(*                                 Printf.printf "analyzeLvals: setResultFalse %s %s\n" gv1 gv2; *)
                                setResultFalse ()
                              end
                        end
                      else
                        begin
(*                           Printf.printf "analyzeLvals: setResultFalse %s %s\n" gv1 gv2; *)
                          setResultFalse ()
                        end
                  in
                    dealWithRenaming true;
                    dealWithRenaming false
                  end
          end
      | (Mem e1, o1), (Mem e2, o2) -> 
          analyzeOffsets o1 o2;
          analyzeExpressions e1 e2
      | _,_ -> setResultFalse ()

  and analyzeOffsets (o1: offset)(o2: offset) : unit = 
    match o1,o2 with 
   
           NoOffset, NoOffset -> ()

         | Field(fi1,o11),Field(fi2,o22) -> 
              analyzeFields fi1 fi2;
              analyzeOffsets o11 o22

         | Index(e1,o11),Index(e2,o22) -> 
              analyzeExpressions e1 e2;
              analyzeOffsets o11 o22

         | _,_ -> setResultFalse ()

  and analyzeFields (f1 : fieldinfo) (f2 : fieldinfo) : unit = 
(*     Printf.printf "analyzefields %s %s\n" f1.fname f2.fname; *)
    if f1.fname <> f2.fname 
    then setResultFalse ()
  in

  let analyzeInstructions (i1: instr) (i2: instr) : unit = 

    match i1,i2 with
        Set (lval1,rhs1,_), Set (lval2,rhs2,_) -> 
          begin
            analyzeLvals lval1 lval2; 
            analyzeExpressions rhs1 rhs2
          end
      | Call (res1,fexpr1,actuals1,_), Call (res2,fexpr2,actuals2,_) ->
          begin
            analyzeExpressions fexpr1 fexpr2;
            try 
              List.iter2 analyzeExpressions actuals1 actuals2
            with Invalid_argument _ -> setResultFalse ()
          end
          (* potential problem with the asm code being different ! *)
    | Asm _, Asm _  -> ()
    | _,_ -> setResultFalse ()

  in

  let rec analyzeStatements (s1 : stmt) (s2 : stmt) : unit =

  match s1.skind, s2.skind with
      Instr il1, Instr il2 ->
        begin
          try 
            List.iter2 analyzeInstructions il1 il2
          with Invalid_argument _ -> setResultFalse ()
        end
    | Return (eo1,_), Return (eo2,_) ->
        begin
          match eo1, eo2 with
            | Some e1, Some e2 ->
                begin
                  analyzeExpressions e1 e2
                end
            | None,None -> ()
            | _,_ -> setResultFalse ()
        end
    | If (e1,b1,b1',_), If (e2,b2,b2',_) ->
        begin
          analyzeExpressions e1 e2;
          analyzeBlocks b1 b2;
          analyzeBlocks b1' b2';
        end
    | Switch (e1,b1,sl1,_), Switch (e2,b2,sl2,_) ->
        begin
          analyzeExpressions e1 e2;
          analyzeBlocks b1 b2;
          begin
            try
              List.iter2 analyzeStatements sl1 sl2
            with Invalid_argument _ -> setResultFalse ()
          end
        end
    | Loop (b1,_,_,_), Loop (b2,_,_,_) -> analyzeBlocks b1 b2
    | Block (b1), Block (b2)  -> analyzeBlocks b1 b2

        (* potential problem with the goto targets being different ! *)
    | Goto (_,_), Goto (_,_) | Break _, Break _ | Continue _, Continue _ -> ()
    | _,_ ->  setResultFalse ()

  and
    analyzeBlocks (b1 : block) (b2 : block) : unit =

    if List.length b1.bstmts <> List.length b2.bstmts 
    then setResultFalse ()
    else 
      begin
        try
          List.iter2 analyzeStatements b1.bstmts b2.bstmts
        with Invalid_argument _ -> setResultFalse ()  
      end
  in

  let createDeclarationMapping (hr : hashOfStringToInt) (vl: varinfo list) : unit =
    List.iter (addVar hr) vl
  in

  let addConstraintsOnFormals (oldFormalsList : varinfo list) (newFormal : varinfo) : unit = 
    List.iter (fun (oldFormal : varinfo) -> if argMatch oldFormal.vname newFormal.vname then
                 addLocalTypeConstraint oldFormal.vtype newFormal.vtype) oldFormalsList
  in
  let () =  
    begin
      List.iter (addConstraintsOnFormals f1.sformals)  f2.sformals;
      createDeclarationMapping f1map f1.sformals;
      createDeclarationMapping f1map f1.slocals;
      createDeclarationMapping f2map f2.sformals;
      createDeclarationMapping f2map f2.slocals;
      try List.iter2 (fun (v1 : varinfo) (v2 : varinfo) -> begin 
                        ignore(addVarinfoConstraint v1 v2);
                        addLocalTypeConstraint v1.vtype v2.vtype end) f1.sformals f2.sformals
      with Invalid_argument _ -> ()
    end
  in
(**************  no constraints are generated for unused formals !!! ******************)
  let checkLocalsTypes () = 
     if (!pass = 2)
     then
        let f2ntl = List.map (fun vi -> (vi.vname, vi.vtype)) in
        if not (nameTypeListMatch (f2ntl f1.slocals) (f2ntl f2.slocals)) 
        then setResultFalse ()   
  in
  let checkBijection (l : (int*int) list) : bool = 
    let isBijection = ref true in
    let firstSet  = List.fold_right (fun (a1,a2) accu -> (a1::accu)) l [] in
    let secondSet = List.fold_right (fun (a1,a2) accu -> (a2::accu)) l [] in
    let countOccurences elem lst = List.fold_right (fun (a) accu -> if a = elem then (accu + 1) else accu) lst 0 in
    let tickEmOff  lst1 lst2 (a1,a2) = 
      let c1 = countOccurences a1 lst1 in
      let c2 = countOccurences a2 lst2 in
        if c1 <> 1 || c2 <> 1 then isBijection := false else ()
    in 
      ignore(List.map (tickEmOff firstSet secondSet) l);
      !isBijection

  in
    begin
      analyzeBlocks f1.sbody f2.sbody;

      (* We flush the constraints anyway; in the worst case we only have constraints on formals 
      and some locals *)
      flushLocalConstraints ();
      (* Printf.printf "CONSTRAINTS:\n"; *)
(*       List.map (fun (a1,a2) -> Printf.printf "[%d,%d]\n" a1 a2) !nameConstraintList; *)

      checkLocalsTypes ();

      if (checkBijection !nameConstraintList)
      then
        begin
          (* Printf.printf "Bijection."; *)
          if (!endResult)
          then
            begin
              begin 
                if (!pass <> 1) && (!someNamesChanged) then 
                  begin
                    Printf.printf "Function \"%s\" : some names changed, but still bijection\n" fname; 
                    functionStatistics#incr "locals/formals changed name";
                  end;
                ()
              end;
              (*Printf.printf "Bijection. Bodies isomorphic\n";*)
              true
            end
          else
            begin
              if (!pass <> 1)  then 
                begin
                  (*Printf.printf "Function \"%s\" : Bijection. Bodies NOT isomorphic.\n" fname;  *)
                  H.add functionCensus fname { f_name = fname; 
                                               f_change = ValChanged;
                                               f_returnValChange = NoChange;
                                               f_argumentsChange = []; };
                  functionStatistics#incr "bijection, bodies not isomorphic";
                end;
              false
            end
        end
      else
        begin 
          if (!pass <> 1)  then 
            begin
              (*Printf.printf "Function \"%s\" : NOT Bijection. Bodies NOT isomorphic.\n" fname;*)
               H.add functionCensus fname { f_name = fname; 
                                            f_change = ValChanged;
                                            f_returnValChange = NoChange;
                                            f_argumentsChange = []; };

              functionStatistics#incr "not bijection, bodies not isomorphic";
            end;
          false
        end
    end

and

  checkGlobalMatchingStatus (n1:string) (n2:string) : int =
  try
    let res = H.find globalMatchingStatus (n1 ^ n2)
    in (*Printf.printf "cache hit (%s)\n" (n1 ^ n2);*) res
  with Not_found -> sNOT_CHECKED_YET

and

  checkGlobalIsomorphismPass1 (g1 : global) (g2 : global) : int =
  let matchResult = 
    match g1,g2 with
      | GVar (v1, i1, l1), GVar (v2, i2, l2) -> checkGVarIsomorphism (v1,i1) (v2,i2)
      | GType (t1, l1), GType (t2, l2) ->       checkTypedefIsomorphism t1 t2
      | GCompTag (c1, l1), GCompTag (c2, l2) -> checkCompinfoIsomorphismPass1 c1 c2
      | GFun (f1, l1), GFun (f2, l2) ->
                begin
                  let bi = checkBodyIsomorphism f1.svar.vname f1 f2 in
                  let fi =  checkFunctionIsomorphism f1.svar.vname f1.svar.vtype f2.svar.vtype in
                    begin
                      (*if bi && (fi = sNO_MATCH) then Printf.printf "False positive\n"*)
                    end;
                    if bi && (fi = sNO_MATCH) then sMATCH else sNO_MATCH
                end
      | _,_ -> sMATCH
  in matchResult

and

  checkGlobalIsomorphismPass2 (g1 : global) (g2 : global) : int =
  
  let n1 = Commoncilutil.getGlobalName g1 and n2 = Commoncilutil.getGlobalName g2 in
  let cs = checkGlobalMatchingStatus n1 n2 in
    if cs = sNOT_CHECKED_YET
    then 
      begin
        let matchResult = 
          match g1,g2 with
               GType (t1, l1), GType (t2, l2) -> checkTypedefIsomorphism t1 t2
            | GCompTag (c1, l1), GCompTag (c2, l2) -> checkCompinfoIsomorphism c1 c2
            | GCompTagDecl (c1, l1), GCompTagDecl (c2, l2) -> sMATCH
            | GEnumTag (e1, l1), GEnumTag (e2, l2) -> checkEnumIsomorphism e1 e2
            | GEnumTagDecl (e1, l1), GEnumTagDecl (e2, l2) -> sMATCH
            | GVarDecl (v1, l1), GVarDecl (v2, l2)-> sMATCH
            | GVar (v1, i1, l1), GVar (v2, i2, l2) -> checkGVarIsomorphism (v1,i1) (v2,i2)
            | GFun (f1, l1), GFun (f2, l2) ->
                begin
                  let bi = checkBodyIsomorphism f1.svar.vname f1 f2 in
                  let fi = checkFunctionIsomorphism f1.svar.vname f1.svar.vtype f2.svar.vtype in
                    begin
                      (*if bi && (fi = sNO_MATCH) then Printf.printf "False positive\n"*)
                    end;
                    (* Printf.printf "f1.svar.vname %s f2.svar.vname %s bi %b fi %d\n" f1.svar.vname f2.svar.vname bi fi; *)
                    if bi && (fi = sMATCH) then sMATCH else sNO_MATCH
                end
            | GAsm (s1, l1), GAsm (s2, l2) -> sMATCH
            | GPragma (a1, l1), GPragma (a2, l2) -> sMATCH
            | GText (s1), GText (s2)-> sMATCH
            | _,_ -> sNO_MATCH
        in 
          begin
            if n1 = n2 
            then
              H.add globalMatchingStatus (n1 ^ n2) matchResult
            else
              begin
                H.add globalMatchingStatus (n1 ^ n2) matchResult;
                H.add globalMatchingStatus (n2 ^ n1) matchResult;
              end;      
            matchResult
          end;
      end
    else (* already checked *)
      cs



let setAs (g : global) (isAdded : bool) : unit =
  let 
    isAddedString = if isAdded then "added" else "deleted" and
    isAddedFact   = if isAdded then Added else Deleted 
  in
    begin
      globalStatistics#incr isAddedString;

      match g with
          GType (t, l) -> 
            begin
              typedefStatistics#incr isAddedString;

              let ch : typedefFact = { t_name = t.tname; t_change = isAddedFact} 
              in  H.add typedefCensus t.tname ch; 
            end

        | GCompTag (c, l) -> 
            begin
              (*Printf.printf "FOO %s %s\n" c.cname isAddedString;*)
              compStatistics#incr isAddedString;

              let ch : compFact = { c_name = c.cname; 
                                         c_isStruct = if c.cstruct then true else false; 
                                         c_change = isAddedFact;
                                         c_fieldsChange = []}
              in  H.add compCensus c.cname ch; 
            end

        | GEnumTag (e, l) -> 
            begin 
              enumStatistics#incr isAddedString;

              let ch : enumFact = { e_name = e.ename; e_change = isAddedFact; e_valsChange = []} 
              in H.add enumCensus e.ename ch
            end

        | GVar  (v, i, l) -> 
            begin 
              varStatistics#incr isAddedString;

              let ch : globalVarFact = { g_name = v.vname; g_change = isAddedFact} 
              in H.add globalVarCensus v.vname ch; 
            end

        | GFun (f, l) -> 
            begin
              functionStatistics#incr isAddedString;
              (* Printf.printf "setAs %s %s\n" isAddedString f.svar.vname; *)
              let ch : functionFact = { f_name = f.svar.vname; 
                                        f_change = isAddedFact; 
                                        f_returnValChange = NoChange;
                                        f_argumentsChange = []}
              in H.add functionCensus f.svar.vname ch; 
            end

    | _ -> () 
    end

let setAsAdded   (g : global) : unit = setAs g true
let setAsDeleted (g : global) : unit = setAs g false


let setStatus (s : string) (statusStore : hashOfStringToInt) (newval : int) : unit =  
  try
    let oldval = H.find statusStore s in 
      begin
        Printf.printf "Changing status  of %s  (%d->%d)\n" s oldval newval;
        H.remove statusStore s;
        H.add    statusStore s newval
      end
  with Not_found -> H.add  statusStore s newval

let computeTotals (hr : hashOfGlobals) (file : string) : unit =
  let doOneGlobal (s : string) (g : global) =
    begin
      globalStatistics#incr file ;
      match g with
      GType (t, l) -> typedefStatistics#incr file
    | GCompTag (c, l) -> 
        (
          compStatistics#incr file;
          (*Printf.printf "FOO %s %s\n" c.cname file;*)
        )
    | GEnumTag (e, l) -> enumStatistics#incr file
    | GVar  (v, i, l) -> varStatistics#incr file
    | GFun (f, l) -> functionStatistics#incr file
    | _ -> () 
    end

  in  H.iter doOneGlobal hr

(* let computeAddedDeleted (s : Statistics.aggregateStatistics) : unit = *)
(*   let f1val = s#get "File1" in *)
(*   let f2val = s#get "File2" in *)
(*   let deleted = s#get "deleted" in *)
(*   let added = s#get "added" (\* (f2val - f1val + deleted) *\) in *)
(*     begin *)
(* (\*        Printf.printf "%d + %d - %d = %d\n" f2val deleted f1val added;  *\) *)
(* (\*        if added >= 0  *\) *)
(* (\*        then *\) *)
(* (\*           s#set "added" added *\) *)
(* (\*        else  *\) *)
(* (\*           begin *\) *)
(* (\*              Printf.printf "File2 : %d  File1 : %d  Deleted : %d => Added : %d\n" f2val f1val deleted added;  *\) *)
(* (\*              raise (Failure "Added < 0 ??") *\) *)
(* (\*           end *\) *)
(*     end *)
let addNaturalConstraints () : unit = 

  let naturalConstraintGenerator tEnv1 tEnv2  =

    let natConGenIterator strVal1 numVal1 = 
      let clist = 
        try
          H.find typeConstraints numVal1 
        with Not_found -> []
      in
        try 
          let numVal2 = H.find tEnv2 strVal1 in
             begin
              if not (List.mem numVal2 clist) 
              then
                begin
                  H.remove typeConstraints numVal1;
                  H.add typeConstraints numVal1 (numVal2::clist);
                end
            end
        with Not_found -> ()
    in H.iter natConGenIterator tEnv1

  in 
    begin
      naturalConstraintGenerator typeEnvironment1 typeEnvironment2;
(*       naturalConstraintGenerator typeEnvironment2 typeEnvironment1; *)
    end

let checkConstraints () : unit = 
  let rec findEquivalent (l : int list) (tv : int) : int =
    match l with
        [] -> -1
      | h::t -> try if (H.find typeEnvironmentReverseMapping1 tv) = (H.find typeEnvironmentReverseMapping2 h) then h else findEquivalent t tv
          with Not_found -> Printf.printf "Bug in findEquivalent ?!\n";1
  in

  let splitConstraints tv1 e tv2 = 
    if e = tv2 
    then
      H.add satisfiedTypeConstraints (tv1,tv2) 0
    else
      H.add unsatisfiedTypeConstraints (tv1,tv2) 0
  in

  let test tv clist = 
    begin
      if (List.length clist) = 1
      then
        begin H.add satisfiedTypeConstraints (tv, List.hd clist) 0; end
      else
        let e = findEquivalent clist tv in
          List.iter (splitConstraints tv e) clist
    end
  in
    H.iter  test typeConstraints

let printConstraints () : unit = 
  let getTypeByVar (tv : int) (ter : hashOfIntToString) : string = 
    try H.find ter tv 
    with Not_found -> Printf.printf "Bug in getTypeByVar ?!\n";""  
  in
  let printConstraint pair ignored = 
    let (x,y) = pair in
    let tn1 = getTypeByVar x typeEnvironmentReverseMapping1 in
    let tn2 = getTypeByVar y typeEnvironmentReverseMapping2 in
      begin
        if tn1 <> tn2 && not (anonMatch tn1 tn2)
        then begin Printf.printf "Type Vars (old,new)"; 
        Printf.printf " (%d,%d) (%s,%s)\n" x y tn1 tn2
          end
      end
  in

    begin
      Printf.printf "Satisfied Constraints (%d) ............\n" (Commonutil.hashtbl_size satisfiedTypeConstraints); 
      H.iter printConstraint satisfiedTypeConstraints;
      Printf.printf "Unsatisfied Constraints (%d) ..........\n" (Commonutil.hashtbl_size unsatisfiedTypeConstraints); 
      H.iter printConstraint unsatisfiedTypeConstraints;
    end

let printDescrVals () =
   let descrList =  
      functionStatistics#sweepDescrVal @
      varStatistics#sweepDescrVal @
      compStatistics#sweepDescrVal @ 
      typedefStatistics#sweepDescrVal @ 
      enumStatistics#sweepDescrVal 
   in
   List.iter (fun (d,v) -> Printf.printf "@@%s,%d@@\n" d v) descrList;;

let printStatistics () : unit = 
   let forceRecomputeAddedDeleted fl stats =
      let deleted = List.length (filterFactList fl Deleted)
      and added =   List.length (filterFactList fl Added) 
      and renamed = List.length (filterFactList fl (Renamed ("",""))) in
      let () = 
        let f1 = stats#get "File1" and
            f2 = stats#get "File2" in
            if (f1 + added - deleted) <> f2 
            then
              raise (Failure(Printf.sprintf "Things don't add up for '%s': f1: %d, f2: %d, added %d, deleted %d (old_a %d, old_d %d)\n " 
                 stats#getStoreName 
                 f1 f2 added deleted
                 (stats#get "added") (stats#get "deleted")
               ))
      in
  
         begin
            stats#set "added" added;
            stats#set "deleted" deleted;
            stats#set "renamed" renamed;
         end
   in
  begin
    computeTotals globalsFile1 "File1";
    computeTotals globalsFile2 "File2";



    forceRecomputeAddedDeleted (List.map (fun x -> (FGlobalVar x)) (Commonutil.hashtbl_values globalVarCensus)) varStatistics;
    forceRecomputeAddedDeleted (List.map (fun x -> (FTypedef x)) (Commonutil.hashtbl_values typedefCensus)) typedefStatistics;
    forceRecomputeAddedDeleted (List.map (fun x -> (FFunction x)) (Commonutil.hashtbl_values functionCensus)) functionStatistics;
    forceRecomputeAddedDeleted (List.map (fun x -> (FComp x)) (Commonutil.hashtbl_values compCensus)) compStatistics;
    forceRecomputeAddedDeleted (List.map (fun x -> (FEnum x)) (Commonutil.hashtbl_values enumCensus)) enumStatistics;

    globalStatistics#set "added" ( functionStatistics#get "added"  
                                   + varStatistics#get "added"
                                   + compStatistics#get "added"
                                   + typedefStatistics#get "added"
                                   + enumStatistics#get "added");

    globalStatistics#set "deleted" ( functionStatistics#get "deleted"  
                                     + varStatistics#get "deleted"
                                     + compStatistics#get "deleted"
                                     + typedefStatistics#get "deleted"
                                     + enumStatistics#get "deleted");

    globalStatistics#set "renamed" ( functionStatistics#get "renamed"  
                                     + varStatistics#get "renamed"
                                     + compStatistics#get "renamed"
                                     + typedefStatistics#get "renamed"
                                     + enumStatistics#get "renamed");

(*     computeAddedDeleted globalStatistics; *)
(*     computeAddedDeleted functionStatistics; *)
(*     computeAddedDeleted varStatistics; *)
(*     computeAddedDeleted compStatistics; *)
(*     computeAddedDeleted typedefStatistics; *)
(*     computeAddedDeleted enumStatistics; *)

    globalStatistics#printAll "g";
    functionStatistics#printAll "f";
    varStatistics#printAll "v";
    compStatistics#printAll "c";
    typedefStatistics#printAll "t";
    enumStatistics#printAll "e"; 

    funargAddedStatistics#printDistribution false;   
    funargDeletedStatistics#printDistribution false;
    funargTypeChangedStatistics#printDistribution false;

    compfieldsAddedStatistics#printDistribution false;
    compfieldsDeletedStatistics#printDistribution false; 
    compfieldsTypeChangeStatistics#printDistribution false;

    enumvalsAddedStatistics#printDistribution false;
    enumvalsDeletedStatistics#printDistribution false;
    enumvalsChangedStatistics#printDistribution false;

    printDescrVals ();
  end

(* global vars and functions are things we 'hang on' to *)


let secondChance (s : string) (hr : hashOfGlobals) =
  let matches = ref false in
  let tst gname _ = 
    if (anonMatch s gname) && (not (H.mem hr s))
    then 
      (matches := true;
       Printf.printf "anonMatch '%s' '%s'\n" s gname
      )
  in 
  H.iter tst hr;
  !matches

let checkOne (direction : char) (srcStatus : hashOfStringToInt) (dest : hashOfGlobals) (s : string) (g2 : global) : unit = 
    if (!pass = 1)
    then
      match g2 with
          (*GVar  (_, _, _) |delete me ASAP*) GFun (_, l) | GCompTag (_,l)-> 
            begin
              (*Printf.printf "Location : %s\n" l.file;*)
              try
                let g1 = H.find dest s in
                ignore(checkGlobalIsomorphismPass1 g1 g2) 
              with Not_found -> ()
            end
        | _ -> (*Printf.printf "global(%s) ignored \n" s*)()
    else
      try
        let g1 = H.find dest s in
        ignore (checkGlobalIsomorphismPass2 g1 g2)
(*            Printf.printf "checkOne : checkGlobalIsomorphismPass2 OK %c global %s\n" direction s *)

      with Not_found -> 
         begin
           let isComp g = (match g with GCompTag _ | GEnumTag _ -> true | _ -> false)  in
           if (isComp g2) && (secondChance s dest)
           then 
             Printf.printf "setting %s as nothing happened\n" s
           else
             (if direction = '+' 
               then 
                 (Printf.printf "setting %s as added\n"s;
                  setAsAdded g2)
               else
                 (Printf.printf "setting %s as deleted\n"s;
                  setAsDeleted g2)
             )
         end
;;

class generateCallGraphVisitor (fi :  fundec) (h : hashOfCallGraph) : cilVisitor = object(self)
  inherit nopCilVisitor

  method vinst i = 
    match i with
        Call(_, Lval(Var vi, NoOffset), _, _) -> 
          begin
            addCallEdge h fi.svar.vname vi.vname;
            DoChildren
          end

      | _ -> DoChildren
end

let generate_call_graph (hg : hashOfGlobals) (h : hashOfCallGraph) =
  let doOneFunction (fi: fundec) = 
    let v = new generateCallGraphVisitor fi h in
      ignore (visitCilBlock v fi.sbody)
  in
    begin
      H.iter (fun x y -> match y with GFun(fi, _) -> doOneFunction fi | _ -> ()) hg;
    end

let doCallGraphAnalysis () = 
let list2string s = List.fold_right (fun x accu -> "\"" ^ x ^ "\"" ^ " " ^ accu) s "" in
let set2list s = StringSet.fold (fun x accu -> x::accu) s [] in
  let doOneCaller caller setRef otherHash = 
    let cmpFun s1 s2 = (s1 = s2) || (anonMatch s1 s2) in
    try
      let otherSetRef = H.find otherHash caller in
      let oldList = set2list !otherSetRef in
      let newList = set2list !setRef in
      let added = Commonutil.list_diff newList oldList cmpFun in
      let deleted = Commonutil.list_diff oldList newList cmpFun in
      let printIfNeeded caller opr l =
        begin
          if l <> []
          then Printf.printf "\n\t\t\t\t %s %d: %s " opr (List.length l) (list2string l)
        end
      in
        begin
          if added <> [] || deleted <> [] 
          then Printf.printf "\n  function \"%s\"" caller; 

          printIfNeeded caller "+" added;
          printIfNeeded caller "-" deleted
        end 
    with Not_found -> ()
  in
  let addCallee h caller callee =
    let setRef = 
      try
        H.find h caller
      with Not_found ->
        let res = ref StringSet.empty in
          H.add h caller res;
          res
    in
    let newSet = StringSet.add callee !setRef in
      H.replace h caller (ref newSet)
  in
  let orderedHash1 = H.create 4099 in
  let orderedHash2 = H.create 4099 in
  let fillHash srcH destH = H.iter (fun (x,y) _ -> addCallee destH x y) srcH in
    begin
      generate_call_graph globalsFile1 functionCallGraph1;
      generate_call_graph globalsFile2 functionCallGraph2;

      fillHash functionCallGraph1 orderedHash1;
      fillHash functionCallGraph2 orderedHash2;


      Printf.printf "\n ----------  Call Graph Changes ----------- \n";

      H.iter (fun x y -> doOneCaller x y orderedHash1) orderedHash2;

      Printf.printf "\n ------------------------------------------ \n";
    end

let fixRenamings () = 

   let r (x, y) = Renamed (x,y) in
   let fixGvarRenaming (x, y) = 
      let census = globalVarCensus in
      try
         match (H.find census x).g_change, (H.find census y).g_change with
              Deleted, Added ->
                 begin
                    H.remove census x;
                    H.remove census y;
                    H.add census y { g_name = y; g_change = r (x, y)};
                 end
            | _ -> ()

      with Not_found -> ()

   and fixFunRenaming (x, y) = 
      let census = functionCensus in
      try
         match (H.find census x).f_change, (H.find census y).f_change with
              Deleted, Added ->
                 begin
                    H.remove census x;
                    (H.find census y).f_change <- r (x, y)
                 end
            | _ -> ()

      with Not_found -> ()

   and fixTypedefRenaming (x, y) = 
      let census = typedefCensus in
      try
         match (H.find census x).t_change, (H.find census y).t_change with
              Deleted, Added ->
                 begin
                    H.remove census x;
                    H.remove census y;
                    H.add census y { t_name = y; t_change = r (x, y)}
                 end
            | _ -> ()

      with Not_found -> ()

   and fixCompRenaming (x, y) = 
      let census = compCensus in
      try
         match (H.find census x).c_change, (H.find census y).c_change with
              Deleted, Added ->
                 begin
                    H.remove census x;
                    (H.find census y).c_change <- r (x, y)
                 end
            | _ -> ()

      with Not_found -> ()

   and fixEnumRenaming (x, y) = 
      let census = enumCensus in
      try
         match (H.find census x).e_change, (H.find census y).e_change with
              Deleted, Added ->
                 begin
                    H.remove census x;
                    (H.find census y).e_change <- r (x, y)
                 end
            | _ ->  () 


      with Not_found -> () 
   in

   List.iter fixGvarRenaming    (gvarPartialMap#getRenamings ());
   List.iter fixFunRenaming     (functionPartialMap#getRenamings ());
   List.iter fixTypedefRenaming (typedefPartialMap#getRenamings ());
   List.iter fixCompRenaming    (compPartialMap#getRenamings ());
   List.iter fixEnumRenaming    (enumPartialMap#getRenamings ())
;;
      

let printFacts () =
  let printProperly (factCategory : string) (fl : fact list) = 
    begin
      Printf.printf "\n\n ----------  %s ----------- \n\n" factCategory;
      let lDeleted = filterFactList fl Deleted
      and lAdded = filterFactList fl Added
      and lTypeChanged = filterFactList fl (TypeChanged ("",""))
      and lValChanged = filterFactList fl ValChanged
      and lNoChange = filterFactList fl NoChange
      and lOther = filterFactList fl (OtherChange "")   
      and lRenamed = filterFactList fl (Renamed ("","")) in
        begin
          List.iter (fun x ->printFact x) lDeleted;
          List.iter (fun x ->printFact x) lAdded;
          List.iter (fun x ->printFact x) lTypeChanged;
          List.iter (fun x ->printFact x) lValChanged;
          List.iter (fun x ->printFact x) lOther;
          List.iter (fun x ->printFact x) lRenamed;
          Printf.printf "%d + %d + %d + %d + %d + %d + %d = %d\n" 
             (List.length lDeleted) 
             (List.length lAdded) 
             (List.length lTypeChanged) 
             (List.length lValChanged) 
             (List.length lNoChange) 
             (List.length lOther) 
             (List.length lRenamed) 

             (List.length fl)
        end
    end
  in
  begin
    printProperly "GLOBAL VARS" (List.map (fun x -> (FGlobalVar x)) (Commonutil.hashtbl_values globalVarCensus));
    printProperly "TYPEDEFS" (List.map (fun x -> (FTypedef x)) (Commonutil.hashtbl_values typedefCensus));
    printProperly "FUNCTIONS" (List.map (fun x -> (FFunction x)) (Commonutil.hashtbl_values functionCensus));
    printProperly "COMPS" (List.map (fun x -> (FComp x)) (Commonutil.hashtbl_values compCensus));
    printProperly "ENUMS" (List.map (fun x -> (FEnum x)) (Commonutil.hashtbl_values enumCensus));
  end

(* Maincil will set this when it boots up *)
let startParsing = ref 0.0

let resetStats () = 
   let resetStats s l = List.iter (fun x -> s#reset x) l in
   begin
      
    (* Set all stats to 0 no matter what, to force them being printed, even if their values are 0 *)

    resetStats varStatistics [("added","VADD");
                              ("deleted", "VDEL");
                              ("renamed", "VREN");
                              ("File1", "VONE");
                              ("File2", "VTWO");
                              ("var type changes","VTC");
                              ("static initializer changes","VSIC")];

    resetStats functionStatistics [("added","FADD");
                                   ("deleted", "FDEL");
                                   ("renamed", "FREN");
                                   ("File1", "FONE");
                                   ("File2", "FTWO");
                                   ("function signature changes","FSC");
                                   ("return types changes","FRTC");
                                   ("arguments type changes","FATC");
                                   ("locals/formals changed name","FLCN");
                                   ("bijection, bodies not isomorphic","FBBNI");
                                   ("not bijection, bodies not isomorphic","FNBNI")];

    resetStats compStatistics [("added","CADD");
                               ("deleted", "CDEL");
                               ("renamed", "CREN");
                               ("File1", "CONE");
                               ("File2", "CTWO");
                               ("fields added","CFA");
                               ("fields deleted","CFD");
                               ("field type changes","CFTC");
                               ("field count changes","CFCC");
                               ("comp count changes","CTC")];


    resetStats typedefStatistics [("added","TADD");
                                  ("deleted", "TDEL");
                                  ("renamed", "TREN");
                                  ("File1", "TONE");
                                  ("File2", "TTWO");
                                  ("base type changes","TBTC")];

    resetStats enumStatistics [("added","EADD");
                               ("deleted", "EDEL");
                               ("renamed", "EREN");
                               ("File1", "EONE");
                               ("File2", "ETWO");
                               ("items added","EIA");
                               ("items deleted","EID");
                               ("val exp changes","EVC");
                               ("item count changes","EIC")];

   end


let doAnalysis (standalone : bool) : unit = 
  let  tEndParsing  = ref 0.0 in
  let  tEndGenerateConstraints = ref 0.0 in
  let  tEndCheckConstraints = ref 0.0 in
  let  tEndPrintStatistics = ref 0.0 in  
  let  tStart = ref 0.0 in
  begin
     if standalone then
     begin
        tStart := !startParsing;
        tEndParsing := (Unix.gettimeofday ());

        Printf.printf "Computing statistics...\n"
     end;

     resetStats ();

     H.iter (checkOne '+' statusStore2 globalsFile1) globalsFile2;

     if standalone then 
     begin
        tEndGenerateConstraints := (Unix.gettimeofday ()); 
     end;

     addNaturalConstraints ();
     checkConstraints ();


     if standalone then 
     begin
        printConstraints (); 
        tEndCheckConstraints := (Unix.gettimeofday ())
     end;

      nextPass ();
      H.iter (checkOne '+' statusStore2 globalsFile1) globalsFile2;
      flipDirection ();
      H.iter (checkOne '-' statusStore1 globalsFile2) globalsFile1;


      (* we might have detected spurious new/old globals due to renamings; fix those *)
      fixRenamings ();

      if standalone then 
      begin
        tEndPrintStatistics := (Unix.gettimeofday ());

        Printf.printf " ------ Facts begin -------------\n";

        doCallGraphAnalysis ();

        printFacts ();
        Printf.printf " ------ Facts end -------------\n";
        printStatistics (); 
        Printf.printf " ------ Analysis ended -------------\n";
        Printf.printf "Parsing :                %.2fs \n" (!tEndParsing -. !tStart);
        Printf.printf "Generating Constraints : %.2fs \n" (!tEndGenerateConstraints -. !tEndParsing);
        Printf.printf "Solving Constraints :    %.2fs \n" (!tEndCheckConstraints -. !tEndGenerateConstraints);
        Printf.printf "Gathering Statistics:    %.2fs \n" (!tEndPrintStatistics-. !tEndCheckConstraints);
        Printf.printf "Total               :    %.2fs \n" (!tEndPrintStatistics-. !tStart);

        raise ImDone
      end;

    end

let addGlobal gStore locStore g : unit = 

    let canonicalName = Commoncilutil.getGlobalName g in 
    begin
       (*Printf.printf "* just added %s\n" canonicalName;*)
       H.add gStore canonicalName g;
       H.add locStore canonicalName (Cil.get_globalLoc g);
    end


(*  Things to do before processing a file *)
let pre (f : Cil.file) (standalone : bool) : unit = () 

(*  Things to do after processing a file *)
let post (f : Cil.file) (standalone : bool) : unit =  
   if (!stage = ProcessFile1Stage) 
   then 
      begin 
      stage := ProcessFile2Stage;
      Printf.printf "Finished processing File1\n"
      end
   else
      doAnalysis standalone

(* this is called for each file in turn; 
   for the second file, post() will compute the diffs *)
let semdiffProcessFile (f : Cil.file) (standalone : bool) : unit = 
  pre f standalone;
  begin
    if !stage = ProcessFile1Stage
    then iterGlobalsFiltered f filterOutGlobalDeclarations (addGlobal globalsFile1 globalLocations)
    else iterGlobalsFiltered f filterOutGlobalDeclarations (addGlobal globalsFile2 globalLocations)
  end;
  post f standalone

(* used by the patch generator to detect functions that have changed between versions *)
let computeFunctionAndTypeChanges (f1 : Cil.file) (f2 : Cil.file) : semdiffResult = 
   (* we might have multiple elems with same key, but patchgen doesn't like that *)
   let uniqifyHash hIn hOut = 
      H.iter (fun x y -> Commonutil.hashtbl_add_once hOut x y) hIn;
      hOut
   in
   begin
      semdiffProcessFile f1 false;
      semdiffProcessFile f2 false;
      {
         sr_vars = uniqifyHash (H.create 127) globalVarCensus;
         sr_functions = uniqifyHash (H.create 127) functionCensus;
         sr_typedefs = uniqifyHash (H.create 127) typedefCensus;
         sr_comps = uniqifyHash (H.create 127) compCensus;
      }
   end

(* 'Semantic' diff feature descriptor *)

let semDiffEnabled = ref false 

let feature : featureDescr = 
  { fd_name = "semdiff";
    fd_enabled = semDiffEnabled;
    fd_description = "find semantic differences between two programs";
    fd_extraopt = [];
    fd_doit = (function (f: file) -> semdiffProcessFile f true);
    fd_post_check = true;
  } 

