(* 
 *
 * Copyright (c) 2004-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Michael Hicks       <mwh@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



open Cil
open Musolver
open Deltasolver


module H = Hashtbl



type effect = 
{ 
   (* ---- updateability effects ----*)
   effect_delta_in    : delta;   (* input effect  *)
   effect_delta_out   : delta;   (* output effect *)
   effect_delta_eigen : delta;   (* function's own delta *)
(*    effect_labels    : delta;   (\* set of labels (function names) *\) *)
   effect_mu        : mu;      (* updatability  *)
   (* ---- VC effects ---- *)
   effect_vc_fwd  : delta;   (* forward flow-sensitive   *)
   effect_vc_eps  : delta;   (* flow-insenstive  *)
   effect_vc_bck  : delta;   (* backward flow-sensitive *)
}

type phi = 
{ 
   (* ---- updateability effects ----*)
   phi_delta  : delta;   (* input effect  *)

   (* ---- VC effects ---- *)
   phi_vc_fwd : delta;   (* forward flow-sensitive   *)
   phi_vc_eps : delta;   (* flow-insenstive  *)
   phi_vc_bck : delta;   (* backward flow-sensitive *)
}

let phi_of_effect eff = { 
  phi_delta  = eff.effect_delta_in;
  phi_vc_fwd = eff.effect_vc_fwd;
  phi_vc_eps = eff.effect_vc_eps;
  phi_vc_bck = eff.effect_vc_bck}
;;


(* Type of hash table to store function effects keyed on a unique effect id *)
type funEffectHash_t = (int, effect) H.t
(* Function name -> function type hash *)
type funTypeHash_t = (string, Cil.typ) H.t

(* replace "" with string_of_int s to print id's *)
let debug_string_of_int s = string_of_int s;;
(* let debug_string_of_int s = "";; *)
 
(* type of hash tables that map from statement ids to input/output effect pairs *)
type statementEffect = (int, (delta * delta)) H.t

(* type of hashtbl for goto targets: maps statement ids to (delta, phi) *) 
type gotoTargetHash = (int, phi) H.t

type constDeltaHash_t = (string, delta) H.t

(* constant nodes in constraint graph can be types/funs/gvars; 
these will appear on __DSU_update() annotations *)
type deltaConstKind = DckType | DckFun | DckGvar
type constNamesHash_t = (string, deltaConstKind) H.t
let int_of_dck  dck = match dck with DckGvar -> 2 | DckFun -> 1 | DckType -> 0;;

(* VC effect kinds *)
type vcKind = VKFwd | VKEps | VKBck | VKChk

(* phi -> (longjmp target, crt_fun_mu) hash *)
type longjmpHash_t = (phi, (Cil.varinfo * mu)) H.t

(* delta id -> capability set hash *)
type deltaSolutionHash_t = (int, CapabilitySet.t) H.t

(* Results that are passed out of the analysis *)
type analysis_result = 
   {
      ar_delta_constraints       : DeltaConstraintSet.t;  
      ar_mu_constraints          : MuConstraintSet.t;
      ar_delta                   : delta;
      ar_mu                      : mu;
      ar_stmtid_to_delta_id      : (int,(int*int)) H.t;
      ar_funEffectHash           : funEffectHash_t;
      ar_funEffWaterMark         : int;
      ar_speculativeUpdateMuHash : (int,int) H.t;
      ar_mu_solution             : (int,bool) H.t;
      ar_delta_solution          : deltaSolutionHash_t;
      ar_constNamesHash          : constNamesHash_t;
      (* --- VC --- *)  
      ar_stmtid_to_eps_tuple     : (int,(int*int*int*int option)) H.t;
      ar_stmtid_to_fun_name      : (int,string) H.t;
      ar_state_xform_eff_id      : int option;
      ar_c_of_id                 : int -> CapabilitySet.t;
   }





(* Misc. functions 
   TODO: sort these out
*)

let noFunction = Cil.emptyFunction "**No FUNCTION**"

(* transforms a hash of constraints (e.g. as used internally by the monad) into a set of constraints *)
let deltaConstraintHashToSet (constraintHash : ((bool * int * int), delta_constraint) H.t) init : DeltaConstraintSet.t = 
  let deltaConstraintHashToSet_fold key deltaConstraint acc = 
    makeSureThereSNoConstOnRight deltaConstraint;
    DeltaConstraintSet.add deltaConstraint acc 
  in
  H.fold deltaConstraintHashToSet_fold constraintHash init

(* transforms a hash of constraints (e.g. as used internally by the monad) into a set of constraints *)
let muConstraintHashToSet (constraintHash : ((mu * mu), unit) H.t) : MuConstraintSet.t = 
  let muConstraintHashToSet_fold muConstraint _ acc = MuConstraintSet.add muConstraint acc in
  H.fold muConstraintHashToSet_fold constraintHash MuConstraintSet.empty



let makeDeltaDebugStrings (d1 : delta) (d2 : delta) : string*string = 
  let d1_id_str = string_of_int d1.delta_id in
  let d2_id_str = string_of_int d2.delta_id in
  let d1str = match d1.delta_kind with DKVar _ -> d1_id_str | DKConst s -> (s ^ "(" ^ d1_id_str ^ ")") in
  let d2str = match d2.delta_kind with DKVar _ -> d2_id_str | DKConst s -> (s ^ "(" ^ d2_id_str ^ ")") in
    d1str,d2str


(*let debugPrint str = Printf.printf str*)
let debugPrint (str:string) : unit = ignore (str)













(*============================================================================*)
(* Monad for the analysis *)

type analysisMonadInput_t =      { 
        ami_delta_constraints         : ((bool * int *int), delta_constraint) H.t;   (* delta constraint hash *)
        ami_mu_constraints            : ((mu * mu), unit) H.t;      (* updatability constraint hash *)
        ami_statement_effects         : statementEffect;            (* mapping from statement id's to effects *)
        ami_fwd_goto_targets          : gotoTargetHash;             (* forward goto targets (sid to bool hash) *)
        ami_mu                        : mu ref;                     (* current updatability *)
        mutable ami_delta             : delta;                      (* current delta *)
        mutable ami_eigen_delta       : delta;                      (* current eigen delta *)
        ami_mu_id_counter             : int ref;                    (* the next unique id for mu variables *)
        ami_fun_effect_id_counter     : int ref;                    (* the next unique id for function effects *)
        ami_funEffectHash             : funEffectHash_t;            (* hash from effect id to function effect *)
        ami_solutionHash              : deltaSolutionHash_t;        (* hash from delta id to capability set *)
        ami_constDeltaHash            : constDeltaHash_t;           (* hash from type names to const deltas.
                                                                       Used so every const delta with the same name
                                                                       has the same delta_id *)
        ami_constNamesHash            : constNamesHash_t;           (* hash from type/fun/gvar name to its kind*) 
        ami_voidptrfun_effect_id      : int;                        (* The id of the effect to associate with funtion 
                                                                       ptrs cast to void *)                    
        mutable ami_current_fd        : Cil.fundec;                 (* fundec of the current function being analysed *)
        ami_stmtid_to_delta_id        : (int,(int*int)) H.t;        (* Hash from stmt id to delta_in_id * delta_out_id *)
        ami_stmt_delta_in_id          : int ref;                    (* Delta id current at start of current statement *)
        ami_mallocFunHash             : Mallocpragma.mallocHash_t;  (* Hash representing set of user specified malloc functions *)
        ami_speculativeUpdateMuHash   : (int,int) H.t;              (* Hash to associate mu variables with the statement id's of 
                                                                       speculative update points *)
        ami_suspect_hash              : (string,(delta*delta)) H.t; (* maps a function name to a constraint 
                                                                    induced by the suspect alias phase *)

        ami_intern_var                : (string, unit) H.t;         (* internal variables *)
        ami_longjmp_hash              : longjmpHash_t;              (* delta_id -> longjmp target hash *)

        (* -- Debuging information -- *)
        mutable ami_loc               : Cil.location;               (* current location in the file *)

        (* --- VC --- *)
                ami_vc_enabled        : bool;                       (* if set to false, no VC constraints are added *)
        mutable ami_vc_fwd            : delta;                      (* current fwd effect *)
        mutable ami_vc_eps            : delta;                      (* current normal effect *)
        mutable ami_vc_bck            : delta;                      (* current bck effect *)
        mutable ami_checkin_eps       : delta option;               (* current eps, if inside a checkin *)

        ami_stmtid_to_vc_ids          : (int,(int*int*int*int option)) H.t; (* Hash from stmt id to (fwd * eps * bck * checkinEps) *)
        ami_stmtid_to_fun_name        : (int,string) H.t;           (* Hash from stmt id to fun name *)

        ami_sweep_hash                : (string,unit) H.t;
        ami_sweeper_enum_syntactic    : Cil.cilVisitor;
        ami_sweeper_gvar_syntactic    : Cil.cilVisitor;
        ami_sweeper_gvar_ptranal      : Cil.cilVisitor;
      }



(* Signature *)

(* Need the signature, to pass it into the constraint generation modules 
   Only keep the minimum interface we can get away with.
*)
module type AnalysisMonadSig = 
sig
  
    type mi = analysisMonadInput_t
    type 'a mo = 'a
    type 'a m = mi -> 'a mo
    val return : 'a -> mi -> 'a
    val ( >>= ) : 'a m -> ('a -> 'b m) -> 'b m

    val amSetUpdatable : unit m
    val amGetDelta : Deltasolver.delta m
    val amGetEigenDelta : Deltasolver.delta m
    val amSetDelta : Deltasolver.delta -> unit m
    val amSetEigenDelta : Deltasolver.delta -> unit m
    val amFreshDeltaVar : Deltasolver.delta m
    val amMakeDeltaConst : string -> deltaConstKind -> Deltasolver.delta m
    val amSetMu : Musolver.mu -> unit m
    val amAddMCLessThanCurrent : Musolver.mu -> unit m
    val amAddDCSubset : Deltasolver.delta -> Deltasolver.delta -> string -> unit m
    val amAddDCCond : Musolver.mu -> Deltasolver.delta -> Deltasolver.delta -> string -> unit m
    val amAddDCSubsetOfCurrent : Deltasolver.delta -> string -> unit m
    val amAddDESubsetOfCurrent : Deltasolver.delta -> string -> unit m
    val amAddDCSupersetOfCurrent : Deltasolver.delta -> string -> unit m
    val amSetVcFwd : Deltasolver.delta -> unit m
    val amSetVcEps : Deltasolver.delta -> unit m
    val amSetVcBck : Deltasolver.delta -> unit m
    val amGetVcFwd : Deltasolver.delta m
    val amGetVcEps : Deltasolver.delta m
    val amGetVcBck : Deltasolver.delta m
    val amSetCheckinEps : Deltasolver.delta option -> unit m
    val amGetCheckinEps : Deltasolver.delta option m
    val amAddVCSubset : Deltasolver.delta -> Deltasolver.delta -> string -> unit m
    val amAddVcSubsetOfCurrent : Deltasolver.delta -> vcKind -> string -> unit m
    val amStoreSpeculativeUpdateMu : int -> unit m
    val getLoc : Cil.location m
end

(* Implementation *)

module AnalysisMonad = 
struct
  (* monad input / output types *)
  type mi = analysisMonadInput_t
        
  type 'a mo = 'a
        
  (* Type of the monad *)
  type 'a m = mi -> 'a mo
    


  (* basic monadic operations *)
  let return (x:'a) = fun (ami:mi) -> x (* {(\* amo_delta = ami.ami_delta;  *\)amo_result = x} *)

  let (>>=) (m : 'a m) (f : 'a -> 'b m) : 'b m = 
   fun (ami : mi) -> 
      let amo = m ami in
      f amo ami
  

     


  (* -- Operations specific to this monad -- *)  

  (* The result type of running this state monad *)
  type monad_run_result = 
      {
        delta_constraints       : DeltaConstraintSet.t;  
        mu_constraints          : MuConstraintSet.t;
        delta                   : delta;
        mu                      : mu;
        stmtid_to_delta_id      : (int,(int*int)) H.t;
        funEffectHash           : funEffectHash_t;
        funEffectWatermark      : int;
        speculativeUpdateMuHash : (int,int) H.t;
        constNamesHash          : constNamesHash_t;
        (* --- VC --- *)
        stmtid_to_vc_ids        : (int,(int*int*int*int option)) H.t;
        stmtid_to_fun_name      : (int,string) H.t;
      }


  (* Standard monad glue *)
  let run 
        (mallocFunHash : Mallocpragma.mallocHash_t) 
        (effectHash : funEffectHash_t)
        (solutionHash : deltaSolutionHash_t)
        (suspectResult : Suspectalias.SAM.mo_result) 
        (internVarHash : (string, unit) H.t) 
        (funEffWaterMark : int)

        (m : 'a m)         = 
    let suspect_hash = suspectResult.Suspectalias.SAM.sar_fname_to_LPS_hash in
    let init_constrs = suspectResult.Suspectalias.SAM.sar_LPC_constraints in
    let sweep_hash = H.create 17 in
    let ami =   { ami_delta_constraints = H.create 65535;
                  ami_mu_constraints = H.create 4099;
                  ami_statement_effects = H.create 133103;
                  ami_fwd_goto_targets = H.create 4099;
                  ami_mu = ref (MuConst true);
                  ami_delta = {delta_loc = Cil.locUnknown; delta_id = -1; delta_kind = DKVar None};
                  ami_eigen_delta = {delta_loc = Cil.locUnknown; delta_id = -3; delta_kind = DKVar None};
                  ami_mu_id_counter = ref 0;
                  ami_fun_effect_id_counter = ref funEffWaterMark;
                  ami_funEffectHash = effectHash; 
                  ami_solutionHash = solutionHash; 
                  ami_constDeltaHash = H.create 4099;
                  ami_constNamesHash = H.create 4099;
                  ami_voidptrfun_effect_id = 0;
                  ami_loc = Cil.locUnknown;
                  ami_current_fd = noFunction; 
                  ami_stmtid_to_delta_id = H.create 4099;
                  ami_mallocFunHash = mallocFunHash;
                  ami_stmt_delta_in_id = ref (-1); 
                  ami_speculativeUpdateMuHash = H.create 4099; 
                  ami_suspect_hash = suspect_hash;
                  ami_longjmp_hash = H.create 7;
                  ami_intern_var = internVarHash;
                  ami_sweep_hash = sweep_hash;
                  ami_sweeper_enum_syntactic = new Updatecilutil.sweepVisitor_enum_syntactic 
                                                        (Updatecilutil.addEnumToHash sweep_hash);

                  ami_sweeper_gvar_syntactic = new Updatecilutil.sweepVisitor_gvar_syntactic 
                                                        (Updatecilutil.addVarToHash sweep_hash);

                  ami_sweeper_gvar_ptranal = new Updatecilutil.sweepVisitorPtranal
                                                        (Updatecilutil.addNameToHash sweep_hash);
                  (* --- VC --- *)
                  ami_vc_enabled = Updateargs.args.Updateargs.arg_vcEnabled;
                  ami_vc_fwd = {delta_loc = Cil.locUnknown; delta_id = -1; delta_kind = DKVar None};
                  ami_vc_eps = {delta_loc = Cil.locUnknown; delta_id = -1; delta_kind = DKVar None};
                  ami_vc_bck = {delta_loc = Cil.locUnknown; delta_id = -1; delta_kind = DKVar None};
                  ami_checkin_eps = None;
                  ami_stmtid_to_vc_ids = H.create 4099;
                  ami_stmtid_to_fun_name = H.create 4099;
                } in
    (* Go *)
    m ami;
    (* done *) 
      { delta_constraints = deltaConstraintHashToSet ami.ami_delta_constraints init_constrs;
        mu_constraints = muConstraintHashToSet ami.ami_mu_constraints; 
(*         delta = amo.amo_delta; *)
        delta = ami.ami_delta;
        mu = !(ami.ami_mu);
        stmtid_to_delta_id = ami.ami_stmtid_to_delta_id;
        funEffectHash = ami.ami_funEffectHash; 
        funEffectWatermark = !(ami.ami_fun_effect_id_counter) + 1; 
        speculativeUpdateMuHash = ami.ami_speculativeUpdateMuHash;
        constNamesHash = ami.ami_constNamesHash;

        (* --- VC --- *)
        stmtid_to_vc_ids = ami.ami_stmtid_to_vc_ids;
        stmtid_to_fun_name = ami.ami_stmtid_to_fun_name;
      }


  (* Set the current mu state to updatable (true) *)
  let amSetUpdatable = fun ami -> ami.ami_mu := MuConst true; return () ami

  (* Set the current mu state to non-updatable (false) *)
  let amSetNonUpdatable = fun ami -> ami.ami_mu := MuConst false; return () ami

  let amGetDelta = fun ami -> return ami.ami_delta ami
  let amGetEigenDelta = fun ami -> return ami.ami_eigen_delta ami

  let amSetDelta (d : delta) = fun ami -> ami.ami_delta <- d; return () ami
  let amSetEigenDelta (d : delta) = fun ami -> ami.ami_eigen_delta <- d; return () ami

  let amGetUpdateConstraint fname =
    let find_opt h k =
      try Some (H.find h k) with Not_found -> None in
    fun ami -> return (find_opt ami.ami_suspect_hash fname) ami

  (* Generate a fresh variable and use it as the default delta *)
  let amFreshDeltaVar : delta m = 
    fun ami -> 
       let d = Deltasolver.freshDelta (DKVar None) ami.ami_loc in 
       return (d) ami

  let amMakeDeltaConst (const_name : string) (const_kind : deltaConstKind): delta m =
    fun ami ->
      let addDck () = Commonutil.hashtbl_add_once ami.ami_constNamesHash const_name const_kind in
      (* Reuse const deltas we've seen before, else generate a fresh one *)
      let delta = 
        try
          H.find ami.ami_constDeltaHash const_name
        with
            Not_found -> 
              let d = Deltasolver.freshDelta (DKConst const_name) ami.ami_loc in
                H.add ami.ami_constDeltaHash const_name d;
		          addDck ();
                ignore(Updatedebug.log [Updatedebug.BTAnalysis] "amMakeDeltaConst '%s' : %d %d\n" const_name d.delta_id (int_of_dck const_kind));
                d
      in 
        return delta ami


  let amFreshMuVar : mu m = 
    fun ami -> 
      incr ami.ami_mu_id_counter;
      return (MuVar !(ami.ami_mu_id_counter)) ami


  let amGetNextFunEffectID : int m = 
    fun ami -> 
      incr ami.ami_fun_effect_id_counter; 
      return !(ami.ami_fun_effect_id_counter) ami

  let amAddFunEffectEntry key effect : unit m =
    fun ami ->
      H.add ami.ami_funEffectHash key effect;
      return () ami

  let amGetFunEffectEntry key : effect m = 
    fun ami ->
      return (
        try
          H.find ami.ami_funEffectHash key
        with Not_found ->
          Updatedebug.abort (Updatedebug.bug "amGetFunEffectEntry: can't find key %d\n" key)
      ) ami

  let amGetMu = fun ami -> return !(ami.ami_mu) ami
  let amSetMu mu = fun ami -> ami.ami_mu := mu; return () ami


  (* Add the mu constraint mu1 < mu2 *)
  let amAddMC (mu1:mu) (mu2:mu) = 
    fun ami ->
      begin
        if not (H.mem ami.ami_mu_constraints (mu1, mu2)) 
        then H.add ami.ami_mu_constraints (mu1, mu2) ();
        (*printConstraint (mu1,mu2);*)
      end;
      return () ami


  (* Add a constraint between mu1 and the current mu *)
  let amAddMCLessThanCurrent (mu1:mu) = 
    amGetMu >>= fun current_mu -> 
      amAddMC mu1 current_mu 

  let amAddMCGreaterThanCurrent (mu1:mu) = 
    amGetMu >>= fun current_mu -> 
      amAddMC current_mu mu1



  (* Add the constraint d1 \subset d2 *)
  let amAddDCSubset (d1 : delta) (d2 : delta) (why : string) : unit m = 
    fun ami -> 
      begin
        if not (H.mem ami.ami_delta_constraints (false, d1.delta_id, d2.delta_id))
        then begin H.add ami.ami_delta_constraints (false, d1.delta_id, d2.delta_id) (DCSubset(d1,d2));
             let d1str,d2str = makeDeltaDebugStrings d1 d2 in 
               ignore(Updatedebug.log [Updatedebug.BTAnalysis] "Adding constraint %s <= %s (%d <= %d) : %s\n" d1str d2str d1.delta_id d2.delta_id why)
        end
      end;
      return () ami


  (* Add a conditional constraint mu => d1 < d2 *)
  let amAddDCCond (mu : mu) (d1 : delta) (d2 : delta) (why : string) : unit m = 
    fun ami ->
      match mu with 
          MuVar mu_id ->
            if not (H.mem ami.ami_delta_constraints (true, d1.delta_id, d2.delta_id))
            then begin H.add ami.ami_delta_constraints (true, d1.delta_id, d2.delta_id) (DCCond(mu_id,(d1,d2)));
              let d1str,d2str = makeDeltaDebugStrings d1 d2 in ignore(Updatedebug.log [Updatedebug.BTAnalysis] "Adding cond constraint %s <= %s (%d <= %d) : %s\n" d1str d2str d1.delta_id d2.delta_id why) end;

            return () ami
              
        | MuConst true ->
            amAddDCSubset d1 d2 why ami
              
        | MuConst false ->
            return () ami
              

  (* Add a constraint between d and the current delta: d \subseteq ami_delta *)
  let amAddDCSubsetOfCurrent (d : delta) (why : string) : unit m = 
    amGetDelta >>= fun current_delta ->
    amAddDCSubset d current_delta why 

  (* Add a constraint between d and the current delta: d \subseteq ami_delta *)
  let amAddDESubsetOfCurrent (d : delta) (why : string) : unit m = 
    amGetEigenDelta >>= fun eigen_delta ->
    amAddDCSubset d eigen_delta why 


  (* Add a constraint between d and the current delta: ami_delta \subseteq d *)
  let amAddDCSupersetOfCurrent (d : delta) (why : string) : unit m = 
    amGetDelta >>= fun current_delta ->
    amAddDCSubset current_delta  d why

  (* Checkin stuff *)
  let amStartCheckinEps : unit m =
    fun ami ->
      let m =          (* first separate delta *)
                          amFreshDeltaVar  
        >>= fun newD   -> amAddDCSubsetOfCurrent newD "amStartCheckinEps"
        >>= fun ()     -> amSetDelta newD 

                       (* then separate checkinEps *) 
        >>= fun ()     -> amFreshDeltaVar  
        >>= fun newCE  -> ami.ami_checkin_eps <- Some newCE;
          return ()
      in m ami

  let amEndCheckinDelta : unit m =
    fun ami ->
      let m =          (* first separate delta *)
                          amFreshDeltaVar  
        >>= fun newD   -> amAddDCSubsetOfCurrent newD "amEndCheckinEps"
        >>= fun ()     -> amSetDelta newD 
        >>= fun ()     -> return ()
      in m ami

  let amEndCheckinEps : unit m =
    fun ami ->
      let m =          
                       (* then reset checkinEps *) 
                          ami.ami_checkin_eps <- None;
                          return ()
      in m ami

  let amGetCheckinEps : delta option m =
    fun ami ->
    return ami.ami_checkin_eps ami

  let amSetCheckinEps (d : delta option) : unit m =
    fun ami ->
    ami.ami_checkin_eps <- d;
    return () ami


  (* get effect of statement *)
  let amGetStatementEffect (sid:int) = 
    fun ami -> 
      return (
        try
          H.find ami.ami_statement_effects sid
        with Not_found ->
          Updatedebug.abort (Updatedebug.bug "amGetFunEffectEntry: can't find sid %d\n" sid)
      ) ami
        

  let amSetStatementEffect (sid:int) (d:delta) = 
    fun ami -> H.add ami.ami_statement_effects sid (d, ami.ami_delta);
      return ()



  (* --- VC --- *)
  let amSetVcFwd (d : delta) = fun ami -> ami.ami_vc_fwd <- d; return () ami
  let amSetVcEps (d : delta) = fun ami -> ami.ami_vc_eps <- d; return () ami
  let amSetVcBck (d : delta) = fun ami -> ami.ami_vc_bck <- d; return () ami


  let amGetVcFwd = fun ami -> return ami.ami_vc_fwd ami
  let amGetVcEps = fun ami -> return ami.ami_vc_eps ami
  let amGetVcBck = fun ami -> return ami.ami_vc_bck ami

  let amAddVCSubset (d1 : delta) (d2 : delta) (why : string) : unit m = 
    fun ami -> 
      if ami.ami_vc_enabled
      then amAddDCSubset d1 d2 why ami
      else return () ami    


  let amAddVcSubsetOfCurrent (d : delta) (vk : vcKind) (why : string) : unit m = 
    (match vk with 
         VKFwd | VKEps | VKBck ->  

           ((match vk with 
                VKFwd -> amGetVcFwd 
              | VKEps -> amGetVcEps 
              | VKBck -> amGetVcBck 
              | VKChk -> assert(false)
           )
           >>= fun current_delta -> 
             ignore(Updatedebug.log [Updatedebug.BTAnalysis] "amAddVcSubsetOfCurrent (%d <= %d)\n" d.delta_id current_delta.delta_id);
             amAddVCSubset d current_delta why)

        | VKChk -> amGetCheckinEps
            >>= fun chki -> match chki with
                None -> return ()
              | Some current_delta -> ignore(Updatedebug.log [Updatedebug.BTAnalysis] "amAddVcSubsetOfCurrent (%d <= %d)\n" d.delta_id current_delta.delta_id);
             amAddVCSubset d current_delta why
    )

  (*               GOTO handling     *)      

  (* save crt input delta just in case we might be the target of a goto *)
  let amAddFwdLabel (sid:int) = 
    fun ami ->

(*       ignore(Updatedebug.log [Updatedebug.BTAnalysis] "amAddFwdLabel: adding id for sid: %d\n" sid); *)

      let p = {phi_delta  = ami.ami_delta;
               phi_vc_fwd = ami.ami_vc_fwd;
               phi_vc_eps = ami.ami_vc_eps;
               phi_vc_bck = ami.ami_vc_bck} 
      in
      H.add ami.ami_fwd_goto_targets sid p;
      return () ami

  (* remove a forward label marker *)
  let amRemoveFwdLabel (sid:int) = 
    fun ami -> 
      H.remove ami.ami_fwd_goto_targets sid;
      return () ami

  (* get the phi associated with a forward label;
     if not found (i.e. forward goto), return current delta;
     this will introduce a trivial constraint d <= d but we don't bother *)

  let amGetFwdPhi (sid:int) : phi option m = 
    fun ami -> 
      try
        let phi = (H.find ami.ami_fwd_goto_targets sid) in
        return (Some phi) ami
      with 
          Not_found -> return None ami




  let amConstrainVoidPtrFunEffect : unit m =
    fun ami ->
      (*
        let m = 
      (* For all single constants (named types) k add the constraint  k \subseteq d_top.
        Associate the effect N;d_top;d_top with the predefined key for void* fun effects in the monad. *)
        amGetFunEffectEntry ami.ami_voidptrfun_effect_id
        >>= fun effect -> return (Commonutil.hashtbl_values (ref ami.ami_constDeltaHash))
      (* Doesn't matter whether we choose the in or out delta capability as they have the same delta id *)
        >>= fun (dl: delta list) -> AMOps.iter (fun (c:delta) -> amAddDCSubset c effect.effect_delta_in) dl
        in 
        m ami
      *)
      return () ami

  (* Create the effect that we use on functions cast to void* *)
  let amCreateVoidPtrFunEffect : unit m =
    fun ami ->
      let m =              
        amFreshDeltaVar       
        >>= fun d_top  -> amAddFunEffectEntry ami.ami_voidptrfun_effect_id
          {
            effect_delta_in  = d_top; 
            effect_delta_out = d_top;
            effect_delta_eigen = d_top;
(*             effect_labels = d_top; *)
            effect_mu = MuConst false;

            (* --- VC --- *)
            effect_vc_fwd  = d_top; 
            effect_vc_eps  = d_top; 
            effect_vc_bck= d_top; 
          }
      in
        m ami


  (* Return the effect id used on functions that are cast to void* *)
  let amGetVoidPtrFunEffectId : int m = 
    fun ami -> return ami.ami_voidptrfun_effect_id ami


  (* Get statistical information *)
  let amGetNumMuConstraints : int m =
    fun ami -> return (Commonutil.hashtbl_size ami.ami_mu_constraints) ami
      
  let amGetNumDeltaConstraints : int m =
    fun ami -> return (Commonutil.hashtbl_size ami.ami_delta_constraints) ami

  (* Get the hash that store the malloc function names *)
  let amGetMallocFunHash : Mallocpragma.mallocHash_t m = 
    fun ami -> return (ami.ami_mallocFunHash) ami


  (* Associate the statement id sid of a speculative update point with the current mu *)
  let amStoreSpeculativeUpdateMu (sid:int)= 
    fun ami ->
      (* Assume current mu is a variable and extract the mu_id *)
      match !(ami.ami_mu) with 
          MuVar mu_id ->
            (*ignore(Updatedebug.log [Updatedebug.BTAlways] "Adding speculative update to mu hash: %d |-> %d\n" sid mu_id);*)
            H.add ami.ami_speculativeUpdateMuHash sid mu_id;
            return () ami

        | _ -> Updatedebug.abort (Updatedebug.bug "amStoreSpeculativeUpdateMu: current mu not a var at statement %d\n" sid)
  


  (* ---- members that make this module a subtype of SyntaxMonad ---- *)

  (* Get/Set for the current location *)
  let getLoc = fun ami -> return ami.ami_loc ami
  let setLoc (loc : Cil.location) = fun ami -> ami.ami_loc <- loc; return () ami

  (* Get/Set the current function being analysed *)
  let unsetCurrentFun = fun ami -> ami.ami_current_fd <- noFunction; return () ami
  let setCurrentFun (current_fd : Cil.fundec) = fun ami -> ami.ami_current_fd <- current_fd; return () ami
  let getCurrentFun = fun ami -> return ami.ami_current_fd ami

  (* Register the start and end of statements so that we can asssociate statement id's to 
     deltas -- needed to print statement effects *)
  let amStmtStart sid = 
    fun ami -> ami.ami_stmt_delta_in_id := ami.ami_delta.delta_id; 
      ignore(Updatedebug.log [Updatedebug.BTAnalysis] "amStmtStart: adding delta id %d for sid: %d\n" !(ami.ami_stmt_delta_in_id) sid);
(*       H.add ami.ami_stmtid_to_vc_ids sid (ami.ami_vc_fwd.delta_id, ami.ami_vc_eps.delta_id, ami.ami_vc_bck.delta_id); *)
      return () ami

  let amStmtEnd sid deltaAtStart vcFwdAtStart = 
    fun ami ->
(*       H.add ami.ami_stmtid_to_delta_id sid (!(ami.ami_stmt_delta_in_id), ami.ami_delta.delta_id); *)
      H.add ami.ami_stmtid_to_delta_id sid (deltaAtStart.delta_id, ami.ami_delta.delta_id);
      ignore(Updatedebug.log [Updatedebug.BTAnalysis] "amStmtEnd: adding ids for sid: %d\n" sid);
      H.add ami.ami_stmtid_to_vc_ids sid 
(*         (ami.ami_vc_fwd.delta_id,  *)
        (vcFwdAtStart.delta_id, 
         ami.ami_vc_eps.delta_id, 
         ami.ami_vc_bck.delta_id,
         (match ami.ami_checkin_eps with 
              None -> None
            | Some d -> Some d.delta_id)
        );
      H.add ami.ami_stmtid_to_fun_name sid ami.ami_current_fd.svar.vname;
      return () ami

  (* return the hash containing all the intern names *)       
  let amGetInternVarHash = fun ami -> return (ami.ami_intern_var) ami      

  (* add (crt_delta, target_jmp_env) pair to the longjmp hash *)
  let amAddLongjmp vi = fun ami -> 
    let p = {phi_delta  = ami.ami_delta;
             phi_vc_fwd = ami.ami_vc_fwd;
             phi_vc_eps = ami.ami_vc_eps;
             phi_vc_bck = ami.ami_vc_bck} 
    in
    H.add ami.ami_longjmp_hash p (vi, !(ami.ami_mu));
    return () ami

  let amGetLongjmpHash = fun ami -> return ami.ami_longjmp_hash ami     

(*   let amResetSweepHash = fun ami -> *)
(*     H.clear ami.ami_sweep_hash; *)
(*     return () ami *)

(*   let amGetSweepedSymbols = fun ami -> *)
(*     return (Commonutil.hashtbl_keys ami.ami_sweep_hash) ami *)

  let amPrintSweepHash e ami = 
    let res = Commonutil.hashtbl_keys ami.ami_sweep_hash in
    ignore(Updatedebug.log [Updatedebug.BTAnalysis] "amSweep (%a):  {%s}\n" d_exp e
    (List.fold_left (fun s n -> s ^ " " ^ n) "" res))

  let amSweepLval accessKind l = fun ami ->
    (* always do the syntactic sweep *)
    ignore (Cil.visitCilExpr ami.ami_sweeper_enum_syntactic (Lval l));
    ignore (Cil.visitCilExpr ami.ami_sweeper_gvar_syntactic (Lval l));

    (if Updateargs.args.Updateargs.arg_pointstoEnabled then 
       begin
         ignore (Cil.visitCilLval ami.ami_sweeper_gvar_ptranal l);
(*          amPrintSweepHash (Lval l) ami *)
       end);

    let res = Commonutil.hashtbl_keys ami.ami_sweep_hash in
    H.clear ami.ami_sweep_hash;  
    return (accessKind,res) ami

  let amSweepExpr accessKind e = fun ami ->
    (* always do the syntactic sweep *)
    ignore (Cil.visitCilExpr ami.ami_sweeper_enum_syntactic e);
    ignore (Cil.visitCilExpr ami.ami_sweeper_gvar_syntactic e);

    (if Updateargs.args.Updateargs.arg_pointstoEnabled then 
       begin
         ignore (Cil.visitCilExpr ami.ami_sweeper_gvar_ptranal e);
(*          amPrintSweepHash e ami *)
       end);

    let res = Commonutil.hashtbl_keys ami.ami_sweep_hash in
    H.clear ami.ami_sweep_hash;  
    return (accessKind,res) ami

end


(* Create a set of monadic operations for this monad *)
module AMOps = Monad.MonadOps(AnalysisMonad)


(* ---- End of analysis monad ---- *)
(*============================================================================*)





(* Common ops for both constraint generators *)

module BaseConstraintGenerator(M:AnalysisMonadSig) = 
struct

  let (>>=) = M.(>>=)
  let return = M.return

  let amTRConcreteUse type_name dt why = 

                        M.amMakeDeltaConst type_name dt
         >>= fun ct ->  M.amAddDCSubsetOfCurrent ct        (why ^ "< delta")
         >>= fun () ->  M.amAddDESubsetOfCurrent ct        (why ^ "< delta_eigen")
         >>= fun () ->  M.amAddVcSubsetOfCurrent ct VKEps  (why ^ "< vc_eps")
         >>= fun () ->  M.amAddVcSubsetOfCurrent ct VKFwd  (why ^ "< vc_fwd")
         >>= fun () ->  M.amAddVcSubsetOfCurrent ct VKBck  (why ^ "< vc_bck")
         >>= fun () ->  M.amAddVcSubsetOfCurrent ct VKChk  (why ^ "< vc_checkin")

  let amTRVCInclude eps prePhi why = 

                        M.amAddVcSubsetOfCurrent eps VKEps  (why ^ "< vc_eps")
         >>= fun () ->  M.amAddVcSubsetOfCurrent eps VKFwd  (why ^ "< vc_fwd")
         >>= fun () ->  M.amAddVcSubsetOfCurrent eps VKChk  (why ^ "< vc_checkin")
         >>= fun () ->  M.amAddVCSubset eps prePhi.phi_vc_bck  (why ^ "< vc_bck")
           

  let amTRDeltaSubset d why = M.amAddDCSubsetOfCurrent d why


  let amTRUpdatePoint suspConstr isSpeculative why =

         (
           match isSpeculative with
               Some sid ->
                 (* Associate a speculative update point with the current mu *)
                 (* We don't constrain mu to be U as this is a speculative udpdate *)
                 M.amStoreSpeculativeUpdateMu sid

             | None ->
                 (* U < current_mu  *)
                 M.amAddMCLessThanCurrent (MuConst true)
         )
             
         >>= fun () ->  M.amFreshDeltaVar                         (* d fresh *)
         >>= fun d  ->  M.amAddDCSubsetOfCurrent d                (* d <= current_d *) (why ^ ": update: d <= current_d")
         >>= fun () ->  M.amSetDelta d                            (* current_d := d *)

         
         >>= fun () ->  suspConstr (* Add constraints to do with suspect aliases *)

(*          (\* --- VC --- *\) *)
         >>= fun () ->  M.amGetVcFwd
         >>= fun f  ->  M.amGetVcBck
         >>= fun b  ->  M.amFreshDeltaVar
         >>= fun nf ->  M.amSetVcFwd nf
         >>= fun () ->  M.amFreshDeltaVar
         >>= fun nb ->  M.amSetVcBck nb
         >>= fun () ->  M.amAddVCSubset f nf (why ^ ": current_fwd < fwd")
         >>= fun () ->  M.amAddVCSubset nb b (why ^ ": bck < current_bck")
           

  let amTRGetPhi =  fun ami ->
    let p = {phi_delta  = ami.ami_delta;
             phi_vc_fwd = ami.ami_vc_fwd;
             phi_vc_eps = ami.ami_vc_eps;
             phi_vc_bck = ami.ami_vc_bck} 
    in
    return p ami

  let amTRSetPhi p = 
                      M.amSetDelta p.phi_delta 
    >>= fun ()     -> M.amSetVcFwd p.phi_vc_fwd
    >>= fun ()     -> M.amSetVcEps p.phi_vc_eps
    >>= fun ()     -> M.amSetVcBck p.phi_vc_bck



  let amTRFreshPhiInternal eps =
                      M.amFreshDeltaVar
    >>= fun d      -> M.amFreshDeltaVar
    >>= fun fwd    -> M.amFreshDeltaVar
    >>= fun bck    -> fun ami ->
      let p = {phi_delta  = d;
               phi_vc_fwd = fwd;
               phi_vc_eps = eps;
               phi_vc_bck = bck} 
      in
      return p ami

  let amTRFreshPhi = 
                      amTRGetPhi
    >>= fun crtPhi -> amTRFreshPhiInternal crtPhi.phi_vc_eps 


  let amTRFreshPhiFreshEps = 
                      M.amFreshDeltaVar
    >>= fun eps    -> amTRFreshPhiInternal eps


  let amTRFreshPhiSpecifiedEps phi =
                      amTRFreshPhiInternal phi.phi_vc_eps

     



  let amTRPhiFlow p1 p2 why =  
                      M.amAddDCSubset p2.phi_delta  p1.phi_delta  why
    >>= fun ()     -> M.amAddVCSubset p1.phi_vc_fwd p2.phi_vc_fwd why
    >>= fun ()     -> M.amAddVCSubset p2.phi_vc_bck p1.phi_vc_bck why

  let amTRPhiFlowFwdOnly p1 p2 why = M.amAddVCSubset p1.phi_vc_fwd p2.phi_vc_fwd why

  let amTRPhiFlowBckOnly p1 p2 why = M.amAddVCSubset p1.phi_vc_bck p2.phi_vc_bck why

  let amTRDeltaFlowOnly p1 p2 why = M.amAddDCSubset p2.phi_delta p1.phi_delta  why

  let amTRDeltaCond mu d why = 
                               M.amGetDelta
            >>= fun cd  ->     M.amAddDCCond mu cd d why       



  let amTRStartFun effect f =
                            M.amFreshDeltaVar
          >>= fun delta  -> M.amSetDelta delta
          >>= fun ()     -> M.amSetEigenDelta effect.effect_delta_eigen
          >>= fun ()     -> M.amAddDCSubset delta effect.effect_delta_in "startFun : delta < delta_in"
          >>= fun ()     -> M.amSetMu effect.effect_mu
 
          (* --- VC --- *)

          >>= fun ()     -> M.amFreshDeltaVar
          >>= fun fwd    -> M.amAddVCSubset effect.effect_vc_fwd fwd "startFun: fun_fwd < fwd"
          >>= fun ()     -> M.amSetVcFwd fwd
          >>= fun ()     -> M.amSetVcEps effect.effect_vc_eps
          >>= fun ()     -> M.amFreshDeltaVar
          >>= fun bck    -> M.amSetVcBck bck
          >>= fun ()     -> M.amSetCheckinEps None

          (* Function types, i.e. fun foo : t1 ---- {foo} -----> t2*)
          (* TODO : This is redundant if we do the points-to analysis, since the function name is included in the funexp, so 
             we get the access when we call the function *)  
          >>= fun _                    -> M.amMakeDeltaConst (Updateconfig.accessPack Updateconfig.SRead f.svar.vname) DckFun
          >>= fun funNameSingletonSet  (* -> M.amAddDCSubset funNameSingletonSet effect.effect_labels "analyzeFunction: fun_name < effect.labels" *)
(*           >>= fun _                     *)-> M.amAddVCSubset funNameSingletonSet effect.effect_vc_eps       "analyzeFunction: fun_name < vc_eps"

  let amTREndFun effect =
 
         (* add equality constraint to force output delta on function to be the same as
            the delta we ended up with *)

                           M.amGetDelta
         >>= fun d      -> ignore(Updatedebug.log [Updatedebug.BTAnalysis] "analyzeFunction(%a): %d\n" d_loc d.delta_loc d.delta_id);
                           M.amAddDCSubsetOfCurrent effect.effect_delta_out  "analyzeFunction: delta_out < d"
         (* --- VC --- *)
         >>= fun ()     -> M.amAddVcSubsetOfCurrent effect.effect_vc_bck  VKBck "analyzeFunction: effect_bck < vc_bck"

end


(* CLASSIC CONSTRAINT GENERATOR *)

(* Our old constraint generator, but now decoupled from
   source code traversal *)

module ClassicConstraintGenerator(AMS:AnalysisMonadSig) =
struct

module M = BaseConstraintGenerator(AMS)

  let (>>=) = AMS.(>>=)

  let amTRConcreteUse type_name dt why =
    M.amTRConcreteUse type_name dt why

  let amTRVCInclude eps prePhi why = 
    M.amTRVCInclude eps prePhi why 

  let amTRDeltaSubset d why =
    M.amTRDeltaSubset d why

  let amTRDeltaCond mu d why =       
    M.amTRDeltaCond mu d why       

  let amTRUpdatePoint suspConstr isSpeculative why =
    M.amTRUpdatePoint suspConstr isSpeculative why

  let amTRStartFun effect f =
    M.amTRStartFun effect f

  let amTREndFun effect =
    M.amTREndFun effect

  let amTRGetPhi =  
    M.amTRGetPhi

  let amTRSetPhi p = 
    M.amTRSetPhi p
     
  let amTRFreshPhi = 
    M.amTRFreshPhi

  let amTRFreshPhiFreshEps = 
    M.amTRFreshPhiFreshEps

  let amTRFreshPhiSpecifiedEps p = 
    M.amTRFreshPhiSpecifiedEps p

  let amTRPhiFlow phi1 phi2 why =  
    M.amTRPhiFlow phi1 phi2 why

  let amTRPhiFlowFwdOnly p1 p2 why = 
    M.amTRPhiFlowFwdOnly p1 p2 why  

  let amTRPhiFlowBckOnly p1 p2 why =
    M.amTRPhiFlowBckOnly p1 p2 why

  let amTRDeltaFlowOnly p1 p2 why =
    M.amTRDeltaFlowOnly p1 p2 why



end


(* SMART CONSTRAINT GENERATOR *)
(* New, nifty, modular constraint generator *)


module SmartConstraintGenerator(AMS:AnalysisMonadSig) =
struct

module M = BaseConstraintGenerator(AMS)

  let (>>=) = AMS.(>>=)
  let return = AMS.return

  let amTRConcreteUse type_name dt why =
    M.amTRConcreteUse type_name dt why

  let amTRVCInclude eps prePhi why = 
    M.amTRVCInclude eps prePhi why 

  let amTRDeltaSubset d why =
    M.amTRDeltaSubset d why

  let amTRDeltaCond mu d why =       
    M.amTRDeltaCond mu d why       

  let amTRUpdatePoint suspConstr isSpeculative why =
    M.amTRUpdatePoint suspConstr isSpeculative why

  let amTRStartFun effect f =
    M.amTRStartFun effect f

  let amTREndFun effect =
    M.amTREndFun effect

  let amTRGetPhi =  
    M.amTRGetPhi

  let amTRSetPhi p = 
    M.amTRSetPhi p
     
  let amTRFreshPhi = 
    M.amTRFreshPhi

  let amTRFreshPhiFreshEps = 
    M.amTRFreshPhiFreshEps

  let amTRFreshPhiSpecifiedEps p = 
    M.amTRFreshPhiSpecifiedEps p

  let amTRPhiFlow phi1 phi2 why =  
    M.amTRPhiFlow phi1 phi2 why

  let amTRPhiFlowFwdOnly p1 p2 why = 
    M.amTRPhiFlowFwdOnly p1 p2 why  

  let amTRPhiFlowBckOnly p1 p2 why =
    M.amTRPhiFlowBckOnly p1 p2 why

  let amTRDeltaFlowOnly p1 p2 why =
    M.amTRDeltaFlowOnly p1 p2 why

end


(* Now that we've decoupled constraint generation from source code 
   traversal, we can switch between the two constraint generation modules
*)
module CG = SmartConstraintGenerator(AnalysisMonad)





(* ---- A specific fresh effect generator for the update analysis ---- *)

(* This module contains all of the machinery that deals with actually putting 
   effects into and out of the type grammar
*)
module AnalysisEffectManipulator =  
struct

  module M = AnalysisMonad


  (* ---- Some helper functions ---- *)

  (* Bind our monad operations locally *)
  let (>>=) = AnalysisMonad.(>>=) 
  let return = AnalysisMonad.return

  let attrIsEffect a = match a with
      Attr(aname, _) when aname = Updateconfig.effectAttrName -> true
    | _ -> false

  let getEffects = List.filter attrIsEffect
  let containsEffects = List.exists attrIsEffect

  (* Generate some fresh constraint variables and put them into the given attribute list 
     If the attr list already has effects, do nothing *)
  let genFreshFunConstraints al : (Cil.attribute list) M.m = 
    if (containsEffects al)
    then 
      (* don't insert the effects again *)
      return al
    else
      begin
   (* generate effect *)
                       AnalysisMonad.amFreshMuVar 
        >>= fun mu  -> AnalysisMonad.amFreshDeltaVar
        >>= fun d   -> AnalysisMonad.amFreshDeltaVar
        >>= fun l   -> AnalysisMonad.amFreshDeltaVar

        (* --- VC --- *)
        >>= fun vc_f -> AnalysisMonad.amFreshDeltaVar
        >>= fun vc_e -> AnalysisMonad.amFreshDeltaVar
        >>= fun vc_b -> AnalysisMonad.amFreshDeltaVar

        >>= fun d' -> AnalysisMonad.amFreshDeltaVar
        >>= fun ei -> AnalysisMonad.amGetNextFunEffectID
        >>= fun key -> AnalysisMonad.amAddFunEffectEntry key {
           effect_delta_in = d; 
           effect_delta_out = d';
           effect_delta_eigen = ei;
(*            effect_labels = l; *)
           effect_mu = mu;

           (* --- VC --- *)
           effect_vc_fwd = vc_f;
           effect_vc_eps = vc_e;
           effect_vc_bck = vc_b;
        }
        >>= fun _   -> 
          return (Cil.addAttribute (Attr(Updateconfig.effectAttrName,[AInt key])) al)
      end


  (* Extract the effect from an attribte list if it exists *)
  let extractFunConstraints al : effect M.m = 
    let getKey () =  
      match Cil.filterAttributes Updateconfig.effectAttrName al with 
          [Attr(name,[AInt key])] when name = Updateconfig.effectAttrName ->
            return key
              
        | [] -> 
            AnalysisMonad.getLoc >>=
              fun loc ->  Updatedebug.abort (Updatedebug.bug 
                          "(%a) extractFunConstraints: error extracting function constraints (no effects in this TFun)" 
                          d_loc loc)
        | _ -> 
            AnalysisMonad.getLoc >>=
              fun loc ->  Updatedebug.abort (Updatedebug.bug 
                          "(%a) extractFunConstraints: error extracting function constraints (more than one effect in this TFun)" 
                          d_loc loc)
    in   
      getKey ()
      >>= fun key -> AnalysisMonad.amGetFunEffectEntry key



  (* check whether the type contains effects *)
  let typeContainsEffects (typ : Cil.typ) : bool =
    let compHash = H.create 4099 in
    let rec typeContainsEffects_inner (typ_inner : Cil.typ) : bool =
      match typ_inner with 
          TVoid _ | TInt _  | TFloat _ | TEnum _ | TBuiltin_va_list _ -> false
            
        | TPtr (t, _) -> typeContainsEffects_inner t
        | TArray(t, _, _) -> typeContainsEffects_inner t
            
        | TFun(tret, targs, isva, al) -> containsEffects al
            
        | TNamed(ti, _) -> typeContainsEffects_inner ti.ttype
            
        | TComp(ci,al) ->  
            if H.mem compHash ci.cname 
            then 
              false
            else 
              begin
                H.add compHash ci.cname true; 
                let field_has_effects (fi : Cil.fieldinfo) : bool = typeContainsEffects_inner fi.ftype in
                  List.exists field_has_effects ci.cfields
              end
    in
      typeContainsEffects_inner typ


  (*----------------------------------------------------------------------------*)
  (* ---- Unused utility functions ---- *)

  (* add effect attr to current type ; die if type already has effects *)
  let addEffectToType (t:Cil.typ) (attr:Cil.attribute) : Cil.typ = 
    (* add recursively behind pointers *)
    let addEffectToType' (t':Cil.typ) (attr:Cil.attribute) : Cil.typ = 
      let ta = Cil.typeAttrs t' in
        if containsEffects ta
        then
          t'
            (*Updatedebug.abort (Updatedebug.bug "addEffectToType : tried to add 2nd effect to type %a\n" d_plaintype t')*)
        else
          begin
            Cil.typeAddAttributes [attr] t'
          end
    in
      match t with
          TPtr(t',a) -> TPtr(addEffectToType' t' attr,a)
        | _ -> addEffectToType' t attr


  (* ---- End of unused utility funcitons ---- *)
  (*----------------------------------------------------------------------------*)



  (* ---- Required functions so we are an instance of Effectinserter.EffectCreator ---- *)
  
  let insertEffectInOuterConstructor (typ:Cil.typ) : Cil.typ M.m =
    match typ with 
        TFun(tret, targs, isva, al) -> 
          genFreshFunConstraints al >>= fun al' -> return (TFun(tret, targs, isva, al'))

      | _ -> return typ
  
end




(* Create a fresh effect inserter for the update analysis *)
module AnalysisGenFreshEffects = Effectinserter.EffectInserter(AnalysisEffectManipulator)(AnalysisMonad)

 


(* ---- End of effect insertion ---- *)
(*============================================================================*)










(* Locally bind the analysis monad basic operations *)
let (>>=) = AnalysisMonad.(>>=)
let return = AnalysisMonad.return

(* shortcut to analysis monad module *)
module AM = AnalysisMonad







         



(* given a type, do a user-specified action on all fun effects in this type *)
let doActionOnFunEffect (typ : Cil.typ) (action : effect -> unit AM.m ) : unit AM.m =
   let compHash = H.create 4099 in
   let rec doActionOnFunEffect' (typ' : Cil.typ) : unit AM.m = 
      match typ' with 
         TVoid _
       | TInt _
       | TFloat _ -> return ()
       | TPtr (t,attr) -> doActionOnFunEffect' t
            
       | TArray(t,e_opt,attr) -> doActionOnFunEffect' t
            
       | TFun(tret,targs,isva,al) ->
            let args_map (s,t,a) = doActionOnFunEffect' t
            in 
                 begin match targs with 
                         Some l -> AMOps.iter args_map l 
                       | None   -> return ()
                      end
            >>= fun ()             -> doActionOnFunEffect' tret
            >>= fun ()             -> AnalysisEffectManipulator.extractFunConstraints al
            >>= fun eff            -> action eff

       | TNamed _ -> return ()
       
       | TComp(ci,al) ->  
           if H.mem compHash ci.cname then return ()
             else
                   begin
              H.add compHash ci.cname true; 
              let field_map (fi : Cil.fieldinfo) : unit AM.m = 
                                    doActionOnFunEffect' fi.ftype
                   in
              AMOps.iter field_map ci.cfields 
              end

       | TEnum _
       | TBuiltin_va_list _ -> return ()
   in 
   doActionOnFunEffect' typ;;

(* Add constraints to force all effects on function types to be N;Top;Top (the void fun ptr effect in the monad) *)
let pacifyFunTypeEffects (typ : Cil.typ) : unit AM.m =
   let action = 
                fun eff            -> AM.amAddMC eff.effect_mu (MuConst false)
            >>= fun ()             -> AM.getLoc
            >>= fun loc            -> return (ignore (Updatedebug.log [Updatedebug.BTAnalysis] "(%a) Pacifying function\n" d_loc loc ))

(*
            >>= fun eff            -> AM.amGetVoidPtrFunEffectId
            >>= fun voidfun_eff_id -> AM.amGetFunEffectEntry voidfun_eff_id
            
            (* Add equality conctraint Top = delta_out *)
            >>= fun voidfun_eff    -> AM.amAddDCSubset eff.effect_delta_out voidfun_eff.effect_delta_out   
            >>= fun ()             -> AM.amAddDCSubset voidfun_eff.effect_delta_out eff.effect_delta_out
            
            (* Add equality conctraint Top = delta_in *)
            >>= fun ()             -> AM.amAddDCSubset eff.effect_delta_in voidfun_eff.effect_delta_in
            >>= fun ()             -> AM.amAddDCSubset voidfun_eff.effect_delta_in eff.effect_delta_in
            
            >>= fun ()             -> AM.amAddMC eff.effect_mu voidfun_eff.effect_mu
*)
   in
      doActionOnFunEffect typ action;;


(*============================================================================*)
(* Arguments to extern functions *)


(* All arguments in exp_arg_list which have function type are forced to have their output
   effect a superset of the external function output effect. This ensures proper effect propagation
   by assuming once we pass a function pointer 'f' to an external function, the external function
   will call 'f'
*)
let constrainArgumentsToExternalFunctions externFunEffect exp_arg_list = 

   let action = fun argumentEffect  -> 
     AM.amAddDCSubset externFunEffect.effect_delta_out argumentEffect.effect_delta_out "constrainArgumentsToExternalFunctions"
   in
   let argument_types = List.map Updatecilutil.properTypeOf exp_arg_list in
   let constrain_one_argument t = doActionOnFunEffect t action in
   AMOps.iter constrain_one_argument argument_types


(* set res to true is an extern var was found *)
class externVarMarker internVarHash (res : bool ref) : cilVisitor = object(self)
  inherit nopCilVisitor

 (* we need to test both Extern and internVarHash to avoid
    - local vars
    - extern decls (definitions from the original program) in patch mode
 *)    
  method vvrbl v = 
     (if v.vstorage = Extern && not (H.mem internVarHash v.vname)
      then
         res := true;
     );
     DoChildren

end

(* given an expresion, return true if it contains an external var *)
let isExternalFunCall (funExpr : Cil.exp) (internVarHash : (string, unit) H.t) = 
   let res = ref false in
   let visitor = new externVarMarker internVarHash res in
   ignore(Cil.visitCilExpr visitor funExpr); (* read-only visitor, so we don't bother with the result *)   
   !res

(* End arguments to extern functions *)
(*============================================================================*)

(*         turn a T[] into a T              *)
let discardArrayType t = match t with
   TArray(ta,_,_) -> ta
 | _ -> t 


module AnalysisSubtypeActions = 
struct
  module M = AnalysisMonad
  module MOps = Monad.MonadOps(AnalysisMonad)

  (* Called when SubtypeCmp performs a standard subtype step *)
  let subtype (t1: Cil.typ) (t2: Cil.typ) : unit M.m =
    match t1,t2 with 
      | TFun(t1',args_opt1,isVA1,attr1), TFun(t2',args_opt2,isVA2,attr2) ->  
                                    AnalysisEffectManipulator.extractFunConstraints attr1 
          >>= fun effect1 ->        AnalysisEffectManipulator.extractFunConstraints attr2 
          >>= fun effect2 ->        M.amAddDCSubset effect2.effect_delta_out effect1.effect_delta_out "subtype-TFun: delta"
          >>= fun ()      ->        

          (* Iulian : why was it the case that t_1 <: t_2 => mu_2 <= mu_1 ?
             shouldn't it be mu_1 <= mu_2 ?
             
             I mean, if t_2 is inferred to be N, t_1 should be N and 
             if t_2 is unconstrained, then t_1 is unconstrained  
             
          *)               
          (*  OLD:           >>= fun _ ->  AM.amAddMC effect2.effect_mu effect1.effect_mu *)
          M.amAddMC effect1.effect_mu effect2.effect_mu
   
          (* Function types:   

             t1 ------ {S1} ----> t1' <:  t2 ------ {S2} -----> t2' 

             if S1 \subseteq S2  
          *)     
(*           >>= fun _ ->  AM.amAddDCSubset effect1.effect_labels effect2.effect_labels "subtype-TFun: labels" *)

          (* --- VC --- *)
          >>= fun () ->  AM.amAddVCSubset effect1.effect_vc_eps  effect2.effect_vc_eps "subtype-TFun: vc_eps"
          >>= fun () ->  AM.amAddVCSubset effect2.effect_vc_fwd  effect1.effect_vc_fwd "subtype-TFun: vc_fwd"
          >>= fun () ->  AM.amAddVCSubset effect2.effect_vc_bck  effect1.effect_vc_bck "subtype-TFun: vc_bck"

      | _ -> return ()



  (* Called when a physical subtype step is performed, i.e. a pointer
     to a record is a subtype of a pointer to the records first field
   *)
  let physicalSubtype (t1: Cil.typ) (t2: Cil.typ) : unit M.m =
    return ()
    
  (* Called when one of the types is external *)
  let externalSubtype (t1: Cil.typ) (t2: Cil.typ) : unit M.m =
    pacifyFunTypeEffects t1 >>= fun () -> 
    pacifyFunTypeEffects t2 >>= fun () ->
    M.getLoc >>= fun loc ->
    MOps.lift (fun () -> ignore (Updatedebug.log [Updatedebug.BTAlways] 
                                "(%a) Analysis:SubtypeConstraints: One type is external %s and %s\n" 
                                d_loc loc (Updatedebug.ppType t1) (Updatedebug.ppType t2)
                             )
           ) (return ())

  (* Called when trying to determine a subtype between comp pointer type
     and another type where the types neither lead to a valid subtype or to a valid 
     physical subtype *)
  let ptrIncompatible (t1: Cil.typ) (t2: Cil.typ) (msg: string) : unit M.m =
    let areWrappers = match t1, t2 with
        TPtr(TComp (ci1,_),_), TPtr(TComp (ci2,_),_) ->
          (ci1.cname = Updateconfig.makeConcreteCompName ci2) ||
          (ci2.cname = Updateconfig.makeConcreteCompName ci1) 
      | _ -> false
(* TODO 7/8/7 : is above checkOK ? *)
(*     if (Wrappers.isWrapperType t1 && Typeequality.uAreTypesEqualNoAttrs (Wrappers.getDataUnionType t1) t2) || *)
(*        (Wrappers.isWrapperType t2 && Typeequality.uAreTypesEqualNoAttrs (Wrappers.getDataUnionType t2) t1)  *)
    in
      if areWrappers
      then (* one is wrapper, one is wrappee; leave them alone *)
        return ()
      else (* kill'em *) 
        pacifyFunTypeEffects t1 >>= fun () -> 
        pacifyFunTypeEffects t2 >>= fun () ->
          M.getLoc >>= fun loc ->
            MOps.lift (fun () -> ignore (Updatedebug.log [Updatedebug.BTAlways] 
                                           "(%a) Analysis:SubtypeConstraints: Incompatible pointers %s and  %s\n" 
                                           d_loc loc (Updatedebug.ppType t1) (Updatedebug.ppType t2)
                             )
                      ) (return ())


  (* Called when C would not accept the subtype *)
  let invalidCSubtype (t1: Cil.typ) (t2: Cil.typ) (msg: string) : unit M.m =
    pacifyFunTypeEffects t1 >>= fun () -> 
    pacifyFunTypeEffects t2 >>= fun () ->
    M.getLoc >>= fun loc ->
    MOps.lift (fun () -> ignore (Updatedebug.log [Updatedebug.BTAlways] 
                                "(%a) Analysis:SubtypeConstraints: invalid C subtype %s and %s\n" 
                                d_loc loc (Updatedebug.ppType t1) (Updatedebug.ppType t2)
                             )
           ) (return ())

  let toVoidPtr t1 t2 = pacifyFunTypeEffects t1
  let fromVoidPtr t1 t2 = pacifyFunTypeEffects t2

end


module AnalysisSubtypeConstraints = Subtype.SubtypeCmp(AnalysisSubtypeActions)(AnalysisMonad)

let subTypeConstraints t1 t2 = AnalysisSubtypeConstraints.subtypeTraverse t1 t2


(* For introducing constraints on update points, called from analysis.ml *)

let suspConstr =
  AM.getCurrentFun >>= fun fd ->
  AM.amGetUpdateConstraint fd.svar.vname >>= fun constr ->
    begin match constr with
      Some (lps,locsG) -> 
   ignore (Updatedebug.log [Updatedebug.BTSuspect] 
        "introducing update constraint: func(%s) ===> %d <= curr, %d <= curr\n"
        fd.svar.vname lps.delta_id locsG.delta_id);
   AM.amAddDCSubsetOfCurrent lps  "suspConstr: L <= current_d" >>= fun ()   -> (*  *)
   AM.amAddDCSubsetOfCurrent locsG "suspConstr: locs(G) <= current_d"    (* locs(G) <= current_d *)
    | None -> 
   ignore (Updatedebug.log [Updatedebug.BTSuspect]
        "missing update constraint: %s\n" fd.svar.vname);
   return ();
    end

(*============================================================================*)
(* ---- Traversal of syntax to gather constraints ---- *)


let flushSweepHash (accessedSyms : (Updateconfig.access_t * string list) list) : unit AM.m = 
(*                                          AM.amGetSweepedSymbols *)
(*                       >>= fun syms   ->  AM.amResetSweepHash *)
(*                       >>= fun ()     ->   *)
  let flatSyms = 
    let flatten_fold acc (access, syms) =
      let adornedSyms = List.rev_map (Updateconfig.accessPack access) syms in
      acc @ adornedSyms
    in
    List.fold_left flatten_fold [] accessedSyms
  in 
  let addOneRef_iter refName =
    CG.amTRConcreteUse refName DckGvar "sweepExpr"
  in
  AMOps.iter addOneRef_iter flatSyms

let sweepLval (accessKind: Updateconfig.access_t) (l : Cil.lval) : (Updateconfig.access_t * string list) AM.m = AM.amSweepLval accessKind l
;;

let sweepExpr (accessKind: Updateconfig.access_t) (e : Cil.exp) : (Updateconfig.access_t * string list) AM.m = AM.amSweepExpr accessKind e
;;


(* traverse instr/stmts to find out what gvars and locations are used *)
let sweepInstrEffects (instr : Cil.instr) : unit AM.m = 
  match instr with 
      Set (lval,exp,loc) -> 
                                         sweepLval Updateconfig.SWrite lval
                      >>= fun alv    ->  sweepExpr Updateconfig.SRead exp
                      >>= fun aex    ->  flushSweepHash [alv; aex]

    | Call (ret,funexp,args,_) -> 
                                                 (* if function is extern 
                                                    - mark its args readwrite ( else read) 
                                                    - don't bother w/the name, it can never be updated *)
 
                                                sweepExpr Updateconfig.SRead funexp
                      >>= fun afex           -> AM.amGetInternVarHash 
                      >>= fun internVarHash  -> (match isExternalFunCall funexp internVarHash with
                                                     true -> return (Updateconfig.SReadWrite, [])
                                                   | false -> return (Updateconfig.SRead, [afex])
                                                )

                      >>= fun (argAccess, afex')  ->  AMOps.map (sweepExpr argAccess) args 
                      >>= fun aargs          -> (match ret with 
                                                     Some lval -> sweepLval Updateconfig.SWrite lval
                                                   | None -> return (Updateconfig.SWrite,[])
                                                )
                      >>= fun aret           ->  flushSweepHash (aret::aargs @ afex')
         
    | Asm (_,_,outputs,inputs,_,_)    -> 
                                        let _, _, outputLvals = Commonutil.splitTriples outputs
                                        and _, _, inputExps   = Commonutil.splitTriples inputs in

                                         AMOps.map (sweepLval Updateconfig.SReadWrite) outputLvals
                      >>= fun aouts   -> AMOps.map (sweepExpr Updateconfig.SReadWrite) inputExps
                      >>= fun ains    -> flushSweepHash (ains @ aouts)
;;

let sweepStmtEffects (stmt : Cil.stmt)  : unit AM.m = 
  match stmt.skind with 

      Return (Some exp,_) ->             
                                         sweepExpr Updateconfig.SRead exp
                      >>= fun aex    ->  flushSweepHash [aex]

    | If (exp,_,_,_)      ->             
                                         sweepExpr Updateconfig.SRead exp
                      >>= fun aex    ->  flushSweepHash [aex]

    | _                   ->             return ()
;;

(*---- end sweeping -----------------------------------------------*)

let rec analyzeLval ((lhost,offset) : Cil.lval) : unit AM.m = 
   match lhost with 
      Mem (e) -> analyzeExp e
    | _       -> return ()



and analyzeExp (exp : Cil.exp) : unit AM.m = 
   (*amGetNumDeltaConstraints >>= fun num_mu ->*)
   let id = return () in
(*    let () = ignore (Updatedebug.log [Updatedebug.BTAlways] "analyzeExp %a:%s\n " d_plainexp exp (Updatedebug.ppType (Updatecilutil.properTypeOf exp))) in *)
   match exp with
        Const _ -> id
        
      | Lval lval -> analyzeLval lval
         
      | SizeOf typ -> id   
                           
      | SizeOfE e -> analyzeExp e
         
      | SizeOfStr str -> id

      | AlignOf typ ->
         AM.getLoc >>=
         fun loc -> ignore(Updatedebug.log [Updatedebug.BTAlways] "WARNING : analyzeExp : (%a) can't deal with AlignOf(%a)!"
                                      d_loc loc d_plaintype typ);
                    id
      
      | AlignOfE exp ->
         AM.getLoc >>= 
         fun loc -> ignore(Updatedebug.log [Updatedebug.BTAlways] "WARNING : analyzeExp : (%a) can't deal with AlignOfE(%a)!"
                                       d_loc loc d_plainexp exp); 
                    id             
                    
      | UnOp (unop, e, typ) -> analyzeExp e

      (* FIXME: C doesn't say which order these get evaluated in!  We need to be more conservative *)
      | BinOp (binop, e1, e2, typ) -> analyzeExp e1 >>= fun _ -> analyzeExp e2
         
      | CastE (typ, e) -> 
                          return (Updatecilutil.properTypeOf e)
           >>= fun e_t -> subTypeConstraints e_t typ
           >>= fun ()  -> analyzeExp e

      | AddrOf lval 
      | StartOf lval -> analyzeLval lval


and analyzeCall (isInTX : bool) (instr : Cil.instr) : unit AM.m = 
(*   let () = ignore (Updatedebug.log [Updatedebug.BTAnalysis] "analyzeCall (isInTX:%b, %a)  called\n" isInTX d_instr instr) in *)

   AM.amGetMallocFunHash >>= fun mallocFunHash ->
   match instr with 

    (* we ignore the arguments to these C-magic-calls *)
    |  Call(_, Lval(Var vi, NoOffset), _, _) when  Commoncilutil.isBuiltin vi.vname -> 
        return ()

    (* When con *)
    | Call(Some(l), Lval(Var vi, NoOffset), [e], _) 
      when  (Updateconfig.isConFunc vi.vname) && (Namedtypes.isConAbsFuncName vi.vname)  ->
         let t = Namedtypes.getAbstractTypeByConAbsName vi.vname in 
         let type_name = Updatecilutil.getAbstractTypeName t  in 
                          analyzeExp e
            >>= fun _  -> subTypeConstraints (Updatecilutil.properTypeOf e) (Updatecilutil.properTypeOfLval l)
            >>= fun _  -> CG.amTRConcreteUse type_name DckType "analyzeCall: con"

    (* when abs *)
    | Call(None, Lval(Var vi, NoOffset), [e; Lval(l)], _) 
      when  (Updateconfig.isAbsFunc vi.vname) && (Namedtypes.isConAbsFuncName vi.vname) ->
         let t = Namedtypes.getAbstractTypeByConAbsName vi.vname in
                       analyzeExp e
         >>= fun () -> analyzeLval l
         (* push the effects of the thing being abs'ed onto the type of the abstract type *)
         >>= fun () -> subTypeConstraints (Updatecilutil.properTypeOf e) t
    
    
(*     (\* A speculative update function *\) *)
(*     | Call(_, Lval(Var vi, NoOffset), _, _) when  Updateconfig.isSpeculativeUpdateFunc vi.vname ->  *)
(*          (\* This is should have matched in analyzeStmt *\) *)
(*                        AM.getLoc *)
(*          >>= fun loc -> Updatedebug.abort (Updatedebug.bug  "(%a) analyzeCall: hit a speculative update point;it should have been hit in analyzeStmt not here.\n" d_loc loc) *)

    
    (* A real update function *)
    | Call(_, Lval(Var vi, NoOffset), _, _) when  Updateconfig.isUpdateFunc vi.vname ->
                       CG.amTRUpdatePoint suspConstr None "analyzeCall: update"

    (* Ignore without any guilt mallocs *)
    | Call(_, Lval(Var vi, NoOffset), _, _) when  Mallocpragma.isMallocFunc mallocFunHash vi.vname -> 
         return ()

    (* Handle constraints for setjmp/longjmp *)        
    | Call(_, Lval(Var vi, NoOffset), _, _) when  Setjmp.isSetjmpFunction vi.vname -> 
                        AM.getLoc
         >>= fun loc -> let () = ignore (Updatedebug.log [Updatedebug.BTAlways] "(%a) setjmp-style function '%s' called\n" d_loc loc vi.vname) in

         return ()

    | Call(_, Lval(Var vi, NoOffset), e :: _, _) when  Setjmp.isLongjumpFunction vi.vname -> 
                        AM.getLoc
         >>= fun loc -> ignore (Updatedebug.log [Updatedebug.BTAlways] "(%a) longjmp-style function '%s' called\n" d_loc loc vi.vname);
                        AM.amAddLongjmp (Commoncilutil.getBaseVarinfo e)

   
    (****** fork() ******)

         (* int pthread_create(pthread_t *restrict thread, *)
         (*               const pthread_attr_t *restrict attr, *)
         (*               void *(\*start_routine)(void*\), void *restrict arg); *)

      (* careful now; exp_func is the thread function (fourth arg to pthread_create) *)                    
    | Call (lval_result_opt, 
            Lval(Var vi, NoOffset), 
            [thread_id; attr; exp_func; arg], loc) 
        when Updateconfig.isForkFunction vi.vname -> (* TODO: deal with function pointers pointing to fork *)
         let (exp_func_ret_t, 
              exp_func_args_t, 
              isva)                 = Updatecilutil.getTypesInFunExp exp_func in
         let numFixedArgs           = List.length exp_func_args_t in 
(*          let numActualArgs          = List.length exp_arg_list in *)
         
         (* Make two different lists for fixed and variable arguments *)
         let ( fixedArgs,
               varArgs )   = Commonutil.list_split numFixedArgs [arg] in 

         (* Create a list of pairs of types (t1,t2) for the fixed arguments that we need to create
            subtype constraints between *)
         let fixed_args_src_target_types = List.combine (List.map Updatecilutil.properTypeOf fixedArgs) exp_func_args_t in
    
              
              (* Do standard function stuff *)
                      analyzeExp exp_func 
         >>= fun _ -> AMOps.iter analyzeExp [arg] 
         >>= fun _ -> 
                begin match lval_result_opt with 
                    None      -> return ()
                  | Some lval -> 
                       (* Analyse the lval and collect constraints on the assignment of the result to the lval *)
                                         analyzeLval lval
                       >>= fun ()     -> return (Updatecilutil.properTypeOfLval lval)
                       >>= fun lval_t -> subTypeConstraints exp_func_ret_t lval_t
                     end
         
         >>= fun _ -> AMOps.iter (fun (t1,t2) -> subTypeConstraints t1 t2) fixed_args_src_target_types 
         >>= fun _ -> return (Updatecilutil.properTypeOf exp_func)
         >>= fun exp_func_t -> 
                    begin match (Cil.unrollType exp_func_t) with 
                       TPtr(TFun(_,_,_,al),_) -> AnalysisEffectManipulator.extractFunConstraints al
            
                     | _              -> Updatedebug.abort (Updatedebug.bug
                                    "analyzeCall: fork: expected function type, but got:\n %a\n" d_plaintype exp_func_t)
                    end
          
         (* If the callee can do an update then we're unconstrained 
            If we're N, the callee must be N too 
            hence the constraint mu_callee <= mu_caller     *)

         >>= fun calleeEffect    -> AM.amAddMCLessThanCurrent calleeEffect.effect_mu


         (* If we're in a TX, make the callee U
            (unless overridden by the above) *)
           
         >>= fun ()              -> (if isInTX && 
                                       (Updateargs.args.Updateargs.arg_updatabilityPolicy = Updateargs.UPTx)
                                     then AM.amAddMC (MuConst true) calleeEffect.effect_mu
                                     else return ())


         >>= fun ()              -> CG.amTRGetPhi
         >>= fun prePhi          -> CG.amTRFreshPhi
         >>= fun postPhi         -> CG.amTRPhiFlow prePhi postPhi "analyzeCall: fork: prePhi --> postPhi"
(*          >>= fun ()              -> CG.amTRDeltaSubset calleeEffect.effect_labels "analyzeCall: labels" *)
         >>= fun ()              -> CG.amTRSetPhi postPhi

(* no effect propagation from parent to child  *)           
(*          >>= fun ()              -> CG.amTRDeltaCond calleeEffect.effect_mu calleeEffect.effect_delta_out "analyzeCall: d < delta_out" *)
        (*  >>= fun ()              -> return (phi_of_effect calleeEffect) *)
        (*  >>= fun calleePhi      -> CG.amTRPhiFlowFwdOnly prePhi  calleePhi "analyzeCall: fork: preFwd < calleeFwd" *)
(*          >>= fun ()              -> CG.amTRPhiFlowBckOnly postPhi calleePhi "analyzeCall: postBck < calleeBck" 

         >>= fun ()              -> CG.amTRVCInclude calleeEffect.effect_vc_eps prePhi "analyzeCall: fork: callee_eps"*)

    (* standard function call *)
    | Call (lval_result_opt, exp_func, exp_arg_list, loc) ->
         let (exp_func_ret_t, 
              exp_func_args_t, 
              isva)                 = Updatecilutil.getTypesInFunExp exp_func in
         let numFixedArgs           = List.length exp_func_args_t in 
(*          let numActualArgs          = List.length exp_arg_list in *)
         
         (* Make two different lists for fixed and variable arguments *)
         let ( fixedArgs,
               varArgs )   = Commonutil.list_split numFixedArgs exp_arg_list in 

         (* Create a list of pairs of types (t1,t2) for the fixed arguments that we need to create
            subtype constraints between *)
         let fixed_args_src_target_types = List.combine (List.map Updatecilutil.properTypeOf fixedArgs) exp_func_args_t in
    
              
              (* Do standard function stuff *)
         (* FIXME: is this the right evaluation order for C? *)
                      analyzeExp exp_func 
         >>= fun _ -> AMOps.iter analyzeExp exp_arg_list 
         >>= fun _ -> 
                begin match lval_result_opt with 
                    None      -> return ()
                  | Some lval -> 
                       (* Analyse the lval and collect constraints on the assignment of the result to the lval *)
                                         analyzeLval lval
                       >>= fun ()     -> return (Updatecilutil.properTypeOfLval lval)
                       >>= fun lval_t -> subTypeConstraints exp_func_ret_t lval_t
                     end
         
         >>= fun _ -> AMOps.iter (fun (t1,t2) -> subTypeConstraints t1 t2) fixed_args_src_target_types 
         >>= fun _ -> return (Updatecilutil.properTypeOf exp_func)
         >>= fun exp_func_t -> 
                    begin match (Cil.unrollType exp_func_t) with 
                       TFun(_,_,_,al) -> AnalysisEffectManipulator.extractFunConstraints al
            
                     | _              -> Updatedebug.abort (Updatedebug.bug
                                    "analyzeCall: expected function type, but got:\n %a\n" d_plaintype exp_func_t)
                    end
          
         (* If the callee can do an update then we're unconstrained 
            If we're N, the callee must be N too 
            hence the constraint mu_callee <= mu_caller     *)

         >>= fun calleeEffect    -> AM.amAddMCLessThanCurrent calleeEffect.effect_mu


         (* If we're in a TX, make the callee U
            (unless overridden by the above) *)
           
         >>= fun ()              -> (if isInTX && 
                                       (Updateargs.args.Updateargs.arg_updatabilityPolicy = Updateargs.UPTx)
                                     then AM.amAddMC (MuConst true) calleeEffect.effect_mu
                                     else return ())


         >>= fun ()              -> CG.amTRGetPhi
         >>= fun prePhi          -> CG.amTRFreshPhi
         >>= fun postPhi         -> CG.amTRPhiFlow prePhi postPhi "analyzeCall: prePhi --> postPhi"
(*          >>= fun ()              -> CG.amTRDeltaSubset calleeEffect.effect_labels "analyzeCall: labels" *)
         >>= fun ()              -> CG.amTRSetPhi postPhi
         >>= fun ()              -> CG.amTRDeltaCond calleeEffect.effect_mu calleeEffect.effect_delta_out "analyzeCall: d < delta_out"
         >>= fun ()              -> return (phi_of_effect calleeEffect)
         >>= fun calleePhi       -> CG.amTRPhiFlowFwdOnly prePhi  calleePhi "analyzeCall: preFwd < calleeFwd"
         >>= fun ()              -> CG.amTRPhiFlowBckOnly postPhi calleePhi "analyzeCall: postBck < calleeBck"

         >>= fun ()              -> CG.amTRVCInclude calleeEffect.effect_vc_eps prePhi "analyzeCall: callee_eps"



           


         (* Constrain function addresses passed to external functions *)
         >>= fun () -> AM.amGetInternVarHash 
         >>= fun internVarHash -> 
                     begin match isExternalFunCall exp_func internVarHash with
                          true -> constrainArgumentsToExternalFunctions calleeEffect exp_arg_list
                        | false -> return ()
                     end

       
    | _ -> Updatedebug.abort (Updatedebug.bug "analyzeCall : no way can be here")

and analyzeInstr (isInTX : bool) (instr : Cil.instr) : unit AM.m = 

                AM.setLoc (Cil.get_instrLoc instr) 

     (* Now add all gvars used in this instr to the current capability 
      * We have to do both a syntactic and a points-to check 
      * since each is not enough in isolation
      *)
          
   >>= fun _ -> sweepInstrEffects instr

   >>= fun _ -> 
      match instr with 
            Set (lval,exp,loc) -> 
                                    analyzeLval lval 
                  >>= fun ()     -> analyzeExp exp 
                  >>= fun ()     -> return (Updatecilutil.properTypeOfLval lval)
                  >>= fun lval_t -> return (Updatecilutil.properTypeOf exp)
                  >>= fun e_t    -> subTypeConstraints e_t lval_t
         
          | Call _ -> analyzeCall isInTX instr
         
          | Asm _ ->
                              AM.getLoc 
               >>= fun loc -> ignore(Updatedebug.log [Updatedebug.BTAnalysis] 
                                        "(%a) WARNING! analyzeInstr: can't deal with Asm at instr level, but you used it!\n\n" 
                                        d_loc loc);
                               return ()


and analyzeStmt (isInTX : bool) (stmt : Cil.stmt) : unit AM.m = 
(*    let () = ignore (Updatedebug.log [Updatedebug.BTAnalysis] "analyzeStmt (isInTX:%b, %a)  called\n" isInTX d_stmt stmt) in *)

   let id = return () in 

   (* check whether the stmt has a checkin label associated with it *)
   let isCheckinStmt = Updateconfig.isCheckinStmt stmt  in

   (if isCheckinStmt
    then
      (ignore(Updatedebug.log [Updatedebug.BTAlways] "analyzeStmt: found check-in at %a, sid %d \n" d_loc (Cil.get_stmtLoc stmt.skind) stmt.sid);
       analyzeCheckinStart) 
      else return ())  >>= fun _ ->
     
   (* check whether the stmt has a TX label associated with it *)
   let isTXStmt = Updateconfig.isTXStmt stmt in
   let () = if isTXStmt then ignore(Updatedebug.log [Updatedebug.BTAlways] "analyzeStmt: found TX at %a \n" d_loc (Cil.get_stmtLoc stmt.skind)) in


     (if isTXStmt && (Updateargs.args.Updateargs.arg_updatabilityPolicy = Updateargs.UPTx)
      then AM.amAddMCLessThanCurrent (MuConst true)
      else return ())  >>= fun _ ->

   AM.setLoc (Cil.get_stmtLoc stmt.skind) >>= fun _ ->
   (* Register the start of the statment, so that the current delta ID can be associated with the stmt id *)
   AM.amStmtStart stmt.sid >>= fun () ->
   analyzeCutPhi >>= fun () ->
     
   AM.amGetDelta >>= fun deltaAtStart ->  
   AM.amGetVcFwd >>= fun vcFwdAtStart ->  
   AM.amAddFwdLabel stmt.sid >>= fun () ->

   sweepStmtEffects stmt >>= fun () ->

   (* if this is a TX, save its Phi *)
   (if isTXStmt  then analyzeTxStart else CG.amTRGetPhi) >>= fun preTxPhi -> 

   begin match stmt.skind with 
    
    (* A speculative update function *)
    | Instr [ Call(_, Lval(Var vi, NoOffset), _, loc) ] when  Updateconfig.isUpdateFunc vi.vname ->
        CG.amTRUpdatePoint suspConstr (Some stmt.sid) "auto-inserted update"
   
    | Instr  il ->     AMOps.iter (analyzeInstr (isTXStmt || isInTX)) il

       
       (* FIX? Seems as if you would have to constrain the output
           effect of the function with the current capability.  On
           some thought, though, it seems as if the capability at the
           conclusion of the block will be more restrictive, so it's
           not actually needed here. *)
   
    | Return (None,loc) -> id
    
    | Return (Some e, loc) ->
                          analyzeExp e  
         >>= fun ()    -> AM.getCurrentFun 
         >>= fun fd    -> return (Updatecilutil.getReturnType fd)
         >>= fun ret_t -> return (Updatecilutil.properTypeOf e)
         >>= fun e_t   -> subTypeConstraints e_t ret_t
       
    | Goto (target, loc) ->
                            CG.amTRGetPhi
      >>= fun currentPhi -> AM.amGetFwdPhi !target.sid 
      >>= fun phio  -> (match phio with

          None -> 
                     ignore(Updatedebug.log [Updatedebug.BTAnalysis] "amDoGoto:  targetSid (%d) not found, must be forward goto\n" !target.sid);
                     return ()

        | Some targetPhi -> 
                     ignore(Updatedebug.log [Updatedebug.BTAnalysis] "amDoGoto:  targetSid (%d) found\n" !target.sid); 
                     CG.amTRPhiFlow currentPhi targetPhi "doGoto : crtPhi ---> fwdPhi" )

(*          AM.amDoGoto !target.sid *)
       
    | Break loc -> 
         Updatedebug.abort (Updatedebug.bug
                           "(%a) analyzeStmt: CIL should have eliminated break!\n" 
                           d_loc loc)
    
    | Continue loc -> 
         Updatedebug.abort (Updatedebug.bug
                           "(%a) analyzeStmt: CIL should have eliminated continue!\n" 
                           d_loc loc)      
    
    | If (eTest,blockT,blockF,loc) ->
              (* --- analyze test expression first --- *)

                            analyzeExp eTest

              (* --- then proceed with the branches --- *)

         (* --- VC --- *)
         >>= fun b_e ->

                            CG.amTRGetPhi
         >>= fun phiE    -> analyzeBlock (isTXStmt || isInTX) blockT

         >>= fun ()      -> CG.amTRGetPhi
         >>= fun phiT    -> CG.amTRSetPhi phiE


         >>= fun ()      -> analyzeBlock (isTXStmt || isInTX) blockF


         >>= fun ()      -> CG.amTRGetPhi
         >>= fun phiF    -> CG.amTRFreshPhi
         >>= fun phiPost -> CG.amTRSetPhi phiPost
         >>= fun ()      -> CG.amTRPhiFlow phiT phiPost "phiT --> phiPost"
         >>= fun ()      -> CG.amTRPhiFlow phiF phiPost "phiF --> phiPost"

(*          CG.amTRIf eTest blockT blockF analyzeExp analyzeBlock *)
      
    | Switch (exp, block,stmt_list,loc) ->
        Updatedebug.abort (Updatedebug.bug 
                          "(%a) analyzeStmt: CIL should have eliminated switch!\n" 
                                    d_loc loc)
       
    | Loop (block,loc,s1,s2) ->

                                 CG.amTRGetPhi
         >>= fun prePhi       -> CG.amTRFreshPhi
         >>= fun blockPhi     -> CG.amTRSetPhi blockPhi
         >>= fun ()           -> CG.amTRPhiFlow prePhi blockPhi "preBlock --> blockStart"

         >>= fun ()           -> analyzeBlock (isTXStmt || isInTX) block

         >>= fun ()           -> CG.amTRGetPhi
         >>= fun blockPhi'    -> CG.amTRFreshPhi
         >>= fun postPhi      -> CG.amTRSetPhi postPhi
         >>= fun ()           -> CG.amTRPhiFlow blockPhi' postPhi "blockEnd --> postBlock"
         >>= fun ()           -> CG.amTRPhiFlow blockPhi' blockPhi "blockEnd --> blockStart"


  
    | Block block -> analyzeBlock (isTXStmt || isInTX) block
       
    | TryFinally  (block1,block2,loc) ->
       Updatedebug.abort (Updatedebug.bug 
                         "(%a) analyzeStmt: __try/__except not supported!\n" 
                         d_loc loc)
  
    | TryExcept (block_try, inst_list, block_except, loc) ->
       Updatedebug.abort (Updatedebug.bug 
                         "(%a) analyzeStmt: __try/__except not supported!\n"
                                   d_loc loc)
   end
   (* Associate the current delta with the output capability of this statement *)
   >>= fun () -> (if isTXStmt  then analyzeTxEnd preTxPhi else return ())  

   >>= fun () -> 
     (if isCheckinStmt
      then 
        (AM.amEndCheckinDelta >>= fun () -> 
         analyzeCutPhi)
      else return ())

   >>= fun () -> AM.amStmtEnd stmt.sid deltaAtStart vcFwdAtStart

   >>= fun () -> 
     (if isCheckinStmt
      then analyzeCheckinEnd 
      else return ())


and analyzeCutPhi : unit AM.m = 
                          CG.amTRGetPhi
   >>= fun prePhi    ->   CG.amTRFreshPhi
   >>= fun phi       ->   CG.amTRSetPhi phi
   >>= fun ()        ->   CG.amTRPhiFlow prePhi phi "analyzeCutPhi"

and analyzeCheckinStart : unit AM.m = 
                          AM.amStartCheckinEps
   >>= fun ()        ->   CG.amTRGetPhi
   >>= fun prePhi    ->   CG.amTRFreshPhi
   >>= fun phi       ->   CG.amTRSetPhi phi
   >>= fun ()        ->   CG.amTRPhiFlow prePhi phi "analyzeCheckinStart"
  
and analyzeCheckinEnd : unit AM.m = 
                          AM.amEndCheckinEps
   >>= fun ()        ->   CG.amTRGetPhi
   >>= fun phi       ->   CG.amTRFreshPhi
   >>= fun postPhi   ->   CG.amTRSetPhi postPhi
   >>= fun ()        ->   CG.amTRPhiFlow phi postPhi "analyzeCheckinEnd"

(* create a new Phi for the transaction, and return the old Phi;
   it will be restored upon tx end *)
and analyzeTxStart : phi AM.m = 

                          AM.getLoc 
   >>= fun loc       ->   (ignore(Updatedebug.log [Updatedebug.BTAnalysis] "(%a) analyzeTxStart: transaction start\n" d_loc loc);
                          CG.amTRGetPhi)
   >>= fun preTxPhi  ->   CG.amTRFreshPhiFreshEps
   >>= fun txPhi     ->   CG.amTRSetPhi txPhi

                          (* if txIsolation flag is set, don't propagate outside effects into the TX body *)
   >>= fun ()        ->   (if (Updateargs.args.Updateargs.arg_txIsolation)
                           then return ()
                           else CG.amTRPhiFlowFwdOnly preTxPhi txPhi "analyzeTxStart : pre-tx fwd < tx fwd"
                          )

   >>= fun ()        ->   CG.amTRDeltaFlowOnly preTxPhi txPhi "analyzeTxStart : tx delta < pre-tx delta"
   >>= fun ()        ->   return preTxPhi
 
and analyzeTxEnd preTxPhi : unit AM.m = 
                          AM.getLoc 
   >>= fun loc       ->   (ignore(Updatedebug.log [Updatedebug.BTAnalysis] "(%a) analyzeTxEnd: transaction end\n" d_loc loc);
                          CG.amTRGetPhi)
   >>= fun txPhi     ->   CG.amTRFreshPhiSpecifiedEps preTxPhi
   >>= fun postTxPhi ->   CG.amTRSetPhi postTxPhi


                          (* if txIsolation flag is set, don't propagate outside effects into the TX body *)
   >>= fun ()        ->   (if (Updateargs.args.Updateargs.arg_txIsolation)
                           then return ()
                           else CG.amTRPhiFlowBckOnly postTxPhi txPhi "analyzeTxEnd: post-tx bck < tx bck"
                          )

   >>= fun ()        ->   CG.amTRPhiFlowBckOnly postTxPhi preTxPhi "analyzeTxEnd: post-tx bck < pre-tx bck" (* can't restore the pre-tx bck, rather make the new BCK a subset of it *)
   >>= fun ()        ->   CG.amTRPhiFlowFwdOnly preTxPhi postTxPhi "analyzeTxEnd: pre-tx fwd < post-tx fwd" (* propagate the pre-tx flow *)
   >>= fun ()        ->   CG.amTRDeltaFlowOnly txPhi postTxPhi "analyzeTxEnd : post-tx delta < tx delta"

and analyzeBlock (isInTX : bool) (b : Cil.block) : unit AM.m =
   AMOps.iter (analyzeStmt isInTX) b.bstmts 
      
  



(* Collect constraints for a function *)
let analyzeFunction (f : Cil.fundec) : unit AM.m =

  let _ = ignore(Updatedebug.log [Updatedebug.BTAnalysis] "Analyzing function %s ...\n" f.svar.vname) in

       AM.setCurrentFun f
   >>= fun _ ->
                match f.svar.vtype with 
                   TFun(_,_,_,al) ->
                                        AnalysisEffectManipulator.extractFunConstraints al

                           >>= fun effect -> ignore(Updatedebug.log [Updatedebug.BTAnalysis] "analyzeFunction(%s): delta_in: %d, delta_out: %d, delta_eigen: %d, vc_fwd : %d, vc_eps : %d, vc_bck : %d\n" 
                                                      f.svar.vname 
                                                      effect.effect_delta_in.delta_id effect.effect_delta_out.delta_id effect.effect_delta_eigen.delta_id
                                                      effect.effect_vc_fwd.delta_id effect.effect_vc_eps.delta_id effect.effect_vc_bck.delta_id );


                                             CG.amTRStartFun effect f
                             
                           >>= fun _      -> analyzeBlock false f.sbody

                           >>= fun ()     -> CG.amTREndFun effect 

                      
                 | _ -> Updatedebug.abort (Updatedebug.bug
                        "analyzeFunction: expected function type, but got:\n %a\n" d_plaintype f.svar.vtype)

   >>= fun _ -> AM.unsetCurrentFun 


(* The following three functions take care of analyzing static initializers *)

let rec analysis_foldLeftCompound (ct: Cil.typ) (initl: (Cil.offset * Cil.init) list) : unit AM.m = 
  let bail s = Updatedebug.abort (Updatedebug.bug "analysis_foldLeftCompound %s" s) in
  let off_l, init_l = List.split initl in
  match unrollType ct with
    TArray(bt, _, _) ->
       AMOps.iter (analyzeInit bt) init_l

  | TComp (comp, _) as tc -> 
       let getTypesForOffset = function
          Field(f, NoOffset) -> (f.ftype, (Commoncilutil.getFieldinfoByName tc f.fname None).ftype)
        | _ -> bail "analysis_foldLeftCompound: malformed initializer"
       in
       let doOneField ((o, i) : (Cil.offset * Cil.init)) : unit AM.m = 
         let field_t,comp_field_t = (getTypesForOffset o) in 
                           subTypeConstraints field_t comp_field_t
            >>= fun () ->  analyzeInit field_t i
       in
       AMOps.iter doOneField initl 

  | _ -> bail "Type of Compound is not array or struct or union"


and analyzeInit (var_t : Cil.typ) (ii : Cil.init) : unit AM.m = 
   begin match ii with
      SingleInit e -> let exp_t = Updatecilutil.properTypeOf e in
                         analyzeExp e
         >>= fun ()   -> subTypeConstraints exp_t var_t
  
    | CompoundInit (t,ol) -> 
                         analysis_foldLeftCompound t ol
         >>= fun ()   -> subTypeConstraints t var_t
   end

let analyzeInitinfo (ii : Cil.initinfo) (var_t : Cil.typ ) : unit AM.m = 
   begin match ii.init with 
      Some i -> 
                       analyzeInit var_t i
         >>= fun () -> return ()
    
    | None   ->        return ()
   end


let crtGlobal = ref 0;;

let analyzeGlobal (g : Cil.global) : unit AM.m = 

  (if Updateargs.args.Updateargs.arg_showProgress then
    AMOps.lift print_string (return ("At global " ^ (incr crtGlobal; string_of_int (!crtGlobal)) ^ "\n"))
  else
    return ()
)
  >>= fun () ->
   (match g with 
        GFun(fd,loc) -> 
           if Namedtypes.isConAbsFuncName fd.svar.vname then
              return ()
           else 
              analyzeFunction fd

       (* analysis-wise, a global init 
          var = exp 
          has to be regarded as a Set(var,exp) 
          We simulate this by forcing a subtype relationship:
               exp_t <: var_t
       *)

      | GVar(vi, ii, loc) -> 
                                AM.setLoc loc 
              >>= fun ()     -> analyzeInitinfo ii vi.vtype 

      | _            -> return ())

let doPointsToAnalysis (file : Cil.file) : unit = 
  ignore(Updatedebug.log [Updatedebug.BTAlways] " Starting Ptranal...\n ");
  (* Ptranal stores its state internally; careful to only call it once *)
(*   Ptranal.print_constraints := true; *)
(*   Ptranal.debug_constraints := true; *)
(*    Ptranal.debug := true;  *)
  Ptranal.analyze_file file;
  Ptranal.compute_results false;
(*   Ptranal.compute_results true; *)
  ignore(Updatedebug.log [Updatedebug.BTAlways] " Ptranal done.\n ")
;;
  
let analyzeFile (file : Cil.file) : unit AM.m =

  (* Do the points-to analysis first *)
  (if Updateargs.args.Updateargs.arg_pointstoEnabled then 
  Stats.time "Stats : pointsTo - doAN" doPointsToAnalysis file);

  (* Now do the safety analysis *)
  AMOps.iter analyzeGlobal file.globals


(* ---- End of traversal of syntax to gather constraints ---- *)
(*============================================================================*)

(* add interprocedural 'goto' constraints due to setjmp/longjmp *)
let purgeLongjmpConstraints (setjmpHash : Setjmp.setjmpHash_t)  : unit AM.m = 

   let purgeLonjmp_iter (longjmpPhi, (env_vi,mu)) =  
   (* in a function with updateability 'mu' , 
      at statement 'delta', we had a 
      longjmp(env,val) 

      now look for statements 
      D;     setjmp(env) ; D'
      and add 
      D' <= delta 
   *)
      try
         let setjmpSid = H.find setjmpHash env_vi in
            begin
                                  AM.amGetFwdPhi setjmpSid 
                 >>= fun phio  -> 
                   (match phio with
                        None -> Updatedebug.abort (Updatedebug.bug "purgeLongjmpConstraints: can't find sid=%d in fwdDelta \n" setjmpSid)
                      | Some setjmpPhi -> CG.amTRPhiFlow setjmpPhi longjmpPhi "purgeLonjmp : crtPhi ---> fwdPhi" )
                           
            end
      with Not_found -> ignore(Updatedebug.log [Updatedebug.BTAlways] " WARNING : purgeLonjmp_iter : can't find destination for '%s' in setjmpHash\n " env_vi.vname);
                        return ()
(*          Updatedebug.abort (Updatedebug.bug "purgeLongjmpConstraints: can't find destination for '%s' in setjmpHash\n" *)
(*                            env_vi.vname)          *)

   in
                              AM.amGetLongjmpHash
   >>= fun longjmpHash     -> let longjmp_hash_contents = Commonutil.hashtbl_key_val_pairs longjmpHash in
                              AMOps.iter purgeLonjmp_iter longjmp_hash_contents;;


(*============================================================================*)



(* Filter statement effects based on the user-specifed effect printing policy *)
let filterPrintedEffects (delta_set : CapabilitySet.t) : CapabilitySet.t = 

   (* this tells us whether we're dealing with a type or function name*)
   let isTypeName (name : string) : bool = 
      try
         ignore (Namedtypes.NamedTypeInfo.isUpdatable name);
         true
      with 
         Not_found -> false
   in
   let notIsTypeName = fun x -> not (isTypeName x) in
   let filteredSet =  
   match Updateargs.args.Updateargs.arg_effectPrintingPolicy with
      Updateargs.EPFull | Updateargs.EPDifferential     -> delta_set
    | Updateargs.EPSolveOnly -> CapabilitySet.empty
    | Updateargs.EPTypes     -> CapabilitySet.filter isTypeName delta_set
    | Updateargs.EPFunctions -> CapabilitySet.filter notIsTypeName delta_set
    | Updateargs.EPNone      -> CapabilitySet.empty
   in
   (* if we want to focus on one type/fun/gvar only, we set the focus in Updateargs *)
   match Updateargs.args.Updateargs.arg_focus with  
       None -> filteredSet
     | Some s -> CapabilitySet.filter (fun x -> (x = s)) filteredSet
;;

   




(*============================================================================*)
(* ---- Printing analysis effects ---- *)

(* these two cap_of_id guys might throw Not_found;
   caveat caller, for the caller must catch it
   if caught in string_of_delta, print NOEFFECT *)

let cap_of_id_lazy solution deltaSolverState id = 
  let res = 
    try
      Deltasolver.lazyQuery deltaSolverState id 
    with Not_found ->
      Deltasolver.CapabilitySet.empty
  in
(*   let () = ignore (Updatedebug.log [Updatedebug.BTAlways] "cap_of_id_lazy: adding %d\n" id) in *)
  Commonutil.hashtbl_add_once solution id res;
  res  
;;

let cap_of_id_eager solution id =
  try
    H.find solution id
  with Not_found ->
    Updatedebug.abort (Updatedebug.bug "cap_of_id_eager: can't find id %d\n" id)
 
;;

let string_of_delta cap_of_id id = 
  try
    let delta_set = cap_of_id id in
      printDelta (filterPrintedEffects delta_set)
  with 
      Not_found -> "NOEFFECT (" ^  (string_of_int id) ^ ")" 
;;

(* install a hook into cil to print capability sets with each statement *)
let installPrintEffectsHook (c_of_id : int -> CapabilitySet.t)
                            (stmtid_to_delta_id : (int,(int*int)) H.t)
                            (stmtid_to_vc_ids : (int,(int*int*int*int option)) H.t)  =  
   let hook s = 
     let sid = s.sid in
(*      let wholeNineYards stmt =  *)
(*        let (d_in_id, d_out_id) = H.find stmtid_to_delta_id stmt.sid  *)
(*        and  (vc_fwd_id, vc_eps_id, vc_bck_id, vc_checkin_eps_id) = H.find stmtid_to_vc_ids stmt.sid  *)
(*        in *)
     let wholeText deltaAnnotation vcAnnotation =   
       "/* " ^ (Updateconfig.delta_test_compat "" ("sid:" ^ (string_of_int sid))) ^
         deltaAnnotation ^
         (Updateconfig.delta_test_compat "" vcAnnotation) ^
         " */"
     in
      (match Updateargs.args.Updateargs.arg_effectPrintingPolicy with
           (* common/fast case first *) 
           Updateargs.EPNone | Updateargs.EPSolveOnly -> ""

           (* print types/function labels only if they're specified in the current effectPrintingPolicy *)   
         | Updateargs.EPFull | Updateargs.EPTypes | Updateargs.EPFunctions ->
            begin
              let cap_of_id delta_id = string_of_delta c_of_id delta_id in
                try
                  let () = ignore (Updatedebug.log [Updatedebug.BTAnalysis] "installPrintEffectsHook: looking up sid: %d\n" sid) in
                  let deltaAnnotation = 
                    let (d_in_id, d_out_id) = H.find stmtid_to_delta_id sid in
                      (Updateconfig.delta_test_compat "" "\n") ^
                        "D(" ^ (debug_string_of_int d_in_id) ^ "): " ^ (cap_of_id d_in_id) ^ 
                        "   *** D'(" ^ (debug_string_of_int d_out_id) ^ "): " ^ (if d_out_id = d_in_id then "NO CHANGE" else (cap_of_id d_out_id)) 
                  in
                  let vcAnnotation =   
                    if Updateargs.args.Updateargs.arg_vcEnabled
                    then
                      let (vc_fwd_id, vc_eps_id, vc_bck_id, vc_checkin_eps_id) = H.find stmtid_to_vc_ids sid in
                        
                        "\nVC-F(" ^ (debug_string_of_int vc_fwd_id) ^ "): " ^ (cap_of_id vc_fwd_id) ^
(*                           "  VC-E(" ^ (debug_string_of_int vc_eps_id) ^ "): " ^ (cap_of_id vc_eps_id) ^ *)
                          "\nVC-B(" ^ (debug_string_of_int vc_bck_id) ^ "): " ^ (cap_of_id vc_bck_id) ^
                          (
                            match vc_checkin_eps_id with 
                                None -> ""
                              | Some eps_id -> "  VC-CI(" ^ (debug_string_of_int eps_id) ^ "): " ^ (cap_of_id eps_id) 
                          )
                    else
                      ""
                  in
                    wholeText deltaAnnotation vcAnnotation
                with 
                    Not_found -> "/* NOEFFECT */" (*Updatedebug.abort (Updatedebug.bug "PrintEffectsHook: could not find effect for statement %d!\n" sid)*)
            end
           (* have to look at prev/next stmts *)   
         | Updateargs.EPDifferential ->
             try
               let deltaThis, deltaThis_id,
                   fwdThis,   fwdThis_id,
                   bckThis,   bckThis_id  = 
                 let (d_in_id, d_out_id) = H.find stmtid_to_delta_id sid in
                 let (vc_fwd_id, vc_eps_id, vc_bck_id, vc_checkin_eps_id) = H.find stmtid_to_vc_ids sid in
                 (c_of_id d_out_id), d_out_id,
                 (c_of_id vc_fwd_id), vc_fwd_id,
                 (c_of_id vc_bck_id), vc_bck_id
               in
               let nPrev = List.length s.preds 
               and nSucc = List.length s.succs 
               in
               let deltaNext,bckNext = 
                   match s.succs with
                       [h] -> 
                         let (d_in_id, d_out_id) = H.find stmtid_to_delta_id h.sid 
                         and  (vc_fwd_id, vc_eps_id, vc_bck_id, vc_checkin_eps_id) = H.find stmtid_to_vc_ids h.sid in 
                           c_of_id d_out_id, c_of_id vc_bck_id
                     | _ -> Deltasolver.CapabilitySet.empty, Deltasolver.CapabilitySet.empty
               and fwdPrev = 
                 match s.preds with
                     [h] -> 
                       let (vc_fwd_id, vc_eps_id, vc_bck_id, vc_checkin_eps_id) = H.find stmtid_to_vc_ids h.sid in 
                         c_of_id vc_fwd_id
                   | _ -> Deltasolver.CapabilitySet.empty
               in
               let deltaAnnotation = 
                 "\nD " ^ 
                   (if nSucc == 1 then "#" else "?") ^ "(" ^ (string_of_int deltaThis_id) ^
                   (printDelta (Deltasolver.CapabilitySet.diff deltaThis deltaNext)) ^ " )"  
               and  vcAnnotation = 
                 "\nVC-F " ^ 
                   (if nPrev == 1 then "#" else "?") ^ "(" ^ (string_of_int fwdThis_id) ^
                   (printDelta (Deltasolver.CapabilitySet.diff fwdThis fwdPrev)) ^ " )"  ^
                 "\nVC-B " ^ 
                   (if nSucc == 1 then "#" else "?") ^ "(" ^ (string_of_int bckThis_id) ^
                   (printDelta (Deltasolver.CapabilitySet.diff bckThis bckNext)) ^ " )" 
               in
               wholeText deltaAnnotation vcAnnotation
                 
             with Not_found ->
               "/* NOEFFECT */"
(* wholeText ("prev: " ^ (string_of_int nPrev) ^ "succ:" ^ (string_of_int nSucc)) "" *)
      )
   in 
   Cil.setStmtPrintEffectHook hook






(* ---- insert function effects as comments ---- *)
let annotateWithFunEffects (funEffectHash : funEffectHash_t) 
                           (funTypeHash : funTypeHash_t) 
                           (c_of_id : int -> CapabilitySet.t) 
                           (mu_solution : (int,bool) H.t) 
                           (fname_to_LPS_hash : (string,delta*delta) H.t)
                           (file:Cil.file)
                           (stateXformEffectId : int option ref) 
                           : unit = 
                           
   let string_of_mu (mu:mu) : string = 
      let ustring_of_bool b = if b then "U" else "N" in
      match mu with 
        MuConst b -> ustring_of_bool b
      | MuVar id  -> 
           try
              (debug_string_of_int id) ^ ":" ^ (ustring_of_bool (H.find mu_solution id))
           with 
              Not_found -> (debug_string_of_int id) ^ ":" ^ (ustring_of_bool true) (*Updatedebug.abort (Updatedebug.bug "annotateWithFunEffects:  function has mu var %d that is not in the solution!\n" id)*)
   in
   
   let annotate_fold (acc:global list) (g:global) = 
      match g with 
         (* only annotate user functions -- Note the definition for update is not added yet, so we don't have to exclude it *)
         GFun(fd,loc) when 
            (not (Namedtypes.isConAbsFuncName fd.svar.vname) )  ->
            ignore(Updatedebug.log [Updatedebug.BTAnalysis] "annotateWithFunEffects : at function %s\n" fd.svar.vname);

            begin match AnalysisEffectManipulator.getEffects (Cil.typeAttrs fd.svar.vtype) with
               [Attr(attrName,[AInt key])] ->
                  begin try
                     (* if it's the state xform, save its id*)
                    let () = if Updateconfig.is_state_xform fd then stateXformEffectId := Some key in
                    let () = H.add funTypeHash fd.svar.vname fd.svar.vtype in (* save fun type now, for the next step in runAnalysis is gonna remove the attrs from the fun type *)
                    let effect = H.find funEffectHash key in
                    let forcePrint = (H.mem Updateargs.args.Updateargs.arg_funToPrint fd.svar.vname) in
                     if (Updateargs.args.Updateargs.arg_effectPrintingPolicy <> Updateargs.EPNone) || forcePrint
                     then begin
                        let cap_of_id delta_id = 
(*                           let () = ignore(Updatedebug.log [Updatedebug.BTAlways] "cap_of_id %d\n" delta_id) in *)
                          string_of_delta c_of_id delta_id in
                        let deltaAnnotation =   
                          let d_in_id = effect.effect_delta_in.delta_id and
                              d_out_id = effect.effect_delta_out.delta_id and
                              d_eigen_id = effect.effect_delta_eigen.delta_id in
                             (* Iulian : change this to BTAlways to see progress as we process each function *)
                          let () = ignore(Updatedebug.log [Updatedebug.BTAnalysis] "annotateWithFunEffects : at function %s (in %d out %d eigen %d)\n" fd.svar.vname effect.effect_delta_in.delta_id effect.effect_delta_out.delta_id effect.effect_delta_eigen.delta_id) in

(*                           let () = ignore(Updatedebug.log [Updatedebug.BTAlways] "will ask %d %d %d\n" d_in_id d_out_id d_eigen_id) in *)

                          let d_in_s = cap_of_id d_in_id and
                              d_out_s= cap_of_id d_out_id and
                              d_eigen_s = cap_of_id d_eigen_id in  

                            "  D(" ^ (debug_string_of_int d_in_id) ^ "): { " ^ (d_in_s) ^ " }" ^
                            "   *** D'(" ^ (debug_string_of_int d_out_id) ^ "): " ^ (if d_out_id = d_in_id then "NO CHANGE" else ("{ " ^ (d_out_s) ^ " }") ) ^ 
                              "  ED(" ^ (debug_string_of_int d_eigen_id) ^ "): { " ^ (d_eigen_s) ^ " }" 
 in
(*                         let () = ignore(Updatedebug.log [Updatedebug.BTAlways] "deltaAnnotation(%s): %s\n" fd.svar.vname deltaAnnotation) in *)

                        (* --- VC --- *)
                        let vcAnnotation =   
                          if Updateargs.args.Updateargs.arg_vcEnabled
                          then
                            let (vc_fwd_id, vc_eps_id, vc_bck_id) = (effect.effect_vc_fwd.delta_id, effect.effect_vc_eps.delta_id, effect.effect_vc_bck.delta_id) in 
                              "\nVC-F(" ^ (debug_string_of_int vc_fwd_id) ^ "): " ^ (cap_of_id vc_fwd_id) ^
                              "  VC-E(" ^ (debug_string_of_int vc_eps_id) ^ "): " ^ (cap_of_id vc_eps_id) ^
                              "  VC-B(" ^ (debug_string_of_int vc_bck_id) ^ "): " ^ (cap_of_id vc_bck_id)
                          else
                            ""
                        in

(*                         let labels_id = effect.effect_labels.delta_id in *)
(*                         let labels_set = c_of_id labels_id in *)
                           (* print types/function labels only if they're specified in the current effectPrintingPolicy *)   
                        let suspAnnotation = 
                          if Updateargs.args.Updateargs.arg_dosuspect
                          then 
                            let (lps_id,locsG_id) = 
                              let (l,locsG) = H.find fname_to_LPS_hash fd.svar.vname in
                              (l.delta_id,locsG.delta_id)
                            in
                             (Updateconfig.delta_test_compat "" "\n") ^ 
                            " L("    ^ (debug_string_of_int lps_id) ^ "): { "   ^ (cap_of_id lps_id) ^ 
                              " }  G(" ^ (debug_string_of_int locsG_id) ^ "): { " ^ (cap_of_id locsG_id) ^ " }"
                          else 
                            ""
                        in
                        let types_string = 
                              match Updateargs.args.Updateargs.arg_effectPrintingPolicy, forcePrint with

                                   Updateargs.EPFull,_ 
                                 | Updateargs.EPTypes,_ ->  
                                     deltaAnnotation ^ (Updateconfig.delta_test_compat "" "\n") ^  
                                     (Updateconfig.delta_test_compat ""  vcAnnotation) ^  
                                     suspAnnotation

                                 | _,true -> deltaAnnotation ^ (Updateconfig.delta_test_compat "" "\n") ^  
                                     (Updateconfig.delta_test_compat ""  vcAnnotation) ^  
                                     suspAnnotation
                                 | _ -> ""
                                      
(*                         and funlabels_string = match Updateargs.args.Updateargs.arg_effectPrintingPolicy with *)

(*                              Updateargs.EPFull  *)
(*                            | Updateargs.EPFunctions -> " Labels(" ^ (debug_string_of_int labels_id) ^ "): { " ^ (cap_of_id labels_id) ^ " }" *)

(*                            | _ -> ""        *)
                        in
                        let fun_header_string =  "/* Mu:" ^ (string_of_mu effect.effect_mu) ^ 
                           types_string ^
(*                            funlabels_string ^ *)
                           " */"  
                        in
                           acc @ [GText( fun_header_string );  
                                  g]
                     end
                     else
                        acc @ [GText("/* Mu:" ^ (string_of_mu effect.effect_mu) ^"*/") ; g]
                  with 
                        Not_found ->
                           (* printSolution solution; *)
                           Updatedebug.abort (Updatedebug.bug "annotateWithFunEffects: Couldn't find something in hash")
                  end
                     
               | _ -> Updatedebug.abort (Updatedebug.bug "annotateWithFunEffects: too many effects or couldn't find effect on function")
            end
         | _ -> acc @ [g]
   in
      file.globals <- List.fold_left annotate_fold [] file.globals

(*============================================================================*)









(*============================================================================*)
(* ---- Conditional constraint solving ---- *)


(* Given a solution to the mu constraints resolve each conditional delta constraint into a 
   regular subset constraint *)
let resolveConditionalConstraints (mu_solution : (int,bool) H.t) (cset : DeltaConstraintSet.t) : DeltaConstraintSet.t =
   let resolve dc cset = 
      match dc with 
        DCCond (muid,(d1,d2)) -> 
           begin try
              if H.find mu_solution muid then begin
       (match (d2.delta_kind) with
       DKConst _ -> let d1str,d2str = makeDeltaDebugStrings d1 d2 in 
       ignore(Updatedebug.log [Updatedebug.BTAnalysis] "Adding constraint %s < %s\n" d1str d2str)
      | _ -> ());
 
                 DeltaConstraintSet.add (DCSubset(d1,d2)) cset
               end
              else 
                 cset
           with
              Not_found -> 
                 (* FIXME: Not sure about this? *)
                 (* If we didn't find it in the solution then assume mu to be true *)
                 (*DeltaConstraintSet.add (DCSubset(d1,d2)) cset*) cset
           end

      | DCSubset(d1,d2) -> 
           DeltaConstraintSet.add dc cset
   in
   DeltaConstraintSet.fold resolve cset DeltaConstraintSet.empty

(*============================================================================*)







(*============================================================================*)
(************************* Debug printers **********************************)
let printSpeculativeUpdateMuHash (speculativeUpdateMuHash : (int,int) H.t) = 
   let printEntry stmt_id mu_id =
      ignore(Updatedebug.log [Updatedebug.BTAnalysis] "%d  |->  %d\n" stmt_id mu_id)
   in
   ignore(Updatedebug.log [Updatedebug.BTAnalysis] " Speculative update mu hash START *********************\n");
   H.iter printEntry speculativeUpdateMuHash;
   ignore(Updatedebug.log [Updatedebug.BTAnalysis] " Speculative update mu hash END ***********************\n")

(*============================================================================*)







(*============================================================================*)
(* ---- Do analysis ---- *)

let printStatistics () : unit AM.m = 
   AM.amGetNumMuConstraints >>= fun num_mu ->
   AM.amGetNumDeltaConstraints >>= fun num_delta ->
   ignore(Updatedebug.log [Updatedebug.BTStatistics] "# Constraints : mu = %d, delta = %d\n" num_mu num_delta);
   return ()

let solveLazily = true;;

(* return the result of running the monad and a solution to the mu constraints *)
let runAnalysis (mallocFunHash : Mallocpragma.mallocHash_t)  
                (effectHash : funEffectHash_t) 
                (funTypeHash : funTypeHash_t) 
                (solutionHash : deltaSolutionHash_t) 
                (setjmpHash : Setjmp.setjmpHash_t) 
                (file : Cil.file) 
                (suspect_result : Suspectalias.SAM.mo_result)
                (internVarHash : (string, unit) H.t)
                (funEffWaterMark : int)

                : analysis_result  = 

  (* need this for an eventual state xform *)
  let monad_result = 
     Stats.time "Stats : Run analysis monad - doAN"
      AM.run mallocFunHash effectHash solutionHash suspect_result internVarHash funEffWaterMark
          (                           (AnalysisGenFreshEffects.genFreshEffectsGlobal file)
            >>= fun _              -> AM.amCreateVoidPtrFunEffect
            >>= fun _              -> (analyzeFile file) 
            >>= fun _              -> AM.amConstrainVoidPtrFunEffect 
            >>= fun _              -> purgeLongjmpConstraints setjmpHash 
            >>= fun _              -> printStatistics ()
          )
   in
   let _ = Updatedebug.log [Updatedebug.BTAnalysis] "Mu constraints start ***********************************\n" in
   let _ = printMuConstraintSet monad_result.AM.mu_constraints in
   let _ = Updatedebug.log [Updatedebug.BTAnalysis] "Mu constraints end *************************************\n" in

   (* Undetermined variables are set to true=U, unless we only insert updates in transactions *)
   let defaultMu = 
     if (Updateargs.args.Updateargs.arg_updatabilityPolicy = Updateargs.UPTx)
     then false
     else true
   in
   let mu_solution = Stats.time "Stats : solveMu - doAN" solveMu monad_result.AM.mu_constraints defaultMu in
   let _ = Updatedebug.log [Updatedebug.BTAnalysis] "Mu solution start ***********************************\n" in
   let () = printMuStats mu_solution in
(*    let () = printMuSolution mu_solution in *)
   let _ = Updatedebug.log [Updatedebug.BTAnalysis] "Mu solution end *************************************\n" in
   let dcs_without_cond = Stats.time "Stats : resolveConditionalConstraints - doAN" resolveConditionalConstraints mu_solution monad_result.AM.delta_constraints in


   let state_xform_eff_id = ref None in 
   let c_of_id = match solveLazily with

       true ->  (* lazy solving *)  
         let deltaSolverState = Stats.time "Stats : solveDelta" Deltasolver.lazyStartSolving dcs_without_cond in
         let () = installPrintEffectsHook (cap_of_id_lazy solutionHash deltaSolverState) monad_result.AM.stmtid_to_delta_id monad_result.AM.stmtid_to_vc_ids in
         let () = Stats.time "Stats : annotateWithFunEffects - doDSU" annotateWithFunEffects monad_result.AM.funEffectHash 
                                                                                    funTypeHash
                                                                                    (cap_of_id_lazy solutionHash deltaSolverState)
                                                                                    mu_solution 
                                                                                    suspect_result.Suspectalias.SAM.sar_fname_to_LPS_hash file
                                                                                    state_xform_eff_id
         in 
(*          let _ = Updatedebug.log [Updatedebug.BTDeltasolver] "Delta solution start ***********************************\n" in *)
         let () = Deltasolver.printSolution solutionHash in
(*          let _ = Updatedebug.log [Updatedebug.BTDeltasolver] "Delta solution end ***********************************\n" in *)
(*          Deltasolver.lazyEndSolving (); *)
         (cap_of_id_lazy solutionHash deltaSolverState)

     | false -> (* eager solving *)  

         let () = Stats.time "Stats : solveDelta" solveDeltaEagerly dcs_without_cond solutionHash in
(*          let _ = Updatedebug.log [Updatedebug.BTDeltasolver] "Delta solution start ***********************************\n" in *)
(*          let () = Deltasolver.printSolution solutionHash in *)
(*          let _ = Updatedebug.log [Updatedebug.BTDeltasolver] "Delta solution end ***********************************\n" in *)
         let () = installPrintEffectsHook (cap_of_id_eager solutionHash) monad_result.AM.stmtid_to_delta_id monad_result.AM.stmtid_to_vc_ids in
         Stats.time "Stats : annotateWithFunEffects - doDSU" annotateWithFunEffects monad_result.AM.funEffectHash 
                                                                                    funTypeHash
                                                                                    (cap_of_id_eager solutionHash) 
                                                                                    mu_solution 
                                                                                    suspect_result.Suspectalias.SAM.sar_fname_to_LPS_hash file
                                                                                    state_xform_eff_id;
         (cap_of_id_eager solutionHash) 
  
   in

   Stats.time "Stats : removeEffectAttrVisitor - doDSU" Cil.visitCilFile (new Updatecilutil.removeEffectAttrVisitor(Updateconfig.effectAttrName)) file;
   printSpeculativeUpdateMuHash monad_result.AM.speculativeUpdateMuHash;
   {
      ar_delta_constraints = monad_result.AM.delta_constraints;       
      ar_mu_constraints = monad_result.AM.mu_constraints;
      ar_delta = monad_result.AM.delta;                 
      ar_mu = monad_result.AM.mu;                     
      ar_stmtid_to_delta_id = monad_result.AM.stmtid_to_delta_id;     
      ar_funEffectHash = monad_result.AM.funEffectHash;
      ar_funEffWaterMark = monad_result.AM.funEffectWatermark;
      ar_speculativeUpdateMuHash = monad_result.AM.speculativeUpdateMuHash;
      ar_mu_solution = mu_solution;           
      ar_delta_solution = solutionHash;
      ar_constNamesHash = monad_result.AM.constNamesHash;
      ar_stmtid_to_eps_tuple = monad_result.AM.stmtid_to_vc_ids;
      ar_stmtid_to_fun_name = monad_result.AM.stmtid_to_fun_name;
      ar_state_xform_eff_id = !state_xform_eff_id;
      ar_c_of_id = c_of_id;
      (* --- VC --- *)  
(*       ar_vc_fwd_constraints  *)
(*       ar_vc_eps_constraints  *)
(*       ar_vc_bck_constraints *)
   }

(*============================================================================*)


