(* 
 *
 * Copyright (c) 2004-2009, 
 * 
 *  Michael Hicks       <mwh@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(*
  Generic monadic folds and iters over CIL syntax




  Notes:
  ------

    - We only process compinfos, varinfos and enuminfos at global level.
      We don;t reprocess them at the expr level.  Use alt handler if you want this 
      behaivour
    - exps and lvals inside of Asm don't get processed. Use alt handler if you want 
      this behaviour.
    - offsets in initialisers don't get traversed by the fold (so FoldFuns.foffset not
      called for them).  This is because they are semantically different from the offsets
      of lvals.  We could supply a separate traversal function for them if we find it 
      necessary.

*)

open Cil
module H = Hashtbl

(* Define the return syntax *)
type ('typ, 'typeinfo,'compinfo, 'varinfo, 'initinfo, 'fundec) tglobal =
  | TGType of 'typeinfo * location    
  | TGCompTag of 'compinfo * location     
  | TGCompTagDecl of 'compinfo * location
  | TGEnumTag of enuminfo * location
  | TGEnumTagDecl of enuminfo * location
  | TGVarDecl of 'varinfo * location
  | TGVar  of 'varinfo * 'initinfo * location
  | TGFun of 'fundec * location           
  | TGAsm of string * location        
  | TGPragma of attribute * location 
  | TGText of string                

type 'typ ttypeinfo = {
  ttname: string;              
  tttype: 'typ;
  ttreferenced: bool;         
}
      
type ('typ,'exp,'typeinfo,'compinfo) ttyp =
    TTVoid of attributes 
  | TTInt of ikind * attributes 
  | TTFloat of fkind * attributes 
  | TTPtr of 'typ * attributes  
  | TTArray of 'typ * 'exp option * attributes
  | TTFun of 'typ * (string * 'typ * attributes) list option * bool * attributes
  | TTNamed of 'typeinfo * attributes 
  | TTComp of 'compinfo * attributes
  | TTEnum of enuminfo * attributes
  | TTBuiltin_va_list of attributes
  | TTCompLoop of compinfo * attributes   (* We pass this when comming across a tcomp info we've seen before *)

type ('typ,'exp,'lval) texp =
    TConst      of constant     
  | TLval       of 'lval          
  | TSizeOf     of 'typ            
  | TSizeOfE    of 'exp             
  | TSizeOfStr  of string
  | TAlignOf    of 'typ              
  | TAlignOfE   of 'exp 
  | TUnOp       of unop * 'exp * 'typ  
  | TBinOp      of binop * 'exp * 'exp * 'typ
  | TCastE      of 'typ * 'exp          
  | TAddrOf     of 'lval  
  | TStartOf    of 'lval   

type ('lhost,'offset) tlval =
    'lhost * 'offset

type ('exp,'varinfo) tlhost = 
  | TVar        of 'varinfo    
  | TMem        of 'exp        

type ('exp,'offset) toffset = 
  | TNoOffset      
  | TField      of fieldinfo * 'offset    
  | TIndex    of 'exp * 'offset

type 'stmtkind tstmt =
    {
      tlabels: label list; 
      tskind: 'stmtkind;    
      tsid: int;           
      tsuccs: stmt list;   
      tpreds: stmt list;   
    } 

type ('exp,'stmt,'instr,'block) tstmtkind = 
  | TInstr  of 'instr list              
  | TReturn of 'exp option * location  
  | TGoto of stmt ref * location     
  | TBreak of location              
  | TContinue of location          
  | TIf of 'exp * 'block * 'block * location 
  | TSwitch of 'exp * 'block * ('stmt list) * location  
  | TLoop of 'block * location * ('stmt option) * ('stmt option) 
  | TBlock of 'block                      
  | TTryFinally of 'block * 'block * location
  | TTryExcept of 'block * ('instr list * 'exp) * 'block * location
      
type ('exp,'lval) tinstr =
    TSet        of 'lval * 'exp * location  
  | TCall       of 'lval option * 'exp * 'exp list * location
  | TAsm        of attributes * string list * (string option * string * lval) list * 
      (string option * string * exp) list * string list * location

type 'stmt tblock = 
    { 
      tbattrs: attributes;  
      tbstmts: 'stmt list;  
    } 

type 'fieldinfo tcompinfo = {
  tcstruct: bool;          
  tcname: string;          
  tckey: int;              
  tcfields: 'fieldinfo list;
  tcattr:   attributes;    
  tcdefined: bool;         
  tcreferenced: bool;      
}

type 'typ tfieldinfo = { 
  tfcomp: compinfo;   
  tfname: string;     
  tftype: 'typ;        
  tfbitfield: int option;   
  tfattr: attributes;       
  tfloc: location;          
}


type 'typ tvarinfo = { 
  tvname: string;	
  tvtype: 'typ;    
  tvattr: attributes;   
  tvstorage: storage;   
  tvglob: bool;	      
  tvinline: bool;
  tvdecl: location;        
  tvid: int;  
  tvaddrof: bool;    
  tvreferenced: bool; 
}

type 'init tinitinfo = {
    tinit : 'init option;
  } 

type ('typ,'exp,'init) tinit = 
  | TSingleInit   of 'exp   
  | TCompoundInit   of 'typ * (offset * 'init) list


type ('varinfo, 'block, 'stmt) tfundec =
    { 
      tsvar:     'varinfo;        
      tsformals: 'varinfo list;   
      tslocals: 'varinfo list;    
      tsmaxid: int;       
      tsbody: 'block;      
      tsmaxstmtid: int option; 
      (* tsallstmts: 'stmt list;  *)
    }











(*============================================================================*)
(* Utility functions that operate on return type data structures *)
(*============================================================================*)

(* Convert a compinfo to a transformer compinfo *)
let ci_to_tci (ci:Cil.compinfo) (fi_rt:'a list) : 'a tcompinfo = 
  {
    tcstruct = ci.cstruct;           
    tcname = ci.cname;          
    tckey = ci.ckey;              
    tcfields = fi_rt;
    tcattr = ci.cattr;    
    tcdefined = ci.cdefined;         
    tcreferenced = ci.creferenced;     
  }

(* Convert a transformer compinfo to a compinfo 
   Write fields back to preserve staring of structure (as fields are mutable)
*)
let tci_to_ci (tci:'a tcompinfo) (fi:fieldinfo list) (ci: compinfo) : unit = 
  ci.cstruct <- tci.tcstruct;           
  ci.cname <- tci.tcname;          
  ci.ckey <- tci.tckey;              
  ci.cfields <- fi;
  ci.cattr <- tci.tcattr;    
  ci.cdefined <- tci.tcdefined;         
  ci.creferenced <- tci.tcreferenced


(* Convert a varinfo  to a transformer varinfo *)
let vi_to_tvi (vi: Cil.varinfo) (typ_rt: 'a) : 'a tvarinfo =
  {  
    tvname = vi.vname;	
    tvtype = typ_rt;    
    tvattr = vi.vattr;   
    tvstorage = vi.vstorage;   
    tvglob = vi.vglob;	      
    tvinline = vi.vinline;
    tvdecl = vi.vdecl;        
    tvid = vi.vid;  
    tvaddrof = vi.vaddrof;    
    tvreferenced = vi.vreferenced; 
  }

(* Convert a transformer varinfo  to a varinfo *)
let tvi_to_vi (tvi: 'a tvarinfo) (typ: Cil.typ) (vi:Cil.varinfo) : unit =  
  vi.vname <- tvi.tvname;	
  vi.vtype <- typ;    
  vi.vattr <- tvi.tvattr;   
  vi.vstorage <- tvi.tvstorage;   
  vi.vglob <- tvi.tvglob;	      
  vi.vinline <- tvi.tvinline;
  vi.vdecl <- tvi.tvdecl;        
  vi.vid <- tvi.tvid;  
  vi.vaddrof <- tvi.tvaddrof;    
  vi.vreferenced <- tvi.tvreferenced


let block_to_tblock (b : Cil.block) (stmts_rt: 'a list) : 'a tblock = 
  { 
    tbattrs = b.battrs;  
    tbstmts = stmts_rt;  
  } 


let tblock_to_block (tb : 'a tblock) (stmts: Cil.stmt list) (b:Cil.block) : unit = 
  b.battrs <- tb.tbattrs;  
  b.bstmts <- stmts



let ci_to_tci (ci: Cil.compinfo) (fil_rt: 'a list) : 'a tcompinfo = 
  {
    tcstruct = ci.cstruct;          
    tcname = ci.cname;          
    tckey = ci.ckey;              
    tcfields = fil_rt;
    tcattr = ci.cattr;    
    tcdefined = ci.cdefined;         
    tcreferenced = ci.creferenced;      
  }
 

let tci_to_ci (tci: 'a tcompinfo) (fil: Cil.fieldinfo list) (ci: Cil.compinfo) : unit = 
  ci.cstruct <- tci.tcstruct;          
  ci.cname <- tci.tcname;          
  ci.ckey <- tci.tckey;              
  ci.cfields <- fil;
  ci.cattr <- tci.tcattr;    
  ci.cdefined <- tci.tcdefined;         
  ci.creferenced <- tci.tcreferenced

 
let fi_to_tfi (fi: Cil.fieldinfo) (ftype_rt: 'a) : 'a tfieldinfo =
  { 
    tfcomp = fi.fcomp;   
    tfname = fi.fname;     
    tftype = ftype_rt;        
    tfbitfield = fi.fbitfield;   
    tfattr = fi.fattr;       
    tfloc = fi.floc;          
  }


let tfi_to_fi (tfi: 'a tfieldinfo) (ftype: Cil.typ) (fi: fieldinfo) : unit =
  fi.fcomp <- tfi.tfcomp;   
  fi.fname <- tfi.tfname;     
  fi.ftype <- ftype;        
  fi.fbitfield <- tfi.tfbitfield;   
  fi.fattr <- tfi.tfattr;       
  fi.floc <- tfi.tfloc


let ti_to_tti (ti: Cil.typeinfo) (t:'a) : 'a ttypeinfo = 
  {
    ttname = ti.tname;              
    tttype = t;
    ttreferenced = ti.treferenced;         
  }


let tti_to_ti (tti: 'a ttypeinfo) (t:Cil.typ) (ti:Cil.typeinfo) : unit = 
    ti.tname <- tti.ttname;              
    ti.ttype <- t;
    ti.treferenced <- tti.ttreferenced


let stmt_to_tstmt (stmt:Cil.stmt) (skind_rt: 'a) : 'a tstmt = 
  {
    tlabels = stmt.labels; 
    tskind = skind_rt;    
    tsid = stmt.sid;           
    tsuccs = stmt.succs;   
    tpreds = stmt.preds;   
  } 

let tstmt_to_stmt (tstmt:'a tstmt) (skind: Cil.stmtkind) (stmt:Cil.stmt) : unit = 
  stmt.labels <- tstmt.tlabels; 
  stmt.skind <- skind;    
  stmt.sid <- tstmt.tsid;           
  stmt.succs <- tstmt.tsuccs;   
  stmt.preds <- tstmt.tpreds




(*============================================================================*)
(* Generic traversal operations *)
(*============================================================================*)



module type ReturnTypes =
sig
  type typ
  type exp
  type lval
  type lhost
  type offset
  type instr
  type stmt
  type stmtkind
  type block 
  type compinfo
  type fieldinfo
  type init
  type initinfo
  type typeinfo
  type varinfo
  type fundec
  type global
end



exception CILTransformEx of string


type process_arg = 
    PATyp of typ | PAExp of exp | PALval of lval | PALhost of lhost | PAOffset of offset
  | PAInstr of instr | PAStmt of stmt | PAStmtkind of stmtkind | PAGlobal of global
  | PAVarinfo of varinfo | PACompinfo of compinfo | PAFieldInfo of fieldinfo
  | PAInitinfo of initinfo | PAInit of init | PAFundec of fundec


module MTransform (M:Monad.SyntaxMonad) (RT:ReturnTypes) = 
struct 

  let (>>=) = M.(>>=)
  let return = M.return 

  module MOps = Monad.MonadOps(M)


  type alt_handlers = 
      {
        should_process: process_arg -> bool;
        dotyp: (Cil.typ -> RT.typ M.m);
        doexp: (Cil.exp -> RT.exp M.m);
        dolval: (Cil.lval -> RT.lval M.m);
        dolhost: (Cil.lhost -> RT.lhost M.m);
        dooffset: (Cil.offset -> lval -> RT.offset M.m);
        doinstr: (Cil.instr -> RT.instr M.m);
        dostmt: Cil.stmt -> RT.stmt M.m;
        dostmtkind: (Cil.stmtkind -> RT.stmtkind M.m);
        doblock: Cil.block -> RT.block M.m;
        docompinfo: Cil.compinfo -> RT.compinfo M.m;
        dofieldinfo: Cil.fieldinfo -> RT.fieldinfo M.m;
        doinit: Cil.init -> RT.init M.m;
        doinitinfo: Cil.initinfo -> RT.initinfo M.m;
        dovarinfo: Cil.varinfo -> RT.varinfo M.m;
        dotypeinfo: Cil.typeinfo -> RT.typeinfo M.m;
        dofundec: Cil.fundec -> RT.fundec M.m;
        doglobal: Cil.global -> RT.global M.m;
      }

  let default_alt_hand hand_name x = 
    raise (CILTransformEx ("No alt handler for " ^ hand_name))
      
  let default_alt_handlers =
    {
      should_process = (fun _ -> true);
      dotyp = default_alt_hand "typ";
      doexp = default_alt_hand "expr";
      dolval = default_alt_hand "lval";
      dolhost = default_alt_hand "lhost";
      dooffset = (fun x -> default_alt_hand "offset");
      doinstr = default_alt_hand "doinstr";
      dostmt = default_alt_hand "dostmt";
      dostmtkind = default_alt_hand "stmtkind";
      doblock = default_alt_hand "block";
      docompinfo = default_alt_hand "compinfo";
      dofieldinfo = default_alt_hand "fieldinfo";
      doinit = default_alt_hand "init";
      doinitinfo = default_alt_hand "initinfo";
      dovarinfo = default_alt_hand "varinfo";
      dotypeinfo = default_alt_hand "typeinfo";
      dofundec = default_alt_hand "fundec";
      doglobal = default_alt_hand "global";
    };;

(*
  module FoldLFuns = 
  struct
    val fTyp: Cil.typ -> RT.typ ttyp -> RT.typ M.m;
    val fExpr: Cil.exp -> (RT.typ,RT.exp,RT.lval) texp -> RT.exp M.m;
    val fLval: Cil.lval -> RT.lval tlval -> RT.lval M.m;
    val fLhost: Cil.lhost -> RT.offset tlhost -> RT.lhost M.m;
    val fOffset: Cil.offset -> RT.offset toffset -> RT.offset M.m;
    val fInstr: Cil.instr -> (RT.exp,RT.lval) tinstr -> RT.instr M.m;
    val fStmt: Cil.stmt -> RT.stmt tstmt -> RT.stmt M.m;
    val fStmtKind: Cil.stmtkind -> (RT.stmt,RT.instr,RT.block) tstmtkind -> RT.stmtkind M.m;
    val fGlobal: Cil.global -> tglobal -> RT.global M.m;
  end
*)

  module type FoldFuns = 
  sig
    val ftyp:  typ -> (RT.typ,RT.exp,typeinfo,compinfo) ttyp -> RT.typ M.m
    val fexp: exp -> (RT.typ,RT.exp,RT.lval) texp -> RT.exp M.m
    val flval: lval -> (RT.lhost,RT.offset) tlval -> RT.lval M.m
    val flhost: lhost -> (RT.exp,RT.varinfo) tlhost -> RT.lhost M.m
    val foffset: offset -> (RT.exp,RT.offset) toffset -> lval -> RT.offset M.m
    val finstr: instr -> (RT.exp,RT.lval) tinstr -> RT.instr M.m
    val fstmtkind: stmtkind -> (RT.exp,RT.stmt,RT.instr,RT.block) tstmtkind -> RT.stmtkind M.m
    val finit:  init -> (RT.typ,RT.exp,RT.init) tinit -> RT.init M.m
    val fglobal: global -> (RT.typ,RT.typeinfo,RT.compinfo,RT.varinfo,RT.initinfo,RT.fundec) tglobal -> RT.global M.m

    (* These are records that need to be updated inplace to preserve sharing,
       so pass the original structure as well *)
    val ffieldinfo: fieldinfo -> RT.typ tfieldinfo -> RT.fieldinfo M.m
    val fcompinfo: compinfo -> RT.fieldinfo tcompinfo -> RT.compinfo M.m
    val fblock : block -> RT.stmt tblock -> RT.block M.m 
    val fstmt: stmt -> RT.stmtkind tstmt -> RT.stmt M.m
    val finitinfo: initinfo -> (RT.init tinitinfo) -> RT.initinfo M.m
    val fvarinfo: varinfo -> RT.typ tvarinfo -> RT.varinfo M.m
    val ffundec: fundec -> (RT.varinfo,RT.block,RT.stmt) tfundec -> RT.fundec M.m
    val ftypeinfo: typeinfo -> (RT.typ ttypeinfo) -> RT.typeinfo M.m
  end





  module MFold (FF:FoldFuns) = 
  struct

    let rec typFoldWithAlt (alt: alt_handlers) (typ': Cil.typ) : RT.typ M.m = 

      let rec typFold' (typ: Cil.typ) : RT.typ M.m = 
        if alt.should_process (PATyp typ) then 
          match typ with 
              TVoid al            -> FF.ftyp typ (TTVoid al)
            | TInt (ikind,al)     -> FF.ftyp typ (TTInt (ikind,al))
            | TFloat (fkind,al)   -> FF.ftyp typ (TTFloat (fkind,al))

            (* Note: Don't process ti and ei here as they have already been 
               processed at global level.
               If we didn't start at the global level we may want to process these here.
               Should it be the default or the exception to do them here? *)
            | TNamed (ti,al)      -> FF.ftyp typ (TTNamed (ti,al))
            | TEnum (ei,al)       -> FF.ftyp typ (TTEnum (ei,al))

            | TBuiltin_va_list al -> FF.ftyp typ (TTBuiltin_va_list al)
                
            | TPtr(t, al)       -> 
                typFold' t >>= fun rt ->
                FF.ftyp typ (TTPtr (rt,al))

            | TArray(t, e_opt, al) -> 
                typFold' t >>= fun rt ->
                MOps.apply_opt (exprFoldWithAlt alt) e_opt >>= fun e_opt_r ->
                FF.ftyp typ (TTArray (rt,e_opt_r,al))
                
            | TComp(ci, al) ->
                (* NOTE: Compinfo processed at global level - we don't process here *)
                FF.ftyp typ (TTComp(ci, al))
                
            | TFun(tret, targs, isva, al) ->
                let args_map (s, t, a) = 
                  typFold' t >>= fun t_r ->
                  return (s,t_r,a)
                in 
                  begin match targs with 
                      Some l -> MOps.map args_map l >>= fun l' -> return (Some l')
                    | None   -> return None
                  end >>= fun targs_r ->
                  typFold' tret >>= fun tret_r ->
                  FF.ftyp typ (TTFun(tret_r,targs_r,isva,al))

        else alt.dotyp typ
      in
        typFold' typ' 
          


    and exprFoldWithAlt (alt: alt_handlers) (exp: Cil.exp) : RT.exp M.m =
      if alt.should_process (PAExp exp) then  
        match exp with
            Const c       -> FF.fexp exp (TConst c) 
          | Lval lval     -> lvalFoldWithAlt alt lval >>= fun lval_r -> FF.fexp exp (TLval lval_r) 
          | SizeOf typ    -> typFoldWithAlt alt typ   >>= fun typ_r -> FF.fexp exp (TSizeOf typ_r)

          | SizeOfE e     -> exprFoldWithAlt alt e >>= fun (exp_r:RT.exp) -> 
                             FF.fexp exp (TSizeOfE exp_r)

          | SizeOfStr str -> FF.fexp exp (TSizeOfStr str)
          | AlignOf typ   -> typFoldWithAlt alt typ   >>= fun typ_r -> FF.fexp exp (TAlignOf typ_r)
          | AlignOfE e    -> exprFoldWithAlt alt e    >>= fun exp_r -> FF.fexp exp (TAlignOfE exp_r)
          | UnOp (unop, e, typ) -> 
              exprFoldWithAlt alt e >>= fun e_r ->
              typFoldWithAlt alt typ >>= fun typ_r ->
              FF.fexp exp(TUnOp(unop,e_r,typ_r))
                  
          | BinOp (binop, e1, e2, typ) -> 
              exprFoldWithAlt alt e1 >>= fun e1_r ->
              exprFoldWithAlt alt e2 >>= fun e2_r ->
              typFoldWithAlt alt typ >>= fun typ_r ->
              FF.fexp exp (TBinOp(binop,e1_r,e2_r,typ_r))
                    
          | CastE (typ, e) -> 
              exprFoldWithAlt alt e >>= fun e_r ->
              typFoldWithAlt alt typ >>= fun typ_r ->
              FF.fexp exp (TCastE(typ_r,e_r))

          | AddrOf lval  -> lvalFoldWithAlt alt lval >>= fun lval_r -> FF.fexp exp (TAddrOf lval_r) 
          | StartOf lval -> lvalFoldWithAlt alt lval >>= fun lval_r -> FF.fexp exp (TStartOf lval_r)
  
      else alt.doexp exp


    and lvalFoldWithAlt (alt: alt_handlers) ((lhost,offset): Cil.lval) : RT.lval M.m =
      if alt.should_process (PALval(lhost,offset))  then 
        lhostFoldWithAlt alt lhost >>= fun lhost_r ->
        offsetFoldWithAlt alt offset (lhost,NoOffset) >>= fun offset_r ->
        FF.flval (lhost,offset) (lhost_r,offset_r)
      else alt.dolval (lhost,offset)


    and lhostFoldWithAlt (alt: alt_handlers) (lhost: Cil.lhost) : RT.lhost M.m =
      if alt.should_process (PALhost lhost) then 
        match lhost with 
            Mem (e) -> exprFoldWithAlt alt e >>= fun e_r -> FF.flhost lhost (TMem e_r)
          | Var(vi) -> viFoldWithAlt alt vi >>= fun vi_r -> FF.flhost lhost (TVar vi_r) 
      else alt.dolhost lhost


    and viFoldWithAlt (alt: alt_handlers) (vi: Cil.varinfo) : RT.varinfo M.m =
      if alt.should_process (PAVarinfo vi) then
        typFoldWithAlt alt vi.vtype >>= fun vtype_r ->
        FF.fvarinfo vi (vi_to_tvi vi vtype_r)
      else alt.dovarinfo vi


    (*
      This one doesn't quite follow the rule for fold functions because of the awkward datastructure
      CIL has for lvalues.  We pass an extra lvalue that the current offset is offsetting from.
    *)
    and offsetFoldWithAlt (alt: alt_handlers) (offset: Cil.offset) ((lhost,loff):Cil.lval) : RT.offset M.m =
      let rec offset_add offset extra_offset = 
        match offset with 
            NoOffset -> extra_offset
          | Field(fi,offset') -> Field(fi, offset_add offset' extra_offset)
          | Index(e,offset')  -> Index(e, offset_add offset' extra_offset)
      in 
        if alt.should_process (PAOffset offset) then 
          match offset with 
              NoOffset -> FF.foffset offset TNoOffset (lhost,loff)        
                
            | Field (fi, offset') -> 
                offsetFoldWithAlt alt offset' (lhost, offset_add loff (Field (fi, NoOffset)))  >>= fun offset'_r ->
                  FF.foffset offset (TField (fi, offset'_r)) (lhost,loff)
            
            | Index (exp, offset') -> 
                exprFoldWithAlt alt exp >>= fun exp_r -> 
                offsetFoldWithAlt alt offset' (lhost, offset_add loff (Index (exp, NoOffset))) >>= fun offset'_r ->
                FF.foffset offset (TIndex (exp_r,offset'_r)) (lhost,loff)

        else alt.dooffset offset (lhost,loff)



    and instrFoldWithAlt (alt: alt_handlers) (instr: Cil.instr) : RT.instr M.m =
      M.setLoc (Cil.get_instrLoc instr) >>= fun () ->
      if alt.should_process (PAInstr instr) then 
        match instr with 
            Set(lval,exp,loc) ->
              lvalFoldWithAlt alt lval >>= fun lval_r ->
              exprFoldWithAlt alt exp  >>= fun exp_r ->
              FF.finstr instr (TSet(lval_r,exp_r,loc))

          | Call(lval_opt, exp, exp_args, loc) ->
	      (* ignore (Updatedebug.log [Updatedebug.BTSuspect] *)
	      (* "fold call: %s:%d %a\n" loc.file loc.line d_instr instr); *)
              begin match lval_opt with 
                  Some lval -> lvalFoldWithAlt alt lval >>= fun lval_r -> return (Some lval_r)
                    | None -> return None
              end >>= fun lval_opt_r ->
              exprFoldWithAlt alt exp >>= fun exp_r ->
              MOps.map (exprFoldWithAlt alt) exp_args >>= fun exp_args_r ->
              FF.finstr instr (TCall(lval_opt_r, exp_r, exp_args_r, loc))

          (* NOTE: Don't process lval and exp inside of asm *)
          | Asm (a,b,c,d,e,f) -> FF.finstr instr (TAsm (a,b,c,d,e,f))

    else alt.doinstr instr


    and blockFoldWithAlt (alt: alt_handlers) (block: Cil.block) : RT.block M.m =
      MOps.map (stmtFoldWithAlt alt) block.bstmts >>= fun bstmts_r ->
      FF.fblock block (block_to_tblock block bstmts_r) 


    and stmtFoldWithAlt (alt: alt_handlers) (stmt: Cil.stmt) : RT.stmt M.m =
      M.setLoc (Cil.get_stmtLoc stmt.skind) >>= fun () ->
      if alt.should_process (PAStmt stmt) then
        stmtkindFoldWithAlt alt stmt.skind >>= fun skind_r ->
        FF.fstmt stmt (stmt_to_tstmt stmt skind_r) 
      else alt.dostmt stmt


    and stmtkindFoldWithAlt (alt: alt_handlers) (skind: Cil.stmtkind) : RT.stmtkind M.m =
      if alt.should_process (PAStmtkind skind) then 
        match skind with 
            Instr il -> MOps.map (instrFoldWithAlt alt) il >>= fun il_r ->
                        FF.fstmtkind skind (TInstr il_r)
 
          | Return (None,loc) -> FF.fstmtkind skind (TReturn (None,loc))
          | Return (Some e, loc) -> exprFoldWithAlt alt e >>= fun e_r -> 
                                    FF.fstmtkind skind (TReturn (Some e_r,loc))

          | Goto (stmt_ref,loc) -> FF.fstmtkind skind (TGoto (stmt_ref,loc)) 
          | Break loc -> FF.fstmtkind skind (TBreak loc)  
          | Continue loc -> FF.fstmtkind skind (TContinue loc)
          | If (eTest,blockT,blockF,loc) ->
              exprFoldWithAlt alt eTest >>= fun eTest_r ->
              blockFoldWithAlt alt blockT >>= fun blockT_r ->
              blockFoldWithAlt alt blockF >>= fun blockF_r ->
              FF.fstmtkind skind (TIf(eTest_r,blockT_r,blockF_r,loc))
      
          | Switch (exp, block,stmt_list,loc) ->
            M.getLoc >>= fun current_loc ->
            Updatedebug.abort (Updatedebug.bug 
                                 "(%a) stmtkindFoldL: CIL should have eliminated switch!\n" 
                                 d_loc current_loc)
       
          | Loop (block,loc,s1_opt,s2_opt) -> 
              blockFoldWithAlt alt block >>= fun block_r ->
              MOps.apply_opt (stmtFoldWithAlt alt) s1_opt >>= fun s1_opt_r ->
              MOps.apply_opt (stmtFoldWithAlt alt) s2_opt >>= fun s2_opt_r ->
              FF.fstmtkind skind (TLoop(block_r,loc,s1_opt_r,s2_opt_r))
  
          | Block block -> blockFoldWithAlt alt block >>= fun block_r ->
                           FF.fstmtkind skind (TBlock block_r)
       
          | TryFinally  (block1,block2,loc) ->
              M.getLoc >>= fun current_loc ->
              Updatedebug.abort (Updatedebug.bug 
                                   "(%a) stmtKindFoldL: __try/__except not supported!\n" 
                                   d_loc current_loc)
  
          | TryExcept (block_try, inst_list, block_except, loc) ->
              M.getLoc >>= fun current_loc ->
              Updatedebug.abort (Updatedebug.bug 
                                   "(%a) stmtKindFoldL: __try/__except not supported!\n"
                                   d_loc current_loc)
   
       else alt.dostmtkind skind


    and fieldinfoFoldWithAlt (alt: alt_handlers) (fi:Cil.fieldinfo) : RT.fieldinfo M.m = 
      if alt.should_process (PAFieldInfo fi) then 
        typFoldWithAlt alt fi.ftype >>= fun ftyp_r ->
        FF.ffieldinfo fi (fi_to_tfi fi ftyp_r)
      else alt.dofieldinfo fi


    and compinfoFoldWithAlt (alt: alt_handlers) (ci:Cil.compinfo) : RT.compinfo M.m = 
      if alt.should_process (PACompinfo ci) then 
        MOps.map (fieldinfoFoldWithAlt alt) ci.cfields >>= fun cfields_r ->
        FF.fcompinfo ci (ci_to_tci ci cfields_r)
      else alt.docompinfo ci


    and initinfoFoldWithAlt (alt: alt_handlers) (ii:Cil.initinfo) : RT.initinfo M.m =
      if alt.should_process (PAInitinfo ii) then 
        match ii with 
            {init=None} -> FF.finitinfo ii {tinit=None}
          | {init=Some i} -> initFoldWithAlt alt i >>= fun i_r ->
                             FF.finitinfo ii {tinit = Some i_r}

      else alt.doinitinfo ii


    and initFoldWithAlt (alt: alt_handlers) (i:Cil.init) : RT.init M.m =
      let doCompoundInit ((o,i) : (Cil.offset*Cil.init)) : (Cil.offset*RT.init) M.m  = 
        (* NOTE: We don't process offsets here.  They are semantically different to 
           lval offsets and we should provide a separate function to iter them if we desire.
           offsetFoldWithAlt alt o >>= fun (o_r:RT.offset) -> *)
        initFoldWithAlt alt i >>= fun (i_r:RT.init) ->
          return (o, i_r)
      in
        if alt.should_process (PAInit i) then 
          match i with 
              SingleInit e -> exprFoldWithAlt alt e >>= fun e_r -> FF.finit i (TSingleInit e_r)
            | CompoundInit (t,comp_init_list) -> 
                MOps.map doCompoundInit comp_init_list >>= fun (comp_init_list_r:(Cil.offset*RT.init) list) ->
                typFoldWithAlt alt t >>= fun (t_r:RT.typ) ->
                FF.finit i (TCompoundInit (t_r,comp_init_list_r))

      else alt.doinit i
      

    and fundecFoldWithAlt (alt: alt_handlers) (fd: Cil.fundec) : RT.fundec M.m =
      if alt.should_process (PAFundec fd) then 
        viFoldWithAlt alt fd.svar >>= fun svar_r ->
        MOps.map (viFoldWithAlt alt) fd.sformals >>= fun sformals_r ->
        MOps.map (viFoldWithAlt alt) fd.slocals >>= fun slocals_r ->
        blockFoldWithAlt alt fd.sbody >>= fun sbody_r ->
	(* MWH: we end up processing everything twice if we do this. *)
	(* Ultimately we should parameterize yes/no *)
        (* MOps.map (stmtFoldWithAlt alt) fd.sallstmts >>= fun allstmts_r -> *)
        FF.ffundec fd
          { 
            tsvar = svar_r;        
            tsformals = sformals_r;   
            tslocals = slocals_r;    
            tsmaxid = fd.smaxid;       
            tsbody = sbody_r;      
            tsmaxstmtid = fd.smaxstmtid; 
            (* tsallstmts = allstmts_r;  *)
          }

      else alt.dofundec fd


    and typeinfoFoldWithAlt (alt:alt_handlers) (ti:Cil.typeinfo) : RT.typeinfo M.m = 
      typFoldWithAlt alt ti.ttype >>= fun typ_r ->
      FF.ftypeinfo ti (ti_to_tti ti typ_r)


    and globalFoldWithAlt (alt: alt_handlers) (global: Cil.global) : RT.global M.m =
      M.setLoc (Cil.get_globalLoc global) >>= fun () ->
      if alt.should_process (PAGlobal global) then 
        match global with 
            GType (ti, loc) -> typeinfoFoldWithAlt alt ti >>= fun tti -> FF.fglobal global (TGType(tti,loc))

          | GCompTag (ci, loc)  -> compinfoFoldWithAlt alt ci >>= fun ci_r -> FF.fglobal global (TGCompTag(ci_r,loc))
          | GCompTagDecl (ci, loc) -> compinfoFoldWithAlt alt ci >>= fun ci_r -> FF.fglobal global (TGCompTagDecl(ci_r,loc))

          | GEnumTag (ei,loc) -> FF.fglobal global (TGEnumTag(ei,loc))
          | GEnumTagDecl (ei,loc) -> FF.fglobal global (TGEnumTagDecl(ei,loc))

          | GVarDecl (vi,loc) -> viFoldWithAlt alt vi >>= fun vi_r -> FF.fglobal global (TGVarDecl(vi_r,loc))

          | GVar  (vi,ii,loc) -> 
              viFoldWithAlt alt vi >>= fun vi_r ->
              initinfoFoldWithAlt alt ii >>= fun ii_r ->
              FF.fglobal global (TGVar(vi_r,ii_r,loc))

          | GFun (fd,loc) -> 
              M.setCurrentFun fd >>= fun () ->
              fundecFoldWithAlt alt fd >>= fun fd_r ->
              M.unsetCurrentFun >>= fun () ->
              FF.fglobal global (TGFun(fd_r,loc))
              
          (* NOTE: don't process exp and lval in asm *)
          | GAsm (a,b)        -> FF.fglobal global (TGAsm (a,b))
          | GPragma (a,b)     -> FF.fglobal global (TGPragma (a,b)) 
          | GText x       -> FF.fglobal global (TGText x)
  
      else alt.doglobal global                 

  end (* MFold *)



(* ---- Derived iterators ---- *)

  (* ---- MAP ---- *)
  module MapDefaultFuns = 
  struct
    let ftyp (typ:typ) (ttyp: (typ,exp,typeinfo,compinfo) ttyp) : typ M.m =
      return (
        match ttyp with 
            TTVoid(al) -> TVoid al
          | TTInt(ik,al)  -> TInt(ik,al)
          | TTFloat(fk,al) -> TFloat(fk,al)
          | TTPtr(t,al) -> TPtr(t,al)
          | TTArray(t,e_opt,al) -> TArray(t,e_opt,al)
          | TTFun(tret,targs,isva,al) -> TFun(tret,targs,isva,al)
          | TTNamed(ti,al) -> TNamed(ti,al)
          | TTComp(ci,al) -> TComp(ci,al)
          | TTEnum(ei,al) -> TEnum(ei,al)
          | TTBuiltin_va_list(al) -> TBuiltin_va_list al
          | TTCompLoop (ci,al) -> TComp(ci,al)
      )

    let fexp (exp:exp) (texp: (typ,exp,lval) texp) : exp M.m = 
      return (
        match texp with 
            TConst c  ->  Const c
          | TLval l  ->  Lval l
          | TSizeOf t  ->  SizeOf t
          | TSizeOfE e  ->  SizeOfE e
          | TSizeOfStr s  ->  SizeOfStr s
          | TAlignOf t  ->  AlignOf t
          | TAlignOfE e  ->  AlignOfE e
          | TUnOp(uop,e,t)  ->  UnOp(uop,e,t)
          | TBinOp(bop,e1,e2,t)  ->  BinOp(bop,e1,e2,t)
          | TCastE(t,e)  ->  CastE(t,e)
          | TAddrOf l  ->  AddrOf l
          | TStartOf l  ->  StartOf l
      )
    
    let flval (lval:lval) (tlval: (lhost,offset) tlval) : lval M.m = return tlval
          
    let flhost (lhost:lhost)(tlhost: (exp,varinfo) tlhost) : lhost M.m =
      return (
        match tlhost with 
          | TVar vi -> Var vi
          | TMem e -> Mem e
      )

    let foffset (offset:offset) (toffset: (exp,offset) toffset) : offset M.m =
      return (
        match toffset with 
          | TNoOffset      -> NoOffset      
          | TField (fi,o)  -> Field (fi,o)
          | TIndex (e,o)   -> Index (e,o)
      )

    let finstr (instr:instr) (tinstr: (exp,lval) tinstr) : instr M.m =
      return (      
        match tinstr with       
            TSet(l,e,loc) -> Set(l,e,loc)
          | TCall(l_opt,e,el,loc) -> Call(l_opt,e,el,loc)
          | TAsm(a,b,c,e,d,f) -> Asm(a,b,c,e,d,f)  
      )

    let fstmt (stmt:stmt) (tstmt: stmtkind tstmt) : stmt M.m =
      stmt.labels <- tstmt.tlabels; 
      stmt.skind <- tstmt.tskind;    
      stmt.sid <- tstmt.tsid;           
      stmt.succs <- tstmt.tsuccs;   
      stmt.preds <- tstmt.tpreds;
      return stmt

    let fstmtkind (skind:stmtkind) (tstmtkind: (exp,stmt,instr,block) tstmtkind) : stmtkind M.m =
      return (
        match tstmtkind with
          | TInstr il  ->  Instr il
          | TReturn(e_opt,loc)  ->  Return(e_opt,loc)
          | TGoto(s_ref,loc)  ->  Goto(s_ref,loc)
          | TBreak loc  ->  Break loc
          | TContinue loc  ->  Continue loc
          | TIf(e,b1,b2,loc)  ->  If(e,b1,b2,loc)
              
          | TLoop(b,loc,s_opt2,s_opt1)  ->  Loop(b,loc,s_opt2,s_opt1)
          | TBlock b  ->  Block b
              
          | TSwitch _
          | TTryFinally _
          | TTryExcept _   -> raise (CILTransformEx "Map hit Switch/TryFinally/TryExcept")
      )

    let fblock (b:block) (tblock: stmt tblock) : block M.m  =
      b.battrs <- tblock.tbattrs;
      b.bstmts <- tblock.tbstmts;
      return b

    let fcompinfo (ci:compinfo) (tcompinfo: fieldinfo tcompinfo) : compinfo M.m =
      MOps.lift (tci_to_ci tcompinfo tcompinfo.tcfields) (return ci) >>= fun () ->
      return ci

    let ffieldinfo (fi:fieldinfo) (tfieldinfo: typ tfieldinfo) : fieldinfo M.m =
      MOps.lift (tfi_to_fi tfieldinfo tfieldinfo.tftype) (return fi) >>= fun () ->
      return fi

    let finit (init:init) (tinit: (typ,exp,init) tinit) : init M.m =
      return (      
        match tinit with 
          | TSingleInit e -> SingleInit e
          | TCompoundInit(t,io_list) -> CompoundInit(t,io_list)
      )

    let finitinfo (ii:initinfo) (tinitinfo: (init tinitinfo)) : initinfo M.m =
      begin
        ii.init <- tinitinfo.tinit;
        return ii
      end

    let fvarinfo (vi: varinfo) (tvarinfo: typ tvarinfo) : varinfo M.m =
      begin
        tvi_to_vi tvarinfo tvarinfo.tvtype vi;
        return vi
      end

    let ftypeinfo (ti:typeinfo) (ttypeinfo: typ ttypeinfo) : typeinfo M.m = 
      tti_to_ti ttypeinfo ttypeinfo.tttype ti;
      return ti

    let ffundec (fd:fundec) (tfundec: (varinfo,block,stmt) tfundec) : fundec M.m =
      begin  
        fd.svar <-  tfundec.tsvar;
        fd.sformals <- tfundec.tsformals;   
        fd.slocals <- tfundec.tslocals;    
        fd.smaxid <- tfundec.tsmaxid;      
        fd.sbody <- tfundec.tsbody;  
        fd.smaxstmtid <- tfundec.tsmaxstmtid; 
        (* fd.sallstmts <- tfundec.tsallstmts;  *)
        return fd
      end

    let fglobal (global:global) (tglobal: (typ,typeinfo,compinfo,varinfo,initinfo,fundec) tglobal) : global M.m =
      return (
        match tglobal with 
          | TGType (ti,loc) -> GType (ti,loc)
          | TGCompTag (ci,loc) -> GCompTag (ci,loc)
          | TGCompTagDecl (ci,loc) -> GCompTagDecl (ci,loc)
          | TGEnumTag (ei,loc) -> GEnumTag (ei,loc)
          | TGEnumTagDecl (ei,loc) -> GEnumTagDecl (ei,loc)
          | TGVarDecl (vi,loc) -> GVarDecl (vi,loc)
          | TGVar(vi,ii,loc) -> GVar(vi,ii,loc)
          | TGFun(fd,loc) -> GFun(fd,loc)
          | TGAsm(s,loc) -> GAsm(s,loc)
          | TGPragma (al,loc) -> GPragma (al,loc)
          | TGText s        -> GText s       
      )

  end


  (**)
  module IterDefaultFuns = 
  struct
    let ftyp (typ:typ) (ttyp: (unit,unit,typeinfo,compinfo) ttyp) : unit M.m = return ()
    let fexp (exp:exp) (texp: (unit,unit,unit) texp) : unit M.m =  return ()
    let flval (lval:lval) (tlval: (unit,unit) tlval) : unit M.m =  return ()
    let flhost (lhost:lhost) (tlhost: (unit,unit) tlhost) : unit M.m = return ()
    let foffset (offset:offset) (toffset: (unit,unit) toffset) : unit M.m = return ()
    let finstr (instr:instr) (tinstr: (unit,unit) tinstr) : unit M.m = return ()
    let fstmt (stmt:stmt) (tstmt: unit tstmt) : unit M.m = return ()
    let fstmtkind (skind:stmtkind) (tstmtkind: (unit,unit,unit,unit) tstmtkind) : unit M.m = return ()
    let fblock (b:block) (tblock: unit tblock) : unit M.m  = return ()
    let fcompinfo (ci:compinfo) (tcompinfo: unit tcompinfo) : unit M.m = return ()
    let ffieldinfo (fi:fieldinfo) (tfieldinfo: unit tfieldinfo) : unit M.m = return ()
    let finit (init:init) (tinit: (unit,unit,unit) tinit) : unit M.m = return ()
    let finitinfo (ii:initinfo) (tinitinfo: (unit tinitinfo)) : unit M.m = return ()
    let fvarinfo (vi: varinfo) (tvarinfo: unit tvarinfo) : unit M.m = return ()
    let ftypeinfo (ti:typeinfo) (ttypeinfo: unit ttypeinfo) : unit M.m =  return ()
    let ffundec (fd:fundec) (tfundec: (unit,unit,unit) tfundec) : unit M.m = return ()
    let fglobal (global:global) (tglobal: (unit,unit,unit,unit,unit,unit) tglobal) : unit M.m = return ()
  end







end (* MTransform *)



(* Return types for building map operation out of fold *)
module MapReturnTypes = 
struct
  type typ       = Cil.typ
  type exp       = Cil.exp
  type lval      = Cil.lval
  type lhost     = Cil.lhost
  type offset    = Cil.offset
  type instr     = Cil.instr
  type stmt      = Cil.stmt
  type stmtkind  = Cil.stmtkind
  type block     = Cil.block
  type compinfo  = Cil.compinfo
  type fieldinfo = Cil.fieldinfo
  type init      = Cil.init
  type initinfo  = Cil.initinfo
  type typeinfo  = Cil.typeinfo
  type varinfo   = Cil.varinfo
  type fundec    = Cil.fundec
  type global    = Cil.global
end


(* Return types for building iter operation out of fold *)
module IterReturnTypes = 
struct
  type typ       = unit
  type exp       = unit
  type lval      = unit
  type lhost     = unit
  type offset    = unit
  type instr     = unit
  type stmt      = unit
  type stmtkind  = unit
  type block     = unit
  type compinfo  = unit
  type fieldinfo = unit
  type init      = unit
  type initinfo  = unit
  type typeinfo  = unit
  type varinfo   = unit
  type fundec    = unit
  type global    = unit
end


 



(* Type fold *)
module TypFold (M:Monad.SyntaxMonad) = 
struct
  let (>>=) = M.(>>=)
  let return = M.return 
    
  module MOps = Monad.MonadOps(M)

  type ('typ,'typeinfo,'compinfo,'fieldinfo) typFoldFuns = 
      {
        tf_typ : ('typ,Cil.exp,'typeinfo,'compinfo) ttyp -> 'typ M.m;
        tf_ti  : typeinfo -> 'typ ttypeinfo -> 'typeinfo M.m;
        tf_ci  : compinfo -> 'fieldinfo tcompinfo -> 'compinfo M.m;
        tf_fi  : fieldinfo -> 'typ tfieldinfo -> 'fieldinfo M.m;
	tf_typ_alt : Cil.typ -> 'typ option;
      }

  let defaultTypFoldFuns = 
    {
      tf_typ = (fun _ -> raise (CILTransformEx "No type fold function!"));
      tf_typ_alt = (fun _ -> None);
      tf_ti = (fun ti tti -> return ti);
      tf_ci = (fun ci tci -> return ci);
      tf_fi = (fun fi tfi -> return fi);
    }

  let iterDefaultTypFoldFuns = 
    {
      tf_typ = (fun _ -> raise (CILTransformEx "No type fold function!"));
      tf_typ_alt = (fun _ -> None);
      tf_ti = (fun ti tti -> return ());
      tf_ci = (fun ci tci -> return ());
      tf_fi = (fun fi tfi -> return ());
    }
    
  (* check whether the type contains effects *)
  let typFold (tff: ('typ,'typeinfo,'compinfo,'fieldinfo) typFoldFuns) (typ : Cil.typ) : 'a M.m=
    let compHash = H.create 13 in  

    let rec typFold' (typ' : Cil.typ) : 'typ M.m =
      let args_map (s,t,al) = typFold' t >>= fun t_r -> return (s, t_r, al) in
      match (tff.tf_typ_alt typ') with
	Some v -> return v
      | None ->
	  begin
      match typ' with 
          TVoid al -> tff.tf_typ (TTVoid al) 
        | TInt (ik,al) -> tff.tf_typ (TTInt (ik,al))
        | TFloat (fk,al) -> tff.tf_typ (TTFloat (fk,al))
        | TEnum (ei,al) -> tff.tf_typ (TTEnum (ei,al))
        | TBuiltin_va_list al -> tff.tf_typ (TTBuiltin_va_list al)
        | TPtr (t, al) -> typFold' t >>= fun t_r -> tff.tf_typ (TTPtr(t_r,al))
        | TArray(t, e_opt, al) -> typFold' t >>= fun t_r -> tff.tf_typ (TTArray(t_r,e_opt,al))
        | TFun(tret, targs, isva, al) -> 
            MOps.apply_opt (MOps.map args_map) targs >>= fun targs_r ->
            typFold' tret >>= fun tret_r ->
            tff.tf_typ (TTFun(tret_r, targs_r, isva, al))
        | TNamed(ti, al) ->
            let tiFold (ti:typeinfo) : 'typeinfo M.m = 
              typFold' ti.ttype >>= fun ttype_r ->
              tff.tf_ti ti (ti_to_tti ti ttype_r)
            in
              tiFold ti >>= fun ti_r -> 
              tff.tf_typ (TTNamed(ti_r,al))

        | TComp(ci,al) -> 
            if H.mem compHash ci.cname 
            then 
              tff.tf_typ (TTCompLoop (ci,al))
            else 
              
              let fiFold (fi:fieldinfo) : 'fieldinfo M.m =
                typFold' fi.ftype >>= fun ftype_r ->
                tff.tf_fi fi (fi_to_tfi fi ftype_r) in

              let ciFold (ci:compinfo) : 'compinfo M.m =  
                begin
                  H.add compHash ci.cname true; 
                  MOps.map fiFold ci.cfields >>= fun cfields_r ->
                  tff.tf_ci ci (ci_to_tci ci cfields_r) 
                end
              in
                ciFold ci >>= fun ci_r ->
                tff.tf_typ (TTComp(ci_r,al))
	  end
    in
      typFold' typ

end
