(* 
 *
 * Copyright (c) 2004-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *  Eric Hardisty       <hardisty@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(* Source transformation to insert con/abs coercions *)

(*

TODO: Cil 1.3.2 renders enums as integers, losing the type information.
      Cil 1.3.6 adds CEnum's to preserve this; need to update enum handling accordingly


TODO: We need a secons pass to remove coercions for non-updatable types.  Should be easy:
         - tmp = abs(e)  --> tmp = e
         - con(e) --> <nothing>

*)


open Cil


(* Monadic plumbing for con/abs insertion ------------------------------ *)
type cai_t = { cai_fd       : Cil.fundec; 
               cai_il       : Cil.instr list;
               cai_loc      : Cil.location ref; 
               cai_ppInstrs : bool ref;
               cai_mallocFunHash : Mallocpragma.mallocHash_t;
               (*cai_vd       : VersionData.version_data;*)
             };;

type 'a cao_t = { cao_val : 'a;
                  cao_il : Cil.instr list };;

type 'a cam = cai_t -> 'a cao_t;;

let return (x:'a) = fun cai -> {cao_val = x; cao_il = cai.cai_il};;

let run (mallocFunHash : Mallocpragma.mallocHash_t) (fd:Cil.fundec) (loc : Cil.location) (m : 'a cam) = 
   let cao = m { cai_fd = fd; cai_il = []; cai_loc = ref loc; cai_ppInstrs = ref false; cai_mallocFunHash = mallocFunHash } in 
   (cao.cao_val, cao.cao_il);;

let (>>=) (m : 'a cam) (f : 'a -> 'b cam) = 
   fun (cai : cai_t) -> 
      (let cao = m cai in
      begin
      if !(cai.cai_ppInstrs) then 
         let len = (List.length cao.cao_il) - (List.length cai.cai_il) in 
         Updatedebug.ppInstrList (Commonutil.list_last_n cao.cao_il len)
      end;
      f cao.cao_val {cai with cai_il = cao.cao_il})

let (>>-) (m1 : unit cam) (m2 : 'b cam) = 
   fun (cai : cai_t) ->
      let cao1 = m1 cai in
      let cao2 = m2 {cai with cai_il = []} in 
      {cao2 with cao_il = cao1.cao_il @ cao2.cao_il};;

let lvalToExp lval = return (Lval lval)

let addIL il = fun cai -> return () {cai with cai_il = cai.cai_il @ il};; 
let getLoc = fun cai -> return !(cai.cai_loc) cai;; 
let setLoc loc = fun cai -> cai.cai_loc := loc; return () cai;; 

let ppInstrs (onOff:bool) = fun cai -> cai.cai_ppInstrs := onOff; return () cai;;

let expToLval e = 
   match e with 
      Lval e -> return e
    | _ -> Updatedebug.abort (Updatedebug.bug "Tried to convert expression %a to lval, but failed\n" d_plainexp e)

let getFd = fun cai -> return cai.cai_fd cai;;
let mkTmp t fd =  return (Cil.makeTempVar fd t)
let mkTmpVar t fd  = return (var (Cil.makeTempVar fd t))

let ulvalToLval ulval = return (Updatecilutil.ulvalToLval ulval)
let lvalToUlval lval = return (Updatecilutil.lvalToUlval lval)
  
let lift f = fun x -> f (); return x


let rec foldlL mop acc = function
      | [] -> return acc
      | x :: xs -> mop acc x >>= fun res -> foldlL mop res xs

let rec foldrL mop init = function
      | [] -> return init
      | x :: xs -> foldrL mop init xs >>= fun res -> mop x res

let mmap mop list = foldrL 
        (fun elm acc -> mop elm >>= fun res -> return (res :: acc)) [] list

let miter (iop : 'a -> unit cam) (l : 'a list) : unit cam = foldlL 
        (fun acc elm -> iop elm >>= fun () -> return ()) () l

let getMallocFunHash : Mallocpragma.mallocHash_t cam = 
   fun cai -> return (cai.cai_mallocFunHash) cai

(*------ end of monadic plumbing ---------- *)





(* Error computing subtypes  *)
exception SubtypeEx of string;;


(* Fresh name generation for function type coertion functions *)
let funCoerFunctionCounter : int ref = ref 0  
let getFreshFunCoerFunctionName () = 
   incr funCoerFunctionCounter; 
   Updateconfig.makeFunCoerFunctionName !funCoerFunctionCounter  

(* Fresh name generation for global variables needed in function type coercion *)
let funCoerGlobalTmpVarCounter : int ref = ref 0;;
let getFreshFunCoerGlobalTmpVarName () =
   incr funCoerGlobalTmpVarCounter;
   Updateconfig.makeFunCoerGlobalTmpVar !funCoerGlobalTmpVarCounter



(* Build a function application *)
let buildCilApplication (tResult:Cil.typ) (funToCall:Cil.fundec) (arg:Cil.exp)
   : (Cil.exp cam) = 
                    getFd               
      >>=           mkTmpVar tResult 
      >>= fun tmp ->
                    addIL [Call(Some tmp, Lval(var funToCall.svar), [arg], !(Cil.currentLoc))] 
      >>= fun _ ->
                    return (Lval tmp)
   


let buildConcretizeLval (lvalToCoerce:Cil.lval) : (Cil.lval cam) =
  (* Create a call to con.  This doesn't return a value, but modifies the structure in place. *)
  let typ = Updatecilutil.properTypeOfLval lvalToCoerce in
  let tAbs,isPtr =
    match typ with
      TNamed _ | TComp _ -> typ,false
    | TPtr ((TNamed _) as t',_) -> t',true
    | TPtr ((TComp _) as t',_) -> t',true
    | TArray ((TNamed _) as t',_,_) -> t',true
    | TArray ((TComp _) as t',_,_) -> t',true
    | _ -> Updatedebug.abort (Updatedebug.bug "buildConcretizeLval: non-abstract type %a for lval %a" 
                                d_type typ d_plainlval lvalToCoerce) 
  in
  let doit () = 
    let tmp_t =
      let tAbs_stripped = Commoncilutil.stripTypeAttributes tAbs in 
      let tAbs_unrolled = Namedtypes.unrollTypeOnce (tAbs_stripped) in
      TPtr( tAbs_unrolled,[]) 
    in
                    getLoc
    >>= fun loc ->  return (Namedtypes.getCon loc tAbs)
    >>= fun con ->  getFd
    >>=             mkTmpVar tmp_t 
    >>= fun tmp -> 
                    addIL 
                    [ 
                       Call(Some tmp, Lval(var con.svar), 
(* MWH: why do we need this cast? *)
(*                       [ CastE(TPtr((Commoncilutil.stripTypeAttributes tAbs),[]), 
                         Cil.mkAddrOf lvalToCoerce)], *)
                       [ if (not isPtr) then Cil.mkAddrOf lvalToCoerce else (Lval lvalToCoerce) ], 
                       !(Cil.currentLoc))
                    ] 
    >>= fun _ ->    return (if (not isPtr) then (Mem(Lval tmp),NoOffset) else tmp)
  in
  
  if Namedtypes.isUpdatable tAbs 
  then
     doit ()
  else
     return lvalToCoerce
   
  
(* Produce some code to concretise the expression eToCoerce from tAbs to tCon *)
let buildConcretizeExp (eToCoerce:Cil.exp) : (Cil.exp cam) =
   let tAbs = Updatecilutil.properTypeOf eToCoerce in
(*      ignore (Updatedebug.log [Updatedebug.BTAlways] "buildConcretizeExp (%a)\n" d_plainexp eToCoerce); *)
   match eToCoerce with
      Lval(lval) ->
         return lval 
         >>= buildConcretizeLval 
         >>= lvalToExp
         
    | StartOf(lval) ->
(*      ignore (Updatedebug.log [Updatedebug.BTAlways] "buildConcretizeExp (%a)\n" d_plainexp eToCoerce); *)
         (* We have to treat this separately because StartOf(x:TArray(t)) == Cast(TPtr(t), x);
            So if x is an Lval, then StartOf(x) should be treated as an lval as well *)
                          return lval >>= buildConcretizeLval 
         >>= fun lval' ->
(*            ignore (Updatedebug.log [Updatedebug.BTAlways] "buildConcretizeExp : %a (%a)" d_plainlval lval' d_plaintype (Updatecilutil.properTypeOfLval lval')); *)
                          return (Lval lval')
        
    | _ -> 
       (* If it's an arbitary expression, assign it to an lval and concretise the lval *)
             getFd
         >>= mkTmpVar tAbs
         >>= fun tmp -> getLoc
         >>= fun loc -> addIL [Set(tmp, eToCoerce, loc)]
         >>- buildConcretizeLval tmp 
         >>= lvalToExp
       
        

    
(* Produce some code to abstract the expression eToCoerce having
   representation type tau to have type tAbs; returns the expression
   with this type.  Assumes tAbs is a TNamed(-). *)

let buildAbstract (tAbs:Cil.typ) (eToCoerce:Cil.exp) : (Cil.exp cam) =
  if Namedtypes.isUpdatable tAbs
  then
    getLoc
    >>= fun loc -> return (Namedtypes.getAbs loc tAbs)
    >>= fun abs -> getFd
    >>=            mkTmpVar tAbs
    >>= fun tmp -> addIL
	[ Call(
          None, 
          Lval(var abs.svar), 
(* MWH: why do we need this cast? *)
(*          [ CastE(Namedtypes.unrollTypeOnce 
		    (Commoncilutil.stripTypeAttributes tAbs),eToCoerce); *)
          [ eToCoerce;
            Cil.mkAddrOf tmp ], 
          !(Cil.currentLoc)
         )
	]
	>>= fun _ -> lvalToExp tmp
  else 
    return eToCoerce

(* Sets the version of a struct when it's allocated.  Assumes tAbs is a TComp. *)

let buildAbstractTComp (tAbs:Cil.typ) (lval:Cil.lval) : (unit cam) =
  if Namedtypes.isUpdatable tAbs
  then
    getLoc
    >>= fun loc -> return (Namedtypes.getAbs loc tAbs)
    >>= fun abs -> 
     addIL
        [ Call(None, Lval(var abs.svar), [Cil.mkAddrOf lval], loc) ]
  else return ()

let rec unrollLval (lval : Cil.lval) : (Cil.lval cam) =
   let typ = Updatecilutil.properTypeOfLval lval in 
   match typ with 
      TNamed (ti,attr) ->
         buildConcretizeLval lval >>= 
         if Namedtypes.isUpdatable typ then unrollLval else return 
         
     | TComp _ ->
         buildConcretizeLval lval
        
     | _ -> return lval

  
let rec unrollExp (exp : Cil.exp) : Cil.exp cam =
   let typ = Updatecilutil.properTypeOf exp in 
   match typ with 
      TNamed (ti,attr) ->
             buildConcretizeExp exp 
         >>= if Namedtypes.isUpdatable typ then unrollExp else return 
        
    | TComp _  ->
        buildConcretizeExp exp 

    | _ -> return exp




let formalParamCounter : int ref = ref 0;;

type cmode = Normal | UnderPtr

(* Create a con/abs coercion code to transform from t1 (i.e. properTypeOf e) to t2 *)
let rec makeSubtypeCoercion (t2:Cil.typ) (mode:cmode) (e : Cil.exp) : Cil.exp cam =
   getLoc >>= fun loc ->
     let e_typ = Updatecilutil.properTypeOf e in
     let t1 = 
       match mode with 
	        Normal -> e_typ
         | UnderPtr -> 
	          (match e_typ with
	               TPtr(t',_) -> t'
	             | _ -> Updatedebug.abort (Updatedebug.bug "(%a) SubtypeEx: bad mode" d_loc loc)
             ) 
     in

   ignore (Updatedebug.log [Updatedebug.BTConabsinsertion] "###### makeSubtypeCoercion[%s]:   %s  <:  %s\n"
	     (match mode with Normal -> "Normal" | UnderPtr -> "UnderPtr")
	     (Updatedebug.ppType t1) (Updatedebug.ppType t2));
   let id = return e in 
   match (t1,t2) with 
      (ty1, ty2) when Ptraliastracking.interchangeableTypes ty1 ty2 -> id
    
    (* do nothing for conversions between different types of float and int *)
    | (TInt _,TInt _) 
    | (TFloat _,TFloat _) 
    | (TInt _,TFloat _) 
    | (TFloat _, TInt _) -> id

    (* we are safe (wrt update correctness) to cast a void* to any other pointer type.
       The reverse cast is not true.  However, we'll rule out those bad cases
       in an earlier phase, so we allow it here. *)
    | TPtr(TVoid _,_), TPtr _ 
    | TPtr _, TPtr (TVoid _,_) -> id

    (* types in bad casts are marked non-updatable before hand *)
    | (TPtr(ty1,attr1),TPtr(ty2,attr2)) -> 
	let cu = Namedtypes.containsUpdatableNamedTypes t1 || Namedtypes.containsUpdatableNamedTypes t2 in 
	if not cu then id
	else 
	  (match mode with 
	    Normal -> makeSubtypeCoercion ty2 UnderPtr e
	  | UnderPtr ->
	      if Ptraliastracking.interchangeableTypes ty1 ty2 then id
	      else 
		(getLoc >>=
		 fun loc -> Updatedebug.abort 
		     (Updatedebug.bug 
			"(%a) SubtypeEx: can't coerce under two pointers that contain updateable types" d_loc loc)))
		  
    | (TArray (ty1,_,attr1),TArray(ty2,_,attr2)) 
    | (TArray (ty1,_,attr1),TPtr(ty2,attr2)) 
    | (TPtr(ty1,attr1),TArray(ty2,_,attr2)) -> id
    | (TInt _, TPtr (t,a)) -> id
    | (TPtr _, TInt _) -> id
             
    (* Can't build abstract types under pointers *)
    | (t1, TNamed(ti2,attr2)) when (Updatecilutil.typeHeight t1) < (Updatecilutil.typeHeight t2) && (mode = Normal) ->
        makeSubtypeCoercion ti2.ttype mode e >>=
        buildAbstract t2 
    
    | (TNamed(ti1,attr1), t2) when (Updatecilutil.typeHeight t1) > (Updatecilutil.typeHeight t2) ->
        (if Namedtypes.isUpdatable t1
        then 
          buildConcretizeExp e 
        else 
          return (CastE(ti1.ttype,e))) 

          >>= makeSubtypeCoercion t2 mode

    | (TNamed(ti1,attr1), TNamed(ti2,attr2)) -> 
        (* Can only fire when t1=t2 or t1 and t2 in different chains *) 
        if ti1.tname = ti2.tname 
           then id
           else  
            (if Namedtypes.isUpdatable t1
             then 
                buildConcretizeExp e 
             else 
                return (CastE(ti1.ttype,e))) 

         >>= makeSubtypeCoercion t2 mode
    
              
    | (TPtr(TFun(ty_res1,args_opt1,isVarargs1,attr1),_),
       TFun(ty_res2,args_opt2,isVarargs2,attr2))

    | (TFun(ty_res1,args_opt1,isVarargs1,attr1),
       TPtr(TFun(ty_res2,args_opt2,isVarargs2,attr2),_))

    | (TFun(ty_res1,args_opt1,isVarargs1,attr1),
       TFun(ty_res2,args_opt2,isVarargs2,attr2)) ->  id

    | (TFun(ty1,args_opt1,isVarargs1,attr1),_)  
    | (_,TFun(ty1,args_opt1,isVarargs1,attr1)) -> 
         getLoc >>=

         fun loc -> Updatedebug.abort (Updatedebug.bug
                                          "(%a) : Warning : makeSubtypeCoercion: tried to coerce function to non function. Types were %a and %a\n"
                                       d_loc loc d_plaintype t1 d_plaintype t2)

              (* In order to analyze the Linux kernel we need the log version below(instead of abort)
               * since these occur very frequently  *)
(*          fun loc ->  ignore (Updatedebug.log [Updatedebug.BTAlways]  *)
(*                                        "(%a) : Warning : makeSubtypeCoercion: tried to coerce function to non function. Types were %a and %a\n" *)
(*                                     d_loc loc d_plaintype t1 d_plaintype t2); *)
(*                  id *)

    | TComp(ci1,attr1), _ ->
	let abort () = 
          getLoc >>=  fun loc ->
	    Updatedebug.abort (Updatedebug.error "(%a) : makeSubtypeCoercion: cannot coerce %a to %a " 
				 d_loc loc d_plaintype t1 d_plaintype t2) in
	(* If these are two structs with the same name, then we're good *)
	(match t2 with
	  (TComp(ci2,attr2)) when ci1.cname = ci2.cname -> id
	    (* Otherwise, this must be a physical subtype; e.g. a cast from a struct to its first field *)
        | _ -> 
	    if mode = UnderPtr then
	      (match ci1.cfields with
		h::tail ->
		  if Ptraliastracking.interchangeableTypes h.ftype t2 then
	            buildConcretizeExp e >>= fun e' -> return (CastE(TPtr(t2,[]),e'))
		  else
                    abort()
	      | _ -> abort())
	    else abort())

     (* Don't support updating enums ... yet *)
     | (TEnum _), _ 
     | _, TEnum _ -> id

     | _,_ -> getLoc >>=  
              fun loc -> Updatedebug.abort (Updatedebug.error "(%a) : makeSubtypeCoercion: illegal coerce %a to %a" d_loc loc d_plaintype t1 d_plaintype t2)
   



let rec insCoerUlval (ulval : Updatecilutil.ulval) : Updatecilutil.ulval cam = 
      match ulval with 
         Updatecilutil.UVar vi -> 
            return (Updatecilutil.UVar vi)
            
       | Updatecilutil.UMem e -> 
              insCoerExp e 
          >>= unrollExp  
          >>= fun e' -> return (Updatecilutil.UMem e')
          
       | Updatecilutil.UField (ulval',fi) -> 
              unrollUlval ulval'
          >>= insCoerUlval 
          >>= ulvalToLval 
          >>= buildConcretizeLval 
          >>= lvalToUlval
          >>= fun ulval'' -> return (Updatecilutil.UField(ulval'', fi))

       | Updatecilutil.UIndex (ulval',e) -> 
              insCoerUlval ulval' 
          >>= fun ulval'' -> unrollUlval ulval''
          >>= fun ulval''' -> return e 
          >>= insCoerExp 
          >>= unrollExp 
          >>= fun e' -> return (Updatecilutil.UIndex(ulval''', e'))

and unrollUlval  (ulval : Updatecilutil.ulval) : Updatecilutil.ulval cam =
   let lval = Updatecilutil.ulvalToLval ulval in
   let typ = Updatecilutil.properTypeOfLval lval in
      match typ with 
         TNamed (ti,attr) ->
                return ulval 
            >>= insCoerUlval
            >>= ulvalToLval
            >>= buildConcretizeLval
            >>= lvalToUlval 
            >>= if Namedtypes.isUpdatable typ then unrollUlval else return 
            
        | TComp _  -> return ulval
           
        | _ -> return ulval
       

and insCoerLvalLval (lval : Cil.lval) : Cil.lval cam = 
       lvalToUlval lval
   >>= insCoerUlval 
   >>= ulvalToLval






(* insert coercion for an lval occurring on the RHS *)
and insCoerLvalExp (lval : Cil.lval) : Cil.lval cam = 
   insCoerLvalLval lval





(* given an expression exp, produce a coerced expression and, possibly, some 
   instructions to be inserted before this expression to actually do the coersion *)
and insCoerExp (exp : Cil.exp) : Cil.exp cam = 
   let id = return exp in
   match exp with
        Const _ -> id
        
      | Lval lval -> 
         insCoerLvalExp lval >>= lvalToExp 
         
      | SizeOf typ -> id   
                           
      | SizeOfE e -> 
             insCoerExp e 
         >>= fun e' -> return (SizeOfE e')
         
      | SizeOfStr str -> id

      | AlignOf typ ->
         getLoc >>=
         fun loc -> ignore (Updatedebug.log [Updatedebug.BTAlways] "(%a) Warning : insCoerExp: can't deal with AlignOf(%a)!"
                                      d_loc loc d_plaintype typ);
                    id
      
      | AlignOfE exp ->
         getLoc >>= 
         fun loc -> ignore (Updatedebug.log [Updatedebug.BTAlways] "(%a) Warning : insCoerExp: can't deal with AlignOfE(%a)!"
                                      d_loc loc d_plainexp exp);
                    id             
                    
      | UnOp (unop, e, typ) -> 
         (* typ should always be a concrete type *) 
         let unrolled_typ = Cil.unrollType typ 
         in 
             insCoerExp e 
         >>= unrollExp 
         >>= fun e' -> return (UnOp(unop, e', unrolled_typ)) 

      | BinOp (binop, e1, e2, typ) ->
         let unrolled_typ = Cil.unrollType typ in 
             insCoerExp e1 
         >>= unrollExp
         >>= fun e1' -> return e2
         >>= insCoerExp  
         >>= unrollExp  
         >>= fun e2' -> return (BinOp (binop, e1', e2', unrolled_typ))
         

      | CastE (typ, e) ->
         getLoc 
         >>= fun loc ->
             if Updatecilutil.isNullInitialiser loc typ e 
             then 
                id
             else
                    insCoerExp e
                >>= makeSubtypeCoercion typ Normal
                >>= fun e' -> return (CastE(typ, e'))

         
      | AddrOf lval -> 
         ignore (Updatedebug.log [Updatedebug.BTConabsinsertion] "Reached AddrOf\n");
             insCoerLvalLval lval 
         >>= fun lval' -> return (AddrOf lval')
         
      | StartOf lval -> 
         ignore (Updatedebug.log [Updatedebug.BTConabsinsertion] 
            "Reached StartOf with \n %a \n of type \n %a\n" 
            d_plainlval lval d_plaintype (Cil.typeOfLval lval) );
            
             insCoerLvalExp lval 
         >>= fun lval' -> return (StartOf lval')
;;

let buildCall lval_result_opt loc specialFunction fname exp_func_ret_t (exp_func : Cil.exp) (fixedArgs : Cil.exp list) (varArgs : Cil.exp list) : unit cam = 
  begin match  lval_result_opt with
      None ->
        addIL [Call (None, exp_func, fixedArgs @ varArgs, loc)]
          
    | Some lval_result ->
        let lval_result_t = Updatecilutil.properTypeOfLval lval_result in
        let insertCastIfNeeded e = Cil.mkCastT e exp_func_ret_t lval_result_t in

          getFd
          >>=                            mkTmpVar exp_func_ret_t
          >>= fun tmp                 -> addIL [Call (Some tmp, exp_func, fixedArgs @ varArgs, loc)]
            >>= fun _                   -> insCoerLvalLval lval_result
              >>= fun coerced_lval_result -> lvalToExp tmp
                >>=                            makeSubtypeCoercion lval_result_t Normal
              >>= fun e_ret               -> 
                (* coerce result, but not if its a special function that needs its result eliding *)
                (if specialFunction then 
                   if Specialfuns.shouldElideReturn fname then 
                     addIL [Set(coerced_lval_result, insertCastIfNeeded(Lval(tmp)), loc)]
                   else
                     addIL [Set(coerced_lval_result, insertCastIfNeeded(e_ret), loc)] 
                 else
                   addIL [Set(coerced_lval_result, insertCastIfNeeded(e_ret), loc)])
  end;;



(* insert coercions into an instruction.  Produce an instruction list whose last
element is the modified instruction we received and the rest are the instructions
to do the coercion *)
let insCoerInstr (instr : Cil.instr) : unit cam = 
   setLoc (get_instrLoc instr) >>= fun _ ->
   getMallocFunHash >>= fun mallocFunHash ->
   match instr with 
      Set (lval,exp,loc) ->
         let lval_t = Updatecilutil.properTypeOfLval lval 
         in 
                  insCoerExp exp 
              >>= makeSubtypeCoercion lval_t Normal
              >>= fun e'    -> insCoerLvalLval lval 
              >>= fun lval' -> addIL [Set (lval',e',loc)]
   


    (* t adest = va_arg(dest, t) *)
    | Call (None, Lval(Var vi, NoOffset), [dest; SizeOf t; adest], loc)
      when vi.vname = "__builtin_va_arg" -> 

       let t_unrolled = Cil.unrollType t in 
       let dest_lval = 
          match stripCasts adest with 
             AddrOf(l) -> l
           | _ -> 
             Updatedebug.abort (Updatedebug.bug 
                                "(%a) insCoerInstr: the destination of a __builtin_va_arg is not an lval!\n"
                                d_loc loc)
       in

       (* arguments passed as var args are unrolled down to a base
          type by cons on their way in, so if they need coercing, it
          will be from this unrolled type. *)
       insCoerExp dest
       >>= fun dest' -> getFd 
       >>= mkTmpVar t_unrolled
       >>= fun tmp -> addIL 
                         [ Call (
                              None, 
                              Lval(Var vi, NoOffset), 
                              [dest'; SizeOf t_unrolled; Cil.mkAddrOf tmp], 
                              loc)
                         ] 
       >>- lvalToExp tmp (* has type t_unrolled *)
       >>= makeSubtypeCoercion (Updatecilutil.properTypeOfLval dest_lval) Normal (* is type of dest_lval *)
       >>= fun e -> insCoerLvalLval dest_lval 
       >>= fun dest_lval' -> addIL [Set (dest_lval', e, loc)]


    (* There is no type information in this va_arg call, so we don't support it *)
    | Call(res, Lval(Var vi, NoOffset), [ ], l) when vi.vname = "__builtin_next_arg" ->
       getLoc >>=
       fun loc -> Updatedebug.abort (Updatedebug.bug 
                                     "(%a) insCoerInstr: found __builtin_next_arg and we don't support this!" 
                                     d_loc loc)

    (* C-magic-call: we deal with arguments and then unroll them *)
    | Call(res, (Lval(Var vi, NoOffset) as exp_func), exp_arg_list, l) 
        when Commoncilutil.isBuiltin vi.vname  ->
        mmap (fun e -> insCoerExp e >>= unrollExp) exp_arg_list
        >>= fun fixedArgs -> 
          let (exp_func_ret_t, 
               exp_func_args_t, 
               isva)  = Updatecilutil.getTypesInFunExp exp_func in
          buildCall res l (Specialfuns.isSpecialFun vi.vname) vi.vname exp_func_ret_t exp_func fixedArgs []

(* addIL [Call(res, Lval(Var vi, NoOffset), exp_arg_list', l)] *)


    (* malloc *)
    (* MWH: should this look more like a regular call, as below? *)
    | Call(lval_result_opt, (Lval(Var vi, NoOffset) as exp_func), args, loc) 
        when Mallocpragma.isMallocFunc mallocFunHash vi.vname ->

          (* malloc functions return a void* and we do not coerce from
	     this to the target pointer type as makes functions
	     non-update-calling in the analysis *)

         let (exp_func_t, exp_func_args_t, _) = Updatecilutil.getTypesInFunExp exp_func in
         let argsAndTypesList = 
           try   
             (List.combine args exp_func_args_t)
           with _ ->

             (* arguments in the call don't match the number in the
                prototype.  Maybe we shouldn't bail here, but rather
                do the fixed and var args separately, the way we do it
                for functions. But most of the time this error comes
                from not having a prototype for malloc().  *)

             let argsTypesList = 
	       List.map (fun x -> Updatedebug.ppType (Updatecilutil.properTypeOf x)) args in
             let argsTypesString = 
	       List.fold_left (fun acc x -> acc ^ " " ^ x) "" argsTypesList in
             Updatedebug.abort (Updatedebug.error 
                                  "(%a) insCoerInstr: malloc-style function '%s' prototype \n(%s)\n not consistent with the arguments \n(%s)\n. Maybe you didn't include <stdlib.h> ?\n" 
                                  d_loc loc vi.vname (Updatedebug.ppType vi.vtype) argsTypesString);

         in 
         (* coerce the args first *)
         mmap 
	   (fun (e,t) -> insCoerExp e  >>= makeSubtypeCoercion t Normal) 
	   argsAndTypesList
         >>= 
         (* then the optional result *) 
	   (fun args' -> 
             begin match lval_result_opt with
	       (* MWH: not returning anything!  Add warning? *)
               None -> addIL [Call(None, exp_func, args', loc)] 
             | Some(l) -> 
		 insCoerLvalLval l 
		 >>= fun result -> 
		   let res_type = Updatecilutil.properTypeOfLval result in
		   addIL [Call(Some result, exp_func, args', loc)]
		 (* Add call to abs() for struct/unions *)
		 >>= fun _ ->
		   (match res_type with
		     TPtr (res_tgt_type,_) ->
		       if (Updatecilutil.isCompType res_tgt_type) then
			 buildAbstractTComp res_tgt_type 
			   (Cil.mkMem (Lval(result)) NoOffset)
		       else return ()
		   | _ -> return () (* FIXME : we this abort in zebra if zcalloc is #pragma malloc'ed*)
(* 		       Updatedebug.abort (Updatedebug.bug "(%a) malloc assigns to non-pointer %a : %a" d_loc loc d_exp (Cil.Lval result) d_type res_type) *)
         )
             end)	   

    | Call (lval_result_opt, exp_func, exp_arg_list, loc) ->

       (* Treat user-inserted DSU_update specially *)
       let args = 
	 match exp_func with
	   (Lval ((Var vi),NoOffset)) when vi.vname = Updateconfig.updateFunctionName ->
	     ignore(Updatedebug.log [Updatedebug.BTAlways] "Processing user-inserted DSU_update\n");
	     if exp_arg_list = [] then [mkString ""]
	     else exp_arg_list
	 | _ -> exp_arg_list in

       let (exp_func_ret_t, 
            exp_func_args_t, 
            isva)                 = Updatecilutil.getTypesInFunExp exp_func in
       let numFixedArgs           = List.length exp_func_args_t in 
       let numActualArgs          = List.length args in

       (* Make two different lists for fixed and variable arguments *)
       let ( fixedArgs,
             varArgs )   = Commonutil.list_split numFixedArgs args in 
      
       (* Determine if this is a special function, extracting it's name *)
       let specialFunction, fname = 
          begin match exp_func with 
             Lval(Var vi, _) -> (Specialfuns.isSpecialFun vi.vname, vi.vname)
           | _ -> (false, "NOT A SPECIAL FUN")
          end
       in

       (* Create a function that coerces the arguments to the function.  If this is a 
          special function, return the function that checks before had it the argument
          should be elided.  The returned function takes as arguments the arg index
          followed by the same arguments as makeSubtypeCoercion *)
       let makeArgSubtypeCoercion = 
          if specialFunction then 
             fun n t2 e -> 
                if Specialfuns.shouldElide fname n then return e else makeSubtypeCoercion t2 Normal e
          else
             fun n t2 e -> makeSubtypeCoercion t2 Normal e
       in
      
       (* given a list of fixed argument expressions and target types for them, compute a monad
          that contains a list of the con-abs inserted arguments *)
       let make_fixed_args (fixedArgs_t : Cil.typ list) (fixedArgs : Cil.exp list) : (Cil.exp list) cam =
          let fixed_args_fold (m : (int * Cil.exp list) cam) (target_t, exp) = 
                 m
             >>= fun (n,al) -> insCoerExp exp
             >>= fun exp' -> makeArgSubtypeCoercion n target_t exp'
             >>= fun exp'' -> return (n+1, (al @ [exp'']))
          in 
             (List.fold_left fixed_args_fold (return (0,[])) (List.combine fixedArgs_t fixedArgs))
             >>= fun (_,args) -> return args
             (*>>= fun x -> ppInstrs false >>= fun _ -> return x*)
       in

       (* the same sort of thing for variable arguments *)
       let make_var_args (varArgs : Cil.exp list) : (Cil.exp list) cam =
          let var_args_fold (m : (Cil.exp list) cam) (exp : Cil.exp) = 
                 m 
             >>= fun al ->   insCoerExp exp
             >>=             unrollExp 
             >>= fun exp' -> return (al @ [exp'])
          in 
             List.fold_left var_args_fold (return []) varArgs
       in

       (*  Print useful information and warnings ------------------------- *)
       getLoc >>=
       fun loc ->
          begin
          ignore(Updatedebug.log [Updatedebug.BTConabsinsertion] "####### (%a) FIXED = %d    VARIABLE = %d   FUNC PARAMS=%d\n" 
             d_loc loc numFixedArgs (numActualArgs-numFixedArgs) numActualArgs);

          if (numFixedArgs = 0) && (numFixedArgs < numActualArgs) 
          then (* probably have no prototype for the function *)
             ignore(Updatedebug.log [Updatedebug.BTAlways] 
                "(%a) Warning!  Calling function with inconsistent number of arguments: it is probably external\n"
                d_loc loc)
          else ();

          (* Report any strangeness in the number of arguments *)
          if (numFixedArgs = 0) && (numFixedArgs < numActualArgs) && (not isva) 
          then 
             (* probably have no prototype for the function *)
             ignore( Updatedebug.log [Updatedebug.BTAlways] 
                                     "(%a) Warning!  Calling function with inconsistent number of arguments: it is probably external\n"
                                     d_loc loc)
             
          else if (numFixedArgs < numActualArgs) && (not isva) 
          then 
             Updatedebug.abort (Updatedebug.bug 
                                "(%a) Bug!  Calling non var arg function with inconsistent number of arguments (and not external!)\n"
                                d_loc loc);
          return ()
          end


       (*------------------------------------------------------------------ *)

 
       >>= fun _ ->           insCoerExp exp_func
       >>=                    unrollExp
       >>= fun exp_func'  ->  return fixedArgs
       >>=                    make_fixed_args exp_func_args_t
       >>= fun fixedArgs' ->  return varArgs
       >>=                    make_var_args
       >>= fun varArgs'   ->  buildCall lval_result_opt loc specialFunction fname exp_func_ret_t exp_func' fixedArgs' varArgs'

    
    | Asm (sl,isvol,outs,ins,clobs,l) -> 
       getLoc >>=
       fun loc ->
          begin
             ignore(Updatedebug.log [Updatedebug.BTAlways] 
                       "(%a) WARNING! insCoerInstr: can't deal with Asm at instruction level \n" d_loc loc);

             let outs_so_list, outs_s_list, outs_lv_list = Commonutil.splitTriples outs
             and ins_so_list, ins_s_list,  ins_e_list    = Commonutil.splitTriples ins
             in    
                                      mmap (fun lv -> return lv >>= insCoerLvalLval >>= unrollLval) outs_lv_list
             >>= fun outs_lv_list' -> mmap (fun e -> insCoerExp e >>= unrollExp) ins_e_list
             >>= fun ins_e_list' -> 
	            let outs' = Commonutil.combineTriples outs_so_list  outs_s_list outs_lv_list'
               and ins'  = Commonutil.combineTriples ins_so_list ins_s_list ins_e_list' in
                   addIL[Asm (sl,isvol,outs',ins',clobs,l)]
          end
             

(* insert abs/con coercions into a statement *)
let rec insCoerStmt (mallocFunHash : Mallocpragma.mallocHash_t) (originatingFun : Cil.fundec) (stmt : Cil.stmt) : Cil.stmt = 
   let stmtLoc = Cil.get_stmtLoc stmt.skind in
   let insCoerStmt' = insCoerStmt mallocFunHash originatingFun in
   (* compute the return type *)
   let skind = stmt.skind in
   let id = ([], skind) in 
   let (il,newSkind) = 
      match stmt.skind with 
       | Instr  il ->
          let inst_list_fold (m : unit cam) (i : Cil.instr) = 
             m >>= fun _ -> insCoerInstr i
          in
          let (_,il') = run mallocFunHash originatingFun stmtLoc (List.fold_left inst_list_fold (return ()) il)
          in
          [], Instr (il')
          
       | Return (None,loc) -> id
       
       | Return (Some e, loc) ->
          let return_type = Updatecilutil.getReturnType originatingFun 
          in
          let e',il = 
             run mallocFunHash originatingFun stmtLoc
                 (     insCoerExp e  
                   >>= makeSubtypeCoercion return_type Normal)
          in 
          il, Return (Some e', loc)
          
       | Goto (target, loc) -> id
          
       | Break loc -> Updatedebug.abort (Updatedebug.bug "(%a) insCoerStmt: CIL should have eliminated break!\n" d_loc stmtLoc)
       
       | Continue loc -> Updatedebug.abort (Updatedebug.bug"(%a) insCoerStmt: CIL should have eliminated continue!\n" d_loc stmtLoc)    
       
       | If (eTest,blockT,blockF,loc) ->
          let (eTest',il) = 
             run mallocFunHash originatingFun stmtLoc
                 (     insCoerExp eTest 
                   >>= unrollExp )
                   (*>>= makeSubtypeCoercion eTest_t (Cil.unrollType eTest_t) )*)
          in
          (* can't coerce eTest to an int, because we don't know the kind of it it might be.
             If the original program was safe (which we assume) then it will be safe to cast
             to the unrolled type *)
          blockT.bstmts <- List.map insCoerStmt' blockT.bstmts;
          blockF.bstmts <- List.map insCoerStmt' blockF.bstmts;
          il, If (eTest',blockT,blockF,loc)
          
       | Switch (exp, block,stmt_list,loc) ->
          Updatedebug.abort (Updatedebug.bug "(%a) insCoerStmt: CIL should have eliminated switch!\n" d_loc stmtLoc)
          
       | Loop (block,loc,s1,s2) ->  
           block.bstmts <- List.map insCoerStmt' block.bstmts;
           ([], Loop (block,loc,s1,s2))

          
       | Block block ->
          block.bstmts <- List.map insCoerStmt' block.bstmts;
          ([], Block block)
          
       | TryFinally  (block1,block2,loc) ->
          Updatedebug.abort (Updatedebug.bug
                                             "(%a) insCoerStmt: __try/__except not supported!\n" 
                                             d_loc stmtLoc)

       | TryExcept (block_try, inst_list, block_except, loc) ->
          Updatedebug.abort (Updatedebug.bug
                                             "(%a) insCoerStmt: __try/__except not supported!\n" 
                                             d_loc stmtLoc)
   in
   match il with 
      [] -> (*  Don't do this to avoid breaking sharing: {stmt with skind = newSkind} *)
            stmt.skind <- newSkind;
            stmt
    | _  -> 
       let stmt_list = [ Commoncilutil.mkStmt (Instr il); 
                         Commoncilutil.mkStmt newSkind] in 
       let block = Cil.mkBlock stmt_list  in

       stmt.skind <- (Block block);
       stmt



let rec insCoerFun (mallocFunHash : Mallocpragma.mallocHash_t) (fd : Cil.fundec) : Cil.fundec = 
   ignore (Updatedebug.log [Updatedebug.BTConabsinsertion] "Now doing function %s\n" fd.svar.vname);
   (* fd's members are mutable, so fd's local variables get added in-place *)
   let coerced_stmt_list = List.map (insCoerStmt mallocFunHash fd) fd.sbody.bstmts in 
   (* add abs() calls to all stack-allocated TComps *)
   let vars_fold (m : unit cam) (v : Cil.varinfo) = 
     m >>= fun _ -> 
       (match v.vtype with
	 TComp _ -> buildAbstractTComp v.vtype ((Var v),NoOffset)
       | _ -> return ()) in
   let (_,il) = 
     run mallocFunHash fd (!Cil.currentLoc) 
       (List.fold_left vars_fold (return ()) fd.slocals) in
   let coerced_stmt_list =
     if il = [] then coerced_stmt_list
     else
       let stmt = Commoncilutil.mkStmt (Instr il) in
       stmt :: coerced_stmt_list in
   fd.sbody.bstmts <- coerced_stmt_list;
   fd


(*
FIXME: see insCoerTopLevelInit comment
let insCoerTopLevelInitExp 
   (eConst : Cil.exp)     (* This expression should be a constant expression *)
   : Cil.exp = 

*)

(*
let insCoerTopLevelInit (ctxType : Cil.typ) (init : Cil.init) :   = 
 FIXME: we need to deal with the case where con/abs coercions are needed
          in top-level definitions, as only constant expressions are allowed here!
          We need to expand these coercions inline.

   match init with 
      None -> g
    | Some(SingleInit e) -> 
       insCoerTopLevelInitExp e
    | Some(CompoundInit (typ, offsetInit list)) ->
*)

(* insert coercions into programs *)
let insCoerProgram (mallocFunHash : Mallocpragma.mallocHash_t) (prog : Cil.global list) : Cil.global list = 
   let insCoerProgram' (g : Cil.global) : Cil.global =
      let globalLoc = get_globalLoc g in
      match g with 
       | GType (ti,loc) -> g
       | GCompTag (ci,loc) -> g
       | GCompTagDecl (ci,loc) -> g
       | GEnumTag (ei, loc) -> g
       | GEnumTagDecl (ei,loc) -> g
       | GVarDecl (vi,loc) -> g
       | GVar (vi,ii,loc) -> 
          (*Cil.errorLoc loc "%a  Type: %a" d_global g d_plaintype vi.vtype; *)
          g (* FIXME: should coerce here! insCoerTopLevelInit vi.vtype ii.init *)
       | GFun (fd,loc) ->
          GFun(insCoerFun mallocFunHash fd, loc)
       | GAsm (asm,loc) -> 
          ignore (Updatedebug.log [Updatedebug.BTAlways] 
             "(%a) Warning : insCoerProgram: can't deal with GAsm at top-level!\n" d_loc globalLoc);
          g
       | GPragma (ai,loc) -> g
       | GText comment -> g           

   in
   List.map insCoerProgram' prog 



let forceConAbsGen (file:Cil.file) : unit =
   Cil.iterGlobals file
      (function 
         GType(ti,loc) -> ignore(Namedtypes.getCon loc (TNamed(ti,[])))
       | GCompTag(ci,loc) -> ignore(Namedtypes.getCon loc (TComp(ci,[])))
       | _ -> ())


type updTypes = {
    mutable types : string list;
    num : int ref;
  }

(* Generate a global definition for a single con/abs *)
(*-----------------------------------------------------------------------*)
let genConAbsGlobalDef (vd : Versiondata.version_data)
                       (t : Cil.typ) 
                       (g : Cil.global) 
                       (numUpd : updTypes) 
                       (numExtern : int ref) 
                       (total : int ref) 
                       : Cil.global list =
(*-----------------------------------------------------------------------*)
   let name = Updatecilutil.getAbstractTypeName t in
   let ca_info = 
      try
         Namedtypes.NamedTypeInfo.getConAbs name 
      with 
         Not_found -> Updatedebug.abort (Updatedebug.bug "genConAbsGlobalDef : can't find ca_info for type %a!\n" d_plaintype t)
   in
      
   if (Namedtypes.isUpdatableByAbsName name) then 

     if Updateargs.args.Updateargs.arg_dostaticconabs || 
        (not (Versiondata.isExistingType vd name)) then 
       begin 
	 ignore (Updatedebug.log [Updatedebug.BTNamedtypes] " updateable |%s|\n" name);

	 incr (numUpd.num);
	 numUpd.types <- name :: numUpd.types;
	 incr total;
	 
	 let con_fun = (GFun (ca_info.Namedtypes.ca_info_con, locUnknown)) in
	 let abs_fun = (GFun (ca_info.Namedtypes.ca_info_abs, locUnknown)) in
	 con_fun :: [abs_fun]
       end
     else
         (* We have an updatable type name that existed before, so generate externs for con/abs *)
       begin
	 ignore (Updatedebug.log [Updatedebug.BTNamedtypes] " updateable |%s| : seen before\n" name);

	 incr (numUpd.num);
	 numUpd.types <- name :: numUpd.types;
         incr total;

         let con_vi = ca_info.Namedtypes.ca_info_con.svar in
         let abs_vi = ca_info.Namedtypes.ca_info_abs.svar in
         con_vi.vstorage <- Cil.Extern;
         abs_vi.vstorage <- Cil.Extern;

         [(GVarDecl (con_vi, locUnknown));
          (GVarDecl (abs_vi, locUnknown));]
       end
   else if (Namedtypes.isExternalTypeName name) then
     begin 
       ignore (Updatedebug.log [Updatedebug.BTNamedtypes] " external |%s|\n" name);
       incr numExtern;       
       []
     end
   else 
     begin 
       ignore (Updatedebug.log [Updatedebug.BTNamedtypes] " non-updateable |%s|\n" name);
       incr total;       
       []
     end

(* Returns a list of globals defining the con/abs functions for the
   named types *)
(*--------------------------------------------------------------------------------------*)
let makeGlobalConAbsList (vd : Versiondata.version_data) (globalList : Cil.global list) = 
(*--------------------------------------------------------------------------------------*)

    let numUpd = { num = ref 0; types = []; } in
    let numExtern = ref 0 in
    let total = ref 0 in  

    let genConAbsGlobalDef_fold acc g = match g with 
         GCompTag(ci,loc) -> (List.rev (genConAbsGlobalDef vd (TComp(ci,[]))  g numUpd numExtern total)) @ acc
       | GType(ti,loc) ->    (List.rev (genConAbsGlobalDef vd (TNamed(ti,[])) g numUpd numExtern total)) @ acc
       | _ -> acc
    in 

    let conAbsList = List.rev (List.fold_left genConAbsGlobalDef_fold [] globalList) in

    ignore(Updatedebug.log [Updatedebug.BTStatistics]
              "# Types : Updatable = %d, Internal = %d, External = %d, Total = %d\n"
              !(numUpd.num) !total !numExtern (!total + !numExtern));
    ignore(Updatedebug.log [Updatedebug.BTStatistics] "Updatable Types : \n");
     List.iter (function n -> 
       ignore(Updatedebug.log [Updatedebug.BTStatistics]
		"  %s\n" n)) numUpd.types;
    conAbsList;;


(* Inserts coercions for all the functions and inserts coercion definitions to the globals list
   as a side effect *)
let default_insert_coercions (mallocFunHash : Mallocpragma.mallocHash_t) 
                             (vd : Versiondata.version_data) 
                             (f : file) : unit =
  begin

    (* force the generation of con/abs coercions (these are cached).  We do this because we don't want 
       to view a type as non-updatable just because we don't use it. *)
    (* MWH: now that con/abs functions are static, should be able to ditch this *)
    forceConAbsGen f;


    (* Add the representation types for comps *)

    let conAbsedGlobals = insCoerProgram mallocFunHash f.globals in
    let splitglobals  = Updatecilutil.splitGlobalsBeforeVars conAbsedGlobals in
    f.globals <- (fst splitglobals) @ 
                 (makeGlobalConAbsList vd f.globals) @ 
                 (snd splitglobals);


    let getRepType_fold acc g =
       match g with 
          GCompTag(ci,loc) -> acc @ ((Namedtypes.getRepTypeDef (TComp(ci,[]))) @ [g])
        | _ -> acc @ [g]
    in 
    f.globals <- List.fold_left getRepType_fold [] f.globals;

  end

