(* 
 *
 * Copyright (c) 2004-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(* 
  Delta constraint solver that throws constraints to Banshee


  C = {(D1,D1'), ..., (Dn,Dn')}
     where D : {id; kind = K(T) | V(solution)}
  H = map from id to node
  
  let N_const = filter C where kind = K
  let N_var   = filter C where kind = V
  
  iter makeVarNode over N_var
  iter makeConstNode over N_const
  iter over C making subtypeEdge from D to D'
  
  Tell banshee doneAdding
  
  iter over N_var calling sweepReachableTypes
  
*)

module H = Hashtbl


(* the kind of delta (constant or variable) *)
type delta_kind_t =   
   DKConst of string              (* constant with type name *)         
 | DKVar      of bool array option   (* variable with array *)

type delta = 
{ 
   delta_loc   : Cil.location;       (* location of effect  *)
   delta_id    : int;            (* unique id of effect *)
   delta_kind  : delta_kind_t    (* kind of delta (constant or variable) *)
}

let empty_delta : delta = 
  {delta_loc = Cil.locUnknown; delta_id = 0; delta_kind = DKVar None};


(* Sets of deltas *)
module Delta_OT  = 
struct
   type t = delta 
   let compare x y = Pervasives.compare x.delta_id y.delta_id
end

module DeltaSet = Set.Make(Delta_OT)


type delta_constraint = DCSubset of delta * delta | DCCond of int*(delta*delta)

(* we don't allow consts on the RHS of a constraint, because Banshee can't grow constants *)
let makeSureThereSNoConstOnRight (deltaConstraint:delta_constraint) : unit =
  let d1,d2 = match deltaConstraint with 
      DCSubset (d1',d2') -> (d1',d2')
    | DCCond (_,(d1',d2')) -> (d1',d2')
  in
    begin
      match d2.delta_kind with 
          DKConst _ ->  Updatedebug.abort (Updatedebug.bug "econd var (id =%d) can't be a const !!!!!" d2.delta_id)
        | _ -> ()
    end

(* Ordered type for constraints on deltas *)
module DeltaConstraint_OT (*: Set.OrderedType*) = 
struct
   type t = delta_constraint
   let compare x y = Pervasives.compare x y
end

module DeltaConstraintSet = Set.Make(DeltaConstraint_OT)


(* Capability sets: sets of named type names *)
module Capability_OT = 
struct 
   type t = string
   let compare x y = Pervasives.compare x y
end

module CapabilitySet = Set.Make(Capability_OT)


(* Given a set of constraints, produce a pair of delta sets.  The first contains
   the constant deltas and the second contains the var deltas *)
let filterNodes (cset : DeltaConstraintSet.t) : (DeltaSet.t * DeltaSet.t * (int,delta) H.t) =
  let id_to_delta = H.create 4099 in
  let add (d:delta) ((dconst_set,dvar_set) : DeltaSet.t * DeltaSet.t) =
    Commonutil.hashtbl_add_once id_to_delta d.delta_id d;
    match d.delta_kind with 
        DKConst s -> (DeltaSet.add d dconst_set, dvar_set)
      | DKVar bao -> (dconst_set, DeltaSet.add d dvar_set)
  in
  let filterNodes_fold (dc:delta_constraint) ((dconst_set, dvar_set) : DeltaSet.t * DeltaSet.t) = 
    match dc with 
        DCSubset(d1,d2) -> add d2 (add d1 (dconst_set, dvar_set))
      | _ -> Updatedebug.abort (Updatedebug.bug "filterNodes: found conditional constraint that should have been eliminated\n")
  in
  let cs, vs = DeltaConstraintSet.fold filterNodes_fold cset (DeltaSet.empty, DeltaSet.empty) in
  (cs, vs, id_to_delta)


(* Tell banshee about a const node *)
let makeConstNode (d : delta) ((orderAdded, numAdded, id_to_node_hash) : ((int,string) H.t) * int * ((int,Bansheegw.node) H.t)) =
   match d.delta_kind with 
      DKConst s -> 
         let node = Bansheegw.makeConstNode (string_of_int d.delta_id) in
         H.add id_to_node_hash d.delta_id node;
         ignore(Updatedebug.log [Updatedebug.BTDeltasolver] "Adding const: %s (id=%d)\n" s d.delta_id);
         H.add orderAdded numAdded s;
         (orderAdded, numAdded+1, id_to_node_hash)
         
    | _ -> Updatedebug.abort (Updatedebug.bug "makeConstNode: Tried to add var node as const node\n")


(* Tell banshee about a var node *)
let makeVarNode (id_to_node_hash : ((int,Bansheegw.node) H.t)) (d : delta) : unit = 
   match d.delta_kind with
      DKVar bao -> 
         ignore(Updatedebug.log [Updatedebug.BTDeltasolver] "adding n%d\n"  d.delta_id);
         let node = Bansheegw.makeVarNode (string_of_int d.delta_id) in
         H.add id_to_node_hash d.delta_id node
         
    | _ -> Updatedebug.abort (Updatedebug.bug "makeVarNode: Tried to add const node as var node\n")

(* monads use this to ensure non-overlapping delta_id sets
   Suspect uses delta_id's 0 and 1 with a sepcial meaning, so start counting at 10 then
 *)
let deltaCounter = ref 10;;

let freshDelta dk loc = 
   incr deltaCounter;
   { delta_loc  = loc; 
     delta_id   = !deltaCounter; 
     delta_kind = dk };;


(*  *)
let getNode (id_to_node_hash : ((int,Bansheegw.node) H.t)) id = 
   try 
      H.find id_to_node_hash id
   with
      Not_found -> Updatedebug.abort (Updatedebug.bug "addSubsetConstraint: Couldn't find node with id %d\n" id)


let addSubsetConstraint (id_to_node_hash : ((int,Bansheegw.node) H.t)) (dc:delta_constraint) =
   match dc with 
      DCSubset(d1,d2) ->
(*         ignore(Updatedebug.log [Updatedebug.BTAlways] "n%d -> n%d; \n"  d1.delta_id  d2.delta_id); *)
         let n1 = getNode id_to_node_hash d1.delta_id in
         let n2 = getNode id_to_node_hash d2.delta_id in
       let _ = Updatedebug.log [Updatedebug.BTDeltasolver] "Adding constraint %d <= %d (%a <= %a)\n" d1.delta_id d2.delta_id Cil.d_loc d1.delta_loc Cil.d_loc d2.delta_loc in
         Bansheegw.subtypeEdge n1 n2
    | _ -> Updatedebug.abort (Updatedebug.bug "addSubsetConstraint: found non-subset constraint")


(* Extract a solution from banshee for delta *)
let makeCapabilitySet (orderAdded : (int,string) H.t)                      (* map from order const given to banshee to type name *) 
                      (numConsts : int)                                    (* size of orderAdded *)
                      (id_to_node_hash : ((int,Bansheegw.node) H.t))       (* hash from delta id to node *)
                      reachable_types
                      (d : delta)                                          (* Current delta to solve for *)
                      : CapabilitySet.t =
(*    ignore(Updatedebug.log [Updatedebug.BTAlways] "makeCapabilitySet: pos %a delta_id %d\n" Cil.d_loc d.delta_loc d.delta_id);  *)
   (* Given the index of the type to consider, add it to the capability set if it is reachable (in the banshee graph) 
      as indicated by 'reachable' *)
   let addTypeIfReachable ((index,reachable_set) : int * CapabilitySet.t) (reachable : int) = 
      if reachable = 1 then 
         let type_name = H.find orderAdded index in
         (index+1, CapabilitySet.add type_name reachable_set)
      else
         (index+1, reachable_set)
   in
   match d.delta_kind with
      DKVar bao -> 
         let n = getNode id_to_node_hash d.delta_id in  
         let () = Bansheegw.sweepReachableTypes n !reachable_types in 
         let (_, reachable_set) = Array.fold_left addTypeIfReachable (0,CapabilitySet.empty) !reachable_types in 
         reachable_set

    | DKConst type_name -> 
         let empty = CapabilitySet.empty in
         let singleton = CapabilitySet.add type_name empty in
         singleton;;
         

let solveOneDelta deltaSolverState id =
  let (const_order_hash, numConsts, id_to_node_hash, low_level_scratchpad, solution, id_to_delta) = deltaSolverState in
  let d = H.find id_to_delta id in  
  try 
    H.find solution d.delta_id

  with Not_found ->
    let dSet = makeCapabilitySet const_order_hash numConsts id_to_node_hash low_level_scratchpad d in
    H.add solution d.delta_id dSet;
    dSet  
;;



let initBansheeAndSolveConstants cset solution =
   let const_nodes, var_nodes, id_to_delta = filterNodes cset in 
   let nC = DeltaSet.cardinal const_nodes in
   let () = Bansheegw.bansheeInit nC in
   let id_to_node_hash = H.create 4099 in 
   let (const_order_hash, numConsts, _) = DeltaSet.fold makeConstNode const_nodes (H.create 4099,0,id_to_node_hash) in
   assert(numConsts = nC);
   let () = DeltaSet.iter (makeVarNode id_to_node_hash) var_nodes in 
   (*let () = List.iter (makeVarNode id_to_node_hash) (DeltaSet.elements var_nodes) in  *)
   let () = DeltaConstraintSet.iter (addSubsetConstraint id_to_node_hash) cset in
   let () = Bansheegw.doneAdding () in

   (* Scratch pad that banshee will write solution to *)
   let low_level_scratchpad = ref (Array.make numConsts 0) in 

   let deltaSolverState = (const_order_hash, numConsts, id_to_node_hash, low_level_scratchpad, solution, id_to_delta) in
   let () = DeltaSet.iter (fun d -> ignore (solveOneDelta deltaSolverState d.delta_id)) const_nodes in 
   deltaSolverState, var_nodes  

(* solve the set of constraints cset filling out solution with a map from delta id's to sets of type names *)
let solveDeltaEagerly 
    (cset : DeltaConstraintSet.t) 
    (solution : (int, CapabilitySet.t) H.t)    (* the hash that we add our solution to.  It maps delta id's to 
                                                  a set of type names *)
    : unit = 

   let deltaSolverState, var_nodes = initBansheeAndSolveConstants cset solution in
   
   let () = DeltaSet.iter (fun d -> ignore (solveOneDelta deltaSolverState d.delta_id)) var_nodes in

   (* Bansheegw.printStats(); *)
   Bansheegw.bansheeReset ()
;;

let lazyStartSolving (cset : DeltaConstraintSet.t) =
   let solution = H.create 4099 in 
   let deltaSolverState,_ = initBansheeAndSolveConstants cset solution in
   deltaSolverState
;; 

let lazyEndSolving () = Bansheegw.bansheeReset ();;

let lazyQuery deltaSolverState d = 
  solveOneDelta deltaSolverState d
;;

(* Convert a given data set to a string *)
let printDelta delta_set = 
   let printType typename acc = 
      acc ^ " " ^ typename
   in
   CapabilitySet.fold printType delta_set ""



(* Convert a DeltaConstraintSet to a string *)
let printConstraint k =
  match k with
    DCSubset (d1,d2) -> Printf.sprintf "(%s)%d <= %d" 
      (match (d1.delta_kind) with DKConst s -> s | _ -> "n") d1.delta_id d2.delta_id
  | DCCond (c,(d1,d2)) -> Printf.sprintf "%d ? %d <= %d" c d1.delta_id d2.delta_id

let printConstraintSet ds =
  let aux k =
    let s = printConstraint k in
    ignore(Updatedebug.log [Updatedebug.BTAlways] "%s\n" s) in
  DeltaConstraintSet.iter aux ds

(* print out a solution *)
let printSolution (solution : (int, CapabilitySet.t) H.t) : unit = 
   ignore(Updatedebug.log [Updatedebug.BTDeltasolver] "*************************************************************\n");
   List.iter 
     (fun key -> 
        let ds = H.find solution key in 
        ignore(Updatedebug.log [Updatedebug.BTDeltasolver] "%d: %s\n" key (printDelta ds))
     ) (List.sort Pervasives.compare (Commonutil.hashtbl_keys solution));

   ignore(Updatedebug.log [Updatedebug.BTDeltasolver] "*************************************************************\n")

(* print solution in tuples of 4 sets (at consecutive indexes) *)
let printSolution_4 solution lbl2ndx tupleEndIndx =
  let rec printBunch n = 
    (if n < tupleEndIndx
     then
       let string_of_indx n = printDelta (H.find solution n) in
           ignore(Updatedebug.log [Updatedebug.BTDeltasolver] "'%s': ({%s}{%s}{%s}{%s})\n" 
                    (H.find lbl2ndx n) 
                    (string_of_indx n) 
                    (string_of_indx (n + 1)) 
                    (string_of_indx (n + 2)) 
                    (string_of_indx (n + 3))
                 );
         printBunch (n + 4)
    );
  in
  printBunch 1;;


(**** conflict stuff *****)  
let conflicts s1 s2 = 
  let conflictSet = CapabilitySet.inter s1 s2 in
  (not (CapabilitySet.is_empty conflictSet), conflictSet)
;; 

let eraseAccessesSet s =
  let addToSet pname acc = 
    let name = Updateconfig.accessErase pname in
    CapabilitySet.add name acc
  in
  CapabilitySet.fold (fun pname acc -> addToSet pname acc) 
                     s
                     CapabilitySet.empty
;;

let getReads s =
  CapabilitySet.fold (fun name acc -> match Updateconfig.accessUnpack name with
                            (Updateconfig.SRead, t) | (Updateconfig.SReadWrite, t) -> CapabilitySet.add t acc
                          | _ -> acc 
                     ) 
                     s
                     CapabilitySet.empty
;;

let getWrites s =
  CapabilitySet.fold (fun name acc -> match Updateconfig.accessUnpack name with
                            (Updateconfig.SWrite, t) | (Updateconfig.SReadWrite, t) -> CapabilitySet.add t acc
                          | _ -> acc 
                     ) 
                     s
                     CapabilitySet.empty
;;


let printSituationAtPoint sid l = 
  ignore(Updatedebug.log [Updatedebug.BTAlways]  "==== update point at %d \n" sid);
  List.iter (fun (conflicts, conflictSet, name) ->
               let conflictString = 
                 if conflicts 
                 then "YES : " ^ printDelta conflictSet
                 else "no"
               in
                 ignore(Updatedebug.log [Updatedebug.BTAlways]  "conflicts %s : %s\n" name conflictString);
            )
    l  
;;
(*************************************** Testing ************************************************)
(* usage : $ bin/cilly --doupdate --testmode --enable-debug=Deltasolver foo.c *)
let testDeltaSolver () =

   let mkDV id      = {delta_loc = Cil.locUnknown; delta_id = id; delta_kind = DKVar None} in
   
   let mkDC id name = {delta_loc = Cil.locUnknown; delta_id = id; delta_kind = DKConst name} in
  
   let count = ref 0 in
  
   let newDV () = incr count; mkDV !count in
   let newDC name = incr count; mkDC !count name in
   
   let cset = ref (DeltaConstraintSet.empty) in

   let flow f d f' d' cset = 
      let cset1 = 
          DeltaConstraintSet.add (DCSubset(d',d)) cset in
          DeltaConstraintSet.add (DCSubset(f',f)) cset1
      in

   let lbl2ndx = H.create 13 in
   let newTuple alphaLabel = 
     let f  = newDV() and
         indx = !count and
         d  = newDV() and
         f' = newDV() and
         d' = newDV() in
     let newCset = flow f d f' d' !cset in
     cset := newCset;
     H.add lbl2ndx indx alphaLabel;   
     f, d, f', d'
   in
   let purgeClist l = 
     let doOne s (x,y) = DeltaConstraintSet.add (DCSubset(x,y)) s in
     cset := (List.fold_left doOne !cset l)
   in 
   let (f_d, d_d, f_d', d_d') = newTuple "d" in
   let (f_i, d_i, f_i', d_i') = newTuple "i" in
   let (f_g, d_g, f_g', d_g') = newTuple "g" in 
   let (f_L, d_L, f_L', d_L') = newTuple "L" in
   let (f_1, d_1, f_1', d_1') = newTuple "1" in
   let (f_2, d_2, f_2', d_2') = newTuple "2" in
   let (f_3, d_3, f_3', d_3') = newTuple "3" in
   let (f_4, d_4, f_4', d_4') = newTuple "4" in
   let (f_5, d_5, f_5', d_5') = newTuple "5" in
   let (f_6, d_6, f_6', d_6') = newTuple "6" in
   let (f_7, d_7, f_7', d_7') = newTuple "7" in
   let (f_8, d_8, f_8', d_8') = newTuple "8" in
   let (f_9, d_9, f_9', d_9') = newTuple "9" in
   let (f_h, d_h, f_h', d_h') = newTuple "h" in

   let tupleEnd = !count in
  
   let dispatch = newDC "dispatch" and   
       getreq   = newDC "getreq" and
       fdtype   = newDC "fdtype" and
       loop     = newDC "loop" and
       decode   = newDC "decode" 
   in   

   let cList = [
                 (dispatch, f_i); (dispatch, d_i);
                 (f_1, f_i); (d_1, d_i);
                 (d_2, d_1);
                 (f_d, f_2); (d_d, d_2); (f_2', f_d');
                 (f_3, f_2'); (d_3, d_2');
                 (f_h,f_3); (d_h,d_3); (f_3',f_h');
                 (d_1',d_3');
                 (f_i', f_1'); (d_i', d_1');
                 (loop, f_L); (loop, d_L);
                 (f_4, f_1); (d_4, d_1);
                 (d_5, d_4);
                 (f_g, f_5); (d_g, d_5); (f_5', f_g'); 
                 (f_d, f_6); (d_d, d_6); (f_6', f_d');
                 (f_6, f_5); (d_6, d_5);
                 (d_4',d_6');
                 (f_L,f_7); (d_L,d_7); (f_7',f_L');
                 (f_L', f_4); (d_L', d_4);
                 (f_L, f_8); (d_L, d_8); (f_8', f_L');
                 (f_9, f_8'); (d_9, d_8');
                 (getreq, f_g); (getreq, d_g);
                 (fdtype, f_d); (fdtype, d_d);
                 (decode, f_d); (decode, d_d);
               ] in
   purgeClist cList;
   
   let sol = H.create 13 in
   solveDeltaEagerly !cset sol;
   printSolution_4 sol lbl2ndx tupleEnd

(* let testDeltaSolver () = *)
(*    let mkDV id      = {delta_loc = Cil.locUnknown; delta_id = id; delta_kind = DKVar None} in *)
   
(*    let mkDC id name = {delta_loc = Cil.locUnknown; delta_id = id; delta_kind = DKConst name} in *)
  
(*    let count = ref 0 in *)
  
(*    let f_out = incr count; mkDV !count in *)
   
(*    let makeIfCall (d_in:delta) cset = *)
(*       let d = incr count; mkDV !count in *)
(*       let d' = incr count; mkDV !count in *)
      
(*       let cset = DeltaConstraintSet.add (DCSubset(d,d_in)) cset in *)
(*       let cset = DeltaConstraintSet.add (DCSubset(d',f_out)) cset in *)
(*       let cset = DeltaConstraintSet.add (DCSubset(d',d_in)) cset in *)
(*       let cset = DeltaConstraintSet.add (DCSubset(d,d')) cset in *)
(*       d,cset *)
(*    in *)
   
(*    let g_out = incr count; mkDV !count in *)

(*    let rec ncalls n d cset = *)
(*       let d',cset = makeIfCall d cset in *)
(*       if n = 1 then *)
(*          DeltaConstraintSet.add (DCSubset(g_out,d')) cset *)
(*       else *)
(*          ncalls (n-1) d' cset *)
(*    in *)
   
(*    let g_in = incr count; mkDV !count in *)
(*    let cset = DeltaConstraintSet.empty in *)
(*    let cset = ncalls 2 g_in cset in *)
(* (\* *)
(*    let cset = DeltaConstraintSet.add (DCSubset(dc1,dv1)) cset in *)
(*    let cset = DeltaConstraintSet.add (DCSubset(dc2,dv1)) cset in *)
(*    let cset = DeltaConstraintSet.add (DCSubset(dv1,dv2)) cset in *)
(*    let cset = DeltaConstraintSet.add (DCSubset(dv2,dv3)) cset in *)
(*    let cset = DeltaConstraintSet.add (DCSubset(dc3,dv3)) cset in *)
(* *\) *)
(*    let sol = H.create 13 in *)
(*    solveDelta cset sol; *)
(*    printSolution sol *)


