(* 
 *
 * Copyright (c) 2004-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(* whenever we apply the subtytpe rule 

   void  * (effect X)  <:    T * (effect Y)

   it'd better be the case that: 
   
   Y \subseteq X
   
   In other words, I can't get out things I haven't put in.
*)


open Cil

module H = Hashtbl

open Deltasolver


(*============================================================================*)
(* Monadic state *)
(*============================================================================*)

let noFunction = Cil.emptyFunction "**No FUNCTION**"

type voidConstraints_t = (delta * delta, Cil.location) H.t
type flowConstraints_t = (delta * delta, unit) H.t

module EAMonad = 
struct
  (* input type for suspect analysis functions *)
  type mi  = 
      { 
         eai_id_to_LPS_hash : (int, delta) H.t;
         eai_LPC_hash : flowConstraints_t;
         eai_const_LPS_hash : (string, delta) H.t;
         eai_void_constraints : voidConstraints_t; (* (X, Y) sets, see note at the top of this file *)
         eai_current_fd : Cil.fundec ref;       (* Current function being processed *)
         eai_current_loc: Cil.location ref;      (* Current source location *)
         eai_malloc_fun_hash: Mallocpragma.mallocHash_t;
      }


  (* output type for suspect analysis functions *)
  type 'a mo = 'a
        

  (* Monad type *)
  type 'a m = mi -> 'a mo
    
    
  (* Basic monad operations *)
  let return (x:'a) : 'a m = fun mi -> x
    
  let (>>=) (m: 'a m) (f: 'a -> 'b m) : 'b m = 
    fun mi -> let mo = m mi in f mo mi

  (* The final result: the set of constraints, and a map from update *)
  type mo_result =
      { 
         mo_eff_constraints : flowConstraints_t; (* flow constraints *)
         mo_eff_void_constraints : voidConstraints_t; (* (X, Y) sets, see note at the top of this file *)
         mo_eff_const_LPS_hash : (string, delta) H.t;
      }

  let run (m: 'a m) malloc_fun_hash : mo_result =
    let mi = 
      {
        eai_id_to_LPS_hash = H.create 4099;
        eai_LPC_hash = H.create 4099;
        eai_const_LPS_hash = H.create 4099;
        eai_void_constraints = H.create 4099;
        eai_current_fd = ref noFunction;
        eai_current_loc = ref Cil.locUnknown;
        eai_malloc_fun_hash = malloc_fun_hash;
      }
    in
      (* go *)
      m mi;
      (* done *)
      { 
         mo_eff_void_constraints = mi.eai_void_constraints;
         mo_eff_constraints = mi.eai_LPC_hash;
         mo_eff_const_LPS_hash = mi.eai_const_LPS_hash;
      }


  (* ---- Function needed to make this a syntax monad ---- *)
 
  (* Set/unset/get the current function (for debugging) *)
  let setCurrentFun (fd: Cil.fundec) : unit m = fun mi -> mi.eai_current_fd := fd; return () mi
  let unsetCurrentFun : unit m = fun mi -> mi.eai_current_fd := noFunction; return () mi
  let getCurrentFun : Cil.fundec m = fun mi -> return !(mi.eai_current_fd) mi

  (* Get/set the current location (for debugging) *)
  let setLoc (loc: Cil.location) : unit m = fun mi -> mi.eai_current_loc := loc; return () mi
  let getLoc : Cil.location m = fun mi -> return !(mi.eai_current_loc) mi

  (* Extract monadic state *)
  let getMallocFunHash () = fun mi -> return mi.eai_malloc_fun_hash mi

  let eamGetConstLPS (type_name : string) : delta m =
    (* First we have to translate the type name to remove the wrapper, if nec. *)
    (* let type_name = Updateconfig.makeAbsTypeName type_name in *)
     fun mi ->
        let lps_const = 
           try
              H.find mi.eai_const_LPS_hash type_name
           with Not_found -> 
             (let lps = Deltasolver.freshDelta (DKConst type_name) !(mi.eai_current_loc) in
             let lpsBogus = Deltasolver.freshDelta (DKVar None) !(mi.eai_current_loc) in
             H.add mi.eai_const_LPS_hash type_name lps;
             H.add mi.eai_LPC_hash (lps,lpsBogus) ();
             H.add mi.eai_id_to_LPS_hash lpsBogus.delta_id lpsBogus;
             lps) 
        in
        return lps_const mi


  let eamAddLPCSubset (lp1 : delta) (lp2 : delta) : unit m = 
    fun mi ->
      H.add mi.eai_LPC_hash (lp1,lp2) ();
      return () mi

  let eamAddVoidConstraint (lp1 : delta) (lp2 : delta) : unit m = 
    fun mi ->
      H.add mi.eai_void_constraints (lp1,lp2) !(mi.eai_current_loc);
      H.add mi.eai_LPC_hash (lp1,lp1) ();
      H.add mi.eai_LPC_hash (lp2,lp2) ();
      return () mi

  let eamAddSuspectTypeToLPS (type_name : string) (lp2 : delta) : unit m  = 
    eamGetConstLPS type_name >>= fun lps1 -> 
    ignore(Updatedebug.log [Updatedebug.BTEffects] 
        "adding LPS suspect %d(%s) <= %d\n" lps1.delta_id type_name lp2.delta_id);
    eamAddLPCSubset lps1 lp2

  let eamFreshLPSVar : delta m = fun mi -> 
    let freshlp = Deltasolver.freshDelta (DKVar None) !(mi.eai_current_loc) in
      (* Printf.eprintf "eamFreshLPSvar: %d\n" !(mi.eai_LPS_counter); *)
      H.add mi.eai_id_to_LPS_hash freshlp.delta_id freshlp;
      return freshlp mi


  (* ---- Monadic operations ---- *)
  (* Look up an LPS given it's id *)
  let eamGetLPSByID (key:int) : delta m = fun mi ->
    try
      return (H.find mi.eai_id_to_LPS_hash key) mi
    with 
        Not_found -> 
          Updatedebug.abort (Updatedebug.bug 
                               "(%a) eamGetLPS: effect %d not found\n" 
                               d_loc (!(mi.eai_current_loc)) key)



end



(* Create the generic monad operations *)
module EAMOps = Monad.MonadOps(EAMonad)




(* The rest of the file uses the EAMonad, so make a shorthand name for it and bind 
   its basic operations locally 
*)
module EAM = EAMonad
let (>>=) = EAMonad.(>>=)
let return = EAMonad.return






(*============================================================================*)
(*
  Fresh effect insertion

  The following code inserts fresh live pointer set variables into
  a syntax tree.

*)
(*============================================================================*)

module EffectCreator =
struct
  (* The monad that this effect inserter works over *)
  module M = EAM
  module MOps = Monad.MonadOps(M)

  let (>>=) = EAM.(>>=)
  let return = EAM.return

  (* helper functions *)

  let attrIsEffect a = match a with
      Attr(aname, _) when aname = Updateconfig.eaVarAttrName -> true
    | _ -> false

  let getEffects = List.filter attrIsEffect
  let containsEffects = List.exists attrIsEffect

  (* Return the LPS in the given attribute list *)
  let extractLPS t al : delta M.m = 
    MOps.lift getEffects (return al) >>= fun al' ->
      match al' with 
          [Attr(aname, [AInt delta_id])] -> M.eamGetLPSByID delta_id 

        | _ -> 
            M.getLoc >>= fun loc ->
              Updatedebug.abort (Updatedebug.bug "(%a) extractLPS: LPS not found in type! %a\n" 
                                   d_loc loc d_plaintype t)

  let extractLPSfromType t : delta M.m =
    extractLPS t (Cil.typeAttrs t)


  let addLPStoAttr (al:Cil.attributes) (tag:string) (lps:delta) : Cil.attributes M.m = 
    return (Cil.addAttribute (Attr(tag, [AInt lps.delta_id])) al)

  (* This function acts non-inductively, inserting a fresh variable in the 
     outer type constructor of: 
       - TInt
       - TFun
       - TPtr
  *)

  let addFreshAttr (al:Cil.attributes) : Cil.attributes M.m = 
     EAM.eamFreshLPSVar >>= fun lps -> addLPStoAttr al Updateconfig.eaVarAttrName lps

  let insertEffectInOuterConstructor (typ: Cil.typ) : Cil.typ M.m = 
    let al = Cil.typeAttrs typ in
      (* Only insert types if we haven't before *)
      if containsEffects al
      then
        return typ
      else
         M.getLoc >>= fun loc ->
        (ignore (Updatedebug.log [Updatedebug.BTEffects] "(%a)Inserting effect in %s\n" d_loc loc (Updatedebug.ppType typ));
        begin 
           addFreshAttr al 
           >>= fun al' -> return (Cil.setTypeAttrs typ al')

(* match typ with  *)
(*             TInt (ik,al)             -> addAttr al >>= fun al' -> return (TInt (ik, al')) *)
(*           | TFun(tret,targs,isva,al) -> addAttr al >>= fun al' -> return (TFun(tret,targs,isva,al')) *)
(*           | TPtr(t,al)               -> addAttr al >>= fun al' -> return (TPtr(t, al')) *)
(*           | TArray(t,e_opt,al)       -> addAttr al >>= fun al' -> return (TArray(t,e_opt,al')) *)
(*           | _                        -> return typ  *)
        end)
    
end



(* Create fresh effect inserter *)
module EAAGenFreshEffects = 
  Effectinserter.EffectInserter(EffectCreator)(EAM)








(*============================================================================*)
(* 
  Main algorithm

  This traverses the syntax collecting constraints as given in the suspect 
  alias analysis. 

*)
(*============================================================================*)



open Ciltransform
module TypFold = Ciltransform.TypFold(EAM)



(*
  Given a type typ and a LPS lps, add constraints to force the lps to contain  
  all of the live pointer sets on all pointer annotations in typ (and
  possibly the function ones too). 
*)

let collectLPSets (getFunEffs:bool) (lps:delta) (typ: Cil.typ) : unit EAM.m =
  let collectLPSetsTyp (ttyp: (unit,Cil.exp,unit,unit) ttyp) : unit EAM.m =
    let includeEffectsInAttr (al: Cil.attributes) : unit EAM.m = 
      EAMOps.lift EffectCreator.getEffects (return al) >>= fun delta_id_list ->
        match delta_id_list with 
            (* There should be exactly one effect in the type *)
            [Attr(aname, [AInt delta_id])] -> 
              EAM.eamGetLPSByID delta_id >>= fun lps' ->
      ignore(Updatedebug.log [Updatedebug.BTEffects] 
          "collectLPsets: (typ)%d <= (set)%d\n" lps'.delta_id lps.delta_id);
              EAM.eamAddLPCSubset lps' lps
              
          | _ -> EAM.getLoc >>= fun loc -> Updatedebug.abort (Updatedebug.bug 
                                                                "(%a) collectLPSets: no effect in int\n" d_loc loc)
    in
    match ttyp with
   TTInt(_, al)
      | TTPtr(_, al)
      | TTArray(_, _, al) -> includeEffectsInAttr al 
      | TTFun(_, _, _, al) -> if getFunEffs then includeEffectsInAttr al else return ()
      | _                 -> return ()
  in            
  let avoidFuns (typ:Cil.typ) : unit option =
    match typ with
      TFun(_,_,_,_) -> 
   (* ignore (Updatedebug.log [Updatedebug.BTEffects] "Ignoring function type %a\n" d_type typ);*)
   Some ()
    | _ -> None
  in
  let collectTypFoldFuns = {TypFold.iterDefaultTypFoldFuns 
                            with TypFold.tf_typ = collectLPSetsTyp;
                  TypFold.tf_typ_alt = avoidFuns;} 
  in
  TypFold.typFold collectTypFoldFuns typ

  (* helper functions *)   
  let getCanonicalTypeName t =
     let res = match t with

         TInt _ | TEnum _ -> 
            let bs = Cil.bitsSizeOf t in
            "int_" ^ (string_of_int bs)
           
       | TFloat _ -> 
            let bs = Cil.bitsSizeOf t in
            "float_" ^ (string_of_int bs)       

       | TPtr _ | TArray _ | TFun _ ->
            "ptr"
     
       | TBuiltin_va_list attr -> "__builtin_va_list"

       | _ -> Updatedebug.abort (Updatedebug.bug "getCanonicalTypeName called for illegal type (%a)\n" 
                                 d_plaintype t)
     in
(*      ignore(Updatedebug.log [Updatedebug.BTEffects] "getCanonicalTypeName :  %s\n" res); *)
     Updateconfig.variable_prefix ^ "_base_type_" ^ res

  let isBaseType typeName = Cil.startsWith (Updateconfig.variable_prefix ^ "_base_type_") typeName


module AnalysisSubtypeActions = 
struct
  module M = EAMonad
  module MOps = Monad.MonadOps(EAMonad)


  (* private function: extract live pointer sets from attributes and add 1<= 2 constraint *)
  let addConstraint (t1:Cil.typ) (al1:Cil.attributes) (t2:Cil.typ) (al2:Cil.attributes) : unit M.m =       
    EffectCreator.extractLPS t1 al1 >>= fun lps1 -> 
    EffectCreator.extractLPS t2 al2 >>= fun lps2 -> 
      ignore(Updatedebug.log [Updatedebug.BTEffects]
        "subtype constraint: %a <= %a ===> %d <= %d\n"
          Cil.d_type t1 Cil.d_type t2 lps1.delta_id lps2.delta_id);
      EAM.eamAddLPCSubset lps1 lps2

  let addVoidConstraint (t1:Cil.typ) (al1:Cil.attributes) (tabs:Cil.typ) : unit M.m =       
    EffectCreator.extractLPS t1 al1 >>= fun lps1 -> 
    EAM.eamGetConstLPS (Updatecilutil.getAbstractTypeName tabs) >>= fun lpsC -> 

      ignore(Updatedebug.log [Updatedebug.BTEffects] 
        "void constraint: %a <= %a ===> %d <= %d\n" 
          Cil.d_type tabs Cil.d_type t1 lpsC.delta_id lps1.delta_id);
      EAM.eamAddVoidConstraint lpsC lps1

  (* Called when SubtypeCmp performs a standard subtype step *)
  let subtype (t1: Cil.typ) (t2: Cil.typ) : unit M.m =
     let compHash = H.create 7 in 

     let rec compSubtype (t1 : Cil.typ) (t2 : Cil.typ) : unit M.m =
        let id = return () in
        let field_iter (fi1, fi2) = localSubtype fi1.ftype fi2.ftype in
           match t1,t2 with 
	             TPtr(TComp(ci1,_),attr1), TPtr(TComp(ci2,_),_) when ci1.cname = ci2.cname && not (H.mem compHash ci1.cname) ->
                    ignore(Updatedebug.log [Updatedebug.BTEffects] "compSubtype %s\n" ci1.cname);
                    H.add compHash ci1.cname true;
                    EAMOps.iter field_iter (List.combine ci1.cfields ci2.cfields)

              | _ -> id

     and localSubtype (t1: Cil.typ) (t2: Cil.typ) : unit M.m =
   ignore(Updatedebug.log [Updatedebug.BTEffects] "localSubtype %s          --->       %s\n" (Updatedebug.ppType t1) (Updatedebug.ppType t2));
        match t1,t2 with 
           | TFun(t1',args_opt1,isVA1,attr1), TFun(t2',args_opt2,isVA2,attr2) ->  
                addConstraint t2 attr2 t1 attr1

           | TPtr(TComp(ci1,_),attr1), TPtr(TComp(ci2,_),_) when ci1.cname = ci2.cname && not (H.mem compHash ci1.cname) ->
                compSubtype t1 t2

           (* Possible bug (may be fixed): missing annotation *)
           | TPtr(t1',attr1), TPtr(t2',attr2) ->
                addConstraint t2 attr2 t1 attr1
   
           | _ -> return ()

  in


   ignore(Updatedebug.log [Updatedebug.BTEffects] "standard %s          --->       %s\n" (Updatedebug.ppType t1) (Updatedebug.ppType t2));
    match t1,t2 with 
      | TFun(t1',args_opt1,isVA1,attr1), TFun(t2',args_opt2,isVA2,attr2) ->  
          addConstraint t2 attr2 t1 attr1

      | TPtr(TComp(ci1,_),attr1), TPtr(TComp(ci2,_),attr2)  ->
           addConstraint t2 attr2 t1 attr1 (* >>= fun () -> *)
(*            compSubtype t1 t2 *)

      (* Possible bug (may be fixed): missing annotation *)
      | TPtr(t1',attr1), TPtr(t2',attr2) ->
          addConstraint t2 attr2 t1 attr1
   
      | _ -> return ()


  (* Called when a physical subtype step is performed, i.e. a pointer
     to a record is a subtype of a pointer to the records first field.

     This is like taking the address of that field, so make the 
     pointer t2 suspect with respect to the (comp) type of t2.
   *)
  let physicalSubtype (t1: Cil.typ) (t2: Cil.typ) : unit M.m =
    match t1,t2 with 
        TPtr((TComp(ci,_) as tcomp),_), TPtr(_,attr2) ->
          EffectCreator.extractLPS t2 attr2 >>= fun lps ->
          EAMonad.eamAddSuspectTypeToLPS (Updatecilutil.getAbstractTypeName tcomp) lps 


      | _ ->
          M.getLoc >>= fun loc ->
          EAMOps.lift (fun () -> ignore (Updatedebug.log [Updatedebug.BTEffects] 
                                         "(%a) Analysis:SubtypeConstraints: physical subtype called for non-record type: %s and %s\n" 
                                         d_loc loc (Updatedebug.ppType t1) (Updatedebug.ppType t2)
                                      )
                    ) (return ())

    
  (* Called when one of the types is external *) 
  let externalSubtype (t1: Cil.typ) (t2: Cil.typ) : unit M.m = (return ())

  (* Called when trying to determine a subtype between comp pointer type
     and another type where the types neither lead to a valid subtype or to a valid 
     physical subtype *)
  let ptrIncompatible (t1: Cil.typ) (t2: Cil.typ) (msg: string) : unit M.m = (return ())

  (* Called when C would not accept the subtype *)
  let invalidCSubtype (t1: Cil.typ) (t2: Cil.typ) (msg: string) : unit M.m = (return ())

  let toVoidPtr t1 t2 =  (*  t1 -> void *        *)
   ignore(Updatedebug.log [Updatedebug.BTEffects] "toVoidPtr %a          --->       %a\n" Cil.d_type t1 Cil.d_type t2);
     match t1, t2 with 
          TPtr((TNamed _ | TComp _) as tabs,attr1), TPtr(_,attr2) ->
                EffectCreator.extractLPS t2 attr2 >>= fun lps ->
                EAMonad.eamAddSuspectTypeToLPS (Updatecilutil.getAbstractTypeName tabs) lps 

        | TPtr(TVoid _,attr1), TPtr(_,attr2) -> addConstraint t1 attr1 t2 attr2

        | TPtr(t1',attr1), TPtr(_,attr2) -> 
                EffectCreator.extractLPS t2 attr2 >>= fun lps ->
                EAMonad.eamAddSuspectTypeToLPS (getCanonicalTypeName t1') lps


        | _ ->  addConstraint t1 (Cil.typeAttrs t1) t2 (Cil.typeAttrs t2) 

  let fromVoidPtr t1 t2 =  (*  void * -> t2       *)
   ignore(Updatedebug.log [Updatedebug.BTEffects] "fromVoidPtr %a          --->       %a\n" Cil.d_type t1 Cil.d_type t2);
    match t1, t2 with 
        TPtr(_,attr1), TPtr((TNamed _ | TComp _) as tabs,attr2)  -> addVoidConstraint t1 attr1 tabs
      | TPtr(_,attr1), TPtr(_,attr2) -> addConstraint t1 attr1 t2 attr2
      | _ -> addConstraint t1 (Cil.typeAttrs t1) t2 (Cil.typeAttrs t2) 

end


module AnalysisSubtypeConstraints = Subtype.SubtypeCmp(AnalysisSubtypeActions)(EAM)

let subTypeConstraints t1 t2 err = 
  try
(*     ignore(Updatedebug.log [Updatedebug.BTEffects] *)
(*         "Doing subtype constraint: %a <= %a : %s\n" *)
(*         Cil.d_type t1 Cil.d_type t2 (Pretty.sprint 80 err)); *)
    AnalysisSubtypeConstraints.subtypeTraverse t1 t2
  with Failure m ->
    (ignore(Updatedebug.abort (Updatedebug.bug "|%s|: %s\n" (Pretty.sprint 80 err) m));
     return ())



open Ciltransform

module SAReturnTypes = 
struct
  type typ       = unit
  type exp       = Cil.typ
  type lval      = Cil.typ
  type lhost     = Cil.typ
  type offset    = unit
  type instr     = unit
  type stmt      = unit
  type stmtkind  = unit
  type block     = unit
  type compinfo  = unit
  type fieldinfo = unit
  type init      = unit
  type initinfo  = unit
  type typeinfo  = unit
  type varinfo   = Cil.typ
  type fundec    = unit
  type global    = unit
end


(* Traversal of syntax *)
module EATransform = Ciltransform.MTransform (EAM) (SAReturnTypes)


module EAFoldFuns = 
struct
  (* Uninteresting map functions *)
  let ftyp (typ:typ) (ttyp: (unit,Cil.typ,typeinfo,compinfo) ttyp) : unit EAM.m = return ()
  let fstmt (stmt:stmt) (tstmt: unit tstmt) : unit EAM.m = return ()
  let fblock (b:block) (tblock: unit tblock) : unit EAM.m  = return ()
  let fcompinfo (ci:compinfo) (tcompinfo: unit tcompinfo) : unit EAM.m = return ()
  let ffieldinfo (fi:fieldinfo) (tfieldinfo: unit tfieldinfo) : unit EAM.m = return ()
  let finit (init:init) (tinit: (unit,Cil.typ,unit) tinit) : unit EAM.m = return ()
  let finitinfo (ii:initinfo) (tinitinfo: (unit tinitinfo)) : unit EAM.m = return ()
  let ftypeinfo (ti:typeinfo) (ttypeinfo: unit ttypeinfo) : unit EAM.m =  return ()
  let foffset (offset:offset) (toffset: (Cil.typ,unit) toffset) (lval:Cil.lval) : unit EAM.m = return ()
  let ffundec (fd:fundec) (tfd: (Cil.typ,unit,unit) tfundec) : unit EAM.m = return ()
  let flhost (lhost:lhost) (tlhost: (Cil.typ,Cil.typ) tlhost) : unit EAM.m = return ()
  let flval ((lhost,offset):lval) ((tlhost_t,toffset): (Cil.typ,unit) tlval) : unit EAM.m = return ()

  (* The cases mentioned here have special processing - i.e. done in an alt hanler.
     This is because they need extra things going *around* the recursive calls *)
  let should_process (pa:Ciltransform.process_arg) = true

  let fexp (exp:exp) (texp: (unit,Cil.typ,Cil.typ) texp) : Cil.typ EAM.m =
    let addEmptyLPSToAttr (al:Cil.attributes) : Cil.attributes EAM.m =
      EffectCreator.addLPStoAttr al Updateconfig.eaVarAttrName Deltasolver.empty_delta in
    let uintConst = addEmptyLPSToAttr [] >>= fun al -> return (TInt(IUInt,al))
    in
      match exp,texp with 
          (* ---- Routine cases ---- *)
          _, TConst c             ->  
                                        EAMOps.lift Updatecilutil.properTypeOf (return exp) 
                     >>= fun t ->       EffectCreator.insertEffectInOuterConstructor t


        | _, TLval lval_t         -> return lval_t

        | _, TSizeOf _  
        | _, TSizeOfE _        
        | _, TSizeOfStr _
        | _, TAlignOf _  
        | _, TAlignOfE _          -> uintConst

        | UnOp (unop, e, typ), _  -> return typ 


        (* ---- Interesting cases ---- *)
        | _, TStartOf lval_t -> 
            begin match (Cil.unrollType lval_t) with 
                TArray(t,e_opt,al) -> return (TPtr(t,al))
              | _            -> 
                  EAM.getLoc >>= fun loc ->
                    Updatedebug.abort (Updatedebug.bug 
                     "(%a) SuspectAliasAnalysis: StartOf lval which is not an array (%a)!"
                     d_loc loc d_type lval_t) 
            end
 
        | AddrOf _, _ -> 
             begin
                EAMOps.lift Updatecilutil.properTypeOf (return exp) >>= fun exp_t ->
                   begin match exp_t with 
                         TPtr( t, al_ptr ) ->  EffectCreator.addFreshAttr al_ptr 
                            >>= fun al_ptr' -> return (TPtr(t, al_ptr'))

                        | _ -> return exp_t
                   end
            end
        | BinOp(binop,e1,e2,typ), TBinOp(_,e1_t,e2_t,ttyp) ->
            subTypeConstraints e1_t typ (Pretty.dprintf "%a" d_exp exp) >>= fun () ->
       (* MWH: do we need both? *)
            subTypeConstraints e2_t typ (Pretty.dprintf "%a" d_exp exp) >>= fun () ->
            return typ

        | CastE(typ,e), TCastE(_,e_t) (* when not (Updatecilutil.isNullInitialiser !(Cil.currentLoc) typ e) *) ->
             ignore(Updatedebug.log [Updatedebug.BTEffects] "Cast %s -> %s\n" (Updatedebug.ppType e_t) (Updatedebug.ppType typ));

          (*   skip null initialisers: 
               T* = (void * ) 0                                             
           *)
          if (Updatecilutil.isNullInitialiser !(Cil.currentLoc) typ e)   
          then 
            return typ
          else
             (               subTypeConstraints e_t typ (Pretty.dprintf "%a" d_exp exp) 
              >>= fun () ->  return typ)

        | _ -> EAM.getLoc >>= fun loc ->
                    Updatedebug.abort (Updatedebug.bug 
                     "(%a) SuspectAliasAnalysis: fexp with unexpexted arg\n%a\n!"
                     d_loc loc d_exp exp) 

  let flval ((lhost,offset):lval) ((tlhost_t,toffset): (Cil.typ,unit) tlval) : Cil.typ EAM.m = 
     return (Updatecilutil.properTypeOfLval ((lhost,offset)))

  let flhost (lhost:lhost) (tlhost: (Cil.typ,Cil.typ) tlhost) : Cil.typ EAM.m =
    let aux t =
      match t with 
   TPtr(t,_)
      | TArray(t,_,_) -> t
      | _ ->  raise (Ciltransform.CILTransformEx "EAFoldFuns: flhost with unexpected arg (2)")
    in
    match lhost,tlhost with 
        Var vi, TVar tvi  -> return vi.vtype
      | _, TMem t         -> return (aux (Cil.unrollType t))
(* raise (Ciltransform.CILTransformEx ("EAFoldFuns: flhost arg"^(Updatedebug.ppType t)) ) *)
      | _ ->  raise (Ciltransform.CILTransformEx "EAFoldFuns: flhost with unexpected arg")

  let fvarinfo (vi: varinfo) (tvarinfo: unit tvarinfo) : Cil.typ EAM.m =
    return vi.vtype
  
  let fglobal (global:global) (tglobal: (unit,unit,unit,Cil.typ,unit,unit) tglobal) : unit EAM.m =
    match tglobal with
      TGVarDecl (typ,loc) 
    | TGVar     (typ,_,loc) -> return ()
    | _ -> return ()

  let finstr (instr:instr) (tinstr: (Cil.typ,Cil.typ) tinstr) : unit EAM.m = 
(*      let _ = ignore(Updatedebug.log [Updatedebug.BTEffects] "finstr '%a'\n" d_instr instr) in *)

     let getOpt (opt: 'a option) (none_val: 'a) = match opt with None -> none_val | Some x -> x in 
        EAM.getMallocFunHash () >>= fun mallocFunHash ->

        match (instr,tinstr) with 

              (_,TSet (lval_t,exp_t,loc))   ->
                 subTypeConstraints exp_t lval_t (Pretty.dprintf "%a" d_instr instr)
            
           | (Call(_, Lval(Var vi, NoOffset), _, _), TCall (lval_result_t_opt, exp_func_t, exp_arg_t_list, loc)) when 
                   Commoncilutil.isBuiltin vi.vname ||
                   ((Updateconfig.isConFunc vi.vname) && (Namedtypes.isConAbsFuncName vi.vname)) || 
                   ((Updateconfig.isAbsFunc vi.vname) && (Namedtypes.isConAbsFuncName vi.vname)) ||
                   Updateconfig.isUpdateFunc vi.vname ||
	                Mallocpragma.isMallocFunc mallocFunHash vi.vname ->


                return ()

           | (Call(_, _, _, _), TCall (lval_result_t_opt, exp_func_t, exp_arg_t_list, loc)) ->

                begin match Cil.unrollType exp_func_t with 
                      (* First: generate subtyping constraints between the *)
                      (* parameters of the function and the arguments provided to *)
                      (* this call *)
                      TPtr(TFun(tret,targs_opt,isva,al),_) | TFun(tret,targs_opt,isva,al) ->
                         let numFixedArgs = List.length (getOpt targs_opt []) in 
                         let ( fixedArgs_t, varArgs_t )   = 
                            Commonutil.list_split numFixedArgs exp_arg_t_list in 
                            (* TODO: deal with variable arguments here?  types are in varArgs_t *)
                         let fixedArgPairs = 
                            List.combine fixedArgs_t 
                               (List.map (fun (s,t,al)-> t)
                                   (getOpt targs_opt [])) in
                            (EAMOps.iter (fun (t1,t2) -> 
                                             subTypeConstraints t1 t2 (Pretty.dprintf "%a" d_instr instr)) fixedArgPairs >>= fun () ->
                                begin match lval_result_t_opt with 
                                      None        -> return ()
                                   | Some lval_t -> subTypeConstraints tret lval_t (Pretty.dprintf "%a" d_instr instr)
                                end) >>= fun () ->
                               ignore(Updatedebug.log [Updatedebug.BTEffects] 
                                         "fun call constraints: %s:%d\n"
                                         loc.file loc.line  );
                               return ()
                                  
                               | _ -> 
                                    EAM.getLoc >>= fun loc -> Updatedebug.abort (Updatedebug.bug 
                                                                                    "(%a) SuspectAnalysis:finstr: in call, function expression does not have a function type : %a\n" 
                                                                                    d_loc loc d_plaintype exp_func_t)
                end
                   
           | _ -> return ()



  let fstmtkind (skind:stmtkind) (tstmtkind: (Cil.typ,unit,unit,unit) tstmtkind) : unit EAM.m = 
    match skind,tstmtkind with 
      | Return (Some e, _),TReturn (Some e_t, loc) ->
          (* Add constraint between type e_t and return type of function *)
          EAM.getCurrentFun >>= fun fd ->
          EAMOps.lift Updatecilutil.getReturnType (return fd) >>=
          fun rt -> subTypeConstraints e_t rt (Pretty.dprintf "return %a" d_exp e)
          
      | _ -> return ()

end


(* build a custom fold based on the functions given above *)
module EAFold = EATransform.MFold (EAFoldFuns)


let rec alt_doglobal (g:Cil.global) : unit EAM.m = match g with

     GFun (fd,loc) ->
        if not (Namedtypes.isConAbsFuncName fd.svar.vname) then
           begin
              EAM.setCurrentFun fd >>= fun () ->
              (* now process the function body *)
              EAFold.fundecFoldWithAlt alt_handlers fd >>= fun fd_r ->
              EAM.unsetCurrentFun >>= fun () ->
              return ()
           end
        else 
           return ()

   | _ -> raise (Ciltransform.CILTransformEx "SAFoldFuns: doglobal with unexpected arg")

and alt_handlers = 
  {
    EATransform.default_alt_handlers with
    EATransform.should_process = EAFoldFuns.should_process; 
    EATransform.doglobal = alt_doglobal;
  };;




let printVoidConstraints delta_solution void_constraints = 
   let print_one_constraint d1 d2 loc = 
      let ds id = 
         try
            H.find delta_solution id
         with Not_found ->
                Updatedebug.abort (Updatedebug.bug "printVoidConstraints : can't find set for delta_id %d in solution hash\n" id)
      in
      let setNonUpd ds = 
         let setNonUpd_iter tn = match Updateargs.args.Updateargs.arg_fromvoid with 

            | Updateargs.UnsoundIgnore -> 
                 ()

            | Updateargs.UnsoundWarn -> 
                 ignore (Updatedebug.log [Updatedebug.BTAlways] "%a Unsafe downcast from void * :  '%s'\n" d_loc loc tn)

            | Updateargs.UnsoundKillTypeShallow  
            | Updateargs.UnsoundKillTypeDeep -> 
                 Namedtypes.setAbsTypeNameNonUpdatable tn loc "Unsafe downcast from void *"

            | Updateargs.UnsoundDie -> 
                 Updatedebug.abort (Updatedebug.error "%a : type '%s' unsafely downcasted from void * and you requested bailing in this case" d_loc loc tn)
         in
         CapabilitySet.iter setNonUpd_iter ds
      in
      let ds1 = ds d1.delta_id 
      and ds2 = ds d2.delta_id in
      let isSubset = CapabilitySet.subset ds1 ds2 in
      let () = ignore(Updatedebug.log [Updatedebug.BTAlways] "printVoidConstraints: %a     %d '%s' <= %d '%s'  subset:%b \n" d_loc loc d1.delta_id (printDelta ds1) d2.delta_id (printDelta ds2) isSubset) in
        if Updateargs.args.Updateargs.arg_currentMode <> Updateargs.NormalMode

        then () 
        else (* Normal Mode *)
          begin
            if not isSubset 
            then setNonUpd ds1
            else 
              begin
                (* careful : we might have smth like I <= {I, int} and we should kill I,
                                      or   smth like I <= {I, J} which is unsound but tolerable *)

                let isUpd x = not (isBaseType x)   (* Namedtypes.isUpdatableByAbsName x *) in
                let allUpdatable = CapabilitySet.for_all isUpd ds2 in
                  if not allUpdatable
                  then setNonUpd ds1
              end
          end
   in
   H.iter (fun (d1,d2) loc -> print_one_constraint d1 d2 loc) void_constraints;;


(*============================================================================*)
(* main *)
(*============================================================================*)

let doEffectsAnalysis (f: Cil.file) (mallocFunHash : Mallocpragma.mallocHash_t) = 
  let m =
    let doGlobal g = EAFold.globalFoldWithAlt alt_handlers g in

    (* Populate the syntax with fresh live pointer sets *)
    EAAGenFreshEffects.genFreshEffectsGlobal f >>= fun () ->
    ( (* Updatedebug.d_file f; *)
    (* do the analysis *)
    EAMOps.iter doGlobal f.globals )
  
  in
  let monad_result = EAM.run m mallocFunHash in
  let () = Stats.time "Stats : removeEffectAttrVisitor - doDSU" Cil.visitCilFile (new Updatecilutil.removeEffectAttrVisitor(Updateconfig.eaVarAttrName)) f in
  let voidptr_dcs = H.fold (fun (d1,d2) () acc ->
                                     (* ignore(Updatedebug.log [Updatedebug.BTAlways] "BOO %d <= %d\n" d1.delta_id d2.delta_id); *)
                                     DeltaConstraintSet.add (DCSubset (d1,d2)) acc) monad_result.EAMonad.mo_eff_constraints DeltaConstraintSet.empty in

  let delta_solution = (H.create 4099) in   
  Stats.time "Stats : solveDelta" solveDeltaEagerly voidptr_dcs delta_solution;
  printVoidConstraints delta_solution monad_result.EAMonad.mo_eff_void_constraints;;


