(* 
 *
 * Copyright (c) 2004-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Michael Hicks       <mwh@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(* A visitor to set all external types non-updatable *)

open Cil

let externString = "/usr/include" (* FIXME: should make into a pragma *)

(* this location is extern if it's in a file in the /usr/include hierarchy *)
let isExtern loc =
  Cil.startsWith externString loc.file


class externalTypesElimVisitor vd : cilVisitor = object(self)
  inherit nopCilVisitor

  method vtype t =
     begin
       match t with
           TComp(ci,_) when not ci.cdefined ->
              begin 
                 ignore (Updatedebug.log [Updatedebug.BTExternaltypeselim] "Eliminating undefined struct %a\n" d_plaintype t);
                 Namedtypes.setTypeExternal t Cil.locUnknown "struct undefined"; ()
              end
              
         | _ -> ()
     end;
     DoChildren


  (* Any types mentionned in the prototype of an external function must be set-external *)
  method vglob g =
     let sizeKnown t = try
        ignore (Cil.bitsSizeOf t);
        true
     with _ -> 
        false
     in

     let doit t loc =
       let mark t = 
          Namedtypes.setTypeExternalDeep t loc "appeared in prototype of external function" in
        match Cil.unrollType t with
            TFun (ret_t, args, _, _) -> 
               let typs_in_fundec = ret_t :: (Commoncilutil.argsToTypesList args) in 
               List.iter mark typs_in_fundec
          | _ -> ()
     in 
   
     match g with 
        GVarDecl(vi,loc) ->
           begin 
           match vi.vstorage with
              (* tricky ! when we're in patch mode some functions defined in 
                 previous versions are now seen as extern *)
              Extern -> 
                 if Updateargs.args.Updateargs.arg_currentMode = Updateargs.PatchMode &&
                    Versiondata.isExistingFunction vd vi.vname
                 then
                    ()
                 else 
                    doit vi.vtype loc

            | _      -> ()
           end;
           SkipChildren

      | GType (ti,loc) ->
          begin
             if isExtern loc then 
                (* this type is extern itself *)
                Namedtypes.setAbsTypeNameExternal ti.tname loc "defined in external header"
             else 
                if (not (sizeKnown (TNamed(ti, [])))) then
                   Namedtypes.setTypeExternal ti.ttype loc "size unknown"
                else
                   if Namedtypes.isAbstractType ti.ttype then
                      if Namedtypes.isExternalType ti.ttype then
                         (* this type is a named tpye for an extern type *)
                         Namedtypes.setTypeExternal ti.ttype loc "rhs of typedef in external header"
          end;
          SkipChildren

      | GCompTag (ci,loc) -> 
          (* FIX: hack to make anon types considered external *)
          let (unupd,reason) = 
            if (isExtern loc) then 
               (true,"defined in external header")
            else if (Commoncilutil.isAnonymous ci.cname) then
               (true,"Cil-generated anonymous definition")
            else if (not (sizeKnown (TComp(ci, [])))) then
               (true, "size unknown")
            else
               (false,"") 
          in
             begin
                if unupd then
                   Namedtypes.setTypeExternal (TComp(ci,[])) loc reason
             end;
             SkipChildren

      | GCompTagDecl (ci,loc) ->
          (* FIX: hack to make anon types considered external *)
          let (unupd,reason) = 
            if (isExtern loc) then 
               (true,"defined in external header")
            else if (Commoncilutil.isAnonymous ci.cname) then
               (true,"Cil-generated anonymous definition")
            else 
               (false,"") 
          in
          if unupd then
            Namedtypes.setTypeExternal (TComp(ci,[])) loc reason;
          SkipChildren

      | _ -> SkipChildren
end


