(* 
 *
 * Copyright (c) 2004-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



open Cil

module H = Hashtbl

(* keeps the names of all internal (defined in this module) functions *)
let internalFunctionsNames : (string, bool) H.t = (H.create 4099)

(* keeps the names of all external functions taking void [*]+ arguments *)
let externalFunctionsVoidStarArgsNames : (string, Cil.typ) H.t = (H.create 4099)


let printExternalFunctionsVoidStarArgs () = 
  begin

    ignore (
      Updatedebug.log [Updatedebug.BTExternalvoidstarfunctions] "--------- external functions taking (void *) arguments : %8d   ------- \n"
              (Commonutil.hashtbl_size externalFunctionsVoidStarArgsNames));

    H.iter (fun name typ -> ignore (Updatedebug.log [Updatedebug.BTExternalvoidstarfunctions] "%s : %s \n" name (Updatedebug.ppType typ)(* d_plaintype typ *)))  
      externalFunctionsVoidStarArgsNames;
  end

class externalFunctionsVoidStarArgsVisitor : cilVisitor = object(self)
  inherit nopCilVisitor

  method vvrbl v =  

    let isExternalFunction fname = not (H.mem internalFunctionsNames fname) in
      
    let hasVoidStarArguments t = 
      (* look for void *, void ** , etc *)
      let rec isVoid  = function
          TVoid _ -> true
        | TPtr(iv',_) -> isVoid iv'
        | _ -> false in
      let isVoidStar ivs = match Cil.unrollTypeDeep ivs with
          TPtr(ivs',_) -> isVoid ivs'
        | _ -> false in

        (* does the function have any void [*]+ args ? *)
        match Cil.unrollType t with
            TFun (_, args, _, _) -> List.exists isVoidStar (Commoncilutil.argsToTypesList args)
          | _ -> false
    in
    begin
      if (isExternalFunction v.vname) &&
         (hasVoidStarArguments v.vtype) 
      then
        Commonutil.hashtbl_add_once externalFunctionsVoidStarArgsNames v.vname v.vtype;
    end;
      DoChildren
end


let default_external_functions_void_star_args (f : file) : unit =
  begin
    (* collect names of internal functions *)
    Cil.iterGlobals f (function GFun(fi, _) -> H.add internalFunctionsNames fi.svar.vname true | _ -> ());

    (* collect external, void [*]+ args functions *)
    visitCilFileSameGlobals (new externalFunctionsVoidStarArgsVisitor) f;
    
    (* print what we've got *)
    printExternalFunctionsVoidStarArgs ();

  end

