(* 
 *
 * Copyright (c) 2004-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Michael Hicks       <mwh@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



open Cil

module H = Hashtbl

type loop_processing_t = LoopNoAction | LoopUpd | LoopTxCheckin | CodeExtract;;



type loopext_ret_t = BreakV of int | ContV of int | ReturnV of int | GotoV of int;;

let retval_of_stmt v = match v with 
      BreakV x -> x
    | ContV x -> x
    | ReturnV x -> x
    | GotoV x -> x
;;

let loop_breakv = BreakV 0;;
let loop_contv  = ContV 1;;
let loop_retv   = ReturnV 2;;

let newExtractionStyle = false;;

let ret v loc = 
  Commoncilutil.mkStmt (Return (Some (Cil.integer (retval_of_stmt v)),loc))
;;

let rec lmem l1 l2 =
  match l1 with
  [] -> None
  | (Label(h,_,_))::t ->
     if (List.mem h l2) then Some h
     else lmem t l2
  | h::t -> lmem t l2

let lfilter l1' l2' =
  let rec aux l1 l2 res =
    match l1 with
      [] -> []
    | ((Label(l,_,_)) as h)::t ->
	if (List.mem l l2) then aux t l2 (h::res)
	else                    aux t l2 res
    | h::t -> aux t l2 (h::res) in
  aux l1' l2' []

let deref vi off = (Mem(Lval(Var vi, NoOffset)), off)
 
(* Makes a struct definition based on the given varinfos *)
let mkLoopStateStruct vlist n =
   let ci = Commoncilutil.makeCompInfo true n in
   let fi = List.map (function vi -> Commoncilutil.makeField ci vi.vname (TPtr(vi.vtype,[]))) vlist in
   ci.cfields <- fi;
   ci

(* converts a local var reference into one through the state struct *)
let loopDataRef state_vi vi off compType =
  let fi = Commoncilutil.getFieldinfoByName compType vi.vname None in
  let dereflval = (Mem(Lval(Var state_vi, NoOffset)), (Field (fi,NoOffset))) in
  (Mem(Lval (dereflval)), off)

let matchSid sopt sid =
  match sopt with
    Some s -> s.sid = sid
  | None -> false


(* collect all labels used in a statement *)
class labelGatherer res : cilVisitor = object(self)
  inherit nopCilVisitor
  method vstmt s =
    let doOneLabel l = match l with 
        Label(s,_,_) -> H.replace res s ()
      | _ -> ()
    in
    List.iter doOneLabel s.labels;
    DoChildren  
  end


(* mirror Cil's strategy for picking target labels when there's more than one *)
let getTargetLabel sref = 
  let rec pickLabel = function
      [] -> None
    | Label (l, _, _) :: _ -> Some l
    | _ :: rest -> pickLabel rest
  in
  pickLabel !sref.labels
;;

(* used by the visitor below; avoids an infinite loop *)
let mode = ref false

(* create the loop function from the original loop body.  It should now
   refer to the state variable for all its local variables, and use
   the return argument to return its actual result.  All break, continue, 
   and return statements return return codes. 
*)
class massageLoopVisitor 
  (cs:stmt option) (* continue statement *)
  (bs:stmt option) (* break statement *)
  (citype:typ)     (* loop_state compinfo type *)
  (state:varinfo)  (* loop_state argument *)
  (retarg: varinfo option) (* return value argument *)
  gotohash
  globalCounter
  internalLabels
  : cilVisitor = object(self)
  inherit nopCilVisitor

  method vstmt s =
    let isInternalTarget sr = 
      match getTargetLabel sr with
          Some l -> 
            if (H.mem internalLabels l)
            then
              ((* ignore(Updatedebug.log [Updatedebug.BTAlways] "isInternalTarget('%s')\n" l); *)
               true)
            else
              false
        | None -> Updatedebug.abort (Updatedebug.bug "massageLoopVisitor (%a): goto without destination !?\n" d_loc !(Cil.currentLoc))
    in
    match s.skind with
    (* All gotos are translated into returns *)
      Goto (sr,loc) ->
        (let sid = (!sr).sid in
        if (matchSid cs sid) 
        then       
          (ChangeTo (ret loop_contv loc))
        else 
          if (matchSid bs sid) 
          then  
            (ChangeTo (ret loop_breakv loc))
          else 
            (
            if newExtractionStyle && (not (isInternalTarget sr))
            then 
              begin
                incr globalCounter;
                H.add gotohash !globalCounter s;
                
                ChangeTo (ret (GotoV !globalCounter) loc)
              end
            else DoChildren)
        )
    (* looking for CIL continue label; delete if found *)
    | Instr [] ->
       (match cs with
          Some cstm -> 
	    (let lbls = List.fold_left 
		(fun res lbl -> 
		  match lbl with 
		    Label (s,_,_) -> s::res
		  | _ -> res) [] s.labels in
            match (lmem cstm.labels lbls) with
              Some _ -> ChangeTo (Commoncilutil.mkStmt (Instr []))
            | None -> DoChildren)
       | None -> DoChildren)
    (* replace returns with assigns+ret *)
    | Return (res,loc) ->
	(* mode is used to avoid an infinite loop when we recurse
	   over the expression that replaces the return statement.
	   If it's true, then this is the second time we've been here
	   and so we bail. *)
	if (!mode) then (mode := false; SkipChildren) else
        let retstm = ret loop_retv loc in
        (match res with
          Some v -> 
            (let retarg_vi = 
	      match retarg with 
		Some vi -> vi
              | None -> Updatedebug.abort (Updatedebug.bug "massageLoopVisitor : bad retarg\n") in
	    (* assign to return argument, if one exists *)
            let asnexp = Set ((deref retarg_vi NoOffset), v, loc) in
            let asnstm = Commoncilutil.mkStmt (Instr [asnexp]) in
            let blk = Commoncilutil.mkStmt (Block (mkBlock [asnstm;retstm])) in
	    mode := true; (* used to avoid an infinite loop when recursing *)
	    ChangeDoChildrenPost (blk,(fun x -> x)))
        | None -> ChangeTo retstm)
    | _ -> DoChildren

  (* replace all local var references to go through state struct *)
  method vlval lv =
    (* we replace all local variables, except for those we introduce *)
    let validLocal vi =
      (vi.vname <> state.vname) &&
      (match retarg with Some ravi -> ravi.vname <> vi.vname | None -> true) in
    match lv with
      Var vi, off -> 
        if vi.vglob then DoChildren (* might have to convert its children, eg.g. array indexes *)
        else 
          (if (validLocal vi) then
	    let newLval = loopDataRef state vi off citype 
	    in ChangeDoChildrenPost (newLval,(fun x -> x)) (* have to convert newLval's children as well *)
	  else SkipChildren)
    | _ -> DoChildren

end


(* generates the call to the extracted loop *)
(* res: varinfo for returned result, 
   nv: loop function name, 
   vlist: loop function args,
   cs: continue stmt, bs: break stmt, loc:loc 
   upd: optional call to update
*)
let callFun nv upd cs bs loc res retarg state_var chkincall gotoHash =
  let var v = ((Var v),NoOffset) in
  let mkgoto sopt =
    match sopt with
      Some s -> Commoncilutil.mkStmt (Goto (ref s, loc))
    | None -> Commoncilutil.mkStmt (Instr []) in
  (* args for the call to the loop function: the return val and the state *)
  let args = 
    let vs = match (retarg) with
    Some vi -> [vi;state_var] | None -> [state_var] in
    List.map (function vi -> Cil.mkAddrOf (var vi)) vs in
  (* the function call *)
  let loopfuncall = [Call (Some (var res), Lval(var nv), args, loc)] in
  let call = Commoncilutil.mkStmt (Instr (chkincall @ loopfuncall)) in
  (* the if check following the call, dispatching on its return code *)
  let check = 
    let breaks = mkgoto bs in
    let conts = mkgoto cs in 
    let rets = 
      let retexp = 
        match retarg with
          Some vi -> Some (Lval (var vi))
        | None -> None in
        Commoncilutil.mkStmt (Return (retexp,loc)) in
    let iftest = (BinOp (Eq,(Lval (var res)),(Cil.integer (retval_of_stmt loop_breakv)),(TInt (IInt,[])))) in
    let elsetest = (BinOp (Eq,(Lval (var res)),(Cil.integer (retval_of_stmt loop_contv)),(TInt (IInt,[])))) in

    let gotostmt =
      let doOne x s acc = 
        let elselsetest = (BinOp (Eq,(Lval (var res)),(Cil.integer (retval_of_stmt (GotoV x))),(TInt (IInt,[])))) in
        Commoncilutil.mkStmt (If (elselsetest,Cil.mkBlock [s],Cil.mkBlock [acc],loc))
      in 
      H.fold doOne gotoHash rets 
    in
    let elsestmt = 
      if newExtractionStyle 
      then Commoncilutil.mkStmt (If (elsetest,Cil.mkBlock [conts],Cil.mkBlock [gotostmt],loc)) 
      else Commoncilutil.mkStmt (If (elsetest,Cil.mkBlock [conts],Cil.mkBlock [rets],loc)) 
    in
    Commoncilutil.mkStmt (If (iftest,Cil.mkBlock [breaks],Cil.mkBlock [elsestmt],loc)) in
  (* if there's a continue statement, add it in before the call *)
  let stms = 
    match upd with
      Some cstm -> [cstm;call;check]
    | None -> [call;check] in
  match cs with
    Some cstm -> Cil.mkBlock (cstm::stms)
  | None -> Cil.mkBlock stms

(* 'pseudo' closure-conversion *)
(* makes a function that 
   - takes local loop variables as arguments (pass by reference) 
   - has body copied verbatim form the loop body 
  Parameters: nStem : the name of the pragma to create
              updateFunction : optional update function to insert a call to
              vlist : the contents of the environment argument
              loops : the actual loop statement to closure convert
              loopfd: the function in which the loop appears
*)
let splitLoop 
      (nStem:string) 
      (action: loop_processing_t) 
      (vlist: varinfo list) 
      (loops:stmt) 
      (loopfd:fundec) 
      loopDataHash 
      doneHash 
      (updateFunction: varinfo) 
      (checkinFunction: varinfo)
      globalCounter
    =
  let n = Updateconfig.freshLoopName nStem in  (* name of the new fun. to create *)
  let ret_type t =
    match t with
      TFun (t,_,_,_) -> t
    | _ -> Updatedebug.abort (Updatedebug.bug "Bad fun type") in
  let bloc,loc,cs,bs = 
    match loops.skind with
        Loop(bloc,loc,cs,bs) when action <> CodeExtract ->
          (bloc,loc,cs,bs)
      | Block(b)  when action = CodeExtract ->
          (b,Cil.locUnknown,None,None)
    | _ -> Updatedebug.abort 
	  (Updatedebug.bug "splitLoop called with non-standard Loop statement ?\n")
  in
      begin
	(* loop function to split out *)
        let f = Cil.emptyFunction n in
	(* set the return type; the argument types are filled out *)
	(* by makeFormalVar below. *)
	f.svar.vtype <- (TFun(Cil.intType, Some [], false, []));

	(* the loop's state struct *)
   let (state_ci, state_type, state_tagdecl, state_var) = 
      try
         H.find loopDataHash loopfd.svar.vname (* have done a loop in this function already; don't generate fresh struct + loop var *)
      with Not_found ->
	      let state_ci' = mkLoopStateStruct vlist (n^"_state") in
	      let state_type' = TComp(state_ci',[]) in
	      let state_tagdecl' = GCompTag(state_ci',Cil.locUnknown) in
	      let state_var' = makeTempVar loopfd ~name:"__loop" (state_type') in
         H.add loopDataHash loopfd.svar.vname (state_ci', state_type', state_tagdecl', state_var');
         (state_ci', state_type', state_tagdecl', state_var')
   in
	(* the return type of the enclosing function *)
   let ret_typ = (ret_type loopfd.svar.vtype) in
	let retarg = 
	  if (isVoidType ret_typ) then None
          else (Some (makeTempVar loopfd ~name:"__loop" ret_typ)) in

	(* the status code form the extracted loop *)
        let res = makeTempVar loopfd ~name:"__loop" (TInt (IInt,[])) in

	(* statements to initialize the state variable *)
	let stms = 
      if H.mem doneHash loopfd.svar.vname 
      then [] (* have done a loop in this function already; no init needed in the enclosing function *)
      else begin
         List.map (function vi ->
	                   let fi = Commoncilutil.getFieldinfoByName state_type vi.vname None in
	                      Commoncilutil.mkStmt 
	                         (Instr [Set ((Var state_var,Field (fi,NoOffset)),
			                                Cil.mkAddrOf (var vi),Cil.locUnknown)])) vlist
      end 
   in

	(* state and return arguments in the loop function, and its type *)
	let make_arg v s =
          let t = TPtr(v.vtype, []) in
          Cil.makeFormalVar f ~where:"^" s t in
	let make_arg_opt vopt s =
	  match vopt with
	      Some vi -> Some (make_arg vi s)
	    | None -> None 
   in
   let state_varptr = make_arg state_var ("__loop_state_" ^ nStem) in
	let retarg_ptr = make_arg_opt retarg "__loop_retval" in

	(* massage the loop function body to refer to the state and return args *)
   let goto_hash = H.create 3 in
   let internalLabels = 
     let res = H.create 3 in
     let vis = new labelGatherer res in
     ignore(visitCilBlock vis bloc);
     (* ignore(Updatedebug.log [Updatedebug.BTAlways] "Internal labels:"); *)
(*      H.iter (fun l () -> ignore(Updatedebug.log [Updatedebug.BTAlways] " %s" l)) res;   *)
(*      ignore(Updatedebug.log [Updatedebug.BTAlways] "\n"); *)
     res  
   in
   let repl_wref_vst = new massageLoopVisitor cs bs state_type state_varptr retarg_ptr goto_hash globalCounter internalLabels in 
   let bloc' = visitCilBlock repl_wref_vst bloc in
	(* Need to add a return statement at the end, for normal iteration *)
        bloc'.bstmts <- 
          (let newBlockS = bloc'.bstmts @ [(ret loop_contv loc)] in
	          match action with
              (* add a "__DSU_CHECKIN_" inside the original loop stsmt *)
              LoopTxCheckin ->
                (let newS' = Commoncilutil.mkStmt (Block (mkBlock newBlockS)) in
                newS'.labels <- [Label(Cilutil.checkinLabelName ^ "_" ^ nStem,Cil.locUnknown,false)];
                [newS'])
	         | _ -> 
                newBlockS
          );   
        f.sbody <- bloc';

        (* the code to replace the loop with *)
	let updCall =
	  (match action with
	       LoopUpd -> 
	         (* FIX? What about the forbidden function list? *)
            (* 	      ignore(Updatedebug.log [Updatedebug.BTAlways] "inserting call to DSU_update()\n"); *)
	         Some (Commoncilutil.mkStmt 
		              (Instr [Call(None,Lval (var updateFunction),[],loc)]))
	     | _ -> None) 
   in
     (* add hoisted check-in *)  
   let checkinCall = match action with
       LoopTxCheckin -> 
         [Call(None, Lval(var checkinFunction), [Cil.mkString nStem], loc)]
     | _ -> []
   in
   let call = callFun f.svar updCall cs bs loc res retarg state_var checkinCall goto_hash in
   let newS = 
     let innerS = { loops with skind = Loop(call,loc,cs,bs) } in
	  match action with
         (* add a "__DSU_TX_"  around original loop stsmt *)
         LoopTxCheckin -> 
           let newS' = Commoncilutil.mkStmt (Block (mkBlock [innerS])) in
           newS'.labels <- [Label(Cilutil.txLabelName ^ "_" ^ nStem,Cil.locUnknown,false)];
           newS'  

       | CodeExtract -> { loops with skind = Block(call) }

	    | _ -> 
           innerS 
	in
	(* return pair: new loop, split out function *)
        (newS, stms, GFun(f, Cil.locUnknown), state_tagdecl)
      end

(* Runs through the program trying to find a loop to split out.
   one tricky part is that the label might be attached to a statement
   preceding the loop, and not the loop itself.  This happens if the
   loop is in a block, or if it was a for loop whose initializer expression
   was extracted by cil. *)
class updateLoopsVisitor 
   (fi: fundec) 
   (lbl : (string * loop_processing_t) ref) 
   (txLoopsHash : (string, loop_processing_t) H.t) 
    newGFunsHash
    newGCompsHash
    loopDataHash
    doneHash
   (initStms : (Cil.stmt list) ref)
   (updateFunction : Cil.varinfo) 
   (checkinFunction : Cil.varinfo)
   globalCounter : cilVisitor = object(self)
      inherit nopCilVisitor

  method vstmt s1 =
    (* try to find the loop in the given statement, and split it into
       a function, a loop replacement, and a state decl, and its initializers *)
    let rec findloop s n action =
      let commonCase a = 
            lbl := ("",a);
            let vs = fi.sformals @ fi.slocals in
            let (newS,newI,newF,newD) = splitLoop n action vs s fi loopDataHash doneHash updateFunction checkinFunction globalCounter in
              if not (H.mem doneHash fi.svar.vname) (* never done this function before *)
              then begin
                H.add doneHash fi.svar.vname ();
                H.add newGCompsHash fi.svar.vname [newD];
	             initStms := !initStms @ newI;
              end;
              H.add newGFunsHash  fi.svar.vname [newF];
              newS
      in
      match s.skind with
	       Block b when action <> CodeExtract ->
	         (match b.bstmts with
	              h::t  -> let h' = findloop h n action in
	                { s with skind = Block { b with bstmts = h'::t }}
	            | [] -> Updatedebug.abort (Updatedebug.bug "pragma referred to label not on block loop %a\n" d_stmt s1)
            )

        | Block b when action = CodeExtract ->
            commonCase CodeExtract

        | Loop(bloc,loc,_,_) ->
            commonCase LoopNoAction

                (* could be a for loop initializer; check next statement *)
        | _ -> if (fst !lbl) <> "" then
            Updatedebug.abort (Updatedebug.bug "pragma referred to label not on loop %a\n" d_stmt s1)
          else (lbl := (n,action); s)
    in
    (* see if this statement's label is marked for extraction, or if
       if a prior one was, and this is the next statement. *)
    let txLabels = Commonutil.hashtbl_keys txLoopsHash in
      match (lmem s1.labels txLabels) with 
          Some n -> 
	         (let s' = findloop s1 n (H.find txLoopsHash n) in 
	            s'.labels <- lfilter s1.labels txLabels;
	            ChangeTo s')
        | None -> 
	         if (fst !lbl) <> "" then 
              (let s' = findloop s1 (fst !lbl) (snd !lbl) in 
	              s'.labels <- lfilter s1.labels txLabels;
	              ChangeTo s')
	         else DoChildren
    end;;

(* Collect all loop pragmas *)
let collectLoopPragmas (f : Cil.file) : (((string, loop_processing_t) H.t) * ((string, string) H.t)) =

   (* Hash of all loops beginning with labels to transform *)
   let loopLabels = H.create 11 
   and hoistedCheckins = H.create 11 in 

   let collectLoopPragmas_iter g = 
     match g with
       GPragma(Attr (name,[AStr lname]),loc) when name = Updateconfig.loopExtractPragma ->
         H.add loopLabels lname LoopNoAction

     | GPragma(Attr (name,[AStr lname]),loc) when name = Updateconfig.loopExtractUpdatePragma ->
         H.add loopLabels lname LoopUpd

     | GPragma(Attr (name,[AStr lname]),loc) when name = Updateconfig.loopTxCheckinPragma ->
         (H.add loopLabels lname LoopTxCheckin;
          H.add hoistedCheckins (Cilutil.checkinLabelName ^ "_" ^ lname) lname)

     | GPragma(Attr (name,[AStr lname]),loc) when name = Updateconfig.codeExtractPragma ->
         if Cilutil.isExtractLabel lname 
         then
           H.add loopLabels lname CodeExtract
         else
           Updatedebug.abort (Updatedebug.error "collectLoopPragmas: illegal name for code extraction label '%s'. It must start with '%s'\n" lname Cilutil.extractLabelName)

     | _ -> ()
   in
   Cil.iterGlobals f collectLoopPragmas_iter;
   (loopLabels, hoistedCheckins);;
   

(* Given a file, loop-closure-convert it *)
let rec loopClosureConvertFile (f : file) 
                               (loopLabels : (string, loop_processing_t) H.t) 
                               (i : int)
                               (updateFunction : varinfo) 
                               (checkinFunction : varinfo)  
                               (globalCounter : int ref) : unit = 

   (* New funs/comps we'll introduce *)
   let newGFunsHash  = H.create 11 in
   let newGCompsHash = H.create 11 in

   (* New loop state type and variables *)
   let loopDataHash = H.create 11 in

   (* Mark if we've extracted a loop in this function already *)
   let doneHash = H.create 11 in

   let getGlobalListForThisKey h k = List.flatten (H.find_all h k) in   

   (* do it for one function only; replace the body in place *)
   let loopClosureConvertFunction (fd : Cil.fundec) : unit =
      let initStms = ref [] in
      let vst = new updateLoopsVisitor fd (ref ("",LoopNoAction)) loopLabels newGFunsHash newGCompsHash loopDataHash doneHash initStms updateFunction checkinFunction globalCounter in
      let vBody = Cil.visitCilBlock vst fd.sbody in
      fd.sbody <- mkBlock (!initStms @ vBody.bstmts)
   in

   (* transform function bodies; put all newly generated globals in newGFunsHash newGCompsHash *)   
   Cil.iterGlobals f (function 
       GFun(fd, _) -> loopClosureConvertFunction fd
     | _ -> ());

   (* add newly generated functions *)
   let flag = ref false in (* whether new functions were introduced *)
   let addNewFuns_fold g acc = 
     match g with
       GFun(fd,_) when H.mem newGFunsHash fd.svar.vname -> 
       (* we have some new functions for this fd; introduce them before the fd *)
       let newFunctions = getGlobalListForThisKey newGFunsHash fd.svar.vname in
       flag := true;
       newFunctions @ (g :: acc)   

        (* remove loop pragmas since we don't need them any more *)         
     | GPragma(Attr (name,[AStr lname]),loc) when name = Updateconfig.loopExtractPragma -> acc

     | _ -> g :: acc
   in
   f.globals <- List.fold_right addNewFuns_fold f.globals [];

   (* add newly generated struct defs *)
   if (!flag) then
     begin
(*        if (i <> 0) then *)
(*          ignore (Updatedebug.abort (Updatedebug.bug *)
(*                    "Loop extraction for nested loop attempted; probably won't work!\n")); *)
       (* get new comp defs from newGCompsHash *)
       let newComps = List.flatten (Commonutil.hashtbl_values newGCompsHash) in
       (* add GCompTags in the type section of the program, i.e. before any var var/fun def/decl *)
       let splitglobals  = Updatecilutil.splitGlobalsBeforeVars f.globals in
       f.globals <- (fst splitglobals) @ newComps @ (snd splitglobals);
       (* go through the whole thing again, in case there were nested loops *)
       loopClosureConvertFile f loopLabels (i+1) updateFunction checkinFunction globalCounter
     end
   else ()
;;



(* Entry point.  Adds pragmas to a global hashtable, and
   processes function bodies to split out loops. *)
let default_update_loops (f : file) : (varinfo * varinfo * varinfo * varinfo * ((string, string) H.t)) =

 (* ---- Creates the update/checkin functions ---- *)
  let (updateFunDecl, updateFunVarinfo) = 
    Updatecilutil.mkDecl Updateconfig.updateFunctionName Updateconfig.updateFunctionType in
  let (checkinFunDecl, checkinFunVarinfo) = 
    Updatecilutil.mkDecl Updateconfig.checkinFunctionName Updateconfig.checkinFunctionType in
  let (startBlockingFunDecl, startBlockingFunVarinfo) = 
    Updatecilutil.mkDecl Updateconfig.startBlockingCallFunctionName Updateconfig.blockingCallFunctionType in
  let (endBlockingFunDecl, endBlockingFunVarinfo) = 
    Updatecilutil.mkDecl Updateconfig.endBlockingCallFunctionName Updateconfig.blockingCallFunctionType

  in
  let blockingFunDecls = 
    if Updateargs.args.Updateargs.arg_wrapBlockingCalls 
    then [ startBlockingFunDecl; endBlockingFunDecl ]
    else []
  in
    (* the update()/checkin() functions will come after global definitions and right before the other functions 
      we could actually insert it at the very beginning *)
  let splitglobals = Updatecilutil.splitGlobalsBeforeFuncs f.globals in
    f.globals <- (fst splitglobals) @ 
                 [updateFunDecl; checkinFunDecl] @
                 blockingFunDecls @ 
                 (snd splitglobals);

   (* we only bother if we've actually found a pragma *)
   let (loopLabels, hoistedCheckins) = collectLoopPragmas f 
   and globalCounter = ref 100 in
   if H.length loopLabels > 0 then
     loopClosureConvertFile f loopLabels 0 updateFunVarinfo checkinFunVarinfo globalCounter;

   (updateFunVarinfo, checkinFunVarinfo, 
    startBlockingFunVarinfo, endBlockingFunVarinfo, 
    hoistedCheckins)
;;
