(* 
 *
 * Copyright (c) 2004-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Manuel Oriol        <moriol@inf.ethz.ch> 
 *  Eric Hardisty       <hardisty@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)

(* TODO (Iulian) : see if we can get rid of this file altogether; it only provides 
   isEqualTyp to typetransformer.ml and we should be able to use some other
   type comparison implementation instead *)

open Pretty
open Cil
open Trace
open String

module H = Hashtbl
module E = Errormsg

 

type hashOfGlobals = (string, global) H.t

(* this hashtable type will store the functions behaviors
 * the first string is the name under which it is existing in the new version
 *    if it does not exist anymore the first string will be prefixed old_
 * the string is the old name (NOT prefixed by old_)
 * the fundec is the old fundec
 * the fundec is the new fundec
 * the boolean indicates if it has changed/new or did not changed/disappeared 
 *)
type hashOfModifiedFunctions = (string, (string option* fundec option* fundec option * bool)) H.t

let stage : int ref = ref 0

let uniqueId : int ref = ref 1

let out : out_channel ref = ref stdout

let oldVersionTag:string  = "v1"
let newVersionTag:string  = "v2"

type functionDirective = 
  ReplacedFunction of (string*string)
  | UnchangedFunction of string
  | NewFunction of string
  | DeletedFunction of string

type listOfDirectives = functionDirective list


(* code used to build the names differentiated for versions *)
let versionSeparator = "_"

let typeTripletList: (Cil.typ*Cil.typ*bool option) list ref = ref []

(*
 * Hashtables that contains the globals from the two versions.
 *
 * Iulian's code 
 **)
let globalsFile1: hashOfGlobals = H.create 65535
let globalsFile2: hashOfGlobals = H.create 65535

(* 
 * This visitor is the one that looks up the global
 * variables and feed the hash tables globalsFile1
 * globalile2.
 *
 * Iulian's code 
 **)
class modifiedFunctionsVisitor = object (self)
  inherit nopCilVisitor

  method private addGlobal (g:global) : unit = 
    let currentStore = if (1 = !stage) then globalsFile1 else globalsFile2 in
      match g with 
	  GFun _ ->
	    let canonicalName = Commoncilutil.getGlobalName g in 
	    H.add currentStore canonicalName g;
	| _ -> ()

  method vglob (g:global) : global list visitAction =  
    self#addGlobal g; DoChildren
      
end


(* Here we begin to define the whole comparison in depth for lists definitions 
 * Note that unlike other functions from the following, this one takes a function 
 * that allows the comparison of elements of the list *)
let rec isEqualList (l1: 'a list)(l2: 'a list)(compFun: 'a -> 'a ->bool):bool = (
   let length1 = List.length l1 in
   let length2 = List.length l2 in
   if (length1 <> length2) 
     then false
     else
       if (length1=0) 
          then true 
          else 
            if (compFun (List.hd l1) (List.hd l2)) 
            then (isEqualList (List.tl l1) (List.tl l2) compFun)
            else false 
  )
let isInTripletList (a1:'a)(a2:'a)(l: ('a*'a*bool option)list):bool =
   let  isInList : bool ref = ref false in 
   let isIt(el:('a*'a*bool option)) =
     (
      match el with 
        (a01,a02,_) -> if ((a1==a01)&&(a2==a02)) then (isInList:= true)
     ) in 
   let _ = List.iter isIt l in
   !isInList

let isInTripletListAndIsChanged (a1:'a)(a2:'a)(l: ('a*'a*bool option)list):bool =
   if (not (isInTripletList a1 a2 l)) then false
   else 
    (
     let  isChanged : bool ref = ref false in 
     let isItAndChanged(el:('a*'a*bool option)) =
      (
       match el with 
         (a01,a02,Some false) -> if ((a1==a01)&&(a2==a02)) then (isChanged:= true)
         | _ -> ()
      ) in 
     let _ = List.iter isItAndChanged l in
     (!isChanged)
    )
let isInTripletListAndIsUnchanged (a1:'a)(a2:'a)(l: ('a*'a*bool option)list):bool =
   if (not (isInTripletList a1 a2 l)) then false
   else 
     (
     let  isUnchanged : bool ref = ref false in 
     let isItAndUnchanged(el:('a*'a*bool option)) =
      (
       match el with 
         (a01,a02,Some true) -> if ((a1==a01)&&(a2==a02)) then (isUnchanged:= true)
         | _ -> ()
      ) in 
     let _ = (List.iter isItAndUnchanged l) in
     (!isUnchanged)
    )

let setInTripletList (a1:typ)(a2:typ)(r: bool) =
   if (not (isInTripletList a1 a2 (!typeTripletList))) then (Printf.printf "Not found\n")
   else 
    (

     let remap(el:('a*'a*bool option)) =
      (
       match el with 
         (a01,a02,_) -> if ((a1==a01)&&(a2==a02)) then (a1,a2,Some r) else el
         
      ) in 
     let _ = (typeTripletList:=(List.rev_map remap (!typeTripletList))) in ()
    )





(* Here we begin to define the whole comparison in depth for typeinfo definitions *)
let rec isEqualTypeinfo (ti1: Cil.typeinfo)(ti2: Cil.typeinfo): bool = 
   ( 
   if (ti1.tname<>ti2.tname) then false else
   (* we ommit treferenced as it constitutes a external factor *)
   (isEqualTypShallow ti1.ttype ti2.ttype)
   )
and

(* Here we begin to define the whole comparison in depth for offset definitions *)
 isEqualOffset (o1: Cil.offset)(o2: Cil.offset): bool = 

    match o1,o2 with 
       NoOffset, NoOffset -> true
       | Field(fi1,o11),Field(fi2,o22) -> (isEqualFieldinfo fi1 fi2) &&  (isEqualOffset o11 o22)
       | Index(e1,o11),Index(e2,o22) -> (isEqualExp e1 e2) &&  (isEqualOffset o11 o22)
       | _,_ -> false

 and 

(* Here we begin to define the whole comparison in depth for lhost definitions *)
 isEqualLhost (lh1: Cil.lhost)(lh2: Cil.lhost): bool = 

    match lh1,lh2 with
       Var(vi1),Var(vi2) -> (isEqualVarinfo vi1 vi2)
       | Mem(e1),Mem(e2) -> (isEqualExp e1 e2)
       | _,_ -> false

 and 

(* Here we begin to define the whole comparison in depth for lval definitions *)
 isEqualLval (lv1: Cil.lval)(lv2: Cil.lval): bool = 

    match lv1,lv2 with
      (lh1,of1),(lh2,of2) -> (isEqualLhost lh1 lh2) && (isEqualOffset of1 of2)

 and 

(* EAH: We need to compare the optional names as well, starting with GCC 3.1 *)
(* Here we begin to define the whole comparison in depth for (string,lval) definitions *)
 isEqualSLv (slv1: (string option*string*Cil.lval))(slv2: (string option*string*Cil.lval)): bool = 
  (
    match slv1,slv2 with
	(so1,s1,lv1),(so2,s2,lv2) -> 
	  begin
	    let iseqso = match so1,so2 with
		Some str1,Some str2 -> str1 = str2
	      | None,None -> true
	      | _,_ -> false
	    in
	      iseqso && (s1 = s2) && (isEqualLval lv1 lv2)
	  end
  )
 and 

(* EAH: We need to compare the optional names as well, starting with GCC 3.1 *)
(* Here we begin to define the whole comparison in depth for (string,exp) definition *)
 isEqualSEx (sex1: (string option*string*Cil.exp))(sex2: (string option*string*Cil.exp)): bool =
  (
    match sex1,sex2 with
	(so1,s1,e1),(so2,s2,e2) ->
	  begin
	    let iseqso = match so1,so2 with
		Some str1,Some str2 -> str1 = str2
	      | None,None -> true
	      | _,_ -> false
	    in 
	      iseqso && (s1 = s2) && (isEqualExp e1 e2)
	  end
   )
 and 

(* Here we begin to define the whole comparison in depth for instr definitions *)
 isEqualInstr (in1: Cil.instr)(in2: Cil.instr): bool = 
   ( 
   match in1,in2 with 
        Set(lv1,e1,_),Set(lv2,e2,_) -> 
           (isEqualLval lv1 lv2) 
        && (isEqualExp e1 e2)

      | Call(Some lv1,e1,le1,_),Call(Some lv2,e2,le2,_) ->
           (isEqualLval lv1 lv2)
        && (isEqualExp e1 e2)
        && (isEqualList le1 le2 isEqualExp)

     | Call(None,e1,le1,_),Call(None,e2,le2,_) -> 
           (isEqualExp e1 e2) 
        && (isEqualList le1 le2 isEqualExp)

     | Asm(la1,ls1,lsl1,lse1,ls11,_),Asm(la2,ls2,lsl2,lse2,ls22,_)->
            (isEqualList la1 la2 isEqualAttribute)
        &&  (isEqualList ls1 ls2 (=))
        &&  (isEqualList lsl1 lsl2 isEqualSLv)
        &&  (isEqualList lse1 lse2 isEqualSEx)
        &&  (isEqualList la1 la2 isEqualAttribute)

     | _,_ -> false
   )
 and

(* Here we begin to define the whole comparison in depth for label definitions *)
 isEqualLabel (la1: Cil.label)(la2: Cil.label): bool = 
   ( 
   match la1,la2 with
     Label(s1,_,b1),Label(s2,_,b2) -> 
         if (b1<>b2) then false else
         (* note that this means that automatically generated labels are not considered *)
         if (b1<>true) then true else
         (s1<>s2)
     | Case(e1,_),Case(e2,_) -> (isEqualExp e1 e2)
     | Default(_),Default(_) -> true
     | _,_ -> false
   )
 and 

(* Here we begin to define the whole comparison in depth for stmtkind definitions *)
 isEqualStmtkind (stk1: Cil.stmtkind)(stk2: Cil.stmtkind): bool = 
   ( 
    match stk1,stk2 with
         Instr(l1),Instr(l2) -> (isEqualList l1 l2 isEqualInstr)
       | Return(Some e1,_),Return(Some e2,_) -> (isEqualExp e1 e2)
       | Return(None,_),Return(None,_) -> true
       | Goto(rs1,_),Goto(rs2,_) -> (isEqualStmt !rs1 !rs2)
       | Break(_),Break(_) -> true
       | Continue(_),Continue(_) -> true
       | If(e1,b1,b11,_),If(e2,b2,b22,_) ->
              (isEqualExp e1 e2)
           && (isEqualBlock b1 b2)
           && (isEqualBlock b11 b22)

       | Switch(e1,b1,ls1,_),Switch(e2,b2,ls2,_) ->
              (isEqualExp e1 e2)
           && (isEqualBlock b1 b2)
           && (isEqualList ls1 ls2 isEqualStmt)

       | Loop(b1,_,Some s1,Some s11),Loop(b2,_,Some s2,Some s22) ->
              (isEqualStmt s1 s2)
           && (isEqualStmt s11 s22)
           && (isEqualBlock b1 b2)

       (* should not be useful *)
       | Loop(b1,_,Some s1,None),Loop(b2,_,Some s2,None) ->
              (isEqualStmt s1 s2)
           && (isEqualBlock b1 b2)

       (* should not be useful *)
       | Loop(b1,_,None,Some s11),Loop(b2,_,None,Some s22) ->
               (isEqualStmt s11 s22)
           &&  (isEqualBlock b1 b2)

       | Loop(b1,_,None,None),Loop(b2,_,None,None) -> (isEqualBlock b1 b2)

       | Block(b1),Block(b2) -> (isEqualBlock b1 b2)

       | TryFinally(b1,b11,_),TryFinally(b2,b22,_) ->
               (isEqualBlock b1 b2)
           &&  (isEqualBlock b11 b22)

       | TryExcept(b1,(l1,e1),b11,_),TryExcept(b2,(l2,e2),b22,_) ->
               (isEqualBlock b1 b2)
           &&  (isEqualBlock b11 b22)
           &&  (isEqualExp e1 e2)
           &&  (isEqualList l1 l2 isEqualInstr)

       | _,_ -> false

   )
 and 

(* Here we begin to define the whole comparison in depth for stmt definitions *)
 isEqualStmt (st1: Cil.stmt)(st2: Cil.stmt): bool = 
      (isEqualList st1.labels st2.labels isEqualLabel) && (isEqualStmtkind st1.skind st2.skind)

   (* we ommit sid *)
   (* we ommit preds and succs *)

and 

(* Here we begin to define the whole comparison in depth for block definitions *)
 isEqualBlock (b1: Cil.block)(b2: Cil.block): bool = 
    (isEqualList b1.battrs b2.battrs isEqualAttribute) && (isEqualList b1.bstmts b2.bstmts isEqualStmt)   
 and 

(* Here we begin to define the whole comparison in depth for ikind definitions *)
 isEqualIkind (ik1: Cil.ikind)(ik2: Cil.ikind): bool = (ik1 = ik2)
 and 

(* Here we begin to define the whole comparison in depth for fkind definitions *)
 isEqualFkind (fk1: Cil.fkind)(fk2: Cil.fkind): bool = (fk1 = fk2)

and 

(* Here we begin to define the whole comparison in depth for int64 *)
 isEqualInt64 (i1:int64)(i2: int64): bool = ((Int64.compare i1 i2) = 0)
 and

(* Here we begin to define the whole comparison in depth for constant definitions *)
 isEqualConstant (c1: Cil.constant)(c2: Cil.constant): bool = 
   ( 
   match c1,c2 with
     CInt64(i1,ik1,Some s1),CInt64(i2,ik2,Some s2) ->
            (isEqualInt64 i1 i2)
         && (isEqualIkind ik1 ik2)

     | CInt64(i1,ik1,None),CInt64(i2,ik2,None) ->
            (isEqualInt64 i1 i2)
         && (isEqualIkind ik1 ik2)
 
     | CStr(s1),CStr(s2) -> (s1 = s2)

     | CWStr(l1),CWStr(l2) -> (isEqualList l1 l2 isEqualInt64)

     | CChr(c1),CChr(c2) -> (c1 = c2)

     | CReal(f1,fk1,Some s1),CReal(f2,fk2,Some s2) ->
            ((Pervasives.compare f1 f2) = 0)
         && (isEqualFkind fk1 fk2)

     | CReal(f1,fk1,None),CReal(f2,fk2,None) ->
            ((Pervasives.compare f1 f2) = 0)
        &&  (isEqualFkind fk1 fk2) 

     | _,_ -> false
   )
 and 

(* Here we begin to define the whole comparison in depth for binop definitions *)
 isEqualBinop (bi1 : Cil.binop) (bi2 : Cil.binop) : bool = (bi1 = bi2) 
 and 

(* Here we begin to define the whole comparison in depth for unop definitions *)
 isEqualUnop (un1 : Cil.unop) (un2 : Cil.unop) : bool = (un1 = un2)

 and 

(* Here we begin to define the whole comparison in depth for attribute definitions *)
 isEqualExp (e1: Cil.exp)(e2: Cil.exp): bool = 
   (
   let e1=  Cil.constFold true e1 in
   let e2=  Cil.constFold true e2 in
   match e1,e2 with
      Const(c1), Const(c2) -> (isEqualConstant c1 c2)
      | Lval(lv1),Lval(lv2) -> (isEqualLval lv1 lv2)
      | SizeOf(t1),SizeOf(t2) -> (isEqualTyp t1 t2)
      | SizeOfE(e1),SizeOfE(e2) -> (isEqualExp e1 e2)
      | SizeOfStr(s1),SizeOfStr(s2) -> (s1 = s2)
      | AlignOf(t1),AlignOf(t2) -> (isEqualTyp t1 t2)
      | AlignOfE(e1),AlignOfE(e2) -> (isEqualExp e1 e2)

      | UnOp(u1,e1,t1),UnOp(u2,e2,t2) -> 
            (isEqualUnop u1 u2)
         && (isEqualExp e1 e2)
         && (isEqualTyp t1 t2)

      | BinOp(b1,e1,e11,t1),BinOp(b2,e2,e22,t2) ->
            (isEqualBinop b1 b2)
         && (isEqualExp e1 e2)
         && (isEqualExp e11 e22)
         && (isEqualTyp t1 t2)
      
      | CastE(t1,e1),CastE(t2,e2) ->
            (isEqualExp e1 e2)
         && (isEqualTyp t1 t2)
          
      | AddrOf(lv1),AddrOf(lv2) -> (isEqualLval lv1 lv2)
      | StartOf(lv1),StartOf(lv2) -> (isEqualLval lv1 lv2)
      | _,_ -> false
   )
 and 

(* Here we begin to define the whole comparison in depth for compinfo definitions *)
 isEqualCompinfo (ci1: Cil.compinfo)(ci2: Cil.compinfo): bool = 
      (ci1.cstruct = ci2.cstruct)
   && (ci1.cname = ci2.cname)
   && (ci1.cdefined = ci2.cdefined)
   && (isEqualList ci1.cfields ci2.cfields isEqualFieldinfo)
   && (isEqualList ci1.cattr ci2.cattr isEqualAttribute)

 and 

(* Here we begin to define the whole comparison in depth for function fields which are triplets of string, exp and location *)
 isEqualSELoc (triplet1: (string*exp*location))(triplet2: (string*exp*location)):bool = 

   match triplet1,triplet2 with
     (s1,e1,loc1),(s2,e2,loc2) -> 
           (s1 = s2)
        && (isEqualExp e1 e2)
 and

(* Here we begin to define the whole comparison in depth for compinfo definitions *)
 isEqualEnuminfo (ei1: Cil.enuminfo)(ei2: Cil.enuminfo): bool = 

       (ei1.ename = ei2.ename)
    && (isEqualList ei1.eitems ei2.eitems isEqualSELoc)
    && (isEqualList ei1.eattr ei1.eattr isEqualAttribute)

 and 

(* Here we begin to define the whole comparison in depth for fieldinfo definitions *)
 isEqualFieldinfo (fi1: Cil.fieldinfo)(fi2: Cil.fieldinfo): bool = 

       (fi1.fname = fi2.fname)

    (* Iulian : changed isEqualTyp to isEqualTypShallow because for fields with 
       abstract types, a change in ftype doesn't need to lead to a change in this TComp
    *)

    && (isEqualTypShallow fi1.ftype fi2.ftype)
    && (isEqualList fi1.fattr fi2.fattr isEqualAttribute)
    && 
    (match fi1.fbitfield,fi2.fbitfield with
       None, None -> true
       | Some i1, Some i2 -> (i1 = i2)
       | _,_ -> false
    (* we ommit location *)
   )
 and 

(* Here we begin to define the whole comparison in depth for function fields which are triplets of string, typ and attributes *)
 isEqualSTLat (triplet1: (string*typ*attributes))(triplet2: (string*typ*attributes)):bool = 

   match triplet1,triplet2 with
     (s1,t1,lat1),(s2,t2,lat2) -> 
           (s1 = s2)
        && (isEqualTyp t1 t2)
        && (isEqualList lat1 lat2 isEqualAttribute)

 and

(* Shallow comparison for type definitions;
   TNamed/TComp/TEnum's are only compared by name 
*)
  isEqualTypShallow (t1: Cil.typ)(t2: Cil.typ): bool =
    match t1,t2 with
        TNamed(ti1, _), TNamed(ti2, _) -> 
          ti1.tname = ti2.tname

      | TComp(ci1, _), TComp(ci2, _) -> 
          ci1.cstruct = ci2.cstruct && ci1.cname = ci2.cname

      | TEnum(ei1, _), TEnum(ei2, _) ->
           ei1.ename = ei2.ename

      | _ -> isEqualTyp t1 t2
 and
(* Here we begin to define the whole comparison in depth for typ definitions *)
  isEqualTyp (t1: Cil.typ)(t2: Cil.typ): bool =
   if (isInTripletListAndIsUnchanged t1 t2 !typeTripletList) then true else 
   if (isInTripletListAndIsChanged t1 t2 !typeTripletList) then false else
   if (isInTripletList t1 t2 !typeTripletList) then true else
   let _ = (typeTripletList:=(t1,t2,None)::!typeTripletList) in
   ( 
   match t1,t2 with
     TVoid(lat1),TVoid(lat2) -> 
         let r=(isEqualList lat1 lat2 isEqualAttribute) in
         let _ = (setInTripletList t1 t2 r) in
         r
     | TInt(ik1,lat1),TInt(ik2,lat2) ->
         let r = (if (not (isEqualIkind ik1 ik2)) then false else
         (isEqualList lat1 lat2 isEqualAttribute)) in
         let _ = (setInTripletList t1 t2 r) in
         r
     | TFloat(fk1,lat1),TFloat(fk2,lat2) ->
         let r = (if (not (isEqualFkind fk1 fk2)) then false else
         (isEqualList lat1 lat2 isEqualAttribute)) in
         let _ = (setInTripletList t1 t2 r) in
         r

     | TPtr(t01,lat1),TPtr(t02,lat2) ->
         let r =(if (not (isEqualTypShallow t01 t02)) then false else
         (isEqualList lat1 lat2 isEqualAttribute)) in
         let _ = (setInTripletList t1 t2 r) in
         r

     | TArray(t01,None,lat1), TArray(t02,None,lat2) ->
         let r = (if (not (isEqualTyp t01 t02)) then false else
         (isEqualList lat1 lat2 isEqualAttribute)) in
         let _ = (setInTripletList t1 t2 r) in
         r

     | TArray(t01,Some e1,lat1), TArray(t02,Some e2,lat2) ->
         let r = (if (not (isEqualTyp t01 t02)) then false else
         if (not (isEqualExp e1 e2)) then false else
         (isEqualList lat1 lat2 isEqualAttribute)) in
         let _ = (setInTripletList t1 t2 r) in
         r

     | TFun(t01,None,b1,lat1),TFun(t02,None,b2,lat2) -> 
         let r =(if (b1<>b2) then false else
         if (not (isEqualTyp t01 t02)) then false else
         (isEqualList lat1 lat2 isEqualAttribute)) in
         let _ = (setInTripletList t1 t2 r) in
         r

     | TFun(t01,Some l1,b1,lat1),TFun(t02,Some l2,b2,lat2) -> 
         let r =(if (b1<>b2) then false else
         if (not (isEqualTyp t01 t02)) then false else
         if (not (isEqualList l1 l2 isEqualSTLat)) then false else
         (isEqualList lat1 lat2 isEqualAttribute)) in
         let _ = (setInTripletList t1 t2 r) in
         r

     | TNamed(ti1,lat1),TNamed(ti2,lat2) ->
         let r = (if (not (isEqualTypeinfo ti1 ti2)) then false else
         (isEqualList lat1 lat2 isEqualAttribute)) in
         let _ = (setInTripletList t1 t2 r) in
         r

     | TComp(ci1,lat1),TComp(ci2,lat2) ->
         let r =(if (not (isEqualCompinfo ci1 ci2)) then false else
         (isEqualList lat1 lat2 isEqualAttribute)) in
         let _ = (setInTripletList t1 t2 r) in
         r

     | TEnum(ei1,lat1),TEnum(ei2,lat2) ->
          isEqualTypShallow t1 t2
(*          let r =(if (not (isEqualEnuminfo ei1 ei2)) then false else  *)
(*          (isEqualList lat1 lat2 isEqualAttribute)) in *)
(*          let _ = (setInTripletList t1 t2 r) in *)
(*          r *)

     | TBuiltin_va_list(lat1),TBuiltin_va_list(lat2) ->
         let r =(isEqualList lat1 lat2 isEqualAttribute) in
         let _ = (setInTripletList t1 t2 r) in
         r
     | _,_ -> 
         let r = false in
         let _ = (setInTripletList t1 t2 r) in
         r
   )
and 


(* Here we begin to define the whole comparison in depth for storage definitions *)
 isEqualStorage (s1 : Cil.storage) (s2 : Cil.storage): bool = (s1 = s2)
 and 

 (* EAH: TSArray is no longer generated with an expression, it uses an int64 
    instead.  Changed the TSArray equality check accordingly *)
 (* Here we begin to define the whole comparison in depth for typ definitions *)
 isEqualTypsig (ts1: Cil.typsig)(ts2: Cil.typsig): bool =

   match ts1,ts2 with
      TSArray(ts01,None, l1), TSArray(ts02, None,l2) -> 
            (isEqualTypsig ts01 ts02)
         && (isEqualList l1 l2 isEqualAttribute)

      | TSArray(ts01,Some e1, l1), TSArray(ts02,Some e2,l2) ->
            (isEqualTypsig ts01 ts02)
         (* && (isEqualExp e1 e2) *)
         && e1 = e2
         && (isEqualList l1 l2 isEqualAttribute)

      | TSPtr(ts01,l1), TSPtr(ts02,l2) ->
            (isEqualTypsig ts01 ts02)
         && (isEqualList l1 l2 isEqualAttribute)

      | TSComp(b1,s1,l1), TSComp(b2,s2,l2) ->
            (b1 = b2)
         && (s1 = s2)
         && (isEqualList l1 l2 isEqualAttribute)
         
      | TSFun(ts01,l1,b1,l01),TSFun(ts02,l2,b2,l02) ->
            (isEqualTypsig ts01 ts02)
         && (isEqualList l1 l2 isEqualTypsig)
         && (b1 = b2)
         && (isEqualList l01 l02 isEqualAttribute)

      | TSEnum(s1,l1),TSEnum(s2,l2) ->
            (s1 = s2)
         && (isEqualList l1 l2 isEqualAttribute)

      | TSBase(t1),TSBase(t2) -> (isEqualTyp t1 t2)

      | _,_ -> false

and

(* Here we begin to define the whole comparison in depth for attrparam definitions *)
  isEqualAttrparam (ap1: Cil.attrparam)(ap2: Cil.attrparam): bool = 

   match ap1, ap2 with
       AInt(i1),AInt(i2) -> (i1 = i2)

     | AStr(s1),AStr(s2) -> (s1 = s2)

     | ACons(s1,l1),ACons(s2,l2) -> 
            (s1 = s2)
         && (isEqualList l1 l2 isEqualAttrparam)

     | ASizeOf(t1),ASizeOf(t2) -> (isEqualTyp t1 t2) 

     | ASizeOfE(ap01),ASizeOfE(ap02) -> (isEqualAttrparam ap01 ap02)

     | ASizeOfS(ts1),ASizeOfS(ts2) -> (isEqualTypsig ts1 ts2)

     | AAlignOf(t1),AAlignOf(t2) -> (isEqualTyp t1 t2)

     | AAlignOfE(ap01),AAlignOfE(ap02) -> (isEqualAttrparam ap01 ap02)

     | AAlignOfS(ts1),AAlignOfS(ts2) -> (isEqualTypsig ts1 ts2)

     | AUnOp(u1,at01), AUnOp(u2,at02) ->
             (isEqualUnop u1 u2)
         &&  (isEqualAttrparam at01 at02)

     | ABinOp(u1,at01,at11), ABinOp(u2,at02,at12) ->
             (isEqualBinop u1 u2)
         &&  (isEqualAttrparam at11 at12)
         &&  (isEqualAttrparam at01 at02)

     | ADot(at01,s1),ADot(at02,s2) ->
             (s1 = s2)
         &&  (isEqualAttrparam at01 at02)

     | _,_ -> false 

 and 

(* Here we begin to define the whole comparison in depth for attribute definitions *)
  isEqualAttribute (at1: Cil.attribute)(at2: Cil.attribute): bool = 
    
   match at1,at2 with
     Attr(name1,l1), Attr(name2,l2) ->
           (name1 = name2)
        && (isEqualList l1 l2 isEqualAttrparam) 
    and 

(* Here we begin to define the whole comparison in depth for varinfo definitions *)
  isEqualVarinfo (vi1: Cil.varinfo)(vi2: Cil.varinfo): bool = 
       (vi1.vname = vi2.vname)
   &&  (isEqualTyp vi1.vtype vi2.vtype)
   &&  (isEqualList vi1.vattr vi2.vattr isEqualAttribute)
   &&  (isEqualStorage vi1.vstorage vi2.vstorage)
   &&  (vi1.vglob = vi2.vglob)
   (* we ommit location, vid, vaddrof, vreferenced *)
   and

   

(* Here we begin to define the whole comparison in depth for function definitions *)
   isEqualFundec (fd1: Cil.fundec)(fd2: Cil.fundec): bool = 

       (isEqualVarinfo fd1.svar fd2.svar)
    && (isEqualList fd1.sformals fd2.sformals isEqualVarinfo)
    && (isEqualList fd1.slocals fd2.slocals isEqualVarinfo)
    && (isEqualList fd1.sallstmts fd2.sallstmts isEqualStmt);;









(*
 * This function does the analysis of the transformed
 * functions and builds a hashtable containing the new name (possibly "NoName")
 * and the two fundecs.
 *  
 **)
let detectModifiedFunctions (l : listOfDirectives) (globalsFile1 : hashOfGlobals) (globalsFile2 : hashOfGlobals) : hashOfModifiedFunctions = ( 

   let hResult: hashOfModifiedFunctions = H.create 65535 in
   
   (* first we will build the list of all function names *)
   let lFunNames: string list ref = ref [] in

   (*
    * this function is here to add the name of the functions to a pool of globals to treat
    * name: the name of the functions
    * global: the processed global 
    * 
    *)
   let addToNames (name: string)(g: Cil.global) =
      (
       match g with
         GFun(fd,_) -> 
           (
            let p (s:string):bool = if (s<>name) then false else true in
            try 
               let _ = List.find p !lFunNames in ()
            with Not_found -> let _ = lFunNames:= name :: !lFunNames in ()
           )
         | _ -> ()
      ) in
  
   let _ = H.iter addToNames globalsFile1 in
   let _ = H.iter addToNames globalsFile2 in

   (* we remove the names that are used in the directives *)
   let removeFromNames (fd: functionDirective) =
      (
       match fd with 
          ReplacedFunction (s1,s2) -> 
             lFunNames := List.filter ((<>) s2) (List.filter ((<>) s1) !lFunNames)
          | UnchangedFunction (s1) 
          | NewFunction (s1) 
          | DeletedFunction (s1) -> lFunNames := List.filter ((<>) s1) !lFunNames 
      ) in
   let _ = List.iter removeFromNames l in
   
   (* initialize hResult with users directives *)
    let initializeHashTable (fd: functionDirective) =
      (
       match fd with 
          ReplacedFunction(s1,s2) -> 
             (try
               let g1 = H.find globalsFile1 s1 in
               let g2 = H.find globalsFile2 s2 in
               (
                match g1,g2 with
                 GFun(fd1,_),GFun(fd2,_) -> H.add hResult s2 (Some s1,Some fd1,Some fd2,true)
                 | _,_ ->()
               )
             with Not_found -> ())
          | UnchangedFunction(s)  ->  
            (try
               let g1 = H.find globalsFile1 s in
               (
                match g1 with
                 GFun(fd1,_) -> H.add hResult s (Some s,Some fd1,Some fd1,false)
                 | _ ->()
               )
             with Not_found -> ())

          | NewFunction(s2) -> 
            (try
               let g2 = H.find globalsFile2 s2 in
               (
                match g2 with
                 GFun(fd2,_) -> H.add hResult s2 (None,None,Some fd2,true)
                 | _ ->()
               )
             with Not_found -> ())

          | DeletedFunction(s1) ->
             (try
               let g1 = H.find globalsFile1 s1 in
               (
                match g1 with
                 GFun(fd1,_) -> H.add hResult ("old_"^s1) (Some s1,Some fd1,None,false)
                 | _ ->()
               )
             with Not_found -> ())


      ) in
        

   let _ = List.iter initializeHashTable l in


   (* complete the hashtable with unspecified functions *) 
   let completeHashTable (fname: string) =
       (
        try 
          let g1 = H.find globalsFile1 fname in
          let g2 = H.find globalsFile2 fname in
          (
           match g1,g2 with
             GFun(fd1,_),GFun(fd2,_) -> H.add hResult fname (Some fname,Some fd1,Some fd2,(not (isEqualFundec fd1 fd2)))
             | _,_ ->()
          )
        with Not_found ->(
         try 
          let g1 = H.find globalsFile1 fname in
          (
           match g1 with
             GFun(fd1,_) -> H.add hResult ("old_"^fname) (Some fname,Some fd1,None,false)
             | _ ->()
          )
        with Not_found -> (
          let g2 = H.find globalsFile2 fname in
          (
           match g2 with
             GFun(fd2,_) -> H.add hResult (fname) (None,None,Some fd2,true)
             | _ ->()
          )
      ))) 
   in
   (* we eventually iterate on all the functions from globalsFile1 *)
   let _ = List.iter completeHashTable !lFunNames in
   (* let _=H.iter doOnePairOfGlobals globalsFile2 in*)
   hResult
   
) ;;

(*
 * Initialization of the analysis.
 *
 * Iulian's code 
 **)

let pre () =  
  begin
    stage := !stage + 1;
    (* Printf.printf "Stage: %d\n" !stage; *)
  end

(*
 * Finalization of the analysis.
 * 
 ***)
let post () =  
  begin
    if (2 = !stage) 
    then 
      let l: listOfDirectives = [] in
   let statsNew = ref 0 in
   let statsChanged = ref 0 in
   let statsDisappeared = ref 0 in
   let statsUnchanged = ref 0 in
   let statsUnexplained = ref 0 in

  
   let printResult(name:string)(change:(string option* fundec option* fundec option * bool)) = 
       match change with
         (None, None, Some(fd2),true) ->
                let _= (statsNew:= !statsNew +1) in
                (Printf.printf "The function %s is new\n" name)
         | (Some(s1), Some(fd1), Some(fd2),true) ->
                let _= (statsChanged:= !statsChanged +1) in
                (Printf.printf "The function %s is a new version of %s\n" name s1)
         | (Some(s1), Some(fd1), None,false) ->
                let _= (statsDisappeared:= !statsDisappeared +1) in
                (Printf.printf "The function %s has disappeared\n" s1)
         | (Some(s1), Some(fd1), Some(fd2),false) ->
                let _= (statsUnchanged:= !statsUnchanged +1) in
                (Printf.printf "The function %s has not changed\n" s1)
         | _ -> 
                let _= (statsUnexplained:= !statsUnexplained +1) in
                (Printf.printf "hum... that is strange...")

 
   in 
     let hashResult = detectModifiedFunctions l globalsFile1 globalsFile2 in 
     let _ = H.iter printResult hashResult in
      
     (Printf.printf "Total result:\n------------------------- \n New Functions: %d\n Functions Modified: %d\n Functions Removed: %d\n Functions Unchanged: %d\n Functions Unexplained: %d\n" !statsNew !statsChanged !statsDisappeared !statsUnchanged !statsUnexplained)
      
  end


(*
 * Bootstrap point of the type transformer feature.
 * 
 * Iulian's code
 **)
let feature : featureDescr = 
  { fd_name = "modifiedfunctions";
    fd_enabled = Cilutil.modifiedfunctionsEnabled;
    fd_description = "Detect and analyse changes in functions from an old version to a new version";
    fd_extraopt = [];
    fd_doit = 
                    (function (f: file) -> 
                       let sdVisitor =  pre (); new modifiedFunctionsVisitor in
                       let res = visitCilFileSameGlobals sdVisitor f
                       in post (); res
                    );
    fd_post_check = true;
  } 

