(* 
 *
 * Copyright (c) 2004-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(*                  Patch generation         *)

(*
 Given two succesive versions of a C program, this feature generates
 a patch that can be fed to the patch compiler; the patch compiler
 in turn will compile it to a shared library

 Invocation:
 $ cilly --dopatchgeneration --merge --patchfile-out=<patch_file> version_n.c version_{n+1}.c

*)


open Cil

module H = Hashtbl
module E = Errormsg

 
type hashOfGlobals = (string, Cil.global) H.t
type hashOfTT = (Cil.typ, string) H.t

(* tells us whether we're processing the first or second file *)
type stage_t = ProcessFile1Stage | ProcessFile2Stage
let stage : stage_t ref = ref ProcessFile1Stage

(* tells whether this is a root (for computing transitive closure) *)
type globalRoot_t = IsRoot | IsNotRoot

(* this will be set in the cmd line via --patchfile-out=<patch_file> *)
let patchFileName : string ref = ref ""

(* this will be set in the cmd line via --pversiondata-in=<vd_file> *)
let versionDataFile : string ref = ref ""

(* when we generate the patch, we need to output the new types/functions
   in the order they appear in the original file; 
   therefore, we keep a reference to the 'new' file, and the globals
   will be processed/output in this order *)


type typeDirective = 
    ReplacedType of (string * string)
  | UnchangedType of string
  | NewType of string
  | DeletedType of string

type functionDirective = 
    ReplacedFunction of (string * string)
  | UnchangedFunction of string
  | NewFunction of string
  | DeletedFunction of string


type listOfTypeDirectives = typeDirective list

(* this hashtable type will store the types
 * the first string is the name under which it is existing in the new version
 *    if it does not exist anymore the first string will be prefixed old_
 * the string is the old name (NOT prefixed by old_)
 * the typ is the old type declaration
 * the typ is the new type declaration
 * the boolean indicates if it has changed/new or did not changed/disappeared 
 * the fundec is the generated type transformer
 *)
type hashOfTransformedTypes = (string, (string option* typ option* typ option * bool * fundec option)) H.t

(* this hashtable type will store the functions behaviors
 * the first string is the name under which it is existing in the new version
 *    if it does not exist anymore the first string will be prefixed old_
 * the string is the old name (NOT prefixed by old_)
 * the fundec is the old fundec
 * the fundec is the new fundec
 * the boolean indicates if it has changed/new or did not changed/disappeared 
 *)
type hashOfModifiedFunctions = (string, (string option* fundec option* fundec option * bool)) H.t


(* code used to build the names differentiated for versions *)
let versionSeparator = "_"
let typeTransformerBaseName = "TT_"

(* hashtables that contains the globals from the two versions *)
let globalsFile1: hashOfGlobals = H.create 65535
let globalsFile2: hashOfGlobals = H.create 65535

(* this helps us distinguish between new and old globals *)
type versionType = OldVersion | NewVersion | MadeUp

(* this helps us distinguish between original globals / globals introduced by us  *)
type nameGeneration_t = GenerateWithSameName | GenerateWithChangedName of string * string 

(* When converting CIL -> CABS, some globals are hard to convert; 
   hence we use the CIL pretty printer to dump the CIL global verbatim
   instead of converting the global CIL -> CABS + CABS pretty print
*)
type cabsOut_t = CabsOutVerbatim of Cil.global


type cabsGlobalKind_t = Cabs_Fundef | Cabs_Typedef | Cabs_Compdef | Cabs_Decdef

(* old and new CIL files *)
let theFiles = ref []

let getPatchFileOutputChannel () =
  try 
    open_out !patchFileName 
  with (Sys_error msg) ->
    Updatedebug.abort (Updatedebug.error "Can't open '%s' for writing the patch: %s\n" !patchFileName msg)


(* Given a list of global descriptors, produce a CIL-C file out of them *)
let dumpPatchToCilFile gl =
  let patchFile = { globals = gl;
                    fileName = !patchFileName;
                    globinit = None;
                    globinitcalled = false} in
  let outChan =  getPatchFileOutputChannel () in
  output_string outChan ("/* Generated by 'cilly --patchgeneration' in CIL output mode */\n");
  Cil.dumpFile Cil.defaultCilPrinter outChan !patchFileName patchFile

(* the type of elements in the cabs hash *)
type cabsGlobalHash_t = (string, Cil.global) H.t

(* Given a list of global descriptors, produce a CABS-C file out of them *)
let dumpPatchToCabsFile gl isRoot =

  (* given a cabs, return a hash on its globals *)
  let getCabsHash (f : Cil.file) : cabsGlobalHash_t = 

    (* iterator over a definition list; add each definition to the given hash *)
    let addCabsDefinition_iter hash definition = 
      let hashAdd globalKind name = 
         let key = match globalKind with
              Cabs_Fundef  -> ("function " ^ name)
            | Cabs_Typedef -> ("typedef " ^ name)
            | Cabs_Compdef -> ("comp " ^ name)
            | Cabs_Decdef  -> ("decdef " ^ name)
         in
           H.add hash key definition
      in
        match definition with
            (* GFun*)
            GFun(fd, _) ->
              let functionName = fd.svar.vname in
(*                 ignore(Updatedebug.log [Updatedebug.BTPatchgen] "FUNDEF %s\n" functionName); *)
              hashAdd Cabs_Fundef functionName

            (* GType*)      
          | GType(ti,_) ->
              let typeName = ti.tname in
(*                 ignore(Updatedebug.log [Updatedebug.BTPatchgen] "TYPEDEF %s\n" typeName); *)
               hashAdd Cabs_Typedef typeName

            (* GCompTag *)
          | GCompTag(ci, _) ->
              let typeName = ci.cname in
(*               ignore(Updatedebug.log [Updatedebug.BTPatchgen] "ONLYTYPEDEF %s\n" typeName); *)
               hashAdd Cabs_Compdef typeName

          | GVar(vi, _,_) ->
              let varName = vi.vname in
               hashAdd Cabs_Decdef varName

          | _ -> ()

    in
    (* create an empty hash and fill it out by iterating over the definition list *)
    let hash = H.create 4099 in
    List.iter (addCabsDefinition_iter hash) f.globals;
    hash
  in

  (* We need to keep track of non-root functions and get rid of their static attributes *)
  let nonRootFunctions = H.create 4099 in
  let elimNonRootStatics_map printDirective = match printDirective with

       (* rewrite CIL fundecs to make them extern *)
       CabsOutVerbatim GVarDecl(vi,loc) when 
          H.mem nonRootFunctions vi.vname &&
          vi.vstorage <> Extern ->

          vi.vstorage <- Extern;
          printDirective 

     | _ -> printDirective
  in

    (* first check whether we have two CABS files in cabsFiles *)
    begin
      match !theFiles with
          [oldFile; newFile] ->
            let cabsHash1 = getCabsHash oldFile 
            and cabsHash2 = getCabsHash newFile in

            (* the keystone of CABS patch generation; this function does the 
               Cil.global -> Cabs.definition list 
               conversion
            *)    
            let generatePatchDefinitions_fold 
              (acc : cabsOut_t list) 
              (g : (Cil.global * versionType * nameGeneration_t)) 
              : cabsOut_t list = 
    
              (* look up a definition in a CABS hash *)
              let retrieveCabsDefinition hash name = 
                begin
                try
                  H.find hash name
                with Not_found ->
                  let hashNr = if hash == cabsHash1 then 1 else 2 in
                  Updatedebug.abort (Updatedebug.bug 
                                       "retrieveCabsDefinition : %s wasn't found in the CABS (%d) file!\n" name hashNr)
                end
              in
(*               let () =  *)
(*                 let global, _, _ = g in  *)
(*                 ignore(Updatedebug.log [Updatedebug.BTPatchgen] "%a\n" d_global global) *)
(*               in *)

              (* compute the req'd CABS defs for this global *)
              let cabsDefListForThisGlobal = 
                (* rewrite a CABS TYPEDEF to bear a new name *)
                let rewriteTypedefName oldTypedef newName = match oldTypedef with
                    GType(ti,loc) -> 
                       GType({ti with tname = newName},loc)

                  | _ -> Updatedebug.abort (Updatedebug.bug "rewriteTypedefName : no way can be here")
                in
                let appropriateHash = function
                    OldVersion -> cabsHash1
                  | NewVersion -> cabsHash2 
                  | _ -> Updatedebug.abort (Updatedebug.bug "appropriateHash : can't call me with MadeUp !")
                in

                match g with 
                    GType(ti,_) as g', ((OldVersion|NewVersion) as ver), GenerateWithSameName ->
                       ignore(Updatedebug.log [Updatedebug.BTPatchgen] "cabsDefListForThisGlobal(GType1) : %a\n" d_shortglobal g');

                      [CabsOutVerbatim (retrieveCabsDefinition (appropriateHash ver) ("typedef " ^ ti.tname))]

                  | GType(ti,_) as g', ((OldVersion|NewVersion) as ver), GenerateWithChangedName(oldName, newName) ->
                       ignore(Updatedebug.log [Updatedebug.BTPatchgen] "cabsDefListForThisGlobal(GType2) : %a\n" d_shortglobal g');
                      let olddef = retrieveCabsDefinition (appropriateHash ver) ("typedef " ^ oldName) in
                      begin
                         match olddef with
                             GType _   -> [CabsOutVerbatim (rewriteTypedefName olddef newName)] 
                           | _ -> Updatedebug.abort (Updatedebug.bug 
                                                      "cabsDefListForThisGlobal : typedef '%s' wasn't found as a typedef in the CIL file!\n" oldName)
                      end

                  | GFun(fd,loc) as g', ((OldVersion|NewVersion) as ver), GenerateWithSameName ->
                       ignore(Updatedebug.log [Updatedebug.BTPatchgen] "cabsDefListForThisGlobal(GFun) : %a\n" d_shortglobal g');

                      begin
                      try
                         let cabs_g = retrieveCabsDefinition (appropriateHash ver) ("function " ^ fd.svar.vname) in
                         let cabs_g' = 
                            if isRoot g'
                            then 
                               cabs_g
                            else 
                               (* non-root; then turn a fundef into an extern declaration *)
                               match cabs_g with 
                                     GFun(fd,loc) ->
                                        fd.svar.vstorage <- Cil.Extern;
                                        (* Add it to non-root hash *)
                                        Commonutil.hashtbl_add_once nonRootFunctions fd.svar.vname ();
                                        (* Turn its definition into an extern declaration *)
                                        GVarDecl(fd.svar,loc)
                                  | _ -> 
                                       Updatedebug.abort (Updatedebug.bug "cabsDefListForThisGlobal(GFun) : got unreasonable global for function '%s'" fd.svar.vname)
                         in
                         [CabsOutVerbatim cabs_g']

                      with Not_found ->
                        Updatedebug.abort (Updatedebug.bug 
                                             "cabsDefListForThisGlobal : function '%s' wasn't found as a FUNDEF in the CABS file!\n" fd.svar.vname)
                      end

                  | GVar(vi,ii,loc) as g', ((OldVersion|NewVersion) as ver), GenerateWithSameName ->
                       ignore(Updatedebug.log [Updatedebug.BTPatchgen] "cabsDefListForThisGlobal(GVar) : %a\n" d_shortglobal g');

                      begin
                      try
                         let cabs_g = retrieveCabsDefinition (appropriateHash ver) ("decdef " ^ vi.vname) in
                         let cabs_g' =
                            if isRoot g'
                            then
                               cabs_g
                            else
                               (* non-root; then turn a fundef into an extern declaration *)
                               let existsInOldVersion g = H.mem globalsFile1 (Commoncilutil.getGlobalName g) in
                               match cabs_g with
                                    GVar(vi,_,loc) ->
                                        (* first filter out static/extern storage specs *)
                                        (* Add it to non-root hash *)
                                        Commonutil.hashtbl_add_once nonRootFunctions vi.vname ();
                                        if existsInOldVersion g' 
                                        then
                                           (* Turn its definition into an extern declaration *)
                                           (vi.vstorage <- Cil.Extern;
                                           GVarDecl(vi,loc))   
                                        else
                                           (* Exists only in the new version; import its definition *)
                                           cabs_g

                                  | _ ->
                                       Updatedebug.abort (Updatedebug.bug "cabsDefListForThisGlobal(GVar) : got unreasonable CABS global for var %s" vi.vname)
                         in
                         [CabsOutVerbatim cabs_g']

                      with Not_found ->
                        Updatedebug.abort (Updatedebug.bug
                                             "cabsDefListForThisGlobal : var '%s' wasn't found as a DECDEF in the CABS file!\n" vi.vname)
                      end


(*                   | GCompTag(ci, _),_,_ when Cil.startsWith "__anon" ci.cname -> [] *)

                  | GCompTag(ci,_) as g', ((OldVersion|NewVersion) as ver), GenerateWithSameName ->
                       ignore(Updatedebug.log [Updatedebug.BTPatchgen] "cabsDefListForThisGlobal(GCompTag1) : %a\n" d_shortglobal g');

                       [CabsOutVerbatim (retrieveCabsDefinition (appropriateHash ver) ("comp " ^ ci.cname))]

                  | GCompTag(ci,loc) as g', ((OldVersion|NewVersion) as ver), GenerateWithChangedName(oldName, newName) ->
                       ignore(Updatedebug.log [Updatedebug.BTPatchgen] "cabsDefListForThisGlobal(GCompTag2) : %a\n" d_shortglobal g');

(*                       ignore (Updatedebug.log [Updatedebug.BTPatchgen] "GCompTag : ci.cname : '%s' oldName : '%s' newName : '%s'\n" ci.cname oldName newName ); *)
                      let olddef = retrieveCabsDefinition (appropriateHash ver) ("comp " ^ oldName) in
                      begin
                        let newdef = 
                          match olddef with
                                GCompTag(ci, loc) -> GCompTag({ci with cname = newName }, loc)

                            | _ -> Updatedebug.abort (Updatedebug.bug 
                                                        "cabsDefListForThisGlobal : ONLYTYPEDEF %s wasn't found as an ONLYTYPEDEF in the CABS file!\n" oldName)
                                
                        in
                          [CabsOutVerbatim newdef] 
                      end

                  | g,_,_ ->
                       ignore(Updatedebug.log [Updatedebug.BTPatchgen] "cabsDefListForThisGlobal verbatim : %a\n" d_shortglobal g);
                       [CabsOutVerbatim g ]

              in
                 (* and join them with the acc *)
                 acc @ cabsDefListForThisGlobal
            in

            (* generate/retrieve the CABS definitions from the globals in the patch *)
            let patchCabsDefinitionList = List.fold_left generatePatchDefinitions_fold [] gl in

            (* fix static declaration of non-root functions *)
            ignore(List.map elimNonRootStatics_map patchCabsDefinitionList);


            (* this pretty much mirrors the printFile in cprint.ml;
               unfortunately, Cprint.print_def is not functional, i.e.
               they use a global reference to the output_channel, so we have to
               duplicate the code here *)
            let myCprintPrintFile cabsDefinitionList = 
              (* use the CIL or CABS pretty printer, depending on the value of cabsOut_t *)
              let printCabsDefinition cabsdef = match cabsdef with
                  CabsOutVerbatim g -> g
              in
              begin
                 let gl = List.map printCabsDefinition cabsDefinitionList in
                 dumpPatchToCilFile gl 
              end
            in
            (* eventually ! go ahead and print to file *)
            myCprintPrintFile patchCabsDefinitionList

        | _ ->
            Updatedebug.abort (Updatedebug.bug "dumpPatchToCabsFile : maincil should have given us exactly two CABS files, but it didn't!\n")
    end  


(* Given a list of global descriptors, produce a C file out of them 
*)
let dumpPatchToFile (gl : (Cil.global * versionType * nameGeneration_t) list) isRoot : unit = 
   dumpPatchToCabsFile gl isRoot


(* constructs a hash containing all functions accessing changed types concretely.
   This functions have to be added to the patch, regardless of whether they've changed or not. 
   Since they access types concretely, we have to get their new versions, otherwise we're not type safe.
*)
let getAffectedFunctions changedTypes (f : Cil.file) : (string, unit) H.t =

   let affectedFunctions = H.create 127 in begin
      try
         begin 
            let vd = Versiondata.loadVersionData !versionDataFile in
            let funEffectHash = vd.Versiondata.vd_an_data.Versiondata.vd_an_funeffect_hash in
            let deltaSolutionHash = vd.Versiondata.vd_an_data.Versiondata.vd_an_solution in
            let funTypeHash = vd.Versiondata.vd_fun_hash in
            let isAffectedFunction concretelyUsedTypesSet = 
               let res = ref false in
                  Deltasolver.CapabilitySet.iter (fun t -> if H.mem changedTypes t then res := true) concretelyUsedTypesSet;
                  !res
            in   
            let doOneFunction fd =
               if H.mem funTypeHash fd.svar.vname then
                  let () = ignore (Updatedebug.log [Updatedebug.BTPatchgen] "getAffectedFunctions: %s is an old one\n" fd.svar.vname) in
                  let originalFunctionType = (H.find funTypeHash fd.svar.vname).Versiondata.vd_fun_type in
                     begin match Analysis.AnalysisEffectManipulator.getEffects (Cil.typeAttrs originalFunctionType) with
                           [Attr(attrName,[AInt key])] ->
                              begin 
                                let failStopFind h id name = 
                                  try
                                    H.find h id 
                                  with 
                                       Not_found ->
                                          (* printSolution solution; *)
                                          Updatedebug.abort (Updatedebug.bug "getAffectedFunctions: Couldn't find delta id %d (%s) in solution hash" id name)
                                in
                                    let effect = 
                                      try
                                      H.find funEffectHash key
                                  with 
                                       Not_found ->
                                          (* printSolution solution; *)
                                          Updatedebug.abort (Updatedebug.bug "getAffectedFunctions: Couldn't find key %d in funEffectHash" key)
                                    in
(*                                     let d_in_id  = effect.Analysis.effect_delta_in.Deltasolver.delta_id in *)
(*                                     let d_out_id = effect.Analysis.effect_delta_out.Deltasolver.delta_id in *)
                                    let d_eigen_id = effect.Analysis.effect_delta_eigen.Deltasolver.delta_id in
(*                                     let d_eps_id = effect.Analysis.effect_vc_eps.Deltasolver.delta_id in *)
(*                                     let d_in_set = failStopFind deltaSolutionHash d_in_id "delta_in" in *)
(*                                     let d_out_set = failStopFind deltaSolutionHash d_out_id "delta_out" in *)
                                    let d_eigen_set = failStopFind deltaSolutionHash d_eigen_id "delta_eigen" in
(*                                     let d_eps_set = failStopFind deltaSolutionHash d_eps_id "d_eps" in *)
(*                                     let d_self_set = Deltasolver.CapabilitySet.diff d_in_set d_out_set in *)
(*                                     let d_self_set = d_eps_set in *)
                                    let d_self_set = d_eigen_set in
(*                                     let d_self_set = d_in_set in *)
                                    let isAffected = if isAffectedFunction d_self_set then (H.add affectedFunctions fd.svar.vname (); true) else false in 
                                       ignore (Updatedebug.log [Updatedebug.BTPatchgen] "###########\n");
                                       Deltasolver.CapabilitySet.iter (fun x  -> ignore (Updatedebug.log [Updatedebug.BTPatchgen] "%s " x)) d_self_set;
                                       ignore (Updatedebug.log [Updatedebug.BTPatchgen] "isAffected %b\n" isAffected)
                              end

                        | _ -> Updatedebug.abort (Updatedebug.bug "annotateWithFunEffects: too many effects or couldn't find effect on function")
                     end
               else 
                  ignore (Updatedebug.log [Updatedebug.BTPatchgen] "getAffectedFunctions: %s is a new one\n" fd.svar.vname)

            in
               List.iter (fun g -> match g with 
                                GFun(fd, _) -> doOneFunction fd
                             | _ -> ())
                  f.globals;
         end
      with _ -> Updatedebug.abort (Updatedebug.error "getAffectedFunctions: failed to load version data from %s\n"  !versionDataFile)

      end;
         affectedFunctions


(* main beast *)
let generatePatch (f : Cil.file) = 

   (* later on we'll have to know which globals are introduced during patching;
      these are called 'roots', since they're the roots for the transitive closure stage
   *)
   let rootHash = H.create 4099 in
   let injectIsRoot l = 
      List.iter (fun (g,_,_) -> H.add rootHash (Commoncilutil.getGlobalNameNonClashing g) ()) l;
      l
   in


  (* pretty print the results of the analysis from the type transformers *)
  let printResultTypes(name:string)(change:(string option* typ option* typ option * bool *fundec option)) = 
     
    (* we define first a function extracting the names of the types *)
(*     let extractName (name: string): string = *)
(*       (\* first of all we define the recursive function that iterates to remove the type *\) *)
(*       let rec tail(name: string) : string = *)
(*         match name.[0] with *)
(*             '"' ->  name  *)
(*           | a ->  (tail (String.sub name 1 ((String.length name)-1))) *)
(*       in *)
(*       (\* we make the call *\) *)
(*       let funcName  = (tail name) in *)
(*         (\* we remove the "" *\) *)
(*       let testValue = String.sub funcName 1 ((String.length funcName)-2) in *)
(*         testValue *)
(*     in *)

    let doSuffix = true and dontSuffix = false in

    (* This function is the one used to output the types themseleves 
     * Note that it also modifies the types names
     **** Note that we do not print the simple types ****
     *)
    let printType (name : string) (t : Cil.typ) (inVersion : versionType) (suffixIt : bool) : (Cil.global * versionType * nameGeneration_t) =    
      let oldVersionTag : string = 
        if inVersion = OldVersion then "old" else "new" in
      let outputName oldName newName = 
        if suffixIt
        then GenerateWithChangedName (oldName, newName)
        else GenerateWithSameName 
      in
(*       let () = ignore(Updatedebug.log [Updatedebug.BTPatchgen] "printType : name '%s' oldVersionTag '%s' suffixIt %b type %a\n" name oldVersionTag suffixIt d_plaintype t) in *)
      let generateNewName (name : string) = 
        if suffixIt 
        then name ^ versionSeparator ^ oldVersionTag 
        else name 
      in
      let globalDescriptor newGlobal originalName newName = 
(*        let () = ignore(Updatedebug.log [Updatedebug.BTPatchgen] "globalDescriptor : originalName '%s' outputName '%s t %a'\n" originalName (generateNewName originalName) Cil.d_plaintype t) in  *)
        (newGlobal, inVersion, outputName originalName newName)
      in
        match t with 
            TNamed(ti,_) ->
              let originalName = ti.tname and newName = generateNewName ti.tname in
              globalDescriptor (GType({ti with tname = newName},locUnknown)) originalName newName

          | TComp(ci,_) -> 
              let originalName = ci.cname and newName = generateNewName ci.cname in
              globalDescriptor (GCompTag({ci with cname = newName},locUnknown)) originalName newName


          | TEnum(ei,_) ->
              let originalName = ei.ename and newName = generateNewName ei.ename in
              globalDescriptor (GEnumTag({ei with ename = newName},locUnknown)) originalName newName

          | _ -> 
              let originalName = name and newName = generateNewName name in
              globalDescriptor (GType({tname = newName; ttype = t; treferenced = true}, locUnknown)) originalName newName
    in
    (* extract the labelled name of the type. e.g. "type \"myInt\"" -> "myInt" *)
    let name = (Commoncilutil.extractGlobalName name) in
    let outputTT t1 t2 tt = 
       (* this is how the output will look like:

          #pragma type_transformer_T T_old T_new
          T_old   definition
          T       definition 
          T_new   definition
          type_transformer_T function definition
       *)
       begin
          let absTypeName = 
             if Namedtypes.isAbstractType t2 
             then Updatecilutil.getAbstractTypeName t2 
             else name
          in
          let ttPragma = (GPragma(Attr (Updateconfig.ttPragmaTypeName,
                                        [ AStr tt.svar.vname; 
                                          AStr absTypeName; 
                                          AStr (absTypeName ^ "_old"); 
                                          AStr (absTypeName ^ "_new")]), Cil.locUnknown),
                          MadeUp,
                          GenerateWithSameName)
          in
          let tOldDefinition = printType name t1 OldVersion doSuffix in 
          let tDefinition    = printType name t2 NewVersion dontSuffix in 
          let tNewDefinition = printType name t2 NewVersion doSuffix in 

          (* fix tt prototype; should be: 
             void tt(t_old * xin, t_new * x_out)
          *)
          let getTypeOfGlobal = function
               GType(ti,_) -> TNamed(ti, []) 
             | GCompTag(ci,_) -> TComp(ci, [])
             | _ -> raise (Failure "")
          in
          let pi1 (x,_,_) = x  in
          let tOld = getTypeOfGlobal (pi1 tOldDefinition) in
          let tNew = getTypeOfGlobal (pi1 tNewDefinition) in

          let varInput:varinfo = makeVarinfo true "xin" (TPtr(tOld,[])) in
          let varOutput:varinfo = makeVarinfo true "xout" (TPtr(tNew,[])) in
          let varNew:varinfo = List.nth tt.sformals 2 in
          let () = Cil.setFormals tt [varInput;varOutput;varNew] in

(*           let () =  *)
(*              let t_old_vi = List.nth tt.sformals 0 in *)
(*              let t_new_vi = List.nth tt.sformals 1 in *)
(*                 t_old_vi.vtype <- TPtr(tOld, []); *)
(*                 t_new_vi.vtype <- TPtr(tNew, []); *)
(*           in *)


          let ttFunctionDefinition = (GFun(tt,locUnknown),
                                      MadeUp,
                                      GenerateWithSameName) 
          in
             if Updateargs.args.Updateargs.arg_gentts = Updateargs.TTGExclusive 
                  (* if only the TTs are to be generated, don't include the type definition *)
             then [ttPragma; tOldDefinition;              tNewDefinition; ttFunctionDefinition ] 
             else [ttPragma; tOldDefinition; tDefinition; tNewDefinition; ttFunctionDefinition ]
       end
    in

        (*
         * finally we make the output of types and type transformers as needed 
         *)
        match change with
            (None, None, Some(t),true,None) -> (* new type *)
              [printType name t NewVersion dontSuffix]
              
          | (Some(s1), Some(t1), Some(t2),true,Some tt) -> (* type modified *)

              (* die if we encounter a type transformer between TNamed(Tcomp) and TNamed(not TComp)*) 
               begin match t1, t2 with
                    TNamed(ti1,_), TNamed(ti2,_) -> 
                       begin match (Cil.unrollType ti1.ttype),(Cil.unrollType ti2.ttype) with
                             TComp(ci1,_), TComp(ci2,_) when ci1.cname = ci2.cname -> 
                                []

                          | _, TComp _| TComp _, _ -> 
                               Updatedebug.abort (Updatedebug.bug "printResultTypes(typedef %s) : base type changed non-trivially between %s and %s. Fix the sources ? " name (Updatedebug.ppType ti1.ttype) (Updatedebug.ppType ti2.ttype))
                              
                          | _ -> outputTT t1 t2 tt
                       end

                  | _ -> outputTT t1 t2 tt
               end

          | (Some(s1), Some(t), None,false, None) -> (* type disappeared *)
              []
          | (Some(s1), Some(t1), Some(t2),false, Some tt) -> (* type did not change *)
              []
          | _ -> 
              Updatedebug.abort (Updatedebug.bug "printResultTypes(%s) : shouldn't be here" name)
 
    in
     
  (* Print out a fun if it's new or has changed *) 
  let printResultFun (name : string) h affectedFunctions g : (Cil.global * versionType * nameGeneration_t) list =
     begin 
     try
        begin 
        match (H.find h name).Facts.f_change with
              
           (* Changed or added; include it*)
             Commondefs.Added 
           | Commondefs.OtherChange _
           | Commondefs.TypeChanged _ 
           | Commondefs.Renamed _ 
           | Commondefs.ValChanged -> 
(*                 ignore(Updatedebug.log [Updatedebug.BTPatchgen] "CHANGED\n"); *)
                injectIsRoot ([(g, NewVersion, GenerateWithSameName)])

           (* Not changed or deleted *)
           | Commondefs.Deleted 
           | Commondefs.NoChange -> 
(*                 ignore(Updatedebug.log [Updatedebug.BTPatchgen] "NOCHANGE\n"); *)
                []
        end
     with
        Not_found -> 
           begin 
           match g with

              (* affectedFunction *) 
               GFun(fd,_) when H.mem affectedFunctions fd.svar.vname -> 
                  injectIsRoot ([(g, NewVersion, GenerateWithSameName)])  

              | _ -> 
                  Updatedebug.abort (Updatedebug.bug "printResultFun(%s) : Not_found\n" name)
           end
     end
  in
  (* Print out a var if it's new or has changed *) 
  let printResultVar  (name : string) h g gvarTTHash : (Cil.global * versionType * nameGeneration_t) list =
     begin 
     try
        begin 
        match (H.find h name).Facts.g_change with
              
           (* Changed or added; include it*)
             Commondefs.Added | Commondefs.Renamed _ ->

                injectIsRoot ([(g, NewVersion, GenerateWithSameName)])

           | Commondefs.ValChanged -> 
                ignore(Updatedebug.log [Updatedebug.BTPatchgen] "CHANGED\n");
                let new_vi = match g with
                     GVar(vi,_,_) -> vi
                   | _ -> Updatedebug.abort (Updatedebug.bug "printResultVar(%s) : this not a Gvar ?" name)

                in
                let old_var_decl = Cil.makeGlobalVar (Updateconfig.makeOldExternVarName new_vi.vname) new_vi.vtype in
                let () = old_var_decl.vstorage <- Cil.Extern in
                let old_var_global = GVarDecl(old_var_decl, Cil.locUnknown) in
                injectIsRoot ([(old_var_global, NewVersion, GenerateWithSameName);
                               (g, NewVersion, GenerateWithSameName)])

           (* Not changed or deleted *)
           | Commondefs.OtherChange _
           | Commondefs.TypeChanged _ 
           | Commondefs.Deleted 
           | Commondefs.NoChange -> 
(*                 ignore(Updatedebug.log [Updatedebug.BTPatchgen] "NOCHANGE\n"); *)
                []
        end
     with
        Not_found -> [] (* semdiff doesn't keep unchanged functions in functionCensus *)
     end
  in

  let semdiffRes = begin 
     match !theFiles with
          [one; two] -> 
              let res = Semdiff.computeFunctionAndTypeChanges one two in
              H.remove res.Semdiff.sr_functions "main"; (* there ! we don't need to bloat the patch by including main() *)
              res
   
        | _ -> Updatedebug.abort (Updatedebug.bug "semdiffRes : should've found two files here \n")
  end
  in

  (* for all GVars whose value has changed, force a type transformer generation *)   
  let gvarTypetransformers = 
     let res = H.create 127 in
     let gatherAllGVarTypes_iter gvarName fact = match fact.Facts.g_change with

           Commondefs.ValChanged ->
              begin
                 let () = ignore (Updatedebug.log [Updatedebug.BTAlways] "gvars whose values have changed : %s\n" gvarName) in
                 let findG hash name version = 
                    try
                       H.find hash name
                    with Not_found -> 
                       Updatedebug.abort (Updatedebug.bug "gatherAllGVarTypes_fold : can't find '%s' in globalHash%d. \n" name version)
                 in

                 let getViOfG g name version = match g with
                       GVar(vi,_,_) -> vi
                    | _ -> Updatedebug.abort (Updatedebug.bug "gatherAllGVarTypes_fold : found '%s' in globalHash%d, but it's not a GVar ? \n" name version)
                 in
                    (*         let getTiOfG g name version = match g with *)
                    (*                     GType(ti,_) -> ti *)
                    (*                   | _ -> Updatedebug.abort (Updatedebug.bug "gatherAllGVarTypes_fold : found '%s' in globalHash%d, but it's not a GType ? \n" name version) *)
                    (*         in *)

                 let canonicalGvarName = (Commoncilutil.getGlobalNameLight_gvar gvarName) in 
                 let v1 = getViOfG (findG globalsFile1 canonicalGvarName 1) canonicalGvarName 1
                 and v2 = getViOfG (findG globalsFile1 canonicalGvarName 2) canonicalGvarName 2 in

                 let typeOfGvarName = Commoncilutil.getGlobalNameLight_gtype (Updateconfig.makeGlobalVarTypeName v2.vname) in
                    
                 let tt = Typeprocessing.doOneTransformer v1.vtype v2.vtype typeOfGvarName in
                    H.add res gvarName (Some typeOfGvarName,Some v1.vtype,Some v2.vtype, true, Some tt)
              end

        | _ -> ()

     in
     
     (* semdiff has put all vars that underwent a change in semdiffRes.Semdiff.sr_vars 
        We have to force TT generation for all of them, and then add them to hashResultTypes below *)   
     H.iter gatherAllGVarTypes_iter semdiffRes.Semdiff.sr_vars;
     res
  in

  (* invoke the two analyses *)
  let changedTypes = H.create 13 in
  let h = Typetransformer.constructTypeTransformers [] globalsFile1 globalsFile2 in
  ignore (Updatedebug.log [Updatedebug.BTAlways] "Changed types: \n");
  H.iter 
    (fun name change ->
      match change with
	        (_,_,Some t,true,Some fd) -> (ignore (Updatedebug.log [Updatedebug.BTAlways] "  %s\n" name);
                                     H.add changedTypes (Updatecilutil.getAbstractTypeName t) ())
      | _ -> ()) h;

  let hashResultTypes : hashOfTransformedTypes = match Updateargs.args.Updateargs.arg_gentts with
       Updateargs.TTGYes ->       h
     | Updateargs.TTGNo ->        H.create 1 
     | Updateargs.TTGExclusive -> h
  in

   if Updateargs.args.Updateargs.arg_gentts = Updateargs.TTGExclusive 
   then 
      begin
         let getTTsOnly name change = match change with
	           (_, _, _, true, Some fd) -> List.map (fun (g,_,_) -> g) (printResultTypes name change)
            | _ -> []
         in
         let ttList = H.fold (fun name change acc -> (getTTsOnly name change) @ acc) hashResultTypes [] in
         dumpPatchToCilFile ttList
      end
   else
      let affectedFunctions = getAffectedFunctions changedTypes f in

(*   let () = H.iter (fun name change -> ignore (Updatedebug.log [Updatedebug.BTPatchgen] "********hashResultTypes  %s\n" name);H.add hashResultTypes name change) gvarTypetransformers in *)

(*     let hashResultFunctions:hashOfModifiedFunctions = Modifiedfunctions.detectModifiedFunctions [] globalsFile1 globalsFile2 in *)

(*     let () = ignore (Updatedebug.log [Updatedebug.BTPatchgen] "|hashResultTypes| = %d,  |hashResultFunctions| = %d\n" (Commonutil.hashtbl_size hashResultTypes) (Commonutil.hashtbl_size hashResultFunctions)) in  *)

    let pi1 (x,_,_) = x in

    (* add the globals generated by patching to file2's global list *)
    let generateNewGlobalsAfterPatch gList = 
       let printPatchTypesAndFunctions_fold acc g =
          let printPatchTypesAndFunctions_one g =
             (* we have >= 0 new globals introduced for this g *)
             let name = Commoncilutil.getGlobalName g in
                ignore (Updatedebug.log [Updatedebug.BTPatchgen] "******** now doing %s\n" name);
                if H.mem hashResultTypes name
                then 
                   (* affected type *)
                   let () = ignore (Updatedebug.log [Updatedebug.BTPatchgen] "******** hashResultTypes %s\n" name) in
                   injectIsRoot (printResultTypes name (H.find hashResultTypes name))
                else 
(*                    if H.mem hashResultFunctions name  *)
                   (* semdiff indexes the functionCensus on bare function name; oh well *)
                   match g with 


                        GFun(fd,_) when ((H.mem semdiffRes.Semdiff.sr_functions fd.svar.vname) || 
                                         (H.mem affectedFunctions fd.svar.vname)) &&      
                                         (fd.svar.vname <> "main" ) ->
                            (* affected function *)
                            let () = ignore (Updatedebug.log [Updatedebug.BTPatchgen] "******** hashResultFunctions %s\n" name) in
                            printResultFun fd.svar.vname semdiffRes.Semdiff.sr_functions affectedFunctions g

                      | GVar(vi,ii,loc) when H.mem semdiffRes.Semdiff.sr_vars vi.vname ->
                            let () = ignore (Updatedebug.log [Updatedebug.BTPatchgen] "******** hashResultVars %s\n" name) in
                            printResultVar vi.vname semdiffRes.Semdiff.sr_vars g gvarTypetransformers

                      | _ ->
                           (* no patch item generated for this global (e.g. because it did not change) *)
                           let () = ignore (Updatedebug.log [Updatedebug.BTPatchgen] "******** no patch item for %s\n" name) in
                           [(g, NewVersion, GenerateWithSameName)]
          in
          let res = printPatchTypesAndFunctions_one g in
(*           let () = ignore (Updatedebug.log [Updatedebug.BTPatchgen] "res has %d elems: " (List.length res)) in *)
(*           let () = List.iter (fun (g,_,_) -> ignore (Updatedebug.log [Updatedebug.BTPatchgen] " %s" (Commoncilutil.getGlobalNameNonClashing g))) res in *)
(*           let () = ignore (Updatedebug.log [Updatedebug.BTPatchgen] "\n") in *)
             if res = [] 
             then acc @ [(g, NewVersion, GenerateWithSameName)]
             else acc @ res
       in
          List.fold_left printPatchTypesAndFunctions_fold [] f.globals
    in

    (* all globals, intermixed; these will be filtered so that only the 
       globals needed to make the patch compile are kept *)

    let newGlobalsAfterPatch = generateNewGlobalsAfterPatch f.globals in

(*      let () = List.iter (fun (g,_,_) -> ignore (Updatedebug.log [Updatedebug.BTPatchgen] "******** newGlobalsAfterPatch %s\n" (Commoncilutil.getGlobalNameNonClashing g))) newGlobalsAfterPatch in  *)
    (* isRoot is a global->bool that's passed to Rmtmps.removeUnusedTemps *)
    let isRoot g = H.mem rootHash (Commoncilutil.getGlobalNameNonClashing g) in

    let h23 = H.create 4099 in
    let addToH23_iter (g, version, name_generation) = 
       let globalName = Commoncilutil.getGlobalNameNonClashing g in
       let () = ignore (Updatedebug.log [Updatedebug.BTPatchgen] "addToH23_iter %s\n" globalName) in
          begin
          try
             ignore (H.find h23 globalName);
             (* we allow decls and pragmas to be repeated *)
             if not (Cil.startsWith "decl" globalName) && not (Cil.startsWith "pragma" globalName)
             then Updatedebug.abort (Updatedebug.bug "addToH23_iter : clash on %s\n" globalName)
          with Not_found ->
             H.add h23 globalName (version, name_generation)
          end
    in

    let retrieveFromH23 g = 
       let globalName = Commoncilutil.getGlobalNameNonClashing g in
          begin
          try
             let (version, name_generation) = H.find h23 globalName in
             (g, version, name_generation)

          with Not_found ->
             Updatedebug.abort (Updatedebug.bug "retrieveFromH23 : %s not found !\n" globalName)
          end
       
    in
    (* remember the version and name_generation for each global *)
    let () = List.iter addToH23_iter newGlobalsAfterPatch in

    (* create a dummy file and run Rmtmps.removeUnusedTemps on it 
       Rmtmps.removeUnusedTemps does transitive closure starting with globlas 
       that pass the isRoot test; all globals that are not reachable from the roots are eliminated
    *)
(*     let () = ignore (Updatedebug.log [Updatedebug.BTPatchgen] "# newGlobalsAfterPatch  = %d\n" (List.length newGlobalsAfterPatch)) in *)
    let dummyFile = {Cil.dummyFile with globals = List.map pi1 newGlobalsAfterPatch} in
(*     let () = ignore (Updatedebug.log [Updatedebug.BTPatchgen] "# rootHashSize  = %d\n" (Commonutil.hashtbl_size rootHash)) in *)
    let () = Rmtmps.removeUnusedTemps ~isRoot:isRoot dummyFile in
(*     let () = ignore (Updatedebug.log [Updatedebug.BTPatchgen] "# after removeUnusedTemps, newGlobalsAfterPatch  = %d\n" (List.length dummyFile.globals)) in *)

    (* now we have only the globals we need; stick back version/name_generation *)
    let dumpableGlobals = List.map retrieveFromH23 dummyFile.globals in
(*     let () = List.iter (fun (g,_,_) -> ignore (Updatedebug.log [Updatedebug.BTPatchgen] "******** dumpableGlobals %s\n" (Commoncilutil.getGlobalName g))) dumpableGlobals in *)

    (* got all the globals we needed, now write them to a C file *)
    dumpPatchToFile dumpableGlobals isRoot




(*  Things to do before processing a file *)
let pre (f : Cil.file) : unit =  
   begin
    (* check for valid patch name *)
    (if !patchFileName = "" 
    then Updatedebug.abort (Updatedebug.error "You must specify a patch output file using --patchfile-out=<file> !\n"));

    (* we need the previous version's vd ; not that this is passed via --pversiondata-in, not --versiondata-in *)  
    (if !versionDataFile = ""
    then Updatedebug.abort (Updatedebug.error "You must specify a VD input file using --pversiondata-in=<file> !\n"));

    (* remove static specifiers *)  
    Removestatics.removeStaticSpecifiers f;
    (* do loop extraction; side-effecting *)
    ignore(Loops.default_update_loops f);
  end

(*  Things to do after processing a file *)
let post (f : Cil.file) =  
   begin
      theFiles := !theFiles @ [f];
      if (!stage = ProcessFile1Stage) 
      then 
         stage := ProcessFile2Stage
      else 
         Stats.time "Stats : doPatchGen" generatePatch f 
   end



(* this is called for each file in turn; 
   for the second file, post() will do the TT inference and printing 
 *)
let patchGenProcessFile (f : Cil.file) : unit = 
   pre f;

(*    Gvartypes.genGlobalVarUniqTypes f; *)

   begin
      if !stage = ProcessFile1Stage
      then Typetransformer.fillGlobalHash f Typetransformer.filterOutGlobalDeclarations globalsFile1
      else Typetransformer.fillGlobalHash f Typetransformer.filterOutGlobalDeclarations globalsFile2
   end;
   post f
  
(*
 * Patch generation feature descriptor
 *)

let doPatchgen = ref false

let feature : featureDescr = 
  { fd_name = "patchgeneration";
    fd_enabled = doPatchgen;
    fd_description = "generate the dynamic patch from an old version to a new version";
    fd_extraopt = [("--patchfile-out", 
                       Arg.String (fun s -> patchFileName := s),
                       "generated patch will be written to this file");

                   ("--pversiondata-in", 
                       Arg.String (fun s -> versionDataFile := s),
                       "name of version data input file"); ];

    fd_doit = (function (f: file) -> patchGenProcessFile f);
    fd_post_check = true;
  } 

