(* 
 *
 * Copyright (c) 2004-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Michael Hicks       <mwh@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *  Eric Hardisty       <hardisty@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(* Check aliasing and mark the named types that are invalidly aliased as non-updatable *)


open Cil

module H = Hashtbl


(* sanity checking *)
let (setT,checkT) =
  let nonupd_hash = H.create 51 in
  let rec aux t loc = 
    let tname = (Updatecilutil.getAbstractTypeName t) in
    if not (Forcedpragma.isForcedUpdatable tname) then 
      begin
	H.replace nonupd_hash t loc;
        (match t with
          TNamed(ti,attr) when Namedtypes.isAbstractType ti.ttype -> 
	    aux ti.ttype loc
        | _ -> ())
      end in
  let check () =
    H.iter (fun t u -> 
      if (Namedtypes.isUpdatableOrNotSeen t) then
        ignore(Updatedebug.log [Updatedebug.BTAlways]
	   "ptrAliasTracking: (%a) needsAliasCheck: spurious bad type %a?\n"
                 d_loc u d_type t)
      else ()) nonupd_hash in
  (aux,check)

let markType t reason =
  match t with
    TNamed(_,_) | TComp (_) -> 
      setT t !Cil.currentLoc;
      (* Namedtypes.setTypeNonUpdatable t (!Cil.currentLoc) reason *)
  | _ -> ()

let rec unwrap t = 
  match t with
    TPtr (t',_) -> unwrap t'
  | TArray (t',_,_) -> unwrap t'
  | _ -> t

(* Assumes we have seen AddrOf(lval), and thus have to invalidate
   types appearing in lval, due to creating an alias "under a con" *)

let markTypeNonUpdatableLval lval =
  let rec markOffset host (pre : Cil.offset) (suff : Cil.offset) =
    let rec append_offset o1 o2 =
      match o1 with
        NoOffset -> o2
      | Field (fi,off') -> Field(fi,append_offset off' o2)
      | Index (e,off') -> Index(e,append_offset off' o2)
    in
    let aux off off' =
      if not (off' = NoOffset) then
        (let new_pre = append_offset pre off in
        let t = Updatecilutil.properTypeOfLval (host,new_pre) in
        markType t "alias created using &";
        markOffset host new_pre off') in
    match suff with
      NoOffset -> ()
    | Field (fi,off') -> aux (Field (fi,NoOffset)) off'
    | Index (exp, off') -> aux (Index(exp,NoOffset)) off'
  in
  (match lval with
    (h, NoOffset) -> () (* This is &h *)
	(* not pointing into the middle of an object, so this is OK *)
  | ((Mem e) as h, off) -> (* This is &(( *e).off) *)
      let e_t = Updatecilutil.properTypeOf e in markType e_t "alias created with &";
      let e_t' = Cil.unrollType e_t in
      (match e_t' with
        TPtr (t,_) -> markType t "alias created with &"
      | _ -> failwith "Mem e: unexpected type!");
      markOffset h NoOffset off
  | ((Var v) as h, off) -> (* This is &(x.fi...) or &(x[i]...) *)
      let e_t = v.vtype in
      (match (Cil.unrollType e_t) with
        TComp _ | TArray _ ->  markType e_t "alias created with &"
      | _ -> failwith "Var v: unexpected type!");
      markOffset h NoOffset off)
        
(* Given two non-structurally-identical types t1 and t2 that are being
   equated (presumably under pointers), we must invalidate any named
   types to which they refer. *)

let markTypeNonUpdatable (t1: Cil.typ) (t2: Cil.typ) = 
  (* Mark t1 and t2 (and any type aliases) as non-updatable,
     if they are named types. *)
  ignore(Updatedebug.log [Updatedebug.BTSuspect] 
	   "type equality invalidating type %a == %a\n"
	   Cil.d_type t1 Cil.d_type t2);
  markType (unwrap t1) "alias due to type equality under references";
  markType (unwrap t2) "alias due to type equality under references";
(*   If, when unrolled, t1 and t2 are not equal, and either one is a *)
(*   compound type, then mark that type as not updateable, and also *)
(*   mark any of the types contained within that type as *)
(*   non-updateable too. *)
  let t1' = Cil.unrollType t1 in
  let t2' = Cil.unrollType t2 in
  if not (Typeequality.uAreTypesEqualNoAttrs t1' t2') then begin
    markType (unwrap t1') "alias due to type equality under references";
    markType (unwrap t2') "alias due to type equality under references"
    (* FIXME: need to mark elements of the compound type as
       non-updateable too! *)
  end

(* Only pointer and array types need alias check.  In this case,
   returns back the target type of the pointer or array.  QUESTION:
   what do I do about typedefs that are pointers?  In some sense I
   don't want to unroll them because if I cast a T to a int * then
   that should be fine--it's just a con if T is updateable, and a
   legal equality if T is not updateable. *)

(* On the other hand, what if I have T == U *, where T is an int * and
   U is an int.  If T is not updateable, then this is equivalent to
   int * = U *, which would seem to reveal representation of U via
   aliasing.  The problem is that I don't know at this point whether T
   or U is updateable or not. *)

(* FIXME: figure out how to deal with this.  For now, we unroll the
   types, and if they're pointers, we just invalidate everything. *)

(* MWH: I think this code is now obsolete, as I've beefed up subtyping
   to take care of the same thing. *)

let needsAliasCheck t_src t_dst =
  let isAlias_tsrc = (Cil.isPointerType t_src) || (Cil.isArrayType t_src) in
  let isAlias_tdst = (Cil.isPointerType t_dst) || (Cil.isArrayType t_dst) in
  let isAbs_tsrc = Namedtypes.isAbstractType t_src in
  let isAbs_tdst = Namedtypes.isAbstractType t_dst in
  let types_eq = Typeequality.uAreTypesEqualNoAttrs t_src t_dst in
(*   ignore (Updatedebug.log [Updatedebug.BTAlways] "needsAliasCheck src=%a (%b,%b) dst=%a (%b,%b) eq=%b\n" d_type t_src isAlias_tsrc isAbs_tsrc d_type t_dst isAlias_tdst isAbs_tdst types_eq); *)
  if isAlias_tsrc && isAlias_tdst then 
    (* src and dst are pointers *)
    if types_eq then false (* identity cast on pointers is safe *)
    else 
      (* pointers are differnt types and not equal *)
      if (not isAbs_tsrc) && (not isAbs_tdst) then 
        (* both non abstract and differ: don't need check *)
        false
      else 
        (* At least one is abstract and they differ: need check *)
        true
  else 
    false (* src and dst are not pointers *)


(* looks for aliasing problems in static initializers. *)
class namedTypesAliasTrackerStaticInitializers : cilVisitor = object(self)
  inherit nopCilVisitor

  method vexpr e = 
    (match e with
      CastE(t,e1) ->  
        let t1 = Updatecilutil.properTypeOf e1 in
        if needsAliasCheck t1 t then markTypeNonUpdatable t1 t
    | AddrOf(lval) -> 
	(* if not Updateargs.args.Updateargs.arg_dosuspect then *)
	(* MWH: We don't handle this is suspectAlias, because we know
           it will always rule out the updatability of the type. *)
	markTypeNonUpdatableLval lval
    | _ -> ());
    DoChildren

end

(* we need this for the arguments of Asm's *)
class killAllTypesInSightTracker : cilVisitor = object(self)
  inherit nopCilVisitor

  method private doType t = match Updateargs.args.Updateargs.arg_assembly with 

     | Updateargs.UnsoundIgnore -> 
          ()

     | Updateargs.UnsoundWarn -> 
          (if Namedtypes.containsUpdatableNamedTypes t 
          then ignore (Updatedebug.log [Updatedebug.BTAlways] "%a : Warning : type '%s' used in inline assembly\n" d_loc !Cil.currentLoc (Updatedebug.ppType t)))

     | Updateargs.UnsoundKillTypeShallow -> 
          Namedtypes.setTypeNonUpdatable t !Cil.currentLoc "argument to Asm"

     | Updateargs.UnsoundKillTypeDeep -> 
          Namedtypes.setTypeNonUpdatableDeep t !Cil.currentLoc "argument to Asm"

     | Updateargs.UnsoundDie -> 
          (if Namedtypes.containsUpdatableNamedTypes t 
          then Updatedebug.abort (Updatedebug.error "%a : type '%s' used in inline assembly and you requested bailing in this case" d_loc !Cil.currentLoc (Updatedebug.ppType t)))


  method vvrbl v = 
     self#doType v.vtype;
     DoChildren

  method vtype t = 
     self#doType t;
     DoChildren

end

(* visitor to remove the abs and cons of unused types *)


class nonUpdatableConAbsRemovalVisitor : cilVisitor = object(self)
  inherit nopCilVisitor

  method vinst i = 
    begin match i with
       Call (result, Lval(Var vi, NoOffset), actuals, loc) ->
           if Namedtypes.isRemovableConAbs vi.vname  
           then 
              raise (Failure "ConAbsRemoval!!")
           else DoChildren
              
       | _  -> DoChildren
    end
end





(* Identify all invalid casts needed and set the types involved to be non-updatable.
   Invalid casts are:
      - any casts between function types where the types are not interchangeable
      - any casts to/from/between pointers where types are not interchangable
      - any casts to/from/between named types where unrolled types are not interchangable 

*)


(* MWH: this is going to be conservative in the case that the two types
   contain some named types that ultimately get made non-updatable. *)

(* determine if two c-types are safely interchangeable wrt conabs insertion *)
let interchangeableTypes (t1:Cil.typ) (t2:Cil.typ) : bool =
   let res = 
   (* Collapse (equate): base types, pointers to functions and functions, and pointers and arrays *)
   let rec collapse (usig:Typeequality.utypsig) : Typeequality.utypsig = 
      match usig with 
         Typeequality.UTSArray(ut,e,al) -> Typeequality.UTSPtr(collapse ut, al) 
       | Typeequality.UTSPtr(Typeequality.UTSFun(ut,utl,b,al), alptr) -> Typeequality.UTSFun(collapse ut,List.map collapse utl,b,al)
       | Typeequality.UTSPtr(ut,al) -> Typeequality.UTSPtr(collapse ut,al)
       | Typeequality.UTSNamed(s,ut,al) -> Typeequality.UTSNamed(s,collapse ut,al)
       | Typeequality.UTSComp(b,s,al) -> Typeequality.UTSComp(b,s,al) 
       | Typeequality.UTSFun(ut,utl,b,al) -> Typeequality.UTSFun(collapse ut,List.map collapse utl,b,al) 
       | Typeequality.UTSEnum(s,al) -> Typeequality.UTSEnum(s,al) 
       | Typeequality.UTSBase(t) ->  Typeequality.UTSBase(TVoid [])    
   in
   let cu = Namedtypes.containsUpdatableNamedTypes t1 || Namedtypes.containsUpdatableNamedTypes t2 in 
   let ce = (collapse (Typeequality.uTypeSigNoAttrs t1)) = (collapse (Typeequality.uTypeSigNoAttrs t2)) in
   (*ignore (Updatedebug.log [Updatedebug.BTAlways] "interchangeableTypes: %s  =  %s : %b(cu) %b(=)\n" (Updatedebug.ppType t1) (Updatedebug.ppType t2) cu ce);*)
   if cu then
      ce
   else true  (* neither type contains any named types, so they are con/abs interchangable *)
   in 
   res
   


let markIfInvalid t1 t2 loc =    
(*   let same = interchangeableTypes t1 t2 in *)
(*   ignore(Updatedebug.log [Updatedebug.BTAlways]
                       "(%a) Checking %a <= %a\n" d_loc !Cil.currentLoc d_type t1 d_type t2);
*)
   let _ = (* kill any aliases to va_list *)
     match t1,t2 with
       TNamed _, t
     | t, TNamed _ ->
	 (match (Cil.unrollType t) with TBuiltin_va_list _ ->
	   Namedtypes.setTypeNonUpdatableDeep t loc "Alias for __builtin_va_list";
	 | _ -> ())
     | _,_ -> () in
   let eq1 = Subtype.isSubtype true t1 t2 in
(*   let eq2 = Subtype.isSubtype true t2 t1 in *)
   let cu = Namedtypes.containsUpdatableNamedTypes t1 ||
     Namedtypes.containsUpdatableNamedTypes t2 in 
   let eq = eq1 || (not cu) in
(*   if (same <> eq) then
     ignore(Updatedebug.log [Updatedebug.BTAlways]
                       "(%a) markIfInvalid: same (%b) != subtype (%b,%b) t1=%a t2=%a\n"
                                 d_loc !Cil.currentLoc same eq1 eq2
                                 d_type t1 d_type t2);
*)
   if (not eq) then 
     (Namedtypes.setTypeNonUpdatableDeep t1 loc "Used in unsafe cast";
     Namedtypes.setTypeNonUpdatableDeep t2 loc "Used in unsafe cast")


class markInvalidSubtypesNonUpdatable 
    (fd:Cil.fundec) 
    (mallocHash : Mallocpragma.mallocHash_t) 
    (asmTracker: cilVisitor)
    : cilVisitor = object(self)
  inherit nopCilVisitor

  method vexpr exp =
    match exp with 
      CastE(t,e1) ->
        let t1 = Updatecilutil.properTypeOf e1 in
	(* MWH: should this alias check also be predicated on whether NULL? *)
        if needsAliasCheck t1 t then markTypeNonUpdatable t1 t;
	(* Allow null initialisers that would otherwise be detected as invalid casts *) 
        if not (Updatecilutil.isNullInitialiser (!Cil.currentLoc) t e1) then 
          markIfInvalid t1 t (!Cil.currentLoc);
        DoChildren
    | AddrOf(lval) -> 
	if not Updateargs.args.Updateargs.arg_dosuspect then 
	  markTypeNonUpdatableLval lval;
	DoChildren
    | _ -> DoChildren

  method vstmt s = 
    match s.skind with 
      Return(Some(e),loc) -> 
        let return_type = Updatecilutil.getReturnType fd in
        let e_type = Updatecilutil.properTypeOf e in
        markIfInvalid e_type return_type loc;
        if needsAliasCheck e_type return_type then
	  markTypeNonUpdatable e_type return_type;
        DoChildren

      | _ -> DoChildren
      
  method vinst i =
    begin match i with
      Set(l,r,loc) ->
        let lt = Updatecilutil.properTypeOfLval l in
        let rt = Updatecilutil.properTypeOf r in
        if needsAliasCheck rt lt
            && (not (Updatecilutil.isNullInitialiser (!Cil.currentLoc) lt r)) then
	  markTypeNonUpdatable rt lt;
        markIfInvalid rt lt loc;
        DoChildren

    (* MWH: note that malloc() no longer treated specially *)
    | Call(lval_ret_opt, e_fun, e_args, loc) ->
	let (exp_func_ret_t, exp_func_args_t, isva) = 
	  Updatecilutil.getTypesInFunExp e_fun in 

        (* check subtyping of arguments *)
	Commonutil.iter2_ignore_trailing 
	  (fun targetType_i e_i ->
	    let e_t = Updatecilutil.properTypeOf e_i in
	    if needsAliasCheck e_t targetType_i
	    then markTypeNonUpdatable e_t targetType_i)
	  exp_func_args_t e_args;
	
	Commonutil.iter2_ignore_trailing 
	  (fun e t -> markIfInvalid (Updatecilutil.properTypeOf e) t loc) 
	  e_args exp_func_args_t;
        
        (* check return subtype *)
        begin match lval_ret_opt with 
          None -> ()  (* nothing to return => nothing to check *)
        | Some(lval) ->
	    (markIfInvalid exp_func_ret_t (Updatecilutil.properTypeOfLval lval) loc;
	     let lt = Updatecilutil.properTypeOfLval lval in
             if needsAliasCheck exp_func_ret_t lt then
	       markTypeNonUpdatable exp_func_ret_t lt)
        end;
	DoChildren

    | Asm (sl,isvol,outs,ins,clobs,l)  -> 
        let outs_opt_s_list, outs_s_list, outs_lv_list = Commonutil.splitTriples outs
        and ins_opt_s_list, ins_s_list, ins_e_list     = Commonutil.splitTriples ins
        in begin
          List.iter (fun e  -> ignore(Cil.visitCilExpr asmTracker e)) ins_e_list;
          List.iter (fun lv -> ignore(Cil.visitCilLval asmTracker lv)) outs_lv_list;
        end;
	DoChildren
     end;
end

(* iter over the globals and make types non-updatable based on suspect usage. *)
let doNamedTypesTracking (mallocHash : Mallocpragma.mallocHash_t) (file:Cil.file) : unit =

   (* we only need to compute this once, since it doesn't depend on the current global *)
   let staticInitializerTracker = new namedTypesAliasTrackerStaticInitializers in

   (* we need this for the arguments of Asm's *)
   let asmTracker = new killAllTypesInSightTracker in

   let namedTypesTrack_iter g = match g with

        GFun(fd,_) -> 
           let functionTracker = 
	     new markInvalidSubtypesNonUpdatable fd mallocHash asmTracker in
           ignore(Cil.visitCilBlock functionTracker fd.sbody)

       | GVar(vi, {init = Some i}, loc) ->
            ignore(Cil.visitCilInit staticInitializerTracker vi NoOffset i)

      | _ -> ()
   in  
   Cil.iterGlobals file namedTypesTrack_iter;

   (* ignore some special types *)
   List.iter (fun t -> Namedtypes.setTypeNonUpdatable t (Cil.locUnknown) "predefined non-updatable type") Updateconfig.typesToIgnore;
 
   checkT();;

