(* 
 *
 * Copyright (c) 2004-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Michael Hicks       <mwh@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(*
  This module is an attempt to put all of the logic for traversing the 
  subtype relation into one place.  It provides a functor SubtypeCmp 
  that calculates a subtype relation and for each subtype decision calls
  one of the functions from a SubtypeCmpActions module.
  

 *)


open Cil


(*=============================================================*)
(* Generic definition of subtyping used throughout *)
(*=============================================================*)

let rec unrollNonUpdType t = 
  match t with
    TNamed (r, _) -> 
      if not (Namedtypes.isUpdatableOrNotSeen t) then
	unrollNonUpdType r.ttype
      else t
  | x -> x

let rec isConst = function
    (Attr("const",[]))::t -> true
  | h::t -> isConst t
  | [] -> false


  (*
    Check it typ1 <: typ2  (includes physical subtyping)
    coercion_ok flag indicates whether or not the types can be coerced
	(e.g. a float to an int)
  *)
let subcnt = ref 0 
let rec isSubtype' (coercion_ok:bool) (typ1 : Cil.typ) (typ2 : Cil.typ) : bool =
  (* called when it might be unsound to treat t1 <= t2 *)
(*   let unsafeAction action = *)
(*     match action with *)
(*       Updateargs.UnsoundWarn ->  *)
(* 	ignore (Updatedebug.log [Updatedebug.BTAlways]  *)
(* 		  "Allowing unsafe subtype %a <= %a\n" d_type typ1 d_type typ2); *)
(* 	true *)
(*     | Updateargs.UnsoundIgnore -> true *)
(*     | _ -> ignore (Updatedebug.log [Updatedebug.BTAlways]  *)
(* 		     "Unsafe subtype %a <= %a\n" d_type typ1 d_type typ2); *)
(* 	false in *)

  match typ1, typ2 with 
      TVoid _, TVoid _ -> true
    | TInt _, TInt _  
    | TEnum _, TInt _ 
    | TInt _, TEnum _ 
    | TPtr _, TInt _ -> (* pointers can be ints *)
	(* FIX: do I need to worry about the alignment being the same, too? *)
	let bitsi1 = Cil.bitsSizeOf typ1 in
	let bitsi2 = Cil.bitsSizeOf typ2 in
	bitsi1 == bitsi2 
    | TFloat (f1,_), TFloat (f2,_) -> f1 == f2 || coercion_ok
    | TInt _, TFloat _
    | TFloat _ , TInt _ -> coercion_ok
    | TEnum _, TEnum _ -> true

    (* can only cast from int to pointer if pointer non-updatable *)
    | TInt _, TPtr (t,a) ->
        ((not (Namedtypes.isAbstractType t)) ||
	(not (Namedtypes.isUpdatableOrNotSeen t))) &&
	isSubtype coercion_ok typ2 typ1

    (* casting void* to t potentially unsafe if it contains updatable types *)
    | TPtr((TVoid _), _), t ->
	(* If it contains updatable types then we don't allow
	   it to be a subtype. *)
	(* MWH: FIXME, unsafe (and inconsistent!) *)
	(* (not (Namedtypes.containsUpdatableNamedTypes t)) ||
	(unsafeAction Updateargs.args.Updateargs.arg_fromvoid) *)
	(* If it contains updatable types then we don't allow
	   it to be a subtype. *)	
(* 	let res = Namedtypes.containsUpdatableNamedTypes t in *)
(* 	not res *)
    true    

    | t, TPtr((TVoid _), _) -> 
	(* If it contains updatable types then we don't allow
	   it to be a subtype. *)	
(* 	let res = Namedtypes.containsUpdatableNamedTypes t in *)
(* 	not res *)
    true        

    (* pointers to functions should be treated the same as functions *)
    | (TPtr((TFun _) as tf1,_)), ((TFun _) as tf2)
    | ((TFun _) as tf1), (TPtr((TFun _) as tf2, _)) -> 
	isSubtype coercion_ok tf1 tf2

    (* MWH: think harder about this *)
    (* t ( * )() is the black hole for functions *)
    | TFun(tret1,None,_,_), TFun(tret2,targs2,_,_) when isSubtype coercion_ok tret1 tret2 -> true
    | TFun(tret1,targs1,_,_), TFun(tret2,None,_,_) when isSubtype coercion_ok tret1 tret2 -> true
          
    | TFun(tret1,targs1,isva1,_), TFun(tret2,targs2,isva2,_) ->
        let returns_are_subtypes = (isSubtype coercion_ok tret1 tret2) in
        let args_are_subtypes =
          let targs1' = Commoncilutil.argsToTypesList targs1 in 
          let targs2' = Commoncilutil.argsToTypesList targs2 in
          let arity_args1 = List.length targs1' in
          let arity_args2 = List.length targs2' in
          arity_args1 = arity_args2 &&
          (* Compare args contravariantly *)
	  (* MWH: should this be coercion_ok here, or false? *)
          List.for_all (fun (t,t') -> isSubtype coercion_ok t' t) (List.combine targs1' targs2') in
        let va_status_matches = (isva1 = isva2) in
        returns_are_subtypes && args_are_subtypes && va_status_matches 

    | TComp(ci1,_),TComp(ci2,_) -> ci1.cname = ci2.cname
	
    | TPtr(t1,attr1), TPtr(t2,attr2) ->
	isWidthSubtype t1 t2 || isPhysSubtype coercion_ok t1 t2 ||
	isPtrSubtype t1 attr1 t2 attr2

    (* Treat TArray and TPtr as the same *)
    | TArray(t1,_,attr1), TArray(t2,_,attr2)
    | TPtr(t1,attr1), TArray(t2,_,attr2)
    | TArray(t1,_,attr1), TPtr(t2,attr2) -> 
	isPtrSubtype t1 attr1 t2 attr2

    (* Make sure the types are same when unrolled. *)
    | TNamed(t1,_), TNamed(t2,_) -> 
	(* names are the same: always good *)
	t1.tname = t2.tname ||
	(* coercions OK: therefore we can unroll and make sure the targets are the same *)
	(coercion_ok &&
	 (let t1' = Cil.unrollType typ1 in
	 let t2' =  Cil.unrollType typ2 in
	 isSubtype coercion_ok t1' t2')) ||
	 (* coercions are not OK: only can unroll non-updatable types *)
	(let t1' = unrollNonUpdType typ1 in
	let t2' = unrollNonUpdType typ2 in
	(t1' != typ1 || t2' != typ2) &&
	isSubtype coercion_ok t1' t2')

    | TNamed(t1,_), t2 ->
	let t1' = if coercion_ok then
	  Cil.unrollType typ1 else unrollNonUpdType typ1 in
        t1' != typ1 && isSubtype coercion_ok t1' t2

    | t1, TNamed(t2,_) ->
	let t2' = if coercion_ok then
	  Cil.unrollType typ2 else unrollNonUpdType typ2 in
	t2' != typ2 && isSubtype coercion_ok t1 t2'

    | TBuiltin_va_list _, TBuiltin_va_list _ -> true

    | _ -> false

and isPtrSubtype t1 attr1 t2 attr2 =
    (* Allow subtyping under const.  Coercions not OK here, as
       subtyping is creating an alias. *)
    isSubtype false t1 t2 && (isConst attr2 || isSubtype false t2 t1)

and isWidthSubtype t1 t2 =
  match t1,t2 with
    TComp(ci1,_),TComp(ci2,_) ->
      (* same struct: OK *)
      if ci1 == ci2 then true
      else (* width subtyping *)
	(try
	  (* width subtyping only allowed if not updatable *)
	  let notupd_t1 = not (Namedtypes.isUpdatableOrNotSeen t1) in
	  let notupd_t2 = not (Namedtypes.isUpdatableOrNotSeen t2) in
(* 	  let arity_ci1 = List.length ci1.cfields in  *)
	  let arity_ci2 = List.length ci2.cfields in 
	  let prefix_ci1 = Commonutil.list_take arity_ci2 ci1.cfields in 
	  let coercion_ok' = false in
	  notupd_t1 && notupd_t2 &&
	  (List.for_all (fun (fi,fi') -> 
	    let ft1 = fi.ftype in
	    let ft2 = fi'.ftype in
	    (* assume we're under a pointer, so field types invariant *)
	    isSubtype coercion_ok' ft1 ft2 &&
	    isSubtype coercion_ok' ft2 ft1)
	     (List.combine prefix_ci1 ci2.cfields))
	with Failure _ -> false)
  | _,_ -> false

and isPhysSubtype coercion_ok t1 t2 =
  match t1 with
    TComp(ci1,_) ->
      (try
	(let ft1 = (List.hd ci1.cfields).ftype in
	isSubtype coercion_ok ft1 t2 &&
	isSubtype coercion_ok t2 ft1)
      with e -> false)
  | TNamed(_,_) when (Updatecilutil.typeHeight t1) > (Updatecilutil.typeHeight t2) ->
      let t1' = if coercion_ok then
	Cil.unrollType t1 else unrollNonUpdType t1 in
      t1' != t1 && isSubtype coercion_ok t1' t2
  | _ -> false

(* This function is for sanity checking and debug printing *)
and isSubtype coercion_ok t1 t2 = 
  let nspaces n =
    let rec aux n sp =
      if n = 0 then sp
      else aux (n-1) "  "^sp in
    if n = 0 then "==> " else
    aux n "     " in
  incr subcnt;
  let sub = isSubtype' coercion_ok t1 t2 in
  decr subcnt;
  ignore (Updatedebug.log [Updatedebug.BTSubtype] "%s|%a| <= |%a| (coer=%b) ? %b\n" (nspaces !subcnt) d_type t1 d_type t2 coercion_ok sub); 
  let eq = Typeequality.uAreTypesEqualNoAttrs t1 t2 in
  if (eq && not sub) then
    Updatedebug.abort (Updatedebug.bug 
                         "isSubtype: %a = %a but is not a subtype" 
                         d_type t1 d_type t2)
  else sub


module H = Hashtbl


module type SubtypeCmpActions = 
sig
  module M : Monad.Monad

  (* Called when SubtypeCmp performs a standard subtype step *)
  val subtype : Cil.typ -> Cil.typ -> unit M.m
    
  (* Called when a physical subtype step is performed, i.e. a pointer
     to a record is a subtype of a pointer to the records first field
   *)
  val physicalSubtype : Cil.typ -> Cil.typ -> unit M.m
    
  (* Called when one of the types is external *)
  val externalSubtype : Cil.typ -> Cil.typ -> unit M.m
    
  (* Called when trying to determine a subtype between comp pointer type
     and another type where the types neither lead to a valid subtype or to a valid 
     physical subtype *)
  val ptrIncompatible : Cil.typ -> Cil.typ -> string -> unit M.m

  (* Called when C would not accept the subtype *)
  val invalidCSubtype : Cil.typ -> Cil.typ -> string -> unit M.m

  (* called when converting to/from void *)
  val toVoidPtr : Cil.typ -> Cil.typ -> unit M.m
  val fromVoidPtr : Cil.typ -> Cil.typ -> unit M.m

  (* TODO: support width subtyping, particularly for alias tracking *)
end






module SubtypeCmp = 
  functor (SCA: SubtypeCmpActions) ->
    functor (M: Monad.Monad with type mi = SCA.M.mi and type 'a mo = 'a SCA.M.mo) ->
struct
  
  let (>>=) = M.(>>=)
  let return = M.return 

  module MOps = Monad.MonadOps(M)

  let subtypeTraverse (t1 : Cil.typ) (t2 : Cil.typ) : unit M.m =
    let compHash = H.create 7 in 
(*     let id = return () in *)
    (* This should correspond with isSubtype above. *)
    let rec subtypeTraverse' (coercion_ok:bool) t1 t2 =       
      match t1,t2 with 
        | TInt _, _
        | TFloat _, _ 
        | _, TInt _ 
        | _, TFloat _ -> SCA.subtype t1 t2 
	      (* MWH: isSubtype has more stuff here; what to do? *)

	(* MWH: should I only call fromVoidPtr if that's enabled? *)
        | (TPtr ((TVoid _), _)), t2' -> SCA.fromVoidPtr t1 t2
        | t1', (TPtr ((TVoid _), _)) -> SCA.toVoidPtr t1 t2
      
        (* pointers to functions should be treated the same as functions *)
        | TPtr((TFun _ as t1'),_), (TFun _) ->  
            SCA.subtype t1 t2 >>= fun () -> subtypeTraverse' coercion_ok t1' t2
        | (TFun _), TPtr((TFun _ as t2'), _) ->  
            SCA.subtype t1 t2 >>= fun () -> subtypeTraverse' coercion_ok t1 t2'
      
        | TFun(t1',args_opt1,isVA1,attr1), TFun(t2',args_opt2,isVA2,attr2) ->  
            (* make the argument types list to generate coercions for -- pair them contra variantly *)
            let args1 = Cil.argsToList args_opt1 in 
            let args2 = Cil.argsToList args_opt2 in 
            let common_args = min (List.length args1) (List.length args2) in 
            
            let fixed_arg_type_pairs = 
              List.combine 
                (Commoncilutil.argsToTypesList (Some(Commonutil.list_take common_args args2)))
                (Commoncilutil.argsToTypesList (Some(Commonutil.list_take common_args args1)))
            in

	    (* FIX: this is odd how it only compares the N args that are the
	       same, as if they could have different arities.  Seems the only
	       way they could is if one had no arguments at all?  Is that
	       handled by this case?  Would seem so, since it takes the 
	       minimum above.  Better to be more direct? *)
            
	    (* setting coercion_ok to false----correct? *)
            (* Add constraints form the arguments to the function *)
            MOps.iter 
              (fun (typ,typ') ->  subtypeTraverse' false typ typ')
              fixed_arg_type_pairs 
            (* add constraints from the results *)
            >>= fun _ ->  subtypeTraverse' coercion_ok t1' t2'
	    (* note the subtyping of the types themselves *)
            >>= fun () -> SCA.subtype t1 t2         
       
        | TComp(ci1,_), TComp(ci2,_) when (not ci1.cdefined) || (not ci2.cdefined) ->
            (* We only allow conversions between TComps (as opposed to pointers to TComps)
               when one or the other structure is external *)
            SCA.externalSubtype t1 t2

        | TPtr(t1',attr1), TPtr(t2',attr2) ->
	    (* see if one is an external type *)
(*
	    let isExternalComp t =
	      match t with
		TComp(ci,_) -> not ci.cdefined
	      | _ -> false in
	    if isExternalComp t1' || isExternalComp t2' then
              SCA.externalSubtype t1 t2
      
	    (* first try width subtyping *)
	    else 
*)
            if (isWidthSubtype t1' t2') then
	      (* NB. ensures that t1' and t2' are not updateable *)
	      (match t1',t2' with
		TComp(ci1,_),TComp(ci2,_) ->
		  (* when comparing fields, their types must be invariant *)
		  let field_iter (fi1, fi2) = 
		    subtypeTraverse' false (fi1.ftype) (fi2.ftype) >>= fun () ->
		    subtypeTraverse' false (fi1.ftype) (fi2.ftype)
		  in
		  (* if part of a recursive type, don't traverse *)
		  if H.mem compHash ci1.cname then return ()
		  else
 		    SCA.subtype t1 t2 >>= fun () ->
                      (H.add compHash ci1.cname true;
                       if (ci1 == ci2) then return ()
                       else
(* 			 let num_fields_ci1 = List.length ci1.cfields in  *)
			 let num_fields_ci2 = List.length ci2.cfields in
			 (* assert(num_fields_ci1 >= num_fields_ci2) *)
			 let prefix_of_ci2_fields = 
			   Commonutil.list_take num_fields_ci2 ci1.cfields in 
			 MOps.iter field_iter 
			   (List.combine prefix_of_ci2_fields ci2.cfields))
	      | _ -> 
		  Updatedebug.abort (Updatedebug.bug 
				       "subtypeTraverse: %a <= %a should be by width subtypes" 
				       d_type t1 d_type t2))

	    (* then try physical subtyping *)
	    else if (isPhysSubtype coercion_ok t1' t2') then 
              (match t1' with
                TComp(ci1,_) ->
                  (try 
                    (let first_field_type = (List.hd ci1.cfields).ftype in
                    SCA.physicalSubtype t1 t2 >>= fun () ->
                    subtypeTraverse' coercion_ok first_field_type t2' >>= fun () ->
		    subtypeTraverse' coercion_ok first_field_type t2')
		  with Failure _ ->  
		    Updatedebug.abort (Updatedebug.bug 
					 "subtypeTraverse: %a <= %a by physical subtypes?!"
					 d_type t1 d_type t2))
	      | TNamed(ti1,attr1) ->
		  (* MWH: not too sure about this ... *)
		  let t1'' = if coercion_ok then
		    Cil.unrollType t1' else unrollNonUpdType t1' in
		  SCA.physicalSubtype t1 t2 >>= fun () ->
		    subtypeTraverse' coercion_ok t1'' t2'

	      | _ -> Updatedebug.abort (Updatedebug.bug 
					  "subtypeTraverse: %a <= %a by physical subtypes?!"
					  d_type t1 d_type t2))

	    (* now try regular pointer subtyping *)
	    else if isPtrSubtype t1' attr1 t2' attr2 then
              SCA.subtype t1 t2 >>= fun () -> 
		subtypeTraverse' coercion_ok t1' t2' >>= fun () ->
                (if (isConst attr2) then return () 
                else subtypeTraverse' coercion_ok t2' t1')

	    (* failed *)
            else
              SCA.ptrIncompatible t1 t2 "subTypeConstraints: incompatible pointers"

        | TArray(t1',_,attr1), TArray(t2',_,attr2)
        | TPtr(t1',attr1), TArray(t2',_,attr2)
        | TArray(t1',_,attr1), TPtr(t2',attr2) ->
	    (* MWH: in markIfInvalid, this is interchangeableTypes t1' t2'.  If fails
	       will call ptrIncompatible which should setTypeNonUpdateableDeep for both types. *)
            if isPtrSubtype t1' attr1 t2' attr2 then
              SCA.subtype t1 t2 >>= fun () -> 
              subtypeTraverse' coercion_ok t1' t2' >>= fun () ->
                (if (isConst attr2) then return () 
                else subtypeTraverse' coercion_ok t2' t1')
            else
              (* We are not a proper subtype, so notify about this fact *)
              SCA.ptrIncompatible t1 t2 "subTypeConstraints: incompatible pointer types"
      
        | TEnum _, _ -> SCA.subtype t1 t2 
        | _, TEnum _ -> SCA.subtype t1 t2 
       
        | TBuiltin_va_list _, TBuiltin_va_list _ -> SCA.subtype t1 t2 
      
	 (* FIX: treat named types here differently?  *)
	 (* That's what we're doing in isSubtype above? *)

        | t1,t2 when Typeequality.uAreTypesEqualNoAttrs t1 t2 -> SCA.subtype t1 t2  
            
        | _ ->
            SCA.invalidCSubtype t1 t2
              "Warning!! subTypeConstraints: PACIFYING!! could not work out subtype constraints"
    in
      (* MWH: making coercion_ok true by default.  Bad idea? *)
      subtypeTraverse' true t1 t2
end





