(* 
 *
 * Copyright (c) 2004-2009, 
 * 
 *  Michael Hicks       <mwh@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(*

   Suspect Alias Analysis

   This file implements an analysis to detect, at each update point, which structures have 
   live pointers to a field of one of their instances.

   Notation
-----------
   In these notes:
      T  -  a composite type name
      live pointer set  -  a set of composite type names
      L  -  variable ranging over live pointer sets
      E  -  variable ranging over a distinguished set of live pointer sets representing the 
            types invalidated by an Gamma-environment
      LI -  an L-index


   High level outline of analysis
---------------------------------

   INPUT:  A program with abs/con inserted
   OUTPUT: A program with LI annotations in types and a hash table 
           mapping each LI to a set of composite type names suspect at that LI.

   The output is used by the update analysis as described below.  

   This is the overall algorithm:

   1. Traverse the syntax and generate fresh live pointer set (L) effect variables in the 
      attributes of pointer and function types.
   2. Traverse the syntax collecting constraints on L of the form:

        T \in L                  where T is a structure name and L a suspect
        L \subseteq L'           where L and L' are live pointer set effect variables
        L \union L' \subset L''  where epsilon and L, L' and L'' are live pointer sets

      Remember to associate each statement ID with the current L variable and E variable
      representing the context, so that they can be associated with update points later.
   3. Solve these constraints via Banshee.
   4. Extract solution.  For every L-variable determine which composite types are
      in that set by asking is T \in L for every type name constant T. 
   5. For each update point calculate the minimum set by unioning together the 
      L and E associated with the statement ID.  Store a map from statement Id's 
      of update points to these sets.
   6. The solution is the sid -> L*E map and the sid -> min_update_annotation map




   Extra from the theory
------------------------

      - Arrays should be treated the same as structs.
      - If one has a pointer to a struct/array then one can use it as a pointer to the first 
        field in that struct via physical subtyping.  If one does, then that pointer 
        should not become suspect as it is just like taking the address of the first field/element.  
      
      Do we need to pacify types when we loose track of them as we do for update analysis?
      If so, where to we apply pacify?

      Casting to/from void* doesn't matter at the top-level * as the associated namedtype
      set L will be transfered to the void*.  Of course, if we have 
         {..., l:int*^L, ...}* <: void* 
      then L is lost.  Any suspect pointer that goes into a void* may of course come out.
      We therefore flatten types cast to void* in the following sense.  If tau*^L is cast to 
      void*^L', then for all pointer types with suspect set L'' in subtypes of tau*^L we 
      insist that L'' \subseteq L'.  This makes L' a conservative estimate on the suspect 
      pointers one might get out of that void pointer.  When taking data from a void pointer 
      we won't know where it came from, and so we have to give the pointers a suspect rating 
      of L'.

      This principle will also apply when casting from a pointer to any compound type into an 
      'incompatible' one.  For example:

      void*^L  to  int ->L' int     then L <= L'
  
      void*^L  to  {x:int*^L''}*'^L  then L <= L' and L <= L''
  
      {x:int*^L, y:int ->L' int}*^L''  to  {}*^L'''
         then L <= L''' and L' <= L''' and L'' <= L'''
      (and when converting in the reverse direction:
         L''' <= L and L''' <= L' and L''' <= L'')

      The general rule to extract from this is:

      tau <: tau'  then for every locations L' in locations(tau'), 
                   we have Union(locations(tau')) <= L'


  
      Width subtyping needs special attention.  For this, we transfer the annotations 
      on the lost types to the outer pointer, but not the kept ones.  E.g.:
       
      {x: int^L1, y: t1 ->L2 t2}*^L3   to   {x: int^L4}*^L5
         then L1 <= L4 and L2,L3 <= L5  (but NOT L1 <= L5)


TODO: 

- Need to add sub type constraints (module AnalysisSubtypeActions);
not currently handling non-normal subtypes.  Look for TODO there, and
documentation above.  At least, we need to (or can) support physical
subtyping.  We would have to change ptraliastracking to not disallow
it, and then handle it here.  This would reveal the con/abs insertion
bug that Gareth mentioned in e-mail.

- Overall: need to unify our various versions of the subtype relation.
Should get ptraliastracking to use it for sure.  There we use various
ad hoc functions to approximate subtyping.

QUESTIONS/OPTS:

Need to process static initializers?  We don't process them now
as they basically would make something non-updatable permanently, so
the types are ruled out in ptraliastracking.

  *)


open Cil

module H = Hashtbl

open Deltasolver


(*============================================================================*)
(* Local types for this module *)
(*============================================================================*)

(* The mode of l-val analyzing:
     LvalSuspect => under a address of operator
     LvalNotSuspect => all other situations
*)
type lvalMode = 
    LvalSuspect of int   (* parm is the id of a live pointer set *) 
  | LvalNotSuspect




(* transforms a hash of constraints (e.g. a s used internally by the monad) into a set of constraints *)
let lpsConstraintHashToSet (constraintHash : (delta_constraint,unit) H.t) : DeltaConstraintSet.t = 
  let aux d ign acc = 
    makeSureThereSNoConstOnRight d;
    DeltaConstraintSet.add d acc 
  in
  H.fold aux constraintHash DeltaConstraintSet.empty


(*============================================================================*)
(* Monadic state *)
(*============================================================================*)

let noFunction = Cil.emptyFunction "**No FUNCTION**"

module SAMonad = 
struct
  (* input type for suspect analysis functions *)
  type mi  = 
      { 

        (* live pointer constraint hash. Includes constraints for
           environment live pointer sets and regular judgement ones.
           Hash from delta_constraints to unit acting as a set of
           delta_constraints.  *)
        sai_LPC_hash : (delta_constraint, unit) H.t;
        
        (* Mapping of locations to LPSs, one per update *)
        sai_fname_to_LPS_hash : (string, delta * delta) H.t;
 
        (* A map from the id's of delta sets to delta sets *)
        sai_id_to_LPS_hash : (int, delta) H.t;

        (* Mapping from type names to constant live pointers.  Used so
           that we only create a constant once and then use the same
           constant when we need it again *) 
	sai_const_LPS_hash : (string, delta) H.t;

        (* The current live pointer set (the current L).  This is
	   essentially an approximation of the suspect pointers
	   on the callstack at the current program point. *)
        sai_LPS : delta ref;

	(* The locations in current environment within a function (the
	   current locs(\Gamma)) Thus, this is the local variables in
	   a function, and the top_Gamma_locs, below *)
	sai_fun_Gamma_locs : delta ref;

	(* The locations in the top-level environment *)
	sai_top_Gamma_locs : delta ref;

        (* Lval mode: suspect/non-suspect *)
        sai_lval_mode : lvalMode ref;

        (* ---- Debug information ---- *)
        sai_current_fd : Cil.fundec ref;       (* Current function being processed *)
        sai_current_loc: Cil.location ref;      (* Current source location *)
      }


  (* output type for suspect analysis functions *)
  type 'a mo = 'a

  (* Monad type *)
  type 'a m = mi -> 'a mo
    
    
  (* Basic monad operations *)
  let return (x:'a) : 'a m = fun mi -> x
    
  let (>>=) (m: 'a m) (f: 'a -> 'b m) : 'b m = 
    fun mi -> let mo = m mi in f mo mi

  (* The final result: the set of constraints, and a map from update *)
  (* locations to constraint variables, and the current counter for *)
  (* choosing delta variables *)
  type mo_result =
      { 
        sar_LPC_constraints : DeltaConstraintSet.t;
        sar_fname_to_LPS_hash : (string, delta * delta) H.t;
      }

  let run (m: 'a m) : mo_result =
    let mi = 
      {
        sai_LPC_hash = H.create 4099;
        sai_fname_to_LPS_hash = H.create 4099;
        sai_id_to_LPS_hash = H.create 4099;
        sai_const_LPS_hash = H.create 4099;
        sai_LPS = ref {delta_loc = Cil.locUnknown; delta_id = -1; delta_kind = DKVar None};
        sai_fun_Gamma_locs = ref {delta_loc = Cil.locUnknown; delta_id = -1; delta_kind = DKVar None};
        sai_top_Gamma_locs = ref {delta_loc = Cil.locUnknown; delta_id = 1; delta_kind = DKVar None};
        sai_lval_mode = ref LvalNotSuspect;
        sai_current_fd = ref noFunction;
        sai_current_loc = ref Cil.locUnknown;
      }
    in
      (* Add empty lps to hash *)
      H.add mi.sai_id_to_LPS_hash 0 Deltasolver.empty_delta;
      H.add mi.sai_id_to_LPS_hash 1 !(mi.sai_top_Gamma_locs);
      (* go *)
      m mi;
      { sar_LPC_constraints = lpsConstraintHashToSet mi.sai_LPC_hash;
	     sar_fname_to_LPS_hash = mi.sai_fname_to_LPS_hash;
      }


  (* Get/set lval mode *)
  let samGetLvalMode : 'lvalMode m = fun mi -> return !(mi.sai_lval_mode) mi
  let samSetLvalMode (lm:lvalMode) : unit m = fun mi -> 
    mi.sai_lval_mode := lm;
    return () mi

  (* Get/set the current LPS sets *)
  let samGetLPS : delta m = fun mi -> return !(mi.sai_LPS) mi
  let samGetFunLocsG : delta m = fun mi -> return !(mi.sai_fun_Gamma_locs) mi
  let samGetTopLocsG : delta m = fun mi -> return !(mi.sai_top_Gamma_locs) mi
  let samSetLPS (d:delta) = fun mi -> mi.sai_LPS := d; return () mi
  let samSetFunLocsG (d:delta) = fun mi -> mi.sai_fun_Gamma_locs := d; return () mi
  let samSetTopLocsG (d:delta) = fun mi -> mi.sai_top_Gamma_locs := d; return () mi

  (* associate current LPS constraints with an update and
     the start of a function *)
  let samMarkUpdate =
    fun mi ->
      ignore(Updatedebug.log [Updatedebug.BTSuspect] 
	       "update constraint: func(%s) ===> (L=%d,G=%d)\n"
	       !(mi.sai_current_fd).svar.vname !(mi.sai_LPS).delta_id !(mi.sai_fun_Gamma_locs).delta_id);
      H.add mi.sai_fname_to_LPS_hash !(mi.sai_current_fd).svar.vname (!(mi.sai_LPS),!(mi.sai_fun_Gamma_locs));
      return () mi

  (* ---- Function needed to make this a syntax monad ---- *)
 
  (* Set/unset/get the current function (for debugging) *)
  let setCurrentFun (fd: Cil.fundec) : unit m = fun mi -> mi.sai_current_fd := fd; return () mi
  let unsetCurrentFun : unit m = fun mi -> mi.sai_current_fd := noFunction; return () mi
  let getCurrentFun : Cil.fundec m = fun mi -> return !(mi.sai_current_fd) mi

  (* Get/set the current location (for debugging) *)
  let setLoc (loc: Cil.location) : unit m = fun mi -> mi.sai_current_loc := loc; return () mi
  let getLoc : Cil.location m = fun mi -> return !(mi.sai_current_loc) mi



  (* ---- Monadic operations ---- *)

  (* Force type_name to be included in lp1.
     Generate new constant for type name only if type_name has not been seen before. 
  *)
      
  (* Add a subset Live Pointer Constraint (LPC): lp1 <= lp2 *)
  let samAddLPCSubset (lp1 : delta) (lp2 : delta) : unit m = 
    fun mi ->
      H.add mi.sai_LPC_hash (DCSubset (lp1,lp2)) ();
       ignore(Updatedebug.log [Updatedebug.BTSuspect] "Suspect adding delta constraint %d <= %d\n" lp1.delta_id lp2.delta_id);

      return () mi

  let samGetConstLPS (type_name : string) : delta m =
    (* First we have to translate the type name to remove the wrapper, if nec. *)
    (* let type_name = Updateconfig.makeAbsTypeName type_name in *)
    fun mi ->
      let lps_const = 
	try
          H.find mi.sai_const_LPS_hash type_name
	with
          Not_found -> 
             let lps = Deltasolver.freshDelta (DKConst type_name) !(mi.sai_current_loc) in
            H.add mi.sai_const_LPS_hash type_name lps;
            lps in
      return lps_const mi

  let samAddSuspectTypeToLPS (type_name : string) (lp1 : delta) (msg:string): unit m  = 
    samGetConstLPS type_name >>= fun lps2 -> 
    ignore(Updatedebug.log [Updatedebug.BTSuspect] 
	     "adding LPS suspect %d(%s) <= (L=%d) %s\n" lps2.delta_id type_name lp1.delta_id msg);
    samAddLPCSubset lps2 lp1
      
  let samFreshLPSVar : delta m = fun mi -> 
    let freshlp = Deltasolver.freshDelta (DKVar None) !(mi.sai_current_loc) in
      (* Printf.eprintf "samFreshLPSvar: %d\n" !(Deltasolver.deltaCounter); *)
      H.add mi.sai_id_to_LPS_hash freshlp.delta_id freshlp;
      return freshlp mi
   
  (* Look up an LPS given it's id *)
  let samGetLPSByID (key:int) : delta m = fun mi ->
    try
      return (H.find mi.sai_id_to_LPS_hash key) mi
    with 
        Not_found -> 
          Updatedebug.abort (Updatedebug.bug 
                               "(%a) samGetLPS: effect %d not found\n" 
                               d_loc (!(mi.sai_current_loc)) key)

  (* return the empty live pointer set *)
  let samGetEmptyLPS : delta m = fun mi -> return Deltasolver.empty_delta mi

end



(* Create the generic monad operations *)
module SAMOps = Monad.MonadOps(SAMonad)




(* The rest of the file uses the SAMonad, so make a shorthand name for it and bind 
   its basic operations locally 
*)
module SAM = SAMonad
let (>>=) = SAMonad.(>>=)
let return = SAMonad.return






(*============================================================================*)
(*
  Fresh effect insertion

  The following code inserts fresh live pointer set variables into
  a syntax tree.

*)
(*============================================================================*)

module EffectCreator =
struct
  (* The monad that this effect inserter works over *)
  module M = SAM
  module MOps = Monad.MonadOps(M)

  let (>>=) = SAM.(>>=)
  let return = SAM.return

  (* helper function *)
  let attrIsEffect a = match a with
      Attr(aname, _) when aname = Updateconfig.saVarAttrName -> true
    | _ -> false

  let getEffects = List.filter attrIsEffect
  let containsEffects = List.exists attrIsEffect

  (* Return the LPS in the given attribute list *)
  let extractLPS t al : delta M.m = 
    MOps.lift getEffects (return al) >>= fun al' ->
      match al' with 
          [Attr(aname, [AInt delta_id])] -> M.samGetLPSByID delta_id 

        | _ -> 
            M.getLoc >>= fun loc ->
              raise (Failure (Pretty.sprint 80 (Pretty.dprintf "(%a) extractLPS: LPS not found in type! %s\n" 
                                   d_loc loc (Updatedebug.ppType t))))

  let extractLPSfromType t : delta M.m =
    extractLPS t (Cil.typeAttrs t)
(*    let tsig = Cil.typeSig t in
    extractLPS t (Cil.typeSigAttrs tsig) *)

  (* add a suspect alias attribute to a an attribute list using the id of the 
     given lps as the contents *)
  let addLPStoAttr (al:Cil.attributes) (tag:string) (lps:delta) : Cil.attributes M.m = 
    return (Cil.addAttribute (Attr(tag, [AInt lps.delta_id])) al)


  (* This function acts non-inductively, inserting a fresh variable in the 
     outer type constructor of: 
       - TInt
       - TFun
       - TPtr
  *)
  let insertEffectInOuterConstructor (typ: Cil.typ) : Cil.typ M.m = 
    let addAttr (al:Cil.attributes) : Cil.attributes M.m = 
      SAM.samFreshLPSVar >>= fun lps -> addLPStoAttr al Updateconfig.saVarAttrName lps
    in
    let al = Cil.typeAttrs typ in
      (* Only insert types if we haven't before *)
      if containsEffects al
      then
        return typ
      else
        (*ignore (Updatedebug.log [Updatedebug.BTAlways] "Inserting effect\n"); *)
        begin match typ with 
            TInt (ik,al)             -> addAttr al >>= fun al' -> return (TInt (ik, al'))
          | TFun(tret,targs,isva,al) -> addAttr al >>= fun al' -> return (TFun(tret,targs,isva,al'))
          | TPtr(t,al)               -> addAttr al >>= fun al' -> return (TPtr(t, al'))
          | TArray(t,e_opt,al)       -> addAttr al >>= fun al' -> return (TArray(t,e_opt,al'))
          | _                        -> return typ 
        end
    
end



(* Create fresh effect inserter *)
module SAAGenFreshEffects = 
  Effectinserter.EffectInserter(EffectCreator)(SAM)








(*============================================================================*)
(* 
  Main algorithm

  This traverses the syntax collecting constraints as given in the suspect 
  alias analysis. 

*)
(*============================================================================*)



open Ciltransform
module TypFold = Ciltransform.TypFold(SAM)



(*
  Given a type typ and a LPS lps, add constraints to force the lps to contain  
  all of the live pointer sets on all pointer annotations in typ (and
  possibly the function ones too). 
*)

let collectLPSets (getFunEffs:bool) (lps:delta) (typ: Cil.typ) : unit SAM.m =
  let collectLPSetsTyp (ttyp: (unit,Cil.exp,unit,unit) ttyp) : unit SAM.m =
    let includeEffectsInAttr (al: Cil.attributes) : unit SAM.m = 
      SAMOps.lift EffectCreator.getEffects (return al) >>= fun delta_id_list ->
        match delta_id_list with 
            (* There should be exactly one effect in the type *)
            [Attr(aname, [AInt delta_id])] -> 
              SAM.samGetLPSByID delta_id >>= fun lps' ->
		ignore(Updatedebug.log [Updatedebug.BTSuspect] 
			 "collectLPsets: (typ)%d <= (set)%d\n" lps'.delta_id lps.delta_id);
              SAM.samAddLPCSubset lps' lps
              
          | _ -> SAM.getLoc >>= fun loc -> Updatedebug.abort (Updatedebug.bug 
                                                                "(%a) collectLPSets: no effect in int\n" d_loc loc)
    in
    match ttyp with
	TTInt(_, al)
      | TTPtr(_, al)
      | TTArray(_, _, al) -> includeEffectsInAttr al 
      | TTFun(_, _, _, al) -> if getFunEffs then includeEffectsInAttr al else return ()
      | _                 -> return ()
  in            
  let avoidFuns (typ:Cil.typ) : unit option =
    match typ with
      TFun(_,_,_,_) -> 
	(* ignore (Updatedebug.log [Updatedebug.BTAlways] "Ignoring function type %a\n" d_type typ);*)
	Some ()
    | _ -> None
  in
  let collectTypFoldFuns = {TypFold.iterDefaultTypFoldFuns 
                            with TypFold.tf_typ = collectLPSetsTyp;
			         TypFold.tf_typ_alt = avoidFuns;} 
  in
  TypFold.typFold collectTypFoldFuns typ


module AnalysisSubtypeActions = 
struct
  module M = SAMonad
  module MOps = Monad.MonadOps(SAMonad)

  (* private function: extract live pointer sets from attributes and add 1<= 2 constraint *)
  let addConstraint (t1:Cil.typ) (al1:Cil.attributes) (t2:Cil.typ) (al2:Cil.attributes) : unit M.m =       
    EffectCreator.extractLPS t1 al1 >>= fun lps1 -> 
    EffectCreator.extractLPS t2 al2 >>= fun lps2 -> 
      ignore(Updatedebug.log [Updatedebug.BTSuspect] 
	     "subtype constraint: %a <= %a\n### %d <= %d\n" 
	       Cil.d_type t1 Cil.d_type t2 lps1.delta_id lps2.delta_id);
      SAM.samAddLPCSubset lps1 lps2


  (* Called when SubtypeCmp performs a standard subtype step *)
  let subtype (t1: Cil.typ) (t2: Cil.typ) : unit M.m =
    match t1,t2 with 
      | TFun(_,_,_,attr1), TFun(_,_,_,attr2) 
      | TPtr(_,attr1), TPtr(_,attr2) ->
(* 	  ignore(Updatedebug.log [Updatedebug.BTSuspect]  *)
(* 		   "subtype: %a <= %a\n"  *)
(* 		   Cil.d_type t1 Cil.d_type t2); *)
	  (* MWH: this was backwards (now fixed)---why? *)
          addConstraint t1 attr1 t2 attr2
      | _ -> return ()


  (* Called when a physical subtype step is performed, i.e. a pointer
     to a record is a subtype of a pointer to the records first field.

     This is like taking the address of that field, so make the 
     pointer t2 suspect with respect to the (comp) type of t2.
   *)
  let physicalSubtype (t1: Cil.typ) (t2: Cil.typ) : unit M.m =
    ignore(Updatedebug.log [Updatedebug.BTSuspect] 
	     "physical subtype: %a <= %a\n" Cil.d_type t1 Cil.d_type t2);
    match t1,t2 with 
      TPtr((TComp(ci,_) as tcomp),_), TPtr(_,attr2) ->
        EffectCreator.extractLPS t2 attr2 >>= fun lps ->
          SAMonad.samAddSuspectTypeToLPS (Updatecilutil.getAbstractTypeName tcomp) lps "(physical subtype)"

    | TPtr((TNamed(ti,_) as tn),_), TPtr(_,attr2) ->
        EffectCreator.extractLPS t2 attr2 >>= fun lps ->
          SAMonad.samAddSuspectTypeToLPS (Updatecilutil.getAbstractTypeName tn) lps "(physical subtype)"

    | _ ->
        M.getLoc >>= fun loc ->
          MOps.lift (fun () -> ignore (Updatedebug.log [Updatedebug.BTAlways] 
                                         "(%a) Analysis:SubtypeConstraints: physical subtype called for non-updatable type: %s and %s\n" 
                                         d_loc loc (Updatedebug.ppType t1) (Updatedebug.ppType t2)
                                      )
                    ) (return ())

    
  (* Called when one of the types is external *)
  let externalSubtype (t1: Cil.typ) (t2: Cil.typ) : unit M.m =
    (* TODO: restrict effects collectLPsets things above to do this. *)
    M.getLoc >>= fun loc ->
    MOps.lift (fun () -> ignore (Updatedebug.log [Updatedebug.BTAlways] 
                                "(%a) Analysis:SubtypeConstraints: One type is external %s and %s\n" 
                                d_loc loc (Updatedebug.ppType t1) (Updatedebug.ppType t2)
                             )
           ) (return ())

  (* Called when trying to determine a subtype between comp pointer type
     and another type where the types neither lead to a valid subtype or to a valid 
     physical subtype *)
  let ptrIncompatible (t1: Cil.typ) (t2: Cil.typ) (msg: string) : unit M.m =
    (* TODO: restrict effects *)
    M.getLoc >>= fun loc ->
    MOps.lift (fun () -> ignore (Updatedebug.log [Updatedebug.BTAlways] 
                                "(%a) Analysis:SubtypeConstraints: Incompatible pointers to records %s and  %s\n" 
                                d_loc loc (Updatedebug.ppType t1) (Updatedebug.ppType t2)
                             )
           ) (return ())

  
  (* Called when C would not accept the subtype *)
  let invalidCSubtype (t1: Cil.typ) (t2: Cil.typ) (msg: string) : unit M.m =
    (* TODO: restrict effects *)
    M.getLoc >>= fun loc ->
    MOps.lift (fun () -> ignore (Updatedebug.log [Updatedebug.BTAlways] 
                                "(%a) Analysis:SubtypeConstraints: invalid C subtype %s and %s\n" 
                                d_loc loc (Updatedebug.ppType t1) (Updatedebug.ppType t2)
                             )
           ) (return ())

  let toVoidPtr t1 t2 =
    (* TODO: restrict effects *)
    (* Note: unrolling here is inherently non-safe; there should have been a CON instead *)
    match (Cil.unrollType t1),(Cil.unrollType t2) with 
        TPtr(_,attr1), TPtr(_,attr2) -> addConstraint t2 attr2 t1 attr1
      | TInt(_,attr1), TPtr(_,attr2) -> addConstraint t2 attr2 t1 attr1
      | _ ->
          M.getLoc >>= fun loc ->
            Updatedebug.abort (Updatedebug.bug 
              "(%a) SuspectAnalysis:toVoidPtr: Non-pointer encountered:\n %s and\n %s\n" 
              d_loc loc (Updatedebug.ppType t1) (Updatedebug.ppType t2))

  let fromVoidPtr t1 t2 = 
    (* Note: unrolling here is inherently non-safe; there should have been a CON instead *)
    match (Cil.unrollType t1),(Cil.unrollType t2) with 
        TPtr(_,attr1), TPtr(_,attr2) -> addConstraint t2 attr2 t1 attr1
      | TPtr(_,attr1), TInt(_,attr2) -> addConstraint t2 attr2 t1 attr1
      | _ ->
          M.getLoc >>= fun loc ->
            Updatedebug.abort (Updatedebug.bug 
              "(%a) SuspectAnalysis:toVoidPtr: Non-pointer encountered:\n %s and\n %s\n" 
              d_loc loc (Updatedebug.ppType t1) (Updatedebug.ppType t2))

end


module AnalysisSubtypeConstraints = Subtype.SubtypeCmp(AnalysisSubtypeActions)(SAM)

let subTypeConstraints t1 t2 err = 
  try
    ignore(Updatedebug.log [Updatedebug.BTSuspect] 
	     "Doing subtype constraint:\n  %s\n  %a <= %a\n"
	     (Pretty.sprint 80 err) Cil.d_type t1 Cil.d_type t2);
    AnalysisSubtypeConstraints.subtypeTraverse t1 t2
  with Failure m ->
    (ignore(Updatedebug.abort (Updatedebug.bug "|%s|: %s\n" (Pretty.sprint 80 err) m));
     return ())



(* let getAttr (typ:Cil.typ) : Cil.attributes SAM.m = *)
(*   let al' =  *)
(*     match typ with    *)
(*         TVoid al  *)
(*       | TInt (_,al)  *)
(*       | TFloat (_,al) *)
(*       | TEnum (_,al)  *)
(*       | TBuiltin_va_list al  *)
(*       | TPtr (_, al)  *)
(*       | TArray(_, _, al)  *)
(*       | TFun(_, _, _, al)  *)
(*       | TNamed(_, al)  *)
(*       | TComp(_,al) -> al *)
(*   in  *)
(*     return al' *)
        


open Ciltransform

module SAReturnTypes = 
struct
  type typ       = unit
  type exp       = Cil.typ
  type lval      = Cil.typ
  type lhost     = Cil.typ
  type offset    = unit
  type instr     = unit
  type stmt      = unit
  type stmtkind  = unit
  type block     = unit
  type compinfo  = unit
  type fieldinfo = unit
  type init      = unit
  type initinfo  = unit
  type typeinfo  = unit
  type varinfo   = Cil.typ
  type fundec    = unit
  type global    = unit
end


(* Traversal of syntax *)
module SATransform = Ciltransform.MTransform (SAM) (SAReturnTypes)


module SAFoldFuns = 
struct
  (* Uninteresting map functions *)
  let ftyp (typ:typ) (ttyp: (unit,Cil.typ,typeinfo,compinfo) ttyp) : unit SAM.m = return ()
  let fstmt (stmt:stmt) (tstmt: unit tstmt) : unit SAM.m = return ()
  let fblock (b:block) (tblock: unit tblock) : unit SAM.m  = return ()
  let fcompinfo (ci:compinfo) (tcompinfo: unit tcompinfo) : unit SAM.m = return ()
  let ffieldinfo (fi:fieldinfo) (tfieldinfo: unit tfieldinfo) : unit SAM.m = return ()
  let finit (init:init) (tinit: (unit,Cil.typ,unit) tinit) : unit SAM.m = return ()
  let finitinfo (ii:initinfo) (tinitinfo: (unit tinitinfo)) : unit SAM.m = return ()
  let ftypeinfo (ti:typeinfo) (ttypeinfo: unit ttypeinfo) : unit SAM.m =  return ()
  let foffset (offset:offset) (toffset: (Cil.typ,unit) toffset) (lval:Cil.lval) : unit SAM.m = return ()

  (* The cases mentioned here have special processing - i.e. done in an alt handler.
     This is because they need extra things going *around* the recursive calls *)
  let should_process (pa:Ciltransform.process_arg) = 
    match pa with
      | PAExp (AddrOf _) -> false
      | PAExp (Lval _)   -> false
      | PAInstr (Call(_, Lval(Var vi, NoOffset), _, _)) when  Commoncilutil.isBuiltin vi.vname -> false 
      | PAOffset _       -> false   (* This gets processed by a manual call in lval *)
      | PAGlobal (GFun (_,_)) -> false
      | _ -> true (* default is to process *)

  let fvarinfo (vi: varinfo) (tvarinfo: unit tvarinfo) : Cil.typ SAM.m =
    return vi.vtype
  
  let fglobal (global:global) (tglobal: (unit,unit,unit,Cil.typ,unit,unit) tglobal) : unit SAM.m =
    let extend_locsG typ =
      SAM.samGetTopLocsG >>= fun locsTopG ->
      (* MWH: just have one top-level epsilon, rather than a lexical one *)
      (* SAM.samFreshLPSVar >>= fun newLocsTopG -> *)
      (* SAM.samSetTopLocsG newLocsTopG >>= fun _ -> *)
      (* SAM.samAddLPCSubset locsTopG newLocsTopG >>= fun _ -> *)
      (* add effects from all local vars and parameters *)
      collectLPSets false locsTopG typ in
    match global,tglobal with
      (GVarDecl (vi,_)),(TGVarDecl (typ,loc))
    | (GVar (vi,_,_)),(TGVar (typ,_,loc)) -> 
	ignore (Updatedebug.log [Updatedebug.BTSuspect] "BEGIN Extending topG: var %s:\n" vi.vname);
	let res = extend_locsG typ in
	ignore (Updatedebug.log [Updatedebug.BTSuspect] "END   Extending topG: var %s:\n" vi.vname);
	res
    | _ -> return ()

  let ffundec (fd:fundec) (tfd: (Cil.typ,unit,unit) tfundec) : unit SAM.m =
    (* add effects from all local vars and parameters *)
    SAM.samGetFunLocsG >>= fun currLocsG ->
      ignore (Updatedebug.log [Updatedebug.BTSuspect] "BEGIN Extending funG(%s)\n" fd.svar.vname);
    SAMOps.iter (collectLPSets false currLocsG) tfd.tsformals >>= fun _ ->
      let res = SAMOps.iter (collectLPSets false currLocsG) tfd.tslocals in
      ignore (Updatedebug.log [Updatedebug.BTSuspect] "END   Extending funG(%s)\n" fd.svar.vname);
      res

  let finstr (instr:instr) (tinstr: (Cil.typ,Cil.typ) tinstr) : unit SAM.m = 
    let getOpt (opt: 'a option) (none_val: 'a) = 
      match opt with None -> none_val | Some x -> x in 
    match (instr,tinstr) with 
      (_,TSet (lval_t,exp_t,loc)) ->
        subTypeConstraints exp_t lval_t (Pretty.dprintf "%a" d_instr instr)
            
    | (Call(_, Lval(Var vi, NoOffset), _, _), TCall (lval_result_t_opt, exp_func_t, exp_arg_t_list, loc)) ->
	if Commoncilutil.isBuiltin vi.vname ||
	   ((Updateconfig.isConFunc vi.vname) && (Namedtypes.isConAbsFuncName vi.vname)) || 
	   ((Updateconfig.isAbsFunc vi.vname) && (Namedtypes.isConAbsFuncName vi.vname)) ||
	   Updateconfig.isUpdateFunc vi.vname then
	  (* TODO: *)
	  (* Mallocpragma.isMallocFunc mallocFunHash vi.vname -> *)
	  return ()
	else
	  begin match exp_func_t with 
	  (* First: generate subtyping constraints between the *)
          (* parameters of the function and the arguments provided to *)
	  (* this call *)
          TPtr(TFun(tret,targs_opt,isva,al),_) | TFun(tret,targs_opt,isva,al) ->
            let numFixedArgs = List.length (getOpt targs_opt []) in 
            let ( fixedArgs_t, varArgs_t )   = 
              Commonutil.list_split numFixedArgs exp_arg_t_list in 
	    (* TODO: deal with variable arguments here?  types are in varArgs_t *)
            let fixedArgPairs = 
	      List.combine fixedArgs_t 
		(List.map (fun (s,t,al)-> t)
		   (getOpt targs_opt [])) in
            (SAMOps.iter (fun (t1,t2) -> 
	      subTypeConstraints t1 t2 (Pretty.dprintf "%a" d_instr instr)) fixedArgPairs >>= fun () ->
	      begin match lval_result_t_opt with 
                None        -> return ()
              | Some lval_t -> subTypeConstraints tret lval_t (Pretty.dprintf "%a" d_instr instr)
              end) >>= fun () ->
	    (* Second: create constraints on the function's LPS *)
	    EffectCreator.extractLPS exp_func_t al >>= fun lps ->
            SAM.samGetFunLocsG >>= fun currLocsG ->
            SAM.samGetLPS >>= fun currLPS ->
              ignore(Updatedebug.log [Updatedebug.BTSuspect] 
                 "BEGIN call to %s in %s:%d where (\\hatL(fun)=%d,L=%d,\\funG=%d)\n"
		       vi.vname loc.file loc.line lps.delta_id currLPS.delta_id currLocsG.delta_id);
            SAM.samAddLPCSubset currLocsG lps >>= fun _ ->
              let res = SAM.samAddLPCSubset currLPS lps in
              ignore(Updatedebug.log [Updatedebug.BTSuspect] 
                 "END   call to %s in %s:%d\n" vi.vname loc.file loc.line);
	    res
		  
        | _ -> 
            SAM.getLoc >>= fun loc -> Updatedebug.abort (Updatedebug.bug 
                   "(%a) SuspectAnalysis:finstr: in call, function expression does not have  not a function type\n" 
                   d_loc loc)
	end

      | _ -> return ()



  let fstmtkind (skind:stmtkind) (tstmtkind: (Cil.typ,unit,unit,unit) tstmtkind) : unit SAM.m = 
    match skind,tstmtkind with 
      | Return (Some e, _),TReturn (Some e_t, loc) ->
          (* Add constraint between type e_t and return type of function *)
          SAM.getCurrentFun >>= fun fd ->
          SAMOps.lift Updatecilutil.getReturnType (return fd) >>=
          fun rt -> subTypeConstraints e_t rt (Pretty.dprintf "return %a" d_exp e)
          
      | _ -> return ()


  let fexp (exp:exp) (texp: (unit,Cil.typ,Cil.typ) texp) : Cil.typ SAM.m =
    let addEmptyLPSToAttr (al:Cil.attributes) : Cil.attributes SAM.m =
      EffectCreator.addLPStoAttr al Updateconfig.saVarAttrName Deltasolver.empty_delta in
    let uintConst = addEmptyLPSToAttr [] >>= fun al -> return (TInt(IUInt,al))
    in
      match exp,texp with 
          (* ---- Routine cases ---- *)
          _, TConst c       -> 
            (* The only constant that could be suspect is int, but as it is a 
               constant, we assume it points nowhere: we make the assumption 
               that people don't hard-wire addresses *)
            SAMOps.lift Updatecilutil.properTypeOf (return exp) >>= fun exp_t ->
              begin match exp_t with 
                  TInt(ik,al) ->
                    addEmptyLPSToAttr al >>= fun al' -> return (TInt(ik,al'))
                | TPtr( TInt( ik, al_int ), al_ptr ) ->
                    addEmptyLPSToAttr al_int >>= fun al_int' -> 
                    addEmptyLPSToAttr al_ptr >>= fun al_ptr' -> 
                      return (TPtr(TInt(ik,al_int'), al_ptr'))

                | _ -> return exp_t
              end
                
        | _, TLval lval_t     -> return lval_t
 
        | _, TSizeOf _  
        | _, TSizeOfE _        
        | _, TSizeOfStr _
        | _, TAlignOf _  
        | _, TAlignOfE _    -> uintConst

        | UnOp (unop, e, typ), _  -> return typ 


        (* ---- Interesting cases ---- *)
        | _, TStartOf lval_t -> 
            begin match (Cil.unrollType lval_t) with 
                TArray(t,e_opt,al) -> return (TPtr(t,al))
              | _            -> 
                  SAM.getLoc >>= fun loc ->
                    Updatedebug.abort (Updatedebug.bug 
                     "(%a) SuspectAliasAnalysis: StartOf lval which is not an array (%a)!"
                     d_loc loc d_type lval_t) 
            end
 
        | AddrOf _, _ -> raise (Ciltransform.CILTransformEx "Bang!") (*  processed in alt handler *)
            
        | BinOp(binop,e1,e2,typ), TBinOp(_,e1_t,e2_t,ttyp) ->
            subTypeConstraints e1_t typ (Pretty.dprintf "%a" d_exp exp) >>= fun () ->
	    (* MWH: do we need both? *)
            subTypeConstraints e2_t typ (Pretty.dprintf "%a" d_exp exp) >>= fun () ->
            return typ
      
        | CastE(typ,e), TCastE(_,e_t) ->
           subTypeConstraints e_t typ (Pretty.dprintf "%a" d_exp exp) >>= fun () ->
            return typ

        | _ -> SAM.getLoc >>= fun loc ->
                    Updatedebug.abort (Updatedebug.bug 
                     "(%a) SuspectAliasAnalysis: fexp with unexpexted arg\n%a\n!"
                     d_loc loc d_exp exp) 

  (* Analogous to Ptraliastracking invalidation of types under & *)
  let rec processOffset (lval_t:Cil.typ) (offset:Cil.offset) : Cil.typ SAM.m = 

    (* If we are suspect at lps, then add the constraint abs_typ \in lps *)
    let addTypeIfSuspect (abs_typ: Cil.typ) : unit SAM.m = 
      SAM.samGetLvalMode >>= fun lval_mode ->
      match lval_mode with 
          LvalNotSuspect ->
            return ()
        | LvalSuspect delta_id ->
            SAM.samGetLPSByID delta_id >>= fun lps ->
            SAM.samAddSuspectTypeToLPS (Updatecilutil.getAbstractTypeName abs_typ) lps "(under &)"
    in

    (* Walk down type and offset collecting constraints if we are suspect.
       When we hit NoOffset, the type is lval_t *)
    match offset, lval_t with
        NoOffset, t  -> return t
 
      | Field(fi,offset'), TNamed(ti,al)  ->
          addTypeIfSuspect lval_t >>= fun () ->
          processOffset ti.ttype offset

      | Field(fi,offset'), TComp(ci,al)  ->
          addTypeIfSuspect lval_t >>= fun () ->
          processOffset fi.ftype offset' (* This assumes that fi is a field of ci *)

      | Index(exp,offset'), TNamed(ti,al) ->
          addTypeIfSuspect lval_t >>= fun () ->
          processOffset ti.ttype offset

      | Index(exp,offset'), TArray(t,_,al) 
      | Index(exp,offset'), TPtr(t,al)   ->
          processOffset t offset'

      | _,_ -> SAM.getLoc >>= fun loc ->
          Updatedebug.abort (Updatedebug.bug 
            "(%a) SuspectAliasAnalysis:processOffset: type doesn't match offset!\n"
            d_loc loc) 


  let flval ((lhost,offset):lval) ((tlhost_t,toffset): (Cil.typ,unit) tlval) : Cil.typ SAM.m = 
    processOffset tlhost_t offset

  let flhost (lhost:lhost) (tlhost: (Cil.typ,Cil.typ) tlhost) : Cil.typ SAM.m =
    let aux t =
      match t with 
	TPtr(t,_)
      | TArray(t,_,_) -> t
      | _ ->  raise (Ciltransform.CILTransformEx "SAFoldFuns: flhost with unexpected arg (2)")
    in
    match lhost,tlhost with 
        Var vi, TVar tvi  -> return vi.vtype
      | _, TMem t         -> return (aux (Cil.unrollType t))
(* raise (Ciltransform.CILTransformEx ("SAFoldFuns: flhost arg"^(Updatedebug.ppType t)) ) *)
      | _ ->  raise (Ciltransform.CILTransformEx "SAFoldFuns: flhost with unexpected arg")

end


(* build a custom fold based on the functions given above *)
module SAFold = SATransform.MFold (SAFoldFuns)



(* Alternative fold handlers for cases identified in SAFoldFuns.should_process 
   as false *)
let rec alt_doexp (exp:Cil.exp) : Cil.typ SAM.m =
  (* We only expect an addrof here *)
  match exp with 
      AddrOf lval -> 
        (* Generate a fresh effect *)
        SAM.samFreshLPSVar >>= fun lps ->
	SAM.samGetLvalMode >>= fun oldmode ->
        SAM.samSetLvalMode (LvalSuspect lps.delta_id) >>= fun () ->
        (* Process lval as prescribed by the fold *)
        SAFold.lvalFoldWithAlt alt_handlers lval >>= fun lval_t ->
        EffectCreator.addLPStoAttr [] Updateconfig.saVarAttrName lps >>= fun al ->
        SAM.samSetLvalMode oldmode >>= fun () ->
        return (TPtr(lval_t,al))
        
    | Lval lval ->
        (* not a suspect lval *)
	SAM.samGetLvalMode >>= fun oldmode ->
        SAM.samSetLvalMode LvalNotSuspect >>= fun () ->
        SAFold.lvalFoldWithAlt alt_handlers lval >>= fun lval_t ->
        SAM.samSetLvalMode oldmode >>= fun () ->
        return lval_t

    | _ -> raise (Ciltransform.CILTransformEx "SAFoldFuns: doexp with unexpected arg")

and alt_dooffset (off:Cil.offset) (lval:Cil.lval) : unit SAM.m = return ()
and alt_doglobal (g:Cil.global) : unit SAM.m = 
  match g with
    GFun (fd,loc) ->
      if not (Namedtypes.isConAbsFuncName fd.svar.vname) then
	begin
	  (* generate a dummy constraint between L and empty delta *)
	  SAM.samGetEmptyLPS >>= fun emptyD ->
	  EffectCreator.extractLPSfromType fd.svar.vtype >>= fun lps ->
	  SAM.samAddLPCSubset emptyD lps >>= fun _ ->
          SAM.samSetLPS lps >>= fun _ ->
          (* fresh variable for current suspect pointers in env *)
          SAM.samFreshLPSVar >>= fun currLocsG ->
          SAM.samSetFunLocsG currLocsG >>= fun _ ->
          (* make this include the current top-level locsG *)
          SAM.samGetTopLocsG >>= fun locsTopG ->
          SAM.samAddLPCSubset locsTopG currLocsG >>= fun _ ->
	  (* We mark the current L and locsG for this function *)
	  (* for later inclusion in the analysis constraints *)
          SAM.setCurrentFun fd >>= fun () ->
	  SAM.samMarkUpdate >>= fun () ->
	  (* now process the function body *)
          SAFold.fundecFoldWithAlt alt_handlers fd >>= fun fd_r ->
          SAM.unsetCurrentFun >>= fun () ->
	  return ()
	end
      else return ()

  | _ -> raise (Ciltransform.CILTransformEx "SAFoldFuns: doglobal with unexpected arg")

and alt_doinstr (g:Cil.instr) : unit SAM.m = return ()

and alt_handlers = 
  {
    SATransform.default_alt_handlers with
    SATransform.should_process = SAFoldFuns.should_process; 
    SATransform.doexp = alt_doexp;
    SATransform.dooffset = alt_dooffset;
    SATransform.doglobal = alt_doglobal;
    SATransform.doinstr = alt_doinstr;
  }



(*============================================================================*)
(* External call points *)
(*============================================================================*)


let doSuspectAliasAnalysis (f: Cil.file) : SAM.mo_result = 
(*   let crtGlobal = ref 0 in *)
  let m =
    let doGlobal g = (* SAMOps.lift print_string (return ("Suspect at global " ^ (incr crtGlobal; string_of_int (!crtGlobal)) ^ "\n")) >>= fun () -> *)
                     SAFold.globalFoldWithAlt alt_handlers g in

    (* Populate the syntax with fresh live pointer sets *)
    SAAGenFreshEffects.genFreshEffectsGlobal f >>= fun () ->
    ((* Updatedebug.d_file f; *)
    (* do the analysis *)
    SAMOps.iter doGlobal f.globals )
  
  in
  let res = SAM.run m in
  Stats.time "Stats : removeEffectAttrVisitor - doDSU" Cil.visitCilFile (new Updatecilutil.removeEffectAttrVisitor(Updateconfig.saVarAttrName)) f;
  if !Updatedebug.bugTypeIsEnabled Updatedebug.BTSuspect then
    Deltasolver.printConstraintSet res.SAM.sar_LPC_constraints;
  res

