(* 
 *
 * Copyright (c) 2004-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(*        Type Indirection      *)


(*  

This module implements the infrastructure for
manipulating abstract types metadata. Essentially, for all updatable types
we need to keep:

- an integer representing the latest version 
- a function pointer pointing to the latest type transformer 

Whenever a program is compiled, for each updatable type T we generate
- a version variable initally set to 0
- an empty TT


Whenever a patch is compiled (see patchcomp.ml), for all types T that have changed,
code will be generated to:

- increment the latest type version
- set the latest TT to the newest version


Updateargs.args.Updateargs.arg_currentMode tells us whether we're running in program or patch mode
so we can act accordingly.

*)

open Cil

module H = Hashtbl


module IndirectTypeInfo = 
(* Straightforward abstraction for storing/retrieving the metadata associated with a type *)
struct

   type indirect_type_info = 
   { 
      iti_latest_tt  : Cil.varinfo;    (* Variable that points to the latest TT *)
      iti_version    : Cil.varinfo;    (* Variable that holds the latest version number *)
      iti_lock       : Cil.varinfo;    (* Variable that holds the per-type mutex *)
   }
   
   (* map between the abstract type name and it associated version and TT variables *)
   let latestTTHash : (string, indirect_type_info) H.t = H.create 4099
   
   let addTypeInfo (tname:string) (iti : indirect_type_info) = 
      H.add latestTTHash tname iti

   let getTypeInfo (tname:string) : indirect_type_info =
      H.find latestTTHash tname

end


(* returns the type transformer varinfo associated with this name *)
let getLatestTypeTransformer tname = (IndirectTypeInfo.getTypeInfo tname).IndirectTypeInfo.iti_latest_tt 

(* returns the version varinfo associated with this name *)
let getVersionVar tname = (IndirectTypeInfo.getTypeInfo tname).IndirectTypeInfo.iti_version

(* returns the lock varinfo associated with this name *)
let getLockVar tname = (IndirectTypeInfo.getTypeInfo tname).IndirectTypeInfo.iti_lock



   
 

(* Create the type of the variable that will point to the latest type transformer.
   For a certain type T, the latest TT has the type:

   void ( * )(T * abs)

   The TTs coerce the data ("abs" here) in place.
*)
let makeTTLatestVarType tname tabs = 
   let funT = TFun( TVoid [],  
                    Some ([("abs", TPtr(tabs,[]), [])]),
                    false,
                    [] ) in 
   TPtr(funT,[])



(* A fold over globals: for each updatable type, generate the latest version/latest TT and add them right after the
   type definition *)
let genTypeInfastructureGlobalFold (vd : Versiondata.version_data) (vi_emptytt : Cil.varinfo)  (g:Cil.global) (acc:Cil.global list) : Cil.global list = 
   (* int __latest_type_version_ = 0; *)
   let genLatestVersionVar tname = 
      let latest_ver_var_name = Updateconfig.makeLatestTypeVersionVarName tname in
      let vi = Cil.makeGlobalVar latest_ver_var_name Updateconfig.wDefaultVersionFieldType in 
      GVar(vi, {init = Some (SingleInit Cil.zero)}, Cil.locUnknown)
   in
   

   (* void ( * tname )(t * ) = empty_tt; *)
   let genLatestTTVar tname t = 
(*       let tt_name = Updateconfig.makeTTName tname 0 in  *)
      let latest_tt_var_name = Updateconfig.makeLatestTTName tname in
      let vi = Cil.makeGlobalVar latest_tt_var_name (makeTTLatestVarType tname t) in 

      (* we need a cast from 
           void( * empty_tt )(void * ) 
         to 
           void( * tname)(t * ) 
      *)
      let cast_to_tt_type e = CastE( (vi.vtype), e ) in
      GVar(vi, {init = Some (SingleInit (cast_to_tt_type ( Lval(var vi_emptytt) ) ))}, Cil.locUnknown)
   in
   
   (* pthread_mutex_t  lock; *)
   let genLockVar tname = 
      let lock_var_name = Updateconfig.makeTypeLockVarName tname in
      let vi = Cil.makeGlobalVar lock_var_name Updateconfig.pthread_mutex_t in 
      GVar(vi, {init = Some (Updateconfig.pthread_mutex_initializer ())}, Cil.locUnknown)
   in
   
   match g with
     (* only TNamed/TComps are updatable, so we're only concerned with GType/GCompTags *) 
      GType(_,loc)
    | GCompTag(_,loc) ->
         let g_t = 
            begin match g with 
               GType(ti,loc)    -> TNamed(ti,[])
             | GCompTag(ci,loc) -> TComp(ci,[])
             | _                -> raise (Failure "Noway we can be here")
            end
         in
         let gtname = Updatecilutil.getAbstractTypeName g_t in 
         let mode = Updateargs.args.Updateargs.arg_currentMode in
         let gLatestVersion = genLatestVersionVar gtname in
         let gLatestTT = genLatestTTVar gtname g_t in
         let gLock = genLockVar gtname in
         let isRegisteredUpdatable t = try Namedtypes.isUpdatable t with Not_found -> false in 
          
         if (mode = Updateargs.NormalMode && (isRegisteredUpdatable g_t)) ||
            (mode = Updateargs.PatchMode && (not (Versiondata.isExistingType vd gtname)) && (isRegisteredUpdatable g_t))
         then 
           (* this is the first version of this type, i.e. an abstract type in the original program
              or a new type introduced in a patch; generate new globals for latest_version and latest_tt *)
            begin match gLatestTT, gLatestVersion, gLock with 
               GVar(vi_latest_TT,_,_), GVar(vi_latest_version,_,_), GVar(vi_lock,_,_) ->
                  IndirectTypeInfo.addTypeInfo gtname 
                     {IndirectTypeInfo.iti_version = vi_latest_version; 
                      IndirectTypeInfo.iti_latest_tt = vi_latest_TT;
                      IndirectTypeInfo.iti_lock = vi_lock};

                  (* if TT locks are disabled, don't emit the static init for the associated mutex *)
                  (if Updateargs.args.Updateargs.arg_ttLocks 
                   then  (g :: gLatestVersion :: gLatestTT :: gLock :: acc)
                   else  (g :: gLatestVersion :: gLatestTT          :: acc)
                  )

             | _ -> raise (Failure "genTypeInfastructureGlobalFold: now way we can be here (1)")
            end
         else if Versiondata.isExistingType vd gtname && 
                 mode = Updateargs.PatchMode &&
                 (isRegisteredUpdatable g_t)     
         then
           (* this is an updated version of a type; 
              then latest_version and latest_tt  must exist in the VD; retrieve them from there *)
            begin match gLatestTT, gLatestVersion, gLock with 
               GVar(vi_latest_TT,_,loc_TT), GVar(vi_latest_version,_,loc_ver), GVar(vi_lock,_,loc_lock) ->
                  IndirectTypeInfo.addTypeInfo gtname 
                     {IndirectTypeInfo.iti_version = vi_latest_version; 
                      IndirectTypeInfo.iti_latest_tt = vi_latest_TT;
                      IndirectTypeInfo.iti_lock = vi_lock};
                  vi_latest_TT.vstorage <- Cil.Extern;
                  vi_latest_version.vstorage <- Cil.Extern;
                  vi_lock.vstorage <- Cil.Extern;
                     g
                  :: (GVarDecl(vi_latest_version, loc_ver))
                  :: (GVarDecl(vi_latest_TT, loc_TT))
                  :: (GVarDecl(vi_lock, loc_lock))
                  :: acc
             | _ -> raise (Failure "genTypeInfastructureGlobalFold: now way we can be here (2)")
            end
         else if not (isRegisteredUpdatable g_t) 
         then 
           (* Not updatable; ignore it *)
           g::acc
         else 
           raise (Failure "genTypeInfastructureGlobalFold: now way we can be here (3)")



    | _ -> g::acc 


(* genTypeInfastructure gets called from main in both patch and normal modes;
   however, vd is only used in patch mode, since in program mode there's no
   existing version data available
 *)
let genTypeInfastructure (vd : Versiondata.version_data) (file : Cil.file) : unit =

   (* Empty TT management *)
   let empty_tt_arg_type = TPtr((TVoid []), []) in
   let empty_tt_type = TFun( TVoid [],  
                             Some ([("abs", empty_tt_arg_type, [])]),
                             false,
                             [] ) in

   (* The empty TT sets the .version to 0; 
      WARNING !!!! We rely on the fact that in wrappers, the .version is at offset 0;
      So whenever the wrapper layout changes, this function has to change as well !!!
   *)
   let make_empty_tt () : Cil.fundec =  
      let empty_tt = Cil.emptyFunction Updateconfig.empty_tt_name in 
        ignore(Cil.makeFormalVar empty_tt "abs" empty_tt_arg_type);
        Cil.setFunctionType empty_tt empty_tt_type;
        empty_tt
   in 
   (* now that we know the empty_tt, add it to the globals and make all tt vars point to it *)
   let genInfastructure (empty_tt_vi : Cil.varinfo) = 
      List.fold_right (genTypeInfastructureGlobalFold vd empty_tt_vi) file.globals []
   in 

   (* In normal mode we create the empty_tt as a normal, empty function *)      
   (* In patch mode we generate the empty_tt function as extern; 
      it must be there already, since we've presumably introduced it in normal mode *)
   match Updateargs.args.Updateargs.arg_currentMode with
      Updateargs.NormalMode  -> 
        let empty_tt = make_empty_tt () in 
         let globals' = genInfastructure empty_tt.svar in
         begin
           file.globals <- [GFun(empty_tt,Cil.locUnknown)] @ globals' 
         end
    
    | Updateargs.PatchMode ->
         let vi = Cil.makeGlobalVar Updateconfig.empty_tt_name empty_tt_type in 
         let globals' = genInfastructure vi in 
         vi.vstorage <- Cil.Extern;
         file.globals <- [GVarDecl(vi, Cil.locUnknown)] @ globals'
   
    | _ -> Updatedebug.abort (Updatedebug.bug "genTypeInfastructure: call not in NormalMode or PatchMode mode!\n")

