(* 
 *
 * Copyright (c) 2004-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Michael Hicks       <mwh@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(* Configuration options and hardwired strings for Proteus *)

open Cil


let variable_prefix = "__DSU_";;
let var_pre_len = String.length variable_prefix;;

(*************************************  CONABSINSERTION  ************************************)

(* We have to ignore some special C-insternal types *)
                     
let gnu_va_list =  TNamed ({tname = "__gnuc_va_list"; ttype = (TBuiltin_va_list []); treferenced = true}, []);;  
let va_list = TNamed ({tname = "va_list"; ttype = gnu_va_list; treferenced = true}, []);;
let typesToIgnore = [TBuiltin_va_list [];
                     gnu_va_list;
                     va_list
                     ];;


(* Create c string name of coercion *)
let conFunctionName = variable_prefix ^ "con_";;
let absFunctionName = variable_prefix ^ "abs_";;


let conAbsFunctionNameLen = 
   let len = String.length conFunctionName in 
   let len' = String.length absFunctionName in
   (* insist the length of abs and con are equal so we can treat them uniformly *)
   if len <> len'
   then Errormsg.s (Errormsg.bug 
           "updateConfig.ml:  Names of con and abs coercions are not the same length!\n")
   else len
   
let makeConName nt =  conFunctionName ^ nt;;
let makeAbsName nt =  absFunctionName ^ nt;;

(* function to test if a function name is an abs/con coercion *)
let isConFunc name = 
   try (String.sub name 0 (conAbsFunctionNameLen) = conFunctionName)
   with Invalid_argument _ -> false

let isAbsFunc name = 
   try (String.sub name 0 (conAbsFunctionNameLen) = absFunctionName)
   with Invalid_argument _ -> false

(* extract the abstract type for a given abs or con coercion name *)
let extractTypeFromCoerName name = 
   String.sub name (conAbsFunctionNameLen) ((String.length name)-conAbsFunctionNameLen);;

(* create a c string name of formal parameters of functions we might create *)
let makeFormalParameter uniqueId = variable_prefix ^ "formal_param_" ^ (string_of_int uniqueId);;

(* Create a c string name for functions we create during function type coercions *)
let makeFunCoerFunctionName uniqueId = variable_prefix ^ "fun_coer_" ^ (string_of_int uniqueId);;

(* create a fresh global variable to act as a temp storage location for the thing to coerce *)
let makeFunCoerGlobalTmpVar uniqueID = variable_prefix ^ "GLOBAL_coer_var_" ^ (string_of_int uniqueID);;


(* make concrete comp name *)
let makeConcreteCompName ci =
   let prefix = variable_prefix ^ "data_" in 
   prefix ^ ci.cname

let makeAbsTypeName tn =
  try 
    if (String.sub tn 0 7) = "struct_" &&
       (String.sub tn 7 var_pre_len) = variable_prefix &&
       (String.sub tn (7+var_pre_len) 5) = "data_" then
      "struct_"^(String.sub tn (12+var_pre_len) ((String.length tn)-(12+var_pre_len)))
    else tn
  with Invalid_argument _ -> tn

(**************************************  WRAPPERS  ******************************************)

(* Wrapper creation options *)

(* prefix for wrapper types : T -> wrapper_T *)
let wTypePrefix = variable_prefix ^ "wrapper_";;
let makeWrapperName nt =  wTypePrefix ^ nt;;

let makeDataUnionName nt =  variable_prefix ^ "udata_" ^ nt;;


(* internal wrappers fields *)
let wVersionFieldName = variable_prefix ^ "version";;
let wDataFieldName = variable_prefix ^ "data";;
let wPDataFieldName = variable_prefix ^ "pdata";;
let wSlopFieldName = variable_prefix ^ "slop";;

let wDataUnionName = variable_prefix ^ "udata";;

let wMinSlopSize = 16;;
let wDefaultSlopRatio = 2.0;;

let wDefaultVersionFieldType = Cil.uintType;;

(* if data is larger than this, use a heap-allocated tmp
   for copying data in the type transformer, rather than a stack-allocated tmp,
   (or else we might blow up the stack)
*)
let maxStackObjectSize = 2048;; 

(* slop type is char[wDefaultSlopSize] *)
let wMakeSlopType slopSize = (TArray(TInt(IChar,[]),Some(Const(CInt64 ((Int64.of_int slopSize) ,IUInt,None))),[]))

(* name and type for mem* protos *)
let memcpyDescriptor = ("memcpy", 
                         (TFun (Cil.voidPtrType, 
                                Some [("dest", Cil.voidPtrType, []);
                                      ("src", TPtr(TVoid [Attr ("const", [])], []),[]);
                                      ("n", intType, [])],
                                false, 
                                [])
                         )
                        );; 

let mallocDescriptor = ("malloc", 
                         (TFun (Cil.voidPtrType, 
                                Some [("n", intType, [])],
                                false, 
                                [])
                         )
                        );; 

let freeDescriptor = ("free", 
                         (TFun (Cil.voidType, 
                                Some [("ptr", Cil.voidPtrType, [])],
                                false, 
                                [])
                         )
                        );; 

(************************************* Analysis *********************************************)

(* name of type attibute used for associating effects with function types *)
let effectAttrName = variable_prefix ^ "EFFECT__";;




(**************************************  LOOPS  *********************************************)

let loopNamePrefix = variable_prefix ^ "loop_"

(* plain loop extraction *)
let loopExtractPragma = variable_prefix ^ "loop";;           
  
(* loop extract + add __DSU_update at the end of the loop *) 
let loopExtractUpdatePragma = variable_prefix ^ "loopupd";;    

(* loop extract + add CHECKIN and TX around extracted function *) 
let loopTxCheckinPragma = variable_prefix ^ "looptc";;

(* code extract *) 
let codeExtractPragma = variable_prefix ^ "extract";;

(* Fresh name generation for tail-recursive functions *)
(* FIX: If this begins with DSU_ as usual, it confuses later stages. *)
let freshLoopNameFixed = "my" ^ loopNamePrefix;;
let freshLoopName pre = freshLoopNameFixed ^ pre;;
let isLoopFun name = Cil.startsWith freshLoopNameFixed name;;
let isLoopStruct name = Cil.startsWith ("struct_" ^ freshLoopNameFixed) name;;

(**************************************  INDIRECTION  ***************************************)

(* Function indirection *)
let makeIndirectionVarName funName version = variable_prefix ^ "indirect_" ^ funName ^ (string_of_int version);;
let makeFunctionWrapperName funName version = variable_prefix ^ "funwrap_" ^ funName (* ^ (string_of_int version) *);;
let makeVersionedFunctionName funName version = funName ^ "__v" ^ (string_of_int version);;
let makeVersionedVarName varName version = varName ^ "__v" ^ (string_of_int version);;
let isWrapperFunction funName = Cil.startsWith (variable_prefix ^ "funwrap_") funName;;
let isWrapperFunctionForMe funName wrapperFunName = Cil.startsWith (variable_prefix ^ "funwrap_" ^ funName) wrapperFunName;;

(* Given the name of a user TT function, produce the name of the in-place TT wrapper *)
let makeInplaceTTName (userTTname:string) : string = 
   variable_prefix ^ userTTname

(**************************************  UPDATE POINTS  *************************************)

(* Update 'keyword' insertion  *)
let updateFunctionName = variable_prefix ^ "update";;


(*         
    int update(char * forbidden_functions);

    Note: this prototype is harcoded in runtime/updatelib.c, so if you change it here, you must
    sync with runtime/updatelib.c too 
*)

let updateFunctionType = TFun(Cil.intType, 
                              Some [("forbidden_functions", Cil.charPtrType, [])], 
                              false, 
                              []) ;;

let semanticUpdateFunctionName = variable_prefix ^ "semantic_update";;
let semanticUpdateFunctionType = TFun(voidType, Some [], false,[]);;

let isUpdateFunc name   = (name = updateFunctionName);;


let checkinFunctionName = variable_prefix ^ "checkin";;
let checkinFunctionType = TFun(TVoid [], 
                              Some [("checkin_id", Cil.intType, [])], 
                              false, 
                              []) ;;

(**************************************  DEBUG  *********************************************)

(* various debug controls; 
   use them carefully, as they produce huge amounts of output *)
let debug_enable_log_messages = true;;
let debug_properType = false;;
let debug_AbstractType = false;;
let debug_AbstractArray = false;;
let debug_coercionGeneration = false;;
let debug_LvalGen = false;;
let debug_delta_test_compat = false;; (* flip it to true when running analysis regression test cases *)
let delta_test_compat s1 s2 = if debug_delta_test_compat then s1 else s2;;

(*********************************** Type transformers ***************************************)

(* prefix to the name of variables that hold a pointer to the latest tt function for a type *)
let latest_tt_var_prefix = variable_prefix ^ "latest_tt_";;

(* prefix for the name of type transformers *)
let tt_var_prefix = variable_prefix ^ "tt_";;

(* Name of the empty type transformer.  The latest type transformer variables are 
   initially set to this. *)
let empty_tt_name = variable_prefix ^ "empty_tt";;

(* is this the empty TT ?*)
let isEmptyTT name = (name = empty_tt_name);;

(* make the string that represents the version of a type transformer *)
let makeTTVersionStr (version:int) = 
   "_v_" ^ (string_of_int version)

(* generate the name of a type transformer *) 
let makeTTName (tname:string) (version:int) = 
   tt_var_prefix ^ tname ^ (makeTTVersionStr version)

(* generate the name of the variable that holds the latest type transformer *)
let makeLatestTTName (tname:string) = 
   latest_tt_var_prefix ^ tname

(* Name of the pragma directive that tells us information about a type transformer *)
let ttPragmaTypeName = variable_prefix ^ "TT"
let ttPragmaGVarName = variable_prefix ^ "TTGVAR"

(********************************** Patch Compilation ****************************************)

let latest_type_version_prefix = variable_prefix ^ "latest_type_version_";;

let makeLatestTypeVersionVarName (tname:string) = 
   latest_type_version_prefix ^ tname;; 

let makeOldExternVarName (varName:string) = variable_prefix ^ "_old_extern" ^ varName;;

let patch_install_function_name = variable_prefix ^ "install_patch";;
let patch_install_function_type = TFun(TVoid [], Some [], false ,[]);;

let state_xform_function_name = variable_prefix ^ "state_xform";;
let is_state_xform fd = fd.svar.vname = state_xform_function_name;;

let state_xform_function_type = TFun(TVoid [], Some [], false ,[]);;

let type_metadata_var_name = variable_prefix ^ "type_metadata";;
let type_metadata_var_type = Cil.charPtrType;;

let def_metadata_var_name = variable_prefix ^ "def_metadata";;
let def_metadata_var_type = Cil.charPtrType;;


(********************************* Version Data **********************************************)


(* This is set as a command line argument now, so we don't need this 
(* Given the name of the file we are processing, generate the name of the version datafile *)
let makeVersionDataFilename (filename : string) : string = 
   filename ^ ".vd"
*)


(********************************* update lib **********************************************)

(* Name of the initialisation function defined in updatelib.c that initialises
   the update notification mechanism *)
let updatelibInitFunName = variable_prefix ^ "init_handler";;

let updatelibInitFunType = TFun(TVoid [], Some [], false ,[]);;

(********************************* add cap **********************************************)

(* fun name + prototype for adding check-in points caps to the hash *)

(*     Note: this prototype is harcoded in runtime/dsu_types.h, so if you change it here, you must *)
(*     sync with runtime/ too  *)

let addcapFunName = variable_prefix ^ "addcap";;

let addcapFunType = TFun(TVoid [], 
                         Some [("checkin_id", Cil.intType, []);
                               ("delta", Cil.charPtrType,[]);
                               ("fwd", Cil.charPtrType,[]);
                               ("bck", Cil.charPtrType,[]);
                               ("info", Cil.charPtrType,[]);
                               ("checkin_label", Cil.charPtrType,[])
                              ],
                         false ,
                         []);;


(******************************** Malloc pragma ********************************)

(* name of the malloc directive *)
let mallocPragmaName = variable_prefix ^ "malloc";;

(* add here all functions that should be considered malloc equivalents by default *)
(* let predefinedMallocs = ["malloc"; "memset"; "free"];; *)
(* MWH: this can't be free() or memset() anymore; it really needs to be malloc() *)
let predefinedMallocs = ["malloc"; "alloca" ; "calloc"];;

(*************************     sizeof() pin down   *****************************)
(* Attention ! sizeof() is a prime suspect if the compiled code segfaults or doesn't work
   in general; so when in trouble, turn sizeofPinDownEnabled to true. 
   It's more conservative, but it can save you some headaches. 
*)
let sizeofPinDownEnabled = true;;


(************************ Suspect alias analysis ******************************)
(* Named of attribute that contains live pointer set variable identifier *)
let saVarAttrName = "SUSPECT_ALIAS"
let saFunAttrName = "SUSPECT_ALIAS_FUN"

(************************ Effect analysis ******************************)
let eaVarAttrName = "VOID_EFFECT"


(*************************     Function Garbage Collector   *****************************)
let addFVersionListFunctionName = variable_prefix ^ "addFVersionList";;

let lockFunctionName = variable_prefix ^ "lock";;
let unlockFunctionName = variable_prefix ^ "unlock";;

let functionVersionListStructName = variable_prefix ^ "FVersionList";;
let fv_itemBaseFunName = ("baseFunName", Cil.charPtrType);;
let fv_itemFirst = ("first", Cil.voidPtrType);;
let fv_itemActiveFunction = ("activeFunction", Cil.voidPtrType);;
let fv_itemOldVersionTotalCount = ("oldVersionsTotalCount", Cil.uintType);;
let fv_itemLength = ("length", Cil.uintType);;

let functionVersionListFields = [
   fv_itemBaseFunName;
   fv_itemFirst;
   fv_itemActiveFunction;
   fv_itemOldVersionTotalCount;
   fv_itemLength];;

(*************************     Global variable types        *****************************)

(* generate the name of a global variable type, based on the variable name *) 
let makeGlobalVarTypeName (var_name : string) : string = 
   variable_prefix ^ "typeof_gvar_" ^ var_name;;

(*************************    Forced (Non)Updatable Types   *****************************)
let forceUpdatablePragmaName = variable_prefix ^ "FORCE_UPDATABLE";;
let forceNonUpdatablePragmaName = variable_prefix ^ "FORCE_NONUPDATABLE";;

(*************************    Transparent typedefs          *****************************)
let enableElimStructTypedefs = true;;

(*************************    setjmp/longjmp                *****************************)
let predefinedSetjmp = ["_setjmp"; "setjmp"; "sigsetjmp"; "__sigsetjmp" ];;
let predefinedLongjmp = ["longjmp"; "siglongjmp" ];;


let stmtHasLabel checkFun stmt = 
  let checkOneLabel = function
      Label(s,l,b) -> checkFun s
    | _ -> false
  in
  List.exists checkOneLabel stmt.labels
;;

let stmtGetLabel checkFun stmt = 
  let checkOneLabel = 
    fun acc lb -> match lb with
        Label(s,l,b) -> if checkFun s then s::acc else acc
      | _ -> acc
  in
  List.fold_left checkOneLabel [] stmt.labels
;;

(*************************    transactions                 *****************************)
(* NB: Cilutil distinguishes (TX|CHECKIN)'s by looking for a label starting with:

 variable_prefix ^ (TX|CHECKIN)

   If you change variable_prefix, you must change Cilutil, too.
*)

let isTXStmt stmt = stmtHasLabel Cilutil.isTxLabel stmt
;;
let isCheckinStmt stmt = stmtHasLabel Cilutil.isCheckinLabel stmt
;;
let getCheckinLabel stmt = stmtGetLabel Cilutil.isCheckinLabel stmt
;;

(*********************  access (read/write/readwrite) ***********************************)
type access_t = SRead | SWrite | SReadWrite | SNone

let accessPack access symbol = match access with
        SRead       -> "<" ^ symbol
      | SWrite      -> ">" ^ symbol
      | SReadWrite  -> "!" ^ symbol
      | SNone       -> Errormsg.s (Errormsg.bug "updateConfig.ml:addAccessToSymbol SNone called\n")
;;

let accessUnpack asymbol = 
  let tail s = String.sub s 1 ((String.length s) - 1) in
  try 
    match String.get asymbol 0 with
        '<' -> SRead, tail asymbol
      | '>' -> SWrite, tail asymbol
      | '!' -> SReadWrite, tail asymbol
      |  _  -> SNone, asymbol
  with 
      Invalid_argument _  -> SNone, ""
;;

let accessErase asymbol = snd (accessUnpack asymbol)
;;

(*********************  multithreading ***********************************)
let isForkFunction name = (name = "pthread_create");;

(* let pthread_mutex_t =  *)
(*   let underlyingUnion = Commoncilutil.makeCompInfo false "__anonunion_pthread_mutex_t_1" in *)
(*   TComp(underlyingUnion, []) *)
(* ;; *)
(* TNamed ({tname = "pthread_mutex_t";  *)
(*                                ttype = (TArray (charPtrType, Some (Cil.integer 24), []));  *)
(*                                treferenced = true}, []);;   *)

(* must make this opaque since we don't have a definition of pthread_mutex_t;
   but we know it's 24 bytes, and the mutex initializer is all 0's
   make it 48 bytes, just in case
*)
let pthread_mutex_t = TArray (charType, Some (Cil.integer 48), [])
;;
let pthread_mutex_initializer () = (Cil.makeZeroInit pthread_mutex_t)
;;

let pthreadLockFunctionName = "pthread_mutex_lock";;
let pthreadUnlockFunctionName = "pthread_mutex_unlock";;
let pthreadLockFunctionType = TFun(Cil.intType, 
                              Some [("mutex", TPtr(pthread_mutex_t,[]), [])], 
                              false, 
                              []) ;;

let makeTypeLockVarName tname = variable_prefix ^ "typelock_" ^ tname;;

(*************************** blocking calls *************************)

(** originally stolen from Capriccio, and added some more **)

let predefinedBlockingCalls = [
(*   "strlen"; *)
(*   "strcpy"; *)
(*   "strcmp"; *)
(*   "strcasecmp"; *)
(*   "strchr"; *)
(*   "strrchr"; *)

(*   "tolower"; *)

(*   "memset"; *)
(*   "memcpy"; *)
(*   "memcmp"; *)
(*   "memchr"; *)

(*   "__errno_location"; *)

(*   "__ctype_b_loc"; *)
  "__xstat";

  "accept";
  "close";
  "connect";
  "creat";
  "dup";  "dup2";
  "fcntl";
  "fstat";
  "fsync";
  "getmsg";
  "ioctl";
  "lockf";
  "lseek";
  "mq_open";
  "msgrcv"; 
  "msgsnd";
  "msync";
  "nanosleep";
  "open";
  "pause";
  "poll";
  "pread";
  "putmsg";
  "pwrite";
  "read"; "readv";
  "recv";  "recvfrom"; "recvmsg";
  "select";
  "semget"; "semctl"; "semop";
  "send"; "sendto"; "sendmsg";
  "sleep";
  "stat";
  "write"; "writev";

  (* Pthread functions *)
  "pthread_mutex_lock";

  "pthread_cond_timedwait";
  "pthread_cond_wait";

  "pthread_rwlock_rdlock";
  "pthread_rwlock_wrlock";

  "pthread_join";

(*   "pthread_cond_signal"; *)
(*   "pthread_mutex_unlock"; *)
(*   "pthread_yield"; *)
 ];;

let startBlockingCallFunctionName = variable_prefix ^ "start_blocking_call";;
let endBlockingCallFunctionName = variable_prefix ^ "end_blocking_call";;
let blockingCallFunctionType = (TFun(voidType, Some [], false,[]));;

