(* 
 *
 * Copyright (c) 2004-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Gareth Stoyle       <gareth.stoyle@cl.cam.ac.uk> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 *)



(* Update points insertion *)

open Cil

module H = Hashtbl

(* 
   We implement the 'update' keyword from theory by a call to the 'update' function.

   There are currently three policies for inserting 'update' keywords:

   UNone : no 'update' keywords are inserted
   -----

   UReturnonly : 'update' keywords are inserted right before the return statements 
   -----------   in functions. The cool thing about CIL is that it inserts return 
                 statements even if the original code lacked them, so we're guaranteed 
                 to get at least one return point per function.

   UTx : 'update' sprinkled in transactions              
   ----

   UFull : 'update' keywords are inserted before each CIL *statement*, i.e. at the 
   -----   beginning of control blocks 

   We might also consider having 'update' keywords inserted before each CIL *instruction*, 
   but that might be too much.

   Updateargs.args.Updateargs.arg_updatePointsPolicy stores the updatePointsPolicy 
   (updatePointsPolicy can be set in the cmd line via --update-points)

*)


(* module that stores information about update points in a file *)
(*====================*)
module UPInfo =
(*====================*)
struct
  (* Info about an update point; used when compiling a patch, 
     to see how many upoints conflict with the patch *)
  type upinfo_entry_t = {
    restr_delta : Deltasolver.CapabilitySet.t;
    restr_fwd   : Deltasolver.CapabilitySet.t;
    restr_bck   : Deltasolver.CapabilitySet.t;
    upFunName   : string;
  }

  type ckinfo_entry_t = {
    pre_delta : Deltasolver.CapabilitySet.t;
    pre_fwd   : Deltasolver.CapabilitySet.t;
    pre_bck   : Deltasolver.CapabilitySet.t;
    post_delta: Deltasolver.CapabilitySet.t;
    post_fwd  : Deltasolver.CapabilitySet.t;
    post_bck  : Deltasolver.CapabilitySet.t;
  }

  type upinfoHash_t = (int, upinfo_entry_t) H.t;;
  type ckinfoHash_t = (int, ckinfo_entry_t) H.t;;
  
  type upinfo_t = { 
    (* if a point is in here, it wasn't dissed by updatability = N *)   
    theUpdateStatements : upinfoHash_t;

    theCheckinPoints : ckinfoHash_t;
   
    (* stateXform effects *)
    mutable stateXformEffects : upinfo_entry_t option; 
  };;
   
  let instance : upinfo_t = {
    theUpdateStatements = H.create 65535;
    theCheckinPoints = H.create 65;
    stateXformEffects   = None;
  };; 

  (*-----------------------------------------------------------------------------------*)
  let addUpdatepoint (sid : int) (ui : upinfo_entry_t) : unit =
  (*-----------------------------------------------------------------------------------*)
    try
      let _ = H.find instance.theUpdateStatements sid in
        Updatedebug.abort (Updatedebug.bug "addUpdateStatement : tried to add already-existing update() statement \n")
   
    with Not_found -> 
      H.add instance.theUpdateStatements sid ui;;
   
  
  (* Remove an update statement from the hash *)
  let removeUpdatePoint (sid : int) : unit =
    H.remove instance.theUpdateStatements sid
  ;;


  (*-----------------------------------------------------------------------------------*)
  let addCheckin (sid : int) (ci : ckinfo_entry_t) : unit =
  (*-----------------------------------------------------------------------------------*)
    try
      let _ = H.find instance.theCheckinPoints sid in
        Updatedebug.abort (Updatedebug.bug "addCheckin : tried to add already-existing checkin point \n")
   
    with Not_found -> 
      H.add instance.theCheckinPoints sid ci;;


end
  
type capStats_t = 
{ 
  mutable vcFwdMoreRestrictive     :  int;
  mutable vcBckMoreRestrictive     :  int;
  mutable vcMoreRestrictive        :  int;
  mutable conflictWithStateXform   :  int;
  mutable totalUpdatePoints        :  int;
};;

(* type checkinIdMapKey = InstrId of int | StmtId of int;; *)

(* type checkinIdToCapMap_t = (checkinIdMapKey, string * string * string * string) H.t;; *)
type checkinIdToCapMap_t = (int, string * string * string * string * string) H.t;;

                            (* funName -> (sid *  checkin_label) list *)
type funNameToCheckinId_t = (  string,    (int *  string) list) H.t;;

(****************************** Insertion of update points ************************)

(* insert updates in a statement *)

let addUpdateToStatementInternal (updateFunction : Cil.varinfo) (atEnd:bool) s = 
  let loc = !Cil.currentLoc in (* this is valid since we're in a visitor *)
    
  (* make an instruction containing the update() call *)
  let callToUpdate = Call(None, Lval(var updateFunction), [], loc) in

  let () = ignore (Updatedebug.log [Updatedebug.BTUpdatepoints] "insertUpdatesSVisitor : inserting update at (%a)\n" d_loc loc) in

  (* pack update () into a statement *)
  let newSInstr = Commoncilutil.mkStmt (Instr [callToUpdate]) in
  let () = ignore(Updatedebug.log [Updatedebug.BTUpdatepoints] "insertUpdatesSVisitor : Generated new Instr statement with ID=%d\n" newSInstr.sid) in

  (* put the statement id into the hash; needed later for computing # of update points *)
(*   let () = UPInfo.addUpdateStatement newSInstr.sid in *)

  (* pack the update() and the original statement into a block *)
  let newSBlock = 
    if atEnd
    then Commoncilutil.mkStmt (Block (mkBlock ([s; newSInstr]))) 
    else Commoncilutil.mkStmt (Block (mkBlock ([newSInstr; s]))) 
  in
  let () = ignore(Updatedebug.log [Updatedebug.BTUpdatepoints] "insertUpdatesSVisitor : Generated new Block statement with ID=%d\n" newSBlock.sid) in
  newSBlock
;;

(* by default add at the beginning *)
let addUpdateToStatement (updateFunction : Cil.varinfo) s = 
  addUpdateToStatementInternal updateFunction false s
;;

let addUpdateToStatementAtTheEnd (updateFunction : Cil.varinfo) s = 
  addUpdateToStatementInternal updateFunction true s
;;

let rec addUpdateToStatementFull updateFunction s =

  let doBlock b = List.map(addUpdateToStatementFull updateFunction) b.bstmts in
  match s.skind with
    
      Instr _ -> addUpdateToStatement updateFunction s

    | If(exp,b1,b2,location) -> {s with skind = (If(exp,
                                   {b1 with bstmts = doBlock b1},
                                   {b2 with bstmts = doBlock b2},
                                   location))}

    | Loop (b,l,so1,so2) -> {s with skind = (Loop({b with bstmts = doBlock b},l,so1,so2))}

    | Block b -> {s with skind = (Block {b with bstmts = doBlock b})} 
    | _ -> s
;;

(* NB: this is called *before* the analysis *)
class insertUpdatesRegardlessVisitor (updateFunction : Cil.varinfo) : cilVisitor = object(self)
  inherit nopCilVisitor
    
  method vstmt s =

    let isReturn = match s.skind with Return (_,_) -> true | _ -> false in

    match Updateargs.args.Updateargs.arg_updatePointsPolicy with
          Updateargs.UNone -> 
            Updatedebug.abort (Updatedebug.bug "insertUpdatesPolicyVisitor : visitor called with UNone ??\n")

        | Updateargs.UReturnonly when not isReturn -> (* update before Returns only; if this is not a Return look at children *)
            DoChildren
              
        | _ -> (* insert them everywhere *)
(*             ChangeDoChildrenPost (addUpdateToStatement updateFunction s, (fun x -> x)) *)
            ChangeTo (addUpdateToStatement updateFunction s)

end

class insertUpdatesRegardlessFullVisitor (updateFunction : Cil.varinfo) : cilVisitor = object(self)
  inherit nopCilVisitor
    
  method vstmt s =

(*             ChangeTo (addUpdateToStatementFull updateFunction s) *)
            ChangeTo (addUpdateToStatement updateFunction s)

end


class insertUpdatesInTXFunctionVisitor (insertRegardlessVisitor : cilVisitor) : cilVisitor = object(self)
  inherit nopCilVisitor
    
  method vstmt s =

    if Updateconfig.isTXStmt s 
    then
      (* insert here and in the children *)
      let newS = Cil.visitCilStmt insertRegardlessVisitor s in
      (* then skip children, as they've been processed above *)  
      ChangeTo newS
    else
      (* descend into the children until we find a tx *)
      DoChildren
end

class checkForTxVisitor (res : bool ref) : cilVisitor = object(self)
  inherit nopCilVisitor
    
  method vstmt s =

    if (Updateconfig.isTXStmt s) 
    then 
      (res := true;
      SkipChildren)
    else
      DoChildren
end

(* add an update after the Tx; doesn't support nested transactions ! *)
class insertUpdatesAtEndOfTxVisitor (updateFunction : Cil.varinfo) : cilVisitor = object(self)
  inherit nopCilVisitor
    
  method vstmt s =

    if Updateconfig.isTXStmt s 
    then
        ChangeTo (addUpdateToStatementAtTheEnd updateFunction s)
    else
      DoChildren
end

(* NB: this is called *before* the analysis *)

(* if the updatabilityPolicy = UPTx  
 * AND the function contains a TX, only insert update points in the TX body 
 * In all other cases, inseet wherever updatepoints policy calls for    
 *)
class insertUpdatesGVisitor (functionHasTx : Cil.fundec -> bool) (updateFunction : Cil.varinfo) : cilVisitor = object(self)
  inherit nopCilVisitor
    
  method vglob g =
    let rv = (new insertUpdatesRegardlessVisitor updateFunction) in
    let rvf = (new insertUpdatesRegardlessFullVisitor updateFunction) in
    let eotxv = (new insertUpdatesAtEndOfTxVisitor updateFunction) in
     match g with 
       GFun(fd,loc) when ((not (Namedtypes.isConAbsFuncName fd.svar.vname)) && 
                          (not (fd.svar.vname = "main")) &&
                          (not (fd.svar.vname = "my__DSU_loop_maincont"))
                         ) ->
         (match Updateargs.args.Updateargs.arg_updatabilityPolicy with

              Updateargs.UPTx when (functionHasTx fd) ->
                ChangeTo [GFun(Cil.visitCilFunction (new insertUpdatesInTXFunctionVisitor rvf) fd, loc)]

            |  _ ->
                 (if (Updateargs.args.Updateargs.arg_updatePointsPolicy = Updateargs.UEndOfTx)
                  then 
                    (
                      ChangeTo [GFun(Cil.visitCilFunction eotxv fd, loc)])
                  else
                    ChangeTo [GFun(Cil.visitCilFunction rv fd, loc)]
                 )
         )

     | _  -> SkipChildren

end;;



(* NB: this is called *before* the analysis *)
let insertUpdatePoints (updFun : varinfo) (f : file) : unit =
  
  let functionHasTx fd = 
    let res = ref false in
    let check4Tx = new checkForTxVisitor res in
    ignore(Cil.visitCilFunction check4Tx fd);
    !res  
  in

  (* create an empty update() function *)    
  let updateFunction = 
    if Updateargs.args.Updateargs.arg_updatePointsPolicy = Updateargs.UEndOfTx
    then Updatecilutil.createFunDecl Updateconfig.semanticUpdateFunctionName Updateconfig.semanticUpdateFunctionType f 
    else 
      (* Updatecilutil.createFunDecl Updateconfig.updateFunctionName Updateconfig.updateFunctionType f  *)
      updFun
  in
  (* insert update() statements in the program, unless directed not to (i.e., updatePointsPolicy = UNone) *)
  (if Updateargs.args.Updateargs.arg_updatePointsPolicy <> Updateargs.UNone
  then (visitCilFile (new insertUpdatesGVisitor functionHasTx updateFunction) f))
;;



(************************** annotate update points, ditch unsafe ones **************************************)

   
class removeUnsafeUpdatePointsVisitor 
   (analysis_solution : Analysis.analysis_result)
   (real_update : Cil.varinfo)
   (removalCount : int ref)
   : cilVisitor = 
object(self)
   inherit nopCilVisitor
   
   
   method vstmt s =
      match s with 
         { skind = Instr [Call(_, Lval(Var vi, NoOffset), arg, loc)]; sid = sid } when Updateconfig.isUpdateFunc vi.vname ->
            let mu_id = 
               begin
               try 
                 Some (H.find analysis_solution.Analysis.ar_speculativeUpdateMuHash sid) 
               with Not_found -> None
(* Updatedebug.abort (Updatedebug.bug  *)
(*                                  "speculativeToRealVisitor: couldn't find mu for speculative update %s at statement %d (%a)\n" vi.vname sid d_loc loc)    *)
               end
            in 
            begin try	
              let mu = match mu_id with
		            Some mu_id -> H.find analysis_solution.Analysis.ar_mu_solution mu_id
		          | None -> true in


(*             let real_update = Instr [Call(None, Lval(Var real_update, NoOffset), arg, loc)] in *)
(*             begin try *)
(*               let mu = H.find analysis_solution.Analysis.ar_mu_solution mu_id in  *)
              if mu = true then 
                  (* the update point is valid, turn it into a real one *)
                  DoChildren

              else begin
                  (* the update point is invalid, remove it *)
                  UPInfo.removeUpdatePoint sid;
                  incr removalCount;
                  s.skind <- Instr [];
                  DoChildren
               end
            with 
               Not_found ->  (* no solution => unconstrained variable => U *) 
(*                   let () = ignore (Updatedebug.log [Updatedebug.BTAlways] "speculativeToRealVisitor (no mu, %a)  called\n" d_stmt s) in *)

                  DoChildren
               
                            
            end
             
       | _ ->
           DoChildren
       
end;; 

let isXname dckHash name knd =
    try
      (H.find dckHash name) = knd
    with
        Not_found -> false
;;
let isGvarName     dckHash (name : string) : bool = isXname dckHash name Analysis.DckGvar;;
let isTypeName     dckHash (name : string) : bool = isXname dckHash name Analysis.DckType;;
let isFunctionName dckHash (name : string) : bool = isXname dckHash name Analysis.DckFun;;

let splitEffects3Ways dckHash effectsSet = 
  let typeSet     = Deltasolver.CapabilitySet.filter (isTypeName dckHash)     effectsSet in
  let functionSet = Deltasolver.CapabilitySet.filter (isFunctionName dckHash) effectsSet in
  let gvarSet     = Deltasolver.CapabilitySet.filter (isGvarName dckHash)     effectsSet in
  (typeSet, functionSet, gvarSet)  
;;

let filterRestriction dckHash d = 
  if (Updateargs.args.Updateargs.arg_restrictTypes &&
        Updateargs.args.Updateargs.arg_restrictFuns &&
        Updateargs.args.Updateargs.arg_restrictGvars)
  then
    (* this (full restriction) is the common case, make it fast *)
    d
  else
    let typeSet, functionSet, gvarSet = splitEffects3Ways dckHash d in
    let l = [ (Updateargs.args.Updateargs.arg_restrictTypes, typeSet); 
              (Updateargs.args.Updateargs.arg_restrictFuns,  functionSet);
              (Updateargs.args.Updateargs.arg_restrictGvars, gvarSet); ]
    and addSetIfNeeded (needed, set) accu = 
      if needed
      then Deltasolver.CapabilitySet.union set accu
      else accu
    in
    List.fold_right addSetIfNeeded l Deltasolver.CapabilitySet.empty
;;

let eraseAccessKinds = Deltasolver.eraseAccessesSet;;
let conflicts s1 s2 = Deltasolver.conflicts s1 s2;;

(* Sync. updates perform a dynamic check : 

   if (\fst \cap \{update digest} = \emptyset or , 
      (\rst \cap \{update digest} = \emptyset

   (i.e. we don't try to update any named symbol 
   that's in our past/future effects)
   then allow the update, otherwise delay it 
   (well, actually 'deny it' at current point and 
   pass it over to upcoming update points) 
*)
class annotateUpdatePointsVisitor 
   (analysis_solution : Analysis.analysis_result)
   (real_update : Cil.varinfo)
   (nonEmptyEffectRemovalCount : int ref)
   (dckHash : (string, Analysis.deltaConstKind) H.t)
   : cilVisitor = 
object(self)
   inherit nopCilVisitor
   
   method vstmt s =
      match s with 
         { skind = Instr [Call(_, Lval(Var vi, NoOffset), _, loc)]; sid = sid } when 
            Updateconfig.isUpdateFunc vi.vname ->

            let delta_id_out, vc_fwd, vc_bck = 
               begin
               try 
                 let din,dout = H.find analysis_solution.Analysis.ar_stmtid_to_delta_id sid in
                 let vf,ve,vb,vci = H.find analysis_solution.Analysis.ar_stmtid_to_eps_tuple sid in
(*                  let _ = ignore (Updatedebug.log [Updatedebug.BTUpdatepoints] "annotateUpdatePointsVisitor : hit update point (speculative = %b) at (%a) sid %d delta_id_in %d delta_id_out %d\n" (Updateconfig.isSpeculativeUpdateFunc vi.vname) d_loc loc sid din dout) in *)
                 let _ = ignore (Updatedebug.log [Updatedebug.BTUpdatepoints] "annotateUpdatePointsVisitor : hit update point at (%a) sid %d delta_id_in %d delta_id_out %d\n" d_loc loc sid din dout) in
dout,vf,vb
               with Not_found -> Updatedebug.abort (Updatedebug.bug 
                                 "annotateUpdatePointsVisitor: couldn't find delta for update %s at statement %d (%a)\n" vi.vname sid d_loc loc)   
               end
            in 
            let funName = 
               begin
               try 
                 H.find analysis_solution.Analysis.ar_stmtid_to_fun_name sid
               with Not_found -> "UnknownFunction??"
                 (* Updatedebug.abort (Updatedebug.bug  *)
(*                                  "annotateUpdatePointsVisitor: couldn't find delta for update %s at statement %d (%a)\n" vi.vname sid d_loc loc)    *)
               end

            in
            begin try
               let dl = analysis_solution.Analysis.ar_c_of_id delta_id_out
               and vf  = analysis_solution.Analysis.ar_c_of_id vc_fwd 
               and vb  = analysis_solution.Analysis.ar_c_of_id vc_bck in
               let delta = eraseAccessKinds dl
               and vfwd  = eraseAccessKinds vf 
               and vbck  = eraseAccessKinds vb 
               in
               let filtered_delta, filtered_vfwd, filtered_vbck = 
                 (* collect only those sets instructed by Updateargs.arg_restrictX *)
                   (filterRestriction dckHash delta, 
                    filterRestriction dckHash vfwd, 
                    filterRestriction dckHash vbck)
               in
                  begin 
                    let real_update_instr  = 
                      let delta_string = Deltasolver.printDelta filtered_delta 
                      and vcfwd_string, vcbck_string = 
                        if Updateargs.args.Updateargs.arg_vcEnabled 
                        then
                          Deltasolver.printDelta filtered_vfwd, Deltasolver.printDelta filtered_vbck 
                        else
                          "", ""
                      in
                      let big_string = Updateconfig.delta_test_compat ""  
                                          (delta_string ^ " | " ^ vcfwd_string ^ " | " ^ vcbck_string
                                           ^ " | "  ^ (string_of_int sid) )
(*                                            ^ " | " (\* ^ (string_of_int cStats.totalUpdatePoints) *\)) *)
                      in
                      Instr [Call(None, Lval(Var real_update, NoOffset), [Cil.mkString big_string], loc)] 
                    in

                    s.skind <- real_update_instr;

                    (* now that we know this is for real, add it to our updatepointsinfo *)
                    UPInfo.addUpdatepoint sid { UPInfo.restr_delta = delta;
                                                UPInfo.restr_fwd   = vfwd;
                                                UPInfo.restr_bck   = vbck;
                                                UPInfo.upFunName   = funName };
                       
                    DoChildren 
                  end
            with 
               Not_found -> DoChildren  (* this shouldn't happen *) 
            end

       | _ -> DoChildren
       
end;; 

let fixDckHash file dckHash = 
  let funNames = H.create 1023 in
  let addIfNeeded name = Commonutil.hashtbl_add_once funNames name () in
  let doOneGlobal g = match g with
	GFun(fd,_) -> addIfNeeded fd.svar.vname
      | _ -> ()
  in
  let () = Cil.iterGlobals file doOneGlobal in
  let newDckHash = H.create 1023 in
  let doOneDckEntry n d = 
    match d with 
	Analysis.DckGvar when H.mem funNames n -> H.add newDckHash n Analysis.DckFun
      | _ -> Commonutil.hashtbl_add_once newDckHash n d
  in
  H.iter doOneDckEntry dckHash;
  newDckHash
;;

let dealWithStateXform ar hash dckHash = 
  match ar.Analysis.ar_state_xform_eff_id with

      None ->  (* no xform *)
        ()
          
    | Some key -> 
        
        let stateXformEff = 
          try 
            H.find ar.Analysis.ar_funEffectHash key
          with Not_found -> 
            Updatedebug.abort (Updatedebug.bug "annotateUpdatePointsVisitor: can't find key %d in funEffectHash\n" key)
        in
        let stateXformEps = 
          let key_id = stateXformEff.Analysis.effect_vc_eps.Deltasolver.delta_id in
            try 
              ar.Analysis.ar_c_of_id key_id
            with Not_found -> 
              Updatedebug.abort (Updatedebug.bug "annotateUpdatePointsVisitor: can't find delta id %d in solution hash\n" key_id)
        in
        let (typeSet, functionSet, gvarSet) = splitEffects3Ways dckHash  (eraseAccessKinds (stateXformEps)) in 
        UPInfo.instance.UPInfo.stateXformEffects <- Some { UPInfo.restr_delta = typeSet;
                                                           UPInfo.restr_fwd   = functionSet;
                                                           UPInfo.restr_bck   = gvarSet;  
                                                           UPInfo.upFunName   = "stateXForm" }  
;;

(* generates a list of Instr's that add capabilities to the global
map via __DSU_addcap *)
let genAddcapInstrList checkinIdToCapMap addcap_vi = 
  let sortedCheckinIds = List.sort Pervasives.compare (Commonutil.hashtbl_keys checkinIdToCapMap) in
  List.map
    (fun checkinId -> 
       let (delta, fwd, bck, info, checkinlabel) = H.find checkinIdToCapMap checkinId in
       (Call(None, 
             Lval (var addcap_vi), 
             [Cil.integer checkinId; Cil.mkString delta; Cil.mkString fwd; Cil.mkString bck; Cil.mkString info; Cil.mkString checkinlabel], 
             Cil.locUnknown)
        )
    )
    sortedCheckinIds
;;


let getPrePostCheckinIds (analysis_solution : Analysis.analysis_result) 
                         (checkinIdToCapMap : checkinIdToCapMap_t)
                         (dckHash : (string, Analysis.deltaConstKind) H.t)
                         loc sid checkinLabel =

  let delta_id_in, delta_id_out, vc_fwd, vc_bck, vc_checkin_eps = 
    begin try 
      let din,dout = H.find analysis_solution.Analysis.ar_stmtid_to_delta_id sid in
      let vf,ve,vb,vci = H.find analysis_solution.Analysis.ar_stmtid_to_eps_tuple sid in
      let _ = ignore (Updatedebug.log [Updatedebug.BTUpdatepoints] "addCheckinToStatement : hit checkin at (%a) sid %d delta_id_in %d delta_id_out %d\n"  d_loc loc sid din dout) in
        (match vci with 
             None -> Updatedebug.abort (Updatedebug.bug 
                                          "addCheckinToStatement: no check-in eps present for checkin statement %d (%a)\n" sid d_loc loc)
           | Some svci -> 
               din, dout, vf, vb, svci)
    with Not_found -> Updatedebug.abort (Updatedebug.bug 
                                           "addCheckinToStatement: couldn't find delta for checkin at statement %d (%a)\n" sid d_loc loc)   
    end
  in 
  let sanitize deltaSet = 
    let exclude x = (Forcedpragma.isForcedUpdatable x) 
(*       || (Updateconfig.isLoopFun x) || (Updateconfig.isLoopStruct x)  *)
    in
    let includ x = not (exclude x) in 
    Deltasolver.CapabilitySet.filter includ (eraseAccessKinds deltaSet)
  in
  let eak deltaSet = Deltasolver.CapabilitySet.filter 
    (fun name -> not ((isTypeName dckHash name) || (Updateconfig.isLoopFun name))) 
    (eraseAccessKinds deltaSet)
  in
    begin try
      let di = analysis_solution.Analysis.ar_c_of_id delta_id_in
      and dl = analysis_solution.Analysis.ar_c_of_id delta_id_out
      and vf = analysis_solution.Analysis.ar_c_of_id vc_fwd 
      and vb = analysis_solution.Analysis.ar_c_of_id vc_bck 
      and ve = analysis_solution.Analysis.ar_c_of_id vc_checkin_eps in

      let delta_in  = sanitize di
      and delta_out = sanitize dl

      and vfwd = eak vf 
      and veps = eak ve 
      and vbck = eak vb 
      in  
      let _ = ignore (Updatedebug.log [Updatedebug.BTUpdatepoints] 
                        "addCheckinToStatement sid %d:  delta_in(%d) '%s' delta_out(%d) '%s' vfwd(%d) '%s' vbck(%d) '%s' veps(%d) '%s'\n" 
                        sid
                        delta_id_in (Deltasolver.printDelta di)
                        delta_id_out (Deltasolver.printDelta dl)
                        vc_fwd (Deltasolver.printDelta vfwd)
                        vc_bck (Deltasolver.printDelta vbck)
                        vc_checkin_eps (Deltasolver.printDelta veps)
                     ) in
        begin 
          let u = Deltasolver.CapabilitySet.union in 

          let      pre_delta,         pre_fwd,       pre_bck = if Updateargs.args.Updateargs.arg_syncUpdates

            then   delta_in,          vfwd,          (u vbck veps)

            else   (u delta_in veps), (u vfwd veps), (u vbck veps)
          in
          let      post_delta,        post_fwd,      post_bck = 

                   delta_out,         (u vfwd veps), vbck
          in
          let pre_delta_string =  Deltasolver.printDelta pre_delta in
          let pre_fwd_string =    Deltasolver.printDelta pre_fwd in
          let pre_bck_string =    Deltasolver.printDelta pre_bck in
          let post_delta_string = Deltasolver.printDelta post_delta in
          let post_fwd_string =   Deltasolver.printDelta post_fwd in
          let post_bck_string =   Deltasolver.printDelta post_bck in

          let preId = (sid * 1000) + 1 in 
          let postId = (sid * 1000) + 2 in 
          let pre_info = " delta:" ^ (string_of_int delta_id_in) ^ " fwd:" ^ (string_of_int vc_fwd) ^ " bck:" ^ (string_of_int vc_bck) in
          let post_info = " delta:" ^ (string_of_int delta_id_out) ^ " fwd:" ^ (string_of_int vc_fwd) ^ " bck:" ^ (string_of_int vc_bck) in
          let () = H.add checkinIdToCapMap preId   (pre_delta_string,  pre_fwd_string,  pre_bck_string,  pre_info,  checkinLabel) in
          let () = H.add checkinIdToCapMap postId  (post_delta_string, post_fwd_string, post_bck_string, post_info, checkinLabel) in

          let () = UPInfo.addCheckin sid { UPInfo.pre_delta  = pre_delta;
                                           UPInfo.pre_fwd    = pre_fwd;
                                           UPInfo.pre_bck    = pre_bck;
                                           UPInfo.post_delta = post_delta;
                                           UPInfo.post_fwd   = post_fwd;
                                           UPInfo.post_bck   = post_bck;
                                         } 
          in


            (*           let pre_str = Updateconfig.delta_test_compat ""   *)
            (*             (pre_delta_string ^ " | " ^ pre_fwd_string ^ " | " ^ pre_bck_string *)
            (*              ^ " | "  ^ (string_of_int sid) ^ "-1" ^ " delta:" ^ (string_of_int delta_id_in) ^ " fwd:" ^ (string_of_int vc_fwd) ^ " bck:" ^ (string_of_int vc_bck) ) in *)
            (*           let post_str = Updateconfig.delta_test_compat ""   *)
            (*             (post_delta_string ^ " | " ^ post_fwd_string ^ " | " ^ post_bck_string *)
            (*              ^ " | "  ^ (string_of_int sid) ^ "-2"  ^ " delta:" ^ (string_of_int delta_id_out) ^ " fwd:" ^ (string_of_int vc_fwd) ^ " bck:" ^ (string_of_int vc_bck) ) in *)
            preId, postId
        end
    with 
        Not_found -> (* this shouldn't happen *) 
          Updatedebug.abort (Updatedebug.bug
                               "annotateUpdatePointsVisitor: couldn't find solution for delta_id %d (%a)\n" delta_id_out d_loc loc)
    end
;;


let addCheckinToStatement (analysis_solution : Analysis.analysis_result) 
                          (checkinFunction : Cil.varinfo) 
                          (hoistedCheckins : (string, string) H.t)
                          (hoistedCheckinAnnotations : (string, int) H.t)
                          (checkinIdToCapMap : checkinIdToCapMap_t)
                          (dckHash : (string, Analysis.deltaConstKind) H.t)
                          s = 
  let sid = s.sid in
  let checkinLabel =
    let labelList = Updateconfig.getCheckinLabel s in  
       let printOneLabel l acc = 
      acc ^ " " ^ l
   in
   List.fold_left printOneLabel "" labelList
  in
  let isHoisted =
    let res = ref None in
    H.iter (fun dsuLabel bareLabel -> 
              if (Updateconfig.stmtHasLabel (fun lb -> lb = dsuLabel) s)
              then res := Some bareLabel) hoistedCheckins;
    !res
  in  
  let loc = !Cil.currentLoc in (* this is valid since we're in a visitor *)
(*   let () = ignore (Updatedebug.log [Updatedebug.BTAlways] " %a\n " d_stmt s) in *)

  let preCheckinId, postCheckinId = getPrePostCheckinIds analysis_solution checkinIdToCapMap dckHash loc sid checkinLabel in  

    
  (* make an instruction containing the update() call *)
  let preCheckinI = match isHoisted with
      None -> 
        [Call(None, Lval(var checkinFunction), [Cil.integer preCheckinId], loc)] 
    | Some lb -> 
        H.add hoistedCheckinAnnotations lb preCheckinId;
        []        
  in
  let postCheckinI = Call(None, Lval(var checkinFunction), [Cil.integer postCheckinId], loc) in

  let () = ignore (Updatedebug.log [Updatedebug.BTUpdatepoints] "addCheckinToStatement : inserting checkins at (%a)\n" d_loc loc) in

  (* pack update () into a statement *)
  let preCheckinS = Commoncilutil.mkStmt (Instr preCheckinI) in
  let postCheckinS = Commoncilutil.mkStmt (Instr [postCheckinI]) in
  let () = ignore(Updatedebug.log [Updatedebug.BTUpdatepoints] "addCheckinToStatement : Generated new Instr statements with preID=%d postID=%d\n" preCheckinS.sid postCheckinS.sid) in

  (* put the statement id into the hash; needed later for computing # of update points *)
(*   let () = UPInfo.addUpdateStatement newSInstr.sid in *)

  (* pack the update() and the original statement into a block *)
  let newSBlock = 
    (* if arg_checkins is false, don't insert checkin()'s *)
    if Updateargs.args.Updateargs.arg_checkins
    then Commoncilutil.mkStmt (Block (mkBlock ([preCheckinS; s; postCheckinS]))) 
    else s
  in
  let () = ignore(Updatedebug.log [Updatedebug.BTUpdatepoints] "addCheckinToStatement : Generated new Block statement with ID=%d\n" newSBlock.sid) in
  newSBlock
;;


class insertCheckinCallsVisitor 
   (analysis_solution : Analysis.analysis_result)
   (checkin_function : Cil.varinfo)
   (hoistedCheckins : (string, string) H.t)
   (hoistedCheckinAnnotations : (string, int) H.t)
   (checkinIdToCapMap : checkinIdToCapMap_t)
   (dckHash : (string, Analysis.deltaConstKind) H.t)
   : cilVisitor = 
object(self)
   inherit nopCilVisitor
   
   method vstmt s =
    if Updateconfig.isCheckinStmt s 
    then
      (* insert here, don't descend; TODO: support nested checkins *)
      ChangeTo (addCheckinToStatement analysis_solution 
                                      checkin_function 
                                      hoistedCheckins 
                                      hoistedCheckinAnnotations
                                      checkinIdToCapMap
                                      dckHash
                                      s)
    else
      (* descend into the children *)
      DoChildren
end;; 

class fixHoistedCheckinAnnotationsVisitor
   (checkin_function : Cil.varinfo)
   (hoistedCheckinAnnotations : (string, int) H.t)
   : cilVisitor = 
object(self)
   inherit nopCilVisitor
   
   method vinst i = match i with
       Call(None, Lval(Var v,NoOffset), [Const(CStr s)], loc) when v.vname = checkin_function.vname ->

         (try
            let newAnnotation = H.find hoistedCheckinAnnotations s in
            ChangeTo [(Call(None, Lval(Var v,NoOffset), [Cil.integer newAnnotation], loc))]
          with Not_found ->
            SkipChildren
         )
     | _ ->
      SkipChildren
end;; 

class collectCheckinsVisitor (crtFun : string ref) (h : funNameToCheckinId_t) : cilVisitor = 
object(self)
  inherit nopCilVisitor
   
  method vstmt s =
    begin match Updateconfig.getCheckinLabel s with
        [] -> (* most common case *)
          ()

      | [l] -> (* one label is ok *) 
          ignore(Updatedebug.log [Updatedebug.BTAlways] "Found checkin '%s' in function '%s'\n" l !crtFun);
          let oldList = 
            try
              H.find h !crtFun
            with Not_found ->
              []
          in (* add one more checkin for this function *)
          H.replace h !crtFun ((s.sid, l)::oldList)

      | _ -> (* more than one is not *)
          Updatedebug.abort (Updatedebug.bug "collectCheckinsVisitor : more than one check-in label on this statement: \n\n%a  \n\n" d_stmt s)
    end;
    DoChildren
end;; 

let collectCheckinSites (file : Cil.file) : funNameToCheckinId_t = 
  let crtFun = ref "" in
  let res = H.create 13 in
  let vis = new collectCheckinsVisitor crtFun res in
  let doOneGlobal = function
      GFun(fd, _) -> 
        crtFun := fd.svar.vname;
        ignore (visitCilFunction vis fd)

    | _ -> ()
  in
  Cil.iterGlobals file doOneGlobal;
  res  
;;

(* Given a solution to the mu-constraints convert all of the speculative update points 
   to real ones iff their associated mu is U (true), otherwise remove them. *)
let annotateUpdatePoints (analysis_solution : Analysis.analysis_result) 
                         (file : Cil.file) 
                         (real_update:Cil.varinfo) : 
                         (UPInfo.upinfo_t * ((string, Analysis.deltaConstKind) H.t)) = 
   (* Was: if Updateargs.args.Updateargs.arg_updatePointsPolicy = Updateargs.UNone then H.create 3 else ...
      Now: Always run this, regardless of policy, as there might be user-inserted update points
   *)

    (* TODO: 7/12/7 code below throws a Not_found in patch comp;  *)

   (* create two statistics counters that are passed to visitors (need to pass by ref. as visitors can't return anything) *)
   let removalCount = ref 0 in
   let nonEmptyEffectRemovalCount = ref 0 in
   let totalBefore = Commonutil.hashtbl_size UPInfo.instance.UPInfo.theUpdateStatements in

   let newDckHash = fixDckHash file analysis_solution.Analysis.ar_constNamesHash in
   visitCilFile (new annotateUpdatePointsVisitor analysis_solution real_update nonEmptyEffectRemovalCount newDckHash) file;
   visitCilFile (new removeUnsafeUpdatePointsVisitor analysis_solution real_update removalCount) file;

   dealWithStateXform analysis_solution UPInfo.instance.UPInfo.theUpdateStatements newDckHash;

   (* visitors remove update points from the hash, so the leftovers are the safe ones *)
   let totalAfter = Commonutil.hashtbl_size UPInfo.instance.UPInfo.theUpdateStatements in
   ignore(Updatedebug.log [Updatedebug.BTStatistics] 
             "# Update points : Safe = %d, Speculative removed = %d, Non-empty effect removed = %d, Total candidates = %d\n" 
             totalAfter !removalCount !nonEmptyEffectRemovalCount totalBefore);
(*    printXformConflicts analysis_solution cStats; *)
(*    ignore(Updatedebug.log [Updatedebug.BTAlways]  *)
(*              "VC Update points : fwd more restrictive = %d, bck more restrictive = %d, fwd+bck more restrictive = %d, conflict with state xform = %d, total = %d\n"    *)
(*              cStats.vcFwdMoreRestrictive *)
(*              cStats.vcBckMoreRestrictive  *)
(*              cStats.vcMoreRestrictive *)
(*              cStats.conflictWithStateXform *)
(*              cStats.totalUpdatePoints); *)
   (UPInfo.instance, newDckHash)
;;
  
let insertCheckins (analysis_solution : Analysis.analysis_result) 
                   (file : Cil.file) 
                   (checkinFun:Cil.varinfo) 
                   (hoistedCheckins : (string, string) H.t) 
                   (dckHash : (string, Analysis.deltaConstKind) H.t)
                   : checkinIdToCapMap_t =
  let hoistedCheckinAnnotations = H.create 11 in
  let checkinIdToCapMap = H.create 11 in
  visitCilFile (new insertCheckinCallsVisitor analysis_solution 
                                              checkinFun 
                                              hoistedCheckins
                                              hoistedCheckinAnnotations
                                              checkinIdToCapMap 
                                              dckHash) file;
  (* fix annotation on hoisted checkins, if any *)
  visitCilFile (new fixHoistedCheckinAnnotationsVisitor checkinFun
                                                        hoistedCheckinAnnotations) file;
  checkinIdToCapMap   
;;






































(*                let conflictsWithStateXform =  *)
(*                  begin match analysis_solution.Analysis.ar_state_xform_eff_id with *)

(*                      None ->  (\* no xform *\) *)
(*                        false *)

(*                    | Some key ->  *)

(*                        let stateXformEff =  *)
(*                          try  *)
(*                            H.find analysis_solution.Analysis.ar_funEffectHash key *)
(*                          with Not_found ->  *)
(*                            Updatedebug.abort (Updatedebug.bug "annotateUpdatePointsVisitor: can't find key %d in funEffectHash\n" key) *)
(*                        in *)
(*                        let stateXformEps =  *)
(*                          let key_id = stateXformEff.Analysis.effect_vc_eps.Deltasolver.delta_id in *)
(*                            try  *)
(*                              analysis_solution.Analysis.ar_c_of_id key_id *)
(*                            with Not_found ->  *)
(*                              Updatedebug.abort (Updatedebug.bug "annotateUpdatePointsVisitor: can't find delta id %d in solution hash\n" key_id) *)
(*                        in *)
(*                          (\* If the transaction's effect \alpha does overlap with the update's effect \epsilon *)
(*                             such that a read in \epsilon conflicts with a write in \alpha then we  *)
(*                             have to delay the update. This is necessary because the state  *)
(*                             transformer code has read a location that was updated since  *)
(*                             the transaction started, violating our assumption that the update  *)
(*                             should appear as if it took place at the beginning or end of the  *)
(*                             transaction. *\) *)
(*                        let conflicts, conflictSet =  *)
(*                          let stateXformConflict epsilon alpha =  *)
(*                            let epsilonReads = Deltasolver.getReads epsilon *)
(*                            and alphaWrites  = Deltasolver.getWrites alpha in *)
(*                              Deltasolver.conflicts epsilonReads alphaWrites *)
(*                          in *)
(*                          stateXformConflict stateXformEps vfwd *)
(*                        in *)
(*                        Deltasolver.printSituationAtPoint sid [(conflicts, conflictSet, "state_xform")];   *)
(*                        conflicts   *)
(*                  end *)
(* (\*                      ((not (Deltasolver.CapabilitySet.is_empty (Deltasolver.CapabilitySet.inter stateXformEps vfwd))) && *\) *)
(* (\*                       (not (Deltasolver.CapabilitySet.is_empty (Deltasolver.CapabilitySet.inter stateXformEps vbck)))) *\) *)
(*                in *)
(*                let () =  *)
(*                  let strictSubset s1 s2 =  *)
(*                    (Deltasolver.CapabilitySet.subset s1 s2) &&  *)
(*                    (Deltasolver.CapabilitySet.cardinal s1 < Deltasolver.CapabilitySet.cardinal s2) *)
(*                  in *)
(*                  let ssf = strictSubset delta vfwd *)
(*                  and ssb = strictSubset delta vbck *)
(*                  in *)
(*                    begin *)
(*                      (if ssf then cStats.vcFwdMoreRestrictive <- cStats.vcFwdMoreRestrictive + 1); *)
(*                      (if ssb then cStats.vcBckMoreRestrictive <- cStats.vcBckMoreRestrictive + 1); *)
(*                      (if (ssf && ssb) then cStats.vcMoreRestrictive <- cStats.vcMoreRestrictive + 1); *)
(*                      (if (conflictsWithStateXform) then cStats.conflictWithStateXform <- cStats.conflictWithStateXform + 1); *)
(*                      (cStats.totalUpdatePoints <- cStats.totalUpdatePoints + 1); *)
(*                    end *)
(*                in   *)
