/* 
 *
 * Copyright (c) 2008-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Michael Hicks       <mwh@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 */


/* 
 * The OS/platform dependent part of the runtime system. 
 * 
 */
#include <dlfcn.h>
#include <signal.h>
#include <time.h>

#include "dsu_debug.h"
#include "dsu_config.h"
#include "fgc.h"
#include "machndep.h"
#include "pmeta.h"

#define MAX_LIBNAME_LENGTH 300

/*  Specifies the signal number to trigger an update. 
 *  N.B. Make sure this doesn't clash with the signals your program uses !!
 */
static int SIGUPDATE = SIGUSR2;

/*  Declaration of sig structure */
static struct sigaction update_action;

/* these come from updatelib */
extern volatile int update_requested;
extern perf_time_t update_request_time;
extern char * metadata_var_name;
extern perf_time_t load_times[];
extern perf_time_t app_times[];
extern char * init_fun_name;
/*
 *  Updating handler. It sets 'update_requested' to 1, so the first subsequent __DSU_update()
 *  will load the patch.
 */

#ifdef DSU_USE_THREADS
#include "mt.h"
#endif // DSU_USE_THREADS

static void updating_request_handler(int signum){

#ifdef DSU_USE_THREADS
/*   pthread_rwlock_wrlock(&map_mtx); */
#endif // DSU_USE_THREADS

  update_requested = 1 ;

#ifdef DSU_USE_THREADS
/*   pthread_rwlock_unlock(&map_mtx); */
#endif // DSU_USE_THREADS

  DSU_DEBUG("updating_request_handler called\n"); 
  __DSU_machndep_get_perf_time(&update_request_time); 

}

/*  This function should be called at the start of any updatable program
 *  in order to trap SIGUPDATE and route it to our updating_request_handler
 */
void __DSU_machndep_init()
{

  update_action.sa_handler = updating_request_handler;
  // MWH: changed from sigemptyset.  This ensures that no application
  // signals will be handled while we're updating; which could cause
  // inconsistent behavior (e.g., if a signal handler was being
  // updated).
  // MWH: do I need to use the SA_RESTART flag?
  sigfillset(&update_action.sa_mask);
  update_action.sa_flags=0;

  // linking SIGUPDATE to updating_request_handler
  sigaction(SIGUPDATE,&update_action,NULL);

}

void (*__DSU_addFVersion_addr)(struct __DSU_FVersionList *l,char *name, unsigned int f_versionNumber, unsigned int p_versionNumber);

/* try to load the patch and return the handle + metadata; if failure, return NULL */

#ifdef NOSPLIT
void * handle;
#endif

handle_t * __DSU_machndep_load_patch(int version, metadata_t * type_metadata, metadata_t * def_metadata)
{

#ifndef NOSPLIT
  handle_t handle;
#endif
  char libname0[MAX_LIBNAME_LENGTH];
  unsigned long long time1, time2;

#ifdef USE_PATCH_METADATA
  char *error;
#endif /* USE_PATCH_METADATA */



  /* gratuitous, but need to keep __DSU_addFVersionList 
     around for when loading the patches. 
     If __DSU_addFVersionList is not used, gcc will throw it away */

  __DSU_addFVersion_addr = &__DSU_addFVersionList;



  /* Path name convention for multiple patches
   *
   *          patch_directory/version_number/patch.so
   */
  snprintf(libname0, MAX_LIBNAME_LENGTH, "%s/%d/%s.patch.so", __DSU_libname, version, __DSU_progname);

  DSU_DEBUG("UPDATELIB::load_patch version %d  : Opening %s ...\n", version, libname0);

  __DSU_machndep_get_perf_time(&time1); 

  /* Open the .patch.so */
  handle = dlopen (libname0, RTLD_LAZY|RTLD_GLOBAL);

  if (!handle) {
    DSU_ERROR("UPDATELIB::load_patch version %d  : %s\n", version, dlerror());
    exit(1);
  }
  DSU_DEBUG("UPDATELIB::load_patch version %d  : Opened.\n", version);

#ifdef USE_PATCH_METADATA
  /* Look up type metadata */
  DSU_DEBUG("UPDATELIB::load_patch version %d  : Searching for symbol %s ...\n", version, type_metadata_var_name);

  *type_metadata = *((metadata_t *)dlsym(handle, type_metadata_var_name));

  if ((error = dlerror()) != NULL)  {
    DSU_ERROR("UPDATELIB::load_patch version %d  : %s\n", version, error);
    exit(1);
  }
  DSU_DEBUG("UPDATELIB::load_patch version %d  : %s found @ %p [%s ]\n", version, type_metadata_var_name, *metadata, *metadata);

  /* Look up def metadata */
  DSU_DEBUG("UPDATELIB::load_patch version %d  : Searching for symbol %s ...\n", version, def_metadata_var_name);

  *def_metadata = *((metadata_t *)dlsym(handle, def_metadata_var_name));

  if ((error = dlerror()) != NULL)  {
    DSU_ERROR("UPDATELIB::load_patch version %d  : %s\n", version, error);
    exit(1);
  }
  DSU_DEBUG("UPDATELIB::load_patch version %d  : %s found @ %p [%s ]\n", version, def_metadata_var_name, *metadata, *metadata);
#endif /* USE_PATCH_METADATA */

  __DSU_machndep_get_perf_time(&time2); 

  DSU_DEBUG("UPDATELIB::load_patch version %d  : load_patch finished in %lld ticks.\n",  version, (time2 - time1));

  load_times[version]= time2 - time1;

  return handle;
}


int __DSU_machndep_apply_patch(handle_t handle, int version)
{
  init_fun_t init_fun;
  unsigned long long time1, time2;
  char *error;

  __DSU_machndep_get_perf_time(&time1); 

  /* Look up __DSU_install_patch */
  DSU_DEBUG("UPDATELIB::apply_patch version %d  : Searching for symbol %s ...\n", version, init_fun_name);

  init_fun = (init_fun_t)dlsym(handle, init_fun_name);
  if ((error = dlerror()) != NULL)  {
    DSU_ERROR("UPDATELIB::apply_patch version %d  : %s\n", version, error);
    exit(1);
  }

  DSU_DEBUG("UPDATELIB::apply_patch version %d  : %s found \n", version, init_fun_name);

  /* invoke __DSU_install_patch */
  DSU_DEBUG("UPDATELIB::apply_patch version %d  : Handle %p, Init_fun %p Calling %s \n", version, 
            (void *)handle, 
            (void *)init_fun, 
            init_fun_name);
  init_fun();

  __DSU_machndep_get_perf_time(&time2); 

  DSU_DEBUG("UPDATELIB::apply_patch version %d  : apply_patch finished in %lld ticks.\n", version,  (time2 - time1));
  app_times[version]= time2 - time1;
  return 1;
}

/* Timing */

void __DSU_machndep_get_perf_time(perf_time_t * tm)
{
  asm volatile( "rdtsc" : "=A" (*tm) ); 
}


/* For x86, we must divide the timestamp counter to clocks per usec. */
/* Luckily, the CIL configuration step has computed this for us
   and cycles_per_usec is defined in the Makefile-generated machndep-from-cil.c
 */
extern double cycles_per_usec;

long long  __DSU_machndep_usec_of_perf_time(perf_time_t tm)
{
  return (long long) ((double)tm / cycles_per_usec);
}

void __DSU_machndep_sleep_nsec(long nsec)
{
  struct timespec ts;

  ts.tv_sec = 0;
  ts.tv_nsec = nsec;

  nanosleep(&ts, NULL);
}


/* Threads */

#ifdef DSU_USE_THREADS
int __DSU_mutex_init(__DSU_mutex_t * mtx)
{
  return pthread_mutex_init(mtx, NULL);
}

int __DSU_mutex_lock(__DSU_mutex_t * mtx)
{
  return pthread_mutex_lock(mtx);
}

int __DSU_mutex_trylock(__DSU_mutex_t * mtx)
{
  return pthread_mutex_trylock(mtx);
}

int __DSU_mutex_unlock(__DSU_mutex_t * mtx)
{
  return pthread_mutex_unlock(mtx);
}

__DSU_thread_t __DSU_thread_self()
{
  return pthread_self();
}

int __DSU_cond_init(__DSU_cond_t * cond)
{
  return pthread_cond_init(cond, NULL);
}

int __DSU_cond_wait(__DSU_cond_t * cond, __DSU_mutex_t * mtx)
{
  return pthread_cond_wait(cond, mtx);
}

int __DSU_cond_broadcast(__DSU_cond_t * cond)
{
  return pthread_cond_broadcast(cond);
}
#endif // DSU_USE_THREADS
