/* 
 *
 * Copyright (c) 2008-2009, 
 *  
 *  Eric Hardisty       <hardisty@cs.umd.edu>
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Michael Hicks       <mwh@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 */


/* 
 * The OS/platform dependent part of the runtime system. 
 * 
 */
#include <signal.h>
#include <time.h>

/* VxWorks location of dlopen and dlsym implementations.  target/usr/src/libdl/rtld.h */
#include <rtld.h>

#include "dsu_debug.h"
#include "dsu_config.h"
#include "fgc.h"
#include "machndep.h"

#define MAX_LIBNAME_LENGTH 300

static int SIGUPDATE = SIGUSR2;

static struct sigaction update_action;

/* these come from updatelib */
extern volatile int update_requested;
extern perf_time_t update_request_time;
extern char * metadata_var_name;
extern perf_time_t load_times[];
extern perf_time_t app_times[];
extern char * init_fun_name;

/* VxWorks seems to support the signal API directly. Good. */


static void updating_request_handler(int signum){
  update_requested = 1 ;
  DSU_DEBUG("updating_request_handler called\n"); 
  __DSU_machndep_get_perf_time(&update_request_time); 

}

void __DSU_machndep_init()
{

  update_action.sa_handler = updating_request_handler;
  sigfillset(&update_action.sa_mask);
  update_action.sa_flags=0;

  sigaction(SIGUPDATE,&update_action,NULL);

}

void (*__DSU_addFVersion_addr)(struct __DSU_FVersionList *l,char *name, unsigned int f_versionNumber, unsigned int p_versionNumber);

/* try to load the patch and return the handle + metadata; if failure, return NULL */

#ifdef NOSPLIT
void * handle;
#endif

/* TODO: copy/port the dynamic loading routines.  Implementations of dlopen
   and dlsym are in target/usr/src/libdl/rltd.[ch] */
handle_t * __DSU_machndep_load_patch(int version, metadata_t * metadata)
{

#ifndef NOSPLIT
  handle_t handle;
#endif
  char libname0[MAX_LIBNAME_LENGTH];
  unsigned long long time1, time2;

#ifdef USE_PATCH_METADATA
  char *error;
#endif /* USE_PATCH_METADATA */



  /* gratuitous, but need to keep __DSU_addFVersionList 
     around for when loading the patches. 
     If __DSU_addFVersionList is not used, gcc will throw it away */

  __DSU_addFVersion_addr = &__DSU_addFVersionList;



  /* Path name convention for multiple patches
   *
   *          patch_directory/version_number/patch.so
   */
  snprintf(libname0, MAX_LIBNAME_LENGTH, "%s/%d/%s.patch.so", __DSU_libname, version, __DSU_progname);

  DSU_DEBUG("UPDATELIB::load_patch version %d  : Opening %s ...\n", version, libname0);

  __DSU_machndep_get_perf_time(&time1); 

  /* Open the .patch.so */
  handle = dlopen (libname0, RTLD_LAZY|RTLD_GLOBAL);

  if (!handle) {
    DSU_ERROR("UPDATELIB::load_patch version %d  : %s\n", version, dlerror());
    exit(1);
  }
  DSU_DEBUG("UPDATELIB::load_patch version %d  : Opened.\n", version);

#ifdef USE_PATCH_METADATA
  /* Look up __DSU_metadata */
  DSU_DEBUG("UPDATELIB::load_patch version %d  : Searching for symbol %s ...\n", version, metadata_var_name);

  *metadata = *((metadata_t *)dlsym(handle, metadata_var_name)); 

  if ((error = dlerror()) != NULL)  {
    DSU_ERROR("UPDATELIB::load_patch version %d  : %s\n", version, error);
    exit(1);
  }
  DSU_DEBUG("UPDATELIB::load_patch version %d  : %s found @ %p [%s ]\n", version, metadata_var_name, *metadata, *metadata);
#endif /* USE_PATCH_METADATA */

  __DSU_machndep_get_perf_time(&time2); 

  DSU_DEBUG("UPDATELIB::load_patch version %d  : load_patch finished in %lld ticks.\n",  version, (time2 - time1));

  load_times[version]= time2 - time1;

  return handle;
}

int __DSU_machndep_apply_patch(handle_t handle, int version)
{
  init_fun_t init_fun;
  unsigned long long time1, time2;
  char *error;

  __DSU_machndep_get_perf_time(&time1); 

  /* Look up __DSU_install_patch */
  DSU_DEBUG("UPDATELIB::apply_patch version %d  : Searching for symbol %s ...\n", version, init_fun_name);

  init_fun = (init_fun_t)dlsym(handle, init_fun_name);
  if ((error = dlerror()) != NULL)  {
    DSU_ERROR("UPDATELIB::apply_patch version %d  : %s\n", version, error);
    exit(1);
  }

  DSU_DEBUG("UPDATELIB::apply_patch version %d  : %s found \n", version, init_fun_name);

  /* invoke __DSU_install_patch */
  DSU_DEBUG("UPDATELIB::apply_patch version %d  : Handle %p, Init_fun %p Calling %s \n", version, 
            (void *)handle, 
            (void *)init_fun, 
            init_fun_name);
  init_fun();

  __DSU_machndep_get_perf_time(&time2); 

  DSU_DEBUG("UPDATELIB::apply_patch version %d  : apply_patch finished in %lld ticks.\n", version,  (time2 - time1));
  app_times[version]= time2 - time1;
  return 1;
}

/* Timing */

/* see target/src/drv/timer/loApicTimer.c */
void __DSU_machndep_get_perf_time(perf_time_t * tm)
{
    /* _WRS_ASM breaks to __asm__ volatile in target/h/tools/gnu/toolMacros.h */
    /* _WRS_ASM( "rdtsc" : "=A" (*tm) );  */
    __asm__ volatile( "rdtsc" : "=A" (*tm) );
}

/* sysClkRateGet() returns ticks per second in target/h/time.h */
long long  __DSU_machndep_usec_of_perf_time(perf_time_t tm)
{
  double cycles_per_usec = sysClkRateGet()/1000000;
  return (long long) ((double)tm / cycles_per_usec);
}

/* VxWorks seems to support nanosleep directly. Good. */
void __DSU_machndep_sleep_nsec(long nsec)
{
  struct timespec ts;

  ts.tv_sec = 0;
  ts.tv_nsec = nsec;

  nanosleep(&ts, NULL);
}


/* Threads; VxWorks seems to support the pthreads API. Good. */

#ifdef DSU_USE_THREADS
int __DSU_mutex_init(__DSU_mutex_t * mtx)
{
  return pthread_mutex_init(mtx, NULL);
}

int __DSU_mutex_lock(__DSU_mutex_t * mtx)
{
  return pthread_mutex_lock(mtx);
}

int __DSU_mutex_trylock(__DSU_mutex_t * mtx)
{
  return pthread_mutex_trylock(mtx);
}

int __DSU_mutex_unlock(__DSU_mutex_t * mtx)
{
  return pthread_mutex_unlock(mtx);
}

__DSU_thread_t __DSU_thread_self()
{
  return pthread_self();
}
#endif // DSU_USE_THREADS
