/* 
 *
 * Copyright (c) 2006-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "machndep.h"
#include "mt.h"
#include "pmeta.h"
#include "thread_map.h"
#include "dsu_types.h"

#ifdef DSU_USE_THREADS

#define SLEEP_NANOSEC 50000000

static int thread_count = 0;
pthread_rwlock_t restriction_mtx;
__DSU_mutex_t update_mtx;

static pthread_key_t tl_key;

#if (RUNTIME_SYSTEM == RUNTIME_SYSTEM_4) || \
    (RUNTIME_SYSTEM == RUNTIME_SYSTEM_5) || \
    (RUNTIME_SYSTEM == RUNTIME_SYSTEM_6) || \
    (RUNTIME_SYSTEM == RUNTIME_SYSTEM_8)
pthread_mutex_t cond_mtx;
pthread_cond_t  cond;
#endif

void __DSU_mt_init()
{
  pthread_rwlock_init(&restriction_mtx, NULL);

  __DSU_mutex_init(&update_mtx);

#if (RUNTIME_SYSTEM == RUNTIME_SYSTEM_5) || (RUNTIME_SYSTEM == RUNTIME_SYSTEM_6)
  __DSU_mutex_init(&cond_mtx);
  pthread_cond_init(&cond, NULL);
#endif

  pthread_key_create(&tl_key, NULL);

  tm_init();
  DSU_DEBUG("__DSU_mt_init()\n");
}

inline static int get_my_index()
{
  tl_data * tl_ptr;
  tl_ptr = pthread_getspecific(tl_key);

  return tl_ptr->index;
}


void thread_register(const char * s)
{
  tl_data * tl_ptr;
  int index;
  
  tl_ptr = pthread_getspecific(tl_key);
  
  if (tl_ptr != NULL) {
    DSU_DEBUG("thread '%s' (%lx): thread local data already inited !?\n", s, __DSU_thread_self());
  }

  tl_ptr = (tl_data *)malloc(sizeof(tl_data));
  pthread_setspecific(tl_key, tl_ptr);

  pthread_rwlock_wrlock(&restriction_mtx);


  thread_count++;
  index = tm_add_thread(s);

  // save index in restriction array for easy use later on;
  tl_ptr->index = index;

  pthread_rwlock_unlock(&restriction_mtx);

  DSU_ERROR("thread '%s' (%lx) registered at index %d, setspecific %p, getspecific %p name back '%s'\n", s, __DSU_thread_self(), index, tl_ptr, pthread_getspecific(tl_key), tm_get_thread_name(get_my_index()));

/*   DSU_ERROR("thread '%s' (%lx) registered, got back '%s'\n",  */
/*             s, __DSU_thread_self(), tm_get_thread_name(__DSU_thread_self())); */
}

void thread_deregister(const char * s)
{
  pthread_rwlock_wrlock(&restriction_mtx);

  thread_count--;
  tm_remove_thread(get_my_index(), s);

  pthread_rwlock_unlock(&restriction_mtx);

  DSU_ERROR("thread '%s' (%lx) deregistered\n", s, __DSU_thread_self());
}


void __DSU_yield(void)
{
  /* In the future, it should be possible to make this a sleep, and
     indeed to imagine an adaptive protocol that moves from yielding
     to sleeping more and more.  Perhaps including some level of
     randomization in the choice of sleep times?  Not sure. */
  sched_yield();
}


void leader()
{
  int retc;
  DSU_DEBUG("checkin: thread '%s' is leader\n", tm_get_thread_name(get_my_index()));

  /* YES, we're first (we're leader), try to do the update */

  pthread_rwlock_wrlock(&restriction_mtx);
  retc = __DSU_update("");

#if (RUNTIME_SYSTEM == RUNTIME_SYSTEM_5) || (RUNTIME_SYSTEM == RUNTIME_SYSTEM_6)
  tm_reset_checkedin_flags();
#endif

  pthread_rwlock_unlock(&restriction_mtx);
      
  __DSU_mutex_unlock(&update_mtx);
/*   if (!retc) */
/*     __DSU_yield(); */
  DSU_DEBUG("checkin: leader '%s' is out\n", tm_get_thread_name(get_my_index()));
}

void follower()
{
  DSU_DEBUG("checkin: thread '%s' is follower\n", tm_get_thread_name(get_my_index()));

/*   if (update_pending) { */

    DSU_DEBUG("checkin: thread '%s' (%lx) waiting for update to finish...\n", tm_get_thread_name(get_my_index()), __DSU_thread_self());

    //__DSU_machndep_sleep_nsec(1000UL * FOLLOWER_WAIT_USEC);
/*     __DSU_yield(); */

    DSU_DEBUG("checkin: thread '%s' (%lx) done with waiting.\n", tm_get_thread_name(get_my_index()), __DSU_thread_self());
/*   } */
    
}

/*
 OK, here's the checkin/sync logic:
 If an update has been requested, the first thread to notice this
 acquires the update_mtx and checks for conflicts
 If other threads get to the check-in point while an update is
 requested/pending, they wait.
 If no update has been requested or is pending, a checkin is just a checkin

 */

extern long long conflicting_checkins;
extern long long protocol_rounds;

void  deal_with_pending_update()
{
  if (update_requested) {
    DSU_DEBUG("checkin: thread at index %d ('%s') : update pending\n", get_my_index(), tm_get_thread_name(get_my_index()));

    /* OK, update requested; are we the first thread to notice this ? */
    if (tm_conflicts_with_one_thread(get_my_index(), get_current_changed_types(), get_current_changed_defs())) {
      conflicting_checkins++;
    }

    if (__DSU_mutex_trylock(&update_mtx) == 0) {
      protocol_rounds++;
      leader();
    }
    else {
      /* NO, we're not first (we're follower) */
      follower();
    }
  }  
  else {
    /* no pending update, do nothing */
    DSU_DEBUG("checkin: thread at index %d ('%s') : no pending update\n", get_my_index(), tm_get_thread_name(get_my_index()));
  }
}


/***************     RS_3     *******************/

/*  RS_3

    We stop for a bit (yield) at the end of each check-in
    block to see whether the other threads are blocked too.
    If all threads are blocked, try to do the update.
    We only check delta's, no VC.

 */

#if (RUNTIME_SYSTEM == RUNTIME_SYSTEM_3)

static int blocked_threads = 0;

void __DSU_start_checkin_sleep(void)
{
  /* NOTE: could use a different mutex to protect blocked_threads and
     thread_count, but then you have to remember to grab the
     restrition_mutex prior to actually doing the update, below. */
  pthread_rwlock_wrlock(&restriction_mtx);

  blocked_threads++;

  if (blocked_threads == thread_count) {
    DSU_DEBUG("jackpot: all %d threads blocked\n", blocked_threads);
    if (update_requested) {
      __DSU_update("");
    }
  }
  pthread_rwlock_unlock(&restriction_mtx);
}

void __DSU_end_checkin_sleep(void)
{
  pthread_rwlock_wrlock(&restriction_mtx);

  blocked_threads--;

  pthread_rwlock_unlock(&restriction_mtx);
}

void __DSU_checkin(int checkin_id)
{
  pthread_rwlock_rdlock(&restriction_mtx);

  tm_checkin_thread(get_my_index(), checkin_id);

  pthread_rwlock_unlock(&restriction_mtx);

  __DSU_start_checkin_sleep();
  __DSU_yield();
/*   __DSU_machndep_sleep_nsec(1000UL *1000UL *1000UL); */
  __DSU_end_checkin_sleep();


  DSU_DEBUG("thread '%s' (%lx) check-in\n", tm_get_thread_name(get_my_index()), __DSU_thread_self());
}

/*******************    RS_1 / RS_2   *******************/
#elif (RUNTIME_SYSTEM == RUNTIME_SYSTEM_1) || (RUNTIME_SYSTEM == RUNTIME_SYSTEM_2)

/*  RS_1  

    Regular update protocol

*/


/*  RS_2               

    Regular update protocol, but we only check delta's, no VC
    (tm_conflicts_with() behavior is ifdef'd on RUNTIME_SYSTEM)

*/


void __DSU_checkin(int checkin_id)
{
  pthread_rwlock_rdlock(&restriction_mtx);

  tm_checkin_thread(get_my_index(), checkin_id);

  pthread_rwlock_unlock(&restriction_mtx);

  deal_with_pending_update();

  DSU_DEBUG("thread '%s' (%lx) check-in\n", tm_get_thread_name(get_my_index()), __DSU_thread_self());
}

#elif (RUNTIME_SYSTEM == RUNTIME_SYSTEM_4)
/***************     RS_4     *******************/

/*  RS_4  

    Check-ins are ignored, unless an update has been requested,
    in which case we wait until all threads have checked in
    their restrictions prior to performing an update.

*/


void __DSU_checkin(int checkin_id)
{

  if (!update_requested)
    return;

  else {
    int index = get_my_index();
    int maybe_update;

    pthread_rwlock_rdlock(&restriction_mtx);

    tm_checkin_thread(index, checkin_id);
    tm_set_checkedin_flag(index);

    maybe_update = tm_all_checked_in();
    pthread_rwlock_unlock(&restriction_mtx);

    if (maybe_update) {
      deal_with_pending_update();
    }
  }

  DSU_DEBUG("thread '%s' (%lx) check-in\n", tm_get_thread_name(get_my_index()), __DSU_thread_self());
}

#elif (RUNTIME_SYSTEM == RUNTIME_SYSTEM_5) || (RUNTIME_SYSTEM == RUNTIME_SYSTEM_6)

/***************     RS_5/RS_6     *******************/

/*  RS_5               

    After an update has been requested, the first non-conflicting
    thread barrier-blocks, waiting for the others.
    (tm_conflicts_with() behavior is ifdef'd on RUNTIME_SYSTEM)

*/



/*  RS_6 = RS_5 + VC check

*/
int threads_checked_in = 0;

void __DSU_checkin(int checkin_id)
{

  if (!update_requested)
    return;

  else {
    int index = get_my_index();

    pthread_rwlock_rdlock(&restriction_mtx);

    tm_checkin_thread(index, checkin_id);

    if (! tm_conflicts_with_one_thread(index, get_current_changed_types(), get_current_changed_defs())) {
      // OK, this thread doesn't have a conflict
      // release read lock
      pthread_rwlock_unlock(&restriction_mtx);
      // acquire it in write mode to bump up threads_checked_in
      pthread_rwlock_wrlock(&restriction_mtx);
      threads_checked_in++;
      if (threads_checked_in == thread_count) {
        // everybody's in, do the update
        __DSU_update("");
        threads_checked_in = 0; // ready for next round
        pthread_cond_broadcast(&cond);
        pthread_rwlock_unlock(&restriction_mtx);
      }
      else {
        // not everybody's in yet, wait
        pthread_rwlock_unlock(&restriction_mtx);
        pthread_mutex_lock(&cond_mtx);
        pthread_cond_wait(&cond, &cond_mtx);
        pthread_mutex_unlock(&cond_mtx);
      }
    }
    else {
      // this thread has a conflict; keep going.
      conflicting_checkins++;
      pthread_rwlock_unlock(&restriction_mtx);
    }
  }

  DSU_DEBUG("thread '%s' (%lx) check-in\n", tm_get_thread_name(get_my_index()), __DSU_thread_self());
}




/***************     RS_7     *******************/

/*  like RS_3, but only __DSU_semantic_update can trigger an update


 */

#elif (RUNTIME_SYSTEM == RUNTIME_SYSTEM_7)

static int blocked_threads = 0;

void __DSU_start_checkin_sleep(void)
{
  /* NOTE: could use a different mutex to protect blocked_threads and
     thread_count, but then you have to remember to grab the
     restrition_mutex prior to actually doing the update, below. */
  pthread_rwlock_wrlock(&restriction_mtx);

  blocked_threads++;

  if (blocked_threads == thread_count) {
    DSU_DEBUG("jackpot: all %d threads blocked\n", blocked_threads);
    if (update_requested) {
      __DSU_update("");
    }
  }
  pthread_rwlock_unlock(&restriction_mtx);
}

void __DSU_end_checkin_sleep(void)
{
  pthread_rwlock_wrlock(&restriction_mtx);

  blocked_threads--;

  pthread_rwlock_unlock(&restriction_mtx);
}

void __DSU_checkin(int checkin_id)
{
  pthread_rwlock_rdlock(&restriction_mtx);

  tm_checkin_thread(get_my_index(), checkin_id);

  pthread_rwlock_unlock(&restriction_mtx);

  DSU_DEBUG("thread '%s' (%lx) check-in\n", tm_get_thread_name(get_my_index()), __DSU_thread_self());
}

void __DSU_semantic_update()
{
  __DSU_start_checkin_sleep();
  __DSU_yield();
/*   __DSU_machndep_sleep_nsec(1000UL *1000UL *1000UL); */
  __DSU_end_checkin_sleep();


  DSU_DEBUG("thread '%s' (%lx) semantic update\n", tm_get_thread_name(get_my_index()), __DSU_thread_self());
}

/***************     RS_8     *******************/

/*  like RS_6, but only __DSU_semantic_update can trigger an update

 */

#elif (RUNTIME_SYSTEM == RUNTIME_SYSTEM_8)


int threads_checked_in = 0;

void __DSU_checkin(int checkin_id)
{

  if (!update_requested)
    return;

  else {
    int index = get_my_index();

    pthread_rwlock_rdlock(&restriction_mtx);

    tm_checkin_thread(index, checkin_id);

    pthread_rwlock_unlock(&restriction_mtx);
  }

  DSU_DEBUG("thread '%s' (%lx) check-in\n", tm_get_thread_name(get_my_index()), __DSU_thread_self());
}

void __DSU_semantic_update()
{

  if (!update_requested)
    return;

  else {
    int index = get_my_index();

    pthread_rwlock_rdlock(&restriction_mtx);

    if (! tm_conflicts_with_one_thread(index, get_current_changed_types(), get_current_changed_defs())) {
      // OK, this thread doesn't have a conflict
      // release read lock
      pthread_rwlock_unlock(&restriction_mtx);
      // acquire it in write mode to bump up threads_checked_in
      pthread_rwlock_wrlock(&restriction_mtx);
      threads_checked_in++;
      if (threads_checked_in == thread_count) {
        // everybody's in, do the update
        __DSU_update("");
        threads_checked_in = 0; // ready for next round
        pthread_cond_broadcast(&cond);
        pthread_rwlock_unlock(&restriction_mtx);
      }
      else {
        // not everybody's in yet, wait
        pthread_rwlock_unlock(&restriction_mtx);
        pthread_mutex_lock(&cond_mtx);
        pthread_cond_wait(&cond, &cond_mtx);
        pthread_mutex_unlock(&cond_mtx);
      }
    }
    else {
      // this thread has a conflict; keep going.
      conflicting_checkins++;
      pthread_rwlock_unlock(&restriction_mtx);
    }
  }

  DSU_DEBUG("thread '%s' (%lx) semantic update\n", tm_get_thread_name(get_my_index()), __DSU_thread_self());
}




#else
#error "You must select a runtime system by defining #RS_n"
#endif





// this is called from DSU_update while holding update_mtx 
int conflicts_with_any_thread(char ** changed_types, char ** changed_defs)
{
  int ret;


/*   pthread_rwlock_wrlock(&restriction_mtx); */

  ret = tm_conflicts_with(changed_types, changed_defs);

/*   pthread_rwlock_unlock(&restriction_mtx); */

  
  //DSU_DEBUG("conflicts_with_any_thread (%s) %d\n", tm_get_thread_name(__DSU_thread_self()), ret);
  return ret;
}


#endif // DSU_USE_THREADS


/************* NOT USED **********************/

/* void __DSU_start_blocking_call(void) */
/* { */
/*   pthread_rwlock_wrlock(&restriction_mtx); */

/*   blocked_threads++; */

/*   if (blocked_threads == thread_count) { */
/*     DSU_ERROR("jackpot: all %d threads blocked\n", blocked_threads); */
/*   } */
/*   pthread_rwlock_unlock(&restriction_mtx); */
/* } */

/* void __DSU_end_blocking_call(void) */
/* { */
/*   pthread_rwlock_wrlock(&restriction_mtx); */

/*   blocked_threads--; */

/*   pthread_rwlock_unlock(&restriction_mtx); */
/* } */
