/* 
 *
 * Copyright (c) 2005-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Chris Hayden        <hayden@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 */



#include <stdlib.h>
#include <string.h>

#include "dsu_config.h"
#include "dsu_debug.h"
#include "pmeta.h"

struct patch_metadata
{
  char ** changed_types; 
  char ** changed_defs; 
};

#define MAX_VERSIONS 20
struct patch_metadata metadata_array[MAX_VERSIONS];

static int current_version = 0;

/* Name of the global variables containing the metadata (should be found in the dynamic patch file).
 * Note: if you change these, then change 
   Updateconfig.[type|def]_metadata_var_name/type too!!!
 */
char * type_metadata_var_name = "__DSU_type_metadata";
char * def_metadata_var_name = "__DSU_def_metadata";


#define MAXTOKENS  4096

/* split string into tokens, return token array */
char **split(const char *string, const char *delim) {
  char **tokens = NULL;
  char *working = NULL;
  char *token = NULL;
  int idx = 0;

  if (string == NULL)
    return NULL;

  tokens  = malloc(sizeof(char *) * MAXTOKENS);
  if(tokens == NULL)
    return NULL;
  working = malloc(sizeof(char) * strlen(string) + 1);
  if(working == NULL)
    return NULL;

  /* to make sure, copy string to a safe place */
  strcpy(working, string);
  for(idx = 0; idx < MAXTOKENS; idx++)
    tokens[idx] = NULL;

  token = strtok(working, delim);
  idx = 0;

  /* always keep the last entry NULL termindated */
  while((idx < (MAXTOKENS - 1)) && (token != NULL)) {
    tokens[idx] = malloc(sizeof(char) * strlen(token) + 1);
    if(tokens[idx] != NULL) {
      strcpy(tokens[idx], token);
      idx++;
      token = strtok(NULL, delim);
    }
  }

  free(working);

  /* this makes deep copies, so don't bother copying forbidden_functions */
  return tokens;
}

/* free token array allocated by split above */
void free_tokens(char **tokens)
{
  int i;

  if (tokens == NULL) 
    return;

  for(i = 0; ((i < MAXTOKENS) && (tokens[i] != NULL)); i++)
    free(tokens[i]);

  free(tokens);
}

int is_present(char * fun_name, char ** forbidden_tokens)
{
  int i;

  for(i = 0; ((i < MAXTOKENS) && (forbidden_tokens[i] != NULL)); i++)
    {
      if (strcmp(fun_name, forbidden_tokens[i]) == 0) {
        DSU_DUMP(" #%s# ", fun_name);
        return 1;
      }
    }  
  
  /* not found */
  return 0;
}

extern void print_threads();

void new_metadata(char * changed_types, char * changed_defs, int version)
{
  char *delim = " ";
  char **ttokens = NULL;
  char **dtokens = NULL;

  DSU_ERROR("new_metadata\n \t types (%p, '%s')\n \t defs (%p, '%s') \n", 
            changed_types, changed_types,
            changed_defs, changed_defs);

  current_version = version;

  /* this makes deep copies, so don't bother copying changed_syms */
  ttokens = split(changed_types, delim);
  metadata_array[version].changed_types = ttokens;

  /* this makes deep copies, so don't bother copying changed_syms */
  dtokens = split(changed_defs, delim);
  metadata_array[version].changed_defs = dtokens;

  if (version > 0) {
    free_tokens(metadata_array[version-1].changed_types);
    free_tokens(metadata_array[version-1].changed_defs);
  }


  print_threads();
}

int disjoint(char **tokens, char ** forbidden_tokens)
{
  int i;
  int ret = 1;

  if (tokens == NULL)
    return 1;

  for(i = 0; tokens[i]; i++)
    {
      if (is_present(tokens[i], forbidden_tokens))
        /* found, not_allowed */
        ret = 0;
    }

  /* no one found, update allowed */
  return ret;
}

/* Return 0 if sets not disjoint, 1 otherwise */
#ifndef DSU_USE_VC
int is_allowed_update(char * modified_functions)
{
  int ret = 0;


/* #ifdef DSU_USE_VC */
/*   char *delim = " "; */
/*   char **tokens = NULL; */
/*   char **dsu_tokens = NULL; */
/*   char **vc_fwd_tokens = NULL; */
/*   char **vc_bck_tokens = NULL; */
/*   char **update_point_id = NULL; */
/*   char * up_id = "??"; */
/*   tokens = split(modified_functions, "|"); // OK, got <DSU | FWD | BCK | ID> strings in tokens[0,1,2,3] */

/*   dsu_tokens      = split(tokens[0], delim); */
/*   vc_fwd_tokens   = split(tokens[1], delim); */
/*   vc_bck_tokens   = split(tokens[2], delim); */
/*   update_point_id = split(tokens[3], delim); */

/*   char ** forbidden_delta_tokens = metadata_array[current_version].changed_symbols; */
/*   char ** forbidden_vc_tokens    = metadata_array[current_version].changed_defs; */

/*   if (update_point_id != NULL) */
/*     up_id = update_point_id[0]; */

/*   if (! disjoint(dsu_tokens, forbidden_delta_tokens)) */
/*     { */
/*       DSU_DEBUG("is_allowed_update @ %s: update forbidden by dsu\n", up_id); */
/*       ret = 0; */
/*       goto end; */
/*     } */

/*   DSU_DEBUG("is_allowed_update @ %s: update permitted by dsu\n", up_id); */

/*   if (disjoint(vc_fwd_tokens, forbidden_vc_tokens)) */
/*     { */
/*       DSU_DEBUG("is_allowed_update @ %s: update permitted by VC forward\n", up_id); */
/*       ret = 1; */
/*       goto end; */
/*     } */
  
/*   if (disjoint(vc_bck_tokens, forbidden_vc_tokens)) */
/*     { */
/*       DSU_DEBUG("is_allowed_update @ %s: update permitted by VC backward\n", up_id); */
/*       ret = 1; */
/*       goto end; */
/*     } */

/*   DSU_DEBUG("is_allowed_update @ %s: update forbidden by VC\n", up_id); */
/*   ret = 0; */

/*  end: */

/*   free_tokens(tokens); */
/*   free_tokens(dsu_tokens); */
/*   free_tokens(vc_fwd_tokens); */
/*   free_tokens(vc_bck_tokens); */
/*   free_tokens(update_point_id); */


/* #else // classic DSU */
  tokens = split(modified_functions, delim);

  ret = disjoint(tokens);

  free_tokens(tokens);

/* #endif //DSU_USE_VC */
 
  return ret;
}
#endif // DSU_USE_VC

void print_string_array(char ** array)
{
  int i; 

  if (array == NULL) return;

  for(i = 0; array[i]; i++)
    {
      DSU_DUMP("%s ", array[i]);
    }
}

void print_string_array_error(char ** array)
{
  int i; 

  if (array == NULL) {
    fprintf(__DSU_debug_file,"NULL");
    return;
  }

  for(i = 0; array[i]; i++) {
    fprintf(__DSU_debug_file,"%s ", array[i]);
  }
}

char ** get_current_changed_types()
{
  return metadata_array[current_version].changed_types;
}


char ** get_current_changed_defs() 
{
  return metadata_array[current_version].changed_defs;
}


/* code below is for testing purposes only */

/*
int expect (int cond, int test_nr)
{
  if (cond)
    printf("Test %d passed\n", test_nr);
  else
    printf("Test %d failed\n", test_nr);
}


typedef char * pchar;
int main(int argc, char ** argv)
{
  int version;
  int tests;

  pchar forbidden = " foo bar baz";

#ifdef DSU_USE_VC
  tests = 5;
  pchar modified[] = {" innie minnie miny | foo bar baz | innie minnie miny ", // OK by dsu/vc_bck 
                      "  | bar | baz ",                                        // fail by vc_fwd/vc_bck
                      " foo | |",                                              // fail by dsu
                      "|  moe | foo bar baz ",                                 // OK by dsu/vc_fwd        
                      "||"};                                                   // OK by all
  int expected_results[] = {1,0,0,1,1};
#else

  tests = 3;

  pchar modified[] = {" innie minnie miny ", " foo ", " "};
  int expected_results[] = {1,0,1};

#endif // DSU_USE_VC


  for (version = 0; version < tests; version++ )
    {
      new_metadata(forbidden, version);
      expect(is_allowed_update(modified[version]) == expected_results[version], version);
    }

  return 0;
  
}
*/
