/* 
 *
 * Copyright (c) 2007-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 */



#include <stdlib.h>
#include <search.h>
#include <string.h>
#include <sys/time.h>
#include "tinymap.h"
#include "pmeta.h"
#include "dsu_debug.h"
#include "machndep.h"
#include "thread_map.h"
#include "mt.h"

#ifdef DSU_USE_THREADS
static void * cap_hash = NULL;


typedef void (*visit_fun_t)(const void *, const VISIT, const int);

typedef struct  {
  int used;
  char * name;
  cid_t checkin_id;
/*   struct timeval last_checkin; */
  int is_checked_in;
} thread_data;

typedef struct  {
  char ** cap_delta;
  char ** cap_fwd;
  char ** cap_bck;
  char ** checkin_info;
  char  * checkin_label;
} cap_data;



thread_data restriction[MAX_THREADS]; // keeps thread info

int restriction_wm = 0; // high watermark

void set_unused(thread_data * td)
{
  td->used = 0;
}

void restriction_iter(void (f)(thread_data * td))
{
  int i;

  for(i = 0; i < restriction_wm; i++) {
    thread_data * td = &restriction[i];
    if (td->used) {
    (*f)(td);
    }
  }
}

int restriction_get_next_unused()
{
  int i;

  for(i = 0; i < MAX_THREADS; i++) {

    if (! (restriction[i].used)) {
      // OK, found an unused entry
      if (i >= restriction_wm) {
        // bump up watermark
        restriction_wm = i + 1;
      }
      return i;
    }
  }

  DSU_ERROR( "restriction_get_next_unused : too many threads\n");
  exit(1);
}

void tm_init()
{
  cap_hash = mCreateMap();

  restriction_iter(set_unused);
}


int tm_add_thread(const char * s)
{
  int index = restriction_get_next_unused();
  thread_data * new_data = &restriction[index];
  size_t name_len = strlen(s);

  // mark used
  new_data->used = 1;

  //init name
  new_data->name = malloc(name_len + 1);
  strncpy(new_data->name, s, name_len);
  new_data->name[name_len] = 0;

  //init checkin_id
  new_data->checkin_id = 0;
  new_data->is_checked_in = 0;

  // init last checkin
/*   gettimeofday(&(new_data->last_checkin), NULL); */

  return index;
}

void __DSU_addcap(cid_t checkin_id, const char * delta, const char * fwd, const char * bck, const char * info, const char * checkin_label)
{
  int found = 0;
  cap_data * cdata = mFind(cap_hash,  checkin_id);

  if (cdata == NULL) {
    //not found
    cdata = malloc(sizeof(cap_data));
  }
  else {
    // found; make way for new caps
    found = 1;
    free_tokens(cdata->cap_delta);
    free_tokens(cdata->cap_fwd);
    free_tokens(cdata->cap_bck);
    free_tokens(cdata->checkin_info);
    free(cdata->checkin_label);
  }

  const char *delim = " ";

  char **delta_tokens = NULL;
  char **fwd_tokens = NULL;
  char **bck_tokens = NULL;
  char **info_tokens = NULL;

  delta_tokens    = split(delta, delim);
  fwd_tokens      = split(fwd, delim);
  bck_tokens      = split(bck, delim);
  info_tokens     = split(info, delim);

/*     DSU_DEBUG("tm_checkin_thread: delta:"); */
/*     print_string_array(delta_tokens); */
/*     DSU_DEBUG("\ntm_checkin_thread: fwd:"); */
/*     print_string_array(delta_tokens); */
/*     DSU_DEBUG("\ntm_checkin_thread: bck:"); */
/*     print_string_array(delta_tokens); */
/*     DSU_DEBUG("\n"); */

/*     update_point_id = split(tokens[3], delim); */


/*   /\* make way for the new ones *\/ */
/*   free_tokens(cdata->cap_delta); */
/*   free_tokens(cdata->cap_fwd); */
/*   free_tokens(cdata->cap_bck); */
/*   free_tokens(cdata->checkin_info); */

    /* and set them */
  cdata->cap_delta    = delta_tokens;
  cdata->cap_fwd      = fwd_tokens;
  cdata->cap_bck      = bck_tokens;
  cdata->checkin_info = info_tokens;

  size_t label_len = strlen(checkin_label);
  cdata->checkin_label = malloc(label_len + 1);
  strncpy(cdata->checkin_label, checkin_label, label_len);
  cdata->checkin_label[label_len] = 0;


  if (!found)
    mAdd(cap_hash, checkin_id, cdata);

  DSU_ERROR("__DSU_addcap (%d, %s)\n", checkin_id, cdata->checkin_label);
}



void tm_remove_thread(int index, const char * s)
{
  int i;
  thread_data * tdata = &restriction[index];

  if (tdata->used) {

    if (tdata->name != NULL)
      free(tdata->name);

    tdata->used = 0;

    // adjust watermark down if necessary
    for(i = MAX_THREADS - 1; i >= 0; i--) {
      if (restriction[i].used) {
        // OK, found first used entry
        restriction_wm = i;
        break;
      }
    }
    
  }
  else {
    DSU_ERROR( "tm_remove_thread: tried to remove non-existent thread at index %d\n", index);
    exit(1);
  }
}

void tm_checkin_thread(int index, cid_t checkin_id)
{
  thread_data * tdata = &restriction[index];

/*   DSU_ERROR_NOPREFIX( "CHECKIN(%d,%d)\n", index, checkin_id); */

  if (tdata->used) {

    /* the checkin we're now at */
    tdata->checkin_id = checkin_id;
    tdata->is_checked_in = 1;

    /* update last checkin*/
/*     gettimeofday(&(tdata->last_checkin), NULL); */
  }
  else {
    DSU_ERROR( "tm_checkin_thread: tried to access non-existent thread at index %d\n", index);
    exit(1);
  }
}

void tm_set_checkedin_flag(int index)
{
  thread_data * tdata = &restriction[index];

  if (tdata->used) {

    tdata->is_checked_in = 1;
  }
  else {
    DSU_ERROR( "tm_set_is_checked_in_flag: tried to access non-existent thread at index %d\n", index);
    exit(1);
  }
}

void reset_checkedin(thread_data * td)
{
  td->is_checked_in = 0;
}

void tm_reset_checkedin_flags()
{
  restriction_iter(reset_checkedin);
}

int tm_all_checked_in()
{
  int i;

  for(i = 0; i < restriction_wm; i++) {
    thread_data * td = &restriction[i];
    if (td->used) {
      // if one is not checked in, that's enough 
      if (!td->is_checked_in) {
        return 0;
      }
    }
  }
  // all checked in
  return 1;
}


char * tm_get_thread_name(int index)
{
  thread_data * tdata = &restriction[index];

  if (tdata->used) {

    return tdata->name;
  }
  else {
    DSU_ERROR( "tm_get_thread_name: tried to access non-existent thread at index %d\n", index);
    exit(1);
  }
}

/** Conflict Detection **************/

void print_conflict(char * threadname, char * setname, char ** cap)
{

/*   DSU_ERROR("current caps: thread '%s',  %s: [ ", threadname, setname); */
/*   print_string_array_error(cap); */
/*   fprintf(__DSU_debug_file,"\n"); */

}


// check conflict with one node(thread)
static void check_cap_conflict(thread_data * td, 
                               char ** changed_types, char ** changed_defs,
                               int * conflict_delta, int * conflict_fwd, int * conflict_bck) 
{

  char * name = td->name;
  cid_t checkin_point = td->checkin_id;

  cap_data * cdata = mFind(cap_hash, checkin_point);

  if (cdata != NULL) {

      if ((cdata->cap_delta != NULL) && (! disjoint((char **)changed_types, cdata->cap_delta)))
        (*conflict_delta)++;


      print_conflict( name, "delta", cdata->cap_delta);

// not all protocols require VC check

// 2,5 only require delta 
#if (RUNTIME_SYSTEM == RUNTIME_SYSTEM_2) || \
    (RUNTIME_SYSTEM == RUNTIME_SYSTEM_5)

      // do nothing

// 1,3,4,6,7,8 require VC as well
#elif (RUNTIME_SYSTEM == RUNTIME_SYSTEM_1) || \
      (RUNTIME_SYSTEM == RUNTIME_SYSTEM_3) || \
      (RUNTIME_SYSTEM == RUNTIME_SYSTEM_4) || \
      (RUNTIME_SYSTEM == RUNTIME_SYSTEM_6) || \
      (RUNTIME_SYSTEM == RUNTIME_SYSTEM_7) || \
      (RUNTIME_SYSTEM == RUNTIME_SYSTEM_8)

      if ((cdata->cap_fwd != NULL) && (! disjoint((char **)changed_defs, cdata->cap_fwd)))
        (*conflict_fwd)++;


      print_conflict( name, "fwd", cdata->cap_fwd);


      if ((cdata->cap_bck != NULL) && (! disjoint((char **)changed_defs, cdata->cap_bck)))
        (*conflict_bck)++;

      print_conflict( name, "bck", cdata->cap_bck);

#else
#error "You must select a runtime system by defining #RS_n"
#endif // RUNTIME_SYSTEM short-circuit
    }
}

int tm_conflicts_with_one_thread(int index, char ** changed_types, char ** changed_defs)
{
  int conflict_delta = 0;
  int conflict_fwd = 0;
  int conflict_bck = 0;


  thread_data * tdata = &restriction[index];

  if (! tdata->used) {
    DSU_ERROR( "tm_conflicts_with_one_thread: tried to access non-existent thread at index %d\n", 
               index);
    exit(1);
  }

  check_cap_conflict(tdata, changed_types, changed_defs, &conflict_delta, &conflict_fwd, &conflict_bck);


#if (RUNTIME_SYSTEM == RUNTIME_SYSTEM_2) || \
    (RUNTIME_SYSTEM == RUNTIME_SYSTEM_5)

  return (conflict_delta != 0);

#elif (RUNTIME_SYSTEM == RUNTIME_SYSTEM_1) || \
      (RUNTIME_SYSTEM == RUNTIME_SYSTEM_3) || \
      (RUNTIME_SYSTEM == RUNTIME_SYSTEM_4) || \
      (RUNTIME_SYSTEM == RUNTIME_SYSTEM_6) || \
      (RUNTIME_SYSTEM == RUNTIME_SYSTEM_7) || \
      (RUNTIME_SYSTEM == RUNTIME_SYSTEM_8)

  return ((conflict_delta != 0) || ((conflict_fwd != 0) && (conflict_bck != 0)));
#else
#error "You must select a runtime system by defining #RS_n"
#endif
}


void print_one_thread(thread_data * td)
{
  char local[19];
  strncpy(local, td->name, 18);
  local[18] = '\0';
  fprintf(__DSU_debug_file,"%-18s ", local); 
}

void print_threads()
{
  fprintf(__DSU_debug_file,"Threads in the system\n"); 
  restriction_iter(print_one_thread);
  fprintf(__DSU_debug_file,"\n"); 
}

/*
  Iterate over the restriction array and check whether each thread is OK with the update.

  Return:
    0 if there's no conflict.
    1 if there is a conflict (or a thread hasn't checked in yet).

 */
int tm_conflicts_with(char ** changed_types, char ** changed_defs)
{
  int i, ret;
  int conflict_delta = 0;
  int conflict_fwd_bck = 0;

  int local_cdelta = 0;
  int local_cfwd = 0;
  int local_cbck = 0;

/*   DSU_DEBUG("tm_conflicts_with: update effect [ "); */
/*   print_string_array(changed_symbols); */
/*   DSU_DUMP("]\n"); */

/*   DSU_DEBUG("tm_conflicts_with: thread checkin points: "); */
/*   iter_map(thread_hash, print_checkin_point); */
/*   DSU_DUMP("]\n"); */


  // iterate, updating the counters in the process


  for (i = 0; i < restriction_wm; i++) {

    thread_data * td = &restriction[i];

    if (td->used) {

      local_cdelta = 0;
      local_cfwd = 0;
      local_cbck = 0;

      /* can't do the conflict check if this thread hasn't hit a check-in point yet */
      if (!td->is_checked_in) {
        DSU_DEBUG( "tm_conflicts_with: can't do the safety check, thread %s hasn't checked in yet !\n", 
                   td->name);
        return 1;
      }

      check_cap_conflict(td, changed_types, changed_defs, &local_cdelta, &local_cfwd, &local_cbck);
/*       DSU_ERROR_NOPREFIX("%c%c%c%-9d       ", */
/*               (local_cdelta == 0) ? '.' : 'x', */
/*               (local_cfwd == 0) ?   '.' : 'x', */
/*               (local_cbck == 0) ?   '.' : 'x', */
/*               td->checkin_id); */
      conflict_delta+=local_cdelta;
      if (local_cfwd && local_cbck)
        conflict_fwd_bck++ ;

    }
  }
/*   DSU_ERROR_NOPREFIX("\n"); */
/*   fflush(__DSU_debug_file);  */

  /* the crux of the matter: update is unsafe if: 
      -  there's a conflict with DELTA, or
      -  there's a conflict with both FWD and BCK
  */ 

#if (RUNTIME_SYSTEM == RUNTIME_SYSTEM_2) || \
    (RUNTIME_SYSTEM == RUNTIME_SYSTEM_5)

  ret = (conflict_delta != 0);

#elif (RUNTIME_SYSTEM == RUNTIME_SYSTEM_1) || \
      (RUNTIME_SYSTEM == RUNTIME_SYSTEM_3) || \
      (RUNTIME_SYSTEM == RUNTIME_SYSTEM_4) || \
      (RUNTIME_SYSTEM == RUNTIME_SYSTEM_6) || \
      (RUNTIME_SYSTEM == RUNTIME_SYSTEM_7) || \
      (RUNTIME_SYSTEM == RUNTIME_SYSTEM_8)

  ret = ((conflict_delta != 0) || ((conflict_fwd_bck != 0)));
#else
#error "You must select a runtime system by defining #RS_n"
#endif

/*   DSU_ERROR("  tm_conflicts_with: number of conflicts: delta %d \t fwd&bck %d\n",  */
/*             conflict_delta, conflict_fwd_bck); */
/*   fflush(__DSU_debug_file);  */

  return ret;
}

#endif // DSU_USE_THREADS



/* // print crt checkin point for all threads */
/* static void print_checkin_point(const void *nodep, const VISIT which, const int depth) */
/* { */
/*   ENTRY * e = *(ENTRY **)nodep; */

/*   if ((which == preorder) || (which == postorder)) */
/*     return; // only visit nodes once */


/*   if (e->data != NULL) */
/*     { */
/*       char * name = ((thread_data *)e->data)->name; */
/* /\*       cid_t checkin_point = ((thread_data *)e->data)->checkin_id; *\/ */

/*       if (name != NULL)  */
/*         { */
/*           struct timeval now; */
/*           struct timeval diff; */
/*           { */
/*             gettimeofday(&now, NULL); */
/*             timersub(&now,  */
/*                      &((thread_data *)e->data)->last_checkin,  */
/*                      &diff); */

/* /\*             DSU_DUMP(" %s:%d (%ld:%6ld)", name, checkin_point, diff.tv_sec, diff.tv_usec); *\/ */
/*           } */
/*         } */
/*       // else: thread reg'd, but not init'd yet */
/*     } */
/* } */


// TEST
/* extern void __DSU_init_handler(); */
/* extern int __DSU_update(char *); */
/* char * libname = ""; */
/* char * progname = ""; */


/* void test_one(__DSU_thread_t tid, char * name) */
/* { */
/*   tm_add_thread(tid, name); */
/*   printf("Added '%s', got back '%s'\n", name, tm_get_thread_name(tid)); */
/* } */

/* int main() */
/* { */
/*   __DSU_init_handler(); */
/*   test_one(0xb7eec6b0, "main"); */
/*   test_one(0xb7eebba0, "stats_thread"); */
/*   test_one(0xb74eaba0, "fserv_thread"); */
/*   test_one(0xb60e8ba0, "slave_thread"); */
/*   test_one(0xb6ae9ba0, "yp_update_thread"); */
/*   test_one(0xb56e7ba0, "handle_connection"); */
/*   test_one(0xb4ce6ba0, "handle_connection"); */
/*   test_one(0xb42e5ba0, "handle_connection"); */
/*   test_one(0xb38e4ba0, "handle_connection"); */

/*   return 0; */
/* } */

/* int main() */
/* { */
/*   __DSU_init_handler(); */
/*   tm_add_thread(0xb7eec6b0, "main"); */
/*   tm_add_thread(0xb7eebba0, "stats_thread"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   tm_checkin_thread(0xb7eebba0, "a | b | c | 0"); */
/*   __DSU_update(""); */
/*   tm_add_thread(0xb74eaba0, "fserv_thread"); */
/*   __DSU_update(""); */
/*   tm_checkin_thread(0xb7eec6b0, "d | e | f | 1"); */
/*   tm_add_thread(0xb60e8ba0, "slave_thread"); */
/*   tm_checkin_thread(0xb7eebba0, "g | h | i | 2"); */

/*   tm_add_thread(0xb6ae9ba0, "yp_update_thread"); */
/*   tm_checkin_thread(0xb6ae9ba0, "g | h | i | 2"); */

/*   tm_add_thread(0xb56e7ba0, "handle_connection"); */
/*   tm_checkin_thread(0xb56e7ba0, "g | h | i | 2"); */

/*   tm_add_thread(0xb4ce6ba0, "handle_connection"); */
/*   tm_checkin_thread(0xb4ce6ba0, "g | h | i | 2"); */
/*   tm_add_thread(0xb42e5ba0, "handle_connection"); */
/*   tm_checkin_thread(0xb42e5ba0, "g | h | i | 2"); */
/*   tm_add_thread(0xb38e4ba0, "handle_connection"); */
/*   tm_checkin_thread(0xb38e4ba0, "g | h | i | 2"); */

/*   return 0; */
/* } */
