/* 
 *
 * Copyright (c) 2004-2009, 
 * 
 *  Iulian Neamtiu      <neamtiu@cs.umd.edu> 
 *  Michael Hicks       <mwh@cs.umd.edu> 
 *  Manuel Oriol        <moriol@inf.ethz.ch> 
 *
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met: 
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 *
 * 3. The names of the contributors may not be used to endorse or promote 
 * products derived from this software without specific prior written 
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE. 
 *
 *
 */

/* 
 * generic library for dynamic loading/updating
 *
 * Note:  To use this library '__DSU_libname' needs to be initialised at the end of the file
 *        to whatever you want to call your dynamic patch file.
 * 
 */
#include <stdio.h>
#include <stdlib.h>

#include "dsu_config.h"
#include "dsu_debug.h"
#include "dsu_types.h"
#include "pmeta.h"
#include "fgc.h"
#include "machndep.h"

#ifdef DSU_USE_THREADS
#include "mt.h"
#endif // DSU_USE_THREADS


/*
 * flag to decide if we are in the first version or not
 */
extern int __DSU_NUPDATES;

/*  Name/type of the function that installs the patch (should be found in the dynamic patch file).
 *  Note: if you change these, then change Updateconfig.patch_install_function_name/type too!!!
 */
char * init_fun_name = "__DSU_install_patch";

/*  Indicates whether an update is available
 *  If you force this to be 1, the first __DSU_update() will go ahead
 *  and try to load the patch !
 */
volatile int update_requested = 0;
perf_time_t update_request_time = 0;

/* file descriptor used for the output */
FILE *__DSU_debug_file = NULL;


volatile int patch_pending = 0;

/*  This function should be called at the start of any updatable program
 *  in order to trap SIGUPDATE and route it to our updating_request_handler
 */
void __DSU_init_handler()
{

  // code to initialise trace

#if DSU_ENABLE_DEBUG == 1 || DSU_ENABLE_ERROR == 1 || DSU_ENABLE_FDEBUG == 1
  __DSU_debug_file = fopen("DSU_output.txt","w");
  if (!__DSU_debug_file) fprintf(stderr,"Couldn't open DSU_output.txt for writing, disabling tracing\n");
  //debug_file = stdout;
#endif

  __DSU_machndep_init();

  // init multi-threading stuff, if needed
#ifdef DSU_USE_THREADS
  __DSU_mt_init();
#endif // DSU_USE_THREADS

}

static int __DSU_CurrentVNumber = 0;

/* the last patch number; after the last patch is loaded, 
   statistics are printed */
static int print_stats_threshold = 20;

/* Function garbage collection */
extern int __DSU_TRIGGER_FGC; 
extern void __DSU_FGC();


perf_time_t load_times[20];
perf_time_t app_times[20];


/* try to load the patch;
   if OK, return the handle, fill out metadata; 
   if failure, return NULL */

handle_t load_patch(int version, metadata_t * type_metadata, metadata_t * def_metadata)
{
  return __DSU_machndep_load_patch(version, type_metadata, def_metadata);
}

/* effectively apply the patch:
   - do fun redirection
   - invoke state transformer 
*/
int apply_patch(handle_t handle, int version)
{
  return __DSU_machndep_apply_patch(handle, version);
}

static inline void print_performance_data()
{
  int i;

  if (__DSU_CurrentVNumber > print_stats_threshold)  
    {
      DSU_ERROR("UPDATELIB::PERFORMANCE\n");

      DSU_ERROR("Load: ");
      for (i = 0; i <= print_stats_threshold; i++)
        DSU_ERROR("%lld usec",  __DSU_machndep_usec_of_perf_time(load_times[i]));
      DSU_ERROR("\n");

      DSU_ERROR("App: ");
      for (i = 0; i <= print_stats_threshold; i++)
        DSU_ERROR("%lld usec",  __DSU_machndep_usec_of_perf_time(app_times[i]));
      DSU_ERROR("\n");

      DSU_ERROR("Total: ");
      for (i = 0; i <= print_stats_threshold; i++)
        DSU_ERROR("%lld usec",  __DSU_machndep_usec_of_perf_time(load_times[i] + app_times[i]));
      DSU_ERROR("\n");
    }
}


#ifdef DSU_USE_THREADS

static volatile   handle_t patch_handle = NULL;
volatile long long conflicting_checkins = 0;
volatile long long protocol_rounds = 0;
/* Note: the prototype of __DSU_update is hardcoded in updateconfig.ml, 
 * so if you change it here, you must sync updateconfig.ml too
 */
int __DSU_update(char * capability)
{
  metadata_t type_metadata, def_metadata;  
  int update_performed = 0;
  perf_time_t update_reached_quiescence_time, update_finished_time;
/*   int tries; */

/*   if(__DSU_TRIGGER_FGC)  */
/*     __DSU_FGC(); */

  DSU_DEBUG("UPDATELIB::__DSU_update %d : entered.\n",  __DSU_CurrentVNumber);

  //OK, an update has been requested.

  if (! patch_pending)
    { // load patch
      patch_pending = 1;

      patch_handle = load_patch(__DSU_CurrentVNumber, &type_metadata, &def_metadata);


#ifdef USE_PATCH_METADATA
      // inform pmeta about our metadata
      new_metadata(type_metadata, def_metadata, __DSU_CurrentVNumber);
#endif /* USE_PATCH_METADATA */
    }

  /* At this point we have the patch ready, but not installed yet
     Check permission
   */

#ifdef USE_PATCH_METADATA

  if (! conflicts_with_any_thread(get_current_changed_types(), get_current_changed_defs())) {
#endif /* USE_PATCH_METADATA */

        __DSU_machndep_get_perf_time(&update_reached_quiescence_time); 

        DSU_ERROR("UPDATELIB::__DSU_update %d  : Update allowed. Quiescence reached in %lld usec, conflicting check-ins %lld, rounds %lld\n",
                  __DSU_CurrentVNumber, 
                  __DSU_machndep_usec_of_perf_time(update_reached_quiescence_time - update_request_time),
                  conflicting_checkins,
                  protocol_rounds
                  );

        conflicting_checkins = 0;
        protocol_rounds = 0;

        apply_patch(patch_handle, __DSU_CurrentVNumber);

        __DSU_machndep_get_perf_time(&update_finished_time); 
        DSU_ERROR("UPDATELIB::__DSU_update %d  : Update applied in %lld usec\n",
                  __DSU_CurrentVNumber, 
                  __DSU_machndep_usec_of_perf_time(update_finished_time - update_reached_quiescence_time));

        patch_pending = 0;
        update_requested = 0;
        update_performed = 1;
        patch_handle = NULL;

        __DSU_CurrentVNumber++;

        print_performance_data();

        /* we're now ready for another update */

#ifdef USE_PATCH_METADATA
      }
      else {
        DSU_DEBUG("UPDATELIB::__DSU_update (crt version %d)  : Update not allowed (capabilities say not safe now).\n", __DSU_CurrentVNumber);

        //__DSU_machndep_sleep_nsec(1000UL * LEADER_WAIT_USEC);
      }

#endif /* USE_PATCH_METADATA */

  return update_performed;
}


#else // ! DSU_USE_THREADS

/*  This is called at every possible update point, so we'd better keep it lite
 *  If there's no pending update, this function is a no-op.
 */


/* Note: the prototype of __DSU_update is hardcoded in updateconfig.ml, 
 * so if you change it here, you must sync updateconfig.ml too
 */
int __DSU_update(char * capability)
{
  handle_t handle = NULL;
  metadata_t type_metadata, def_metadata;  
  int update_performed = 0;
  unsigned long long update_apply_time;

  DSU_DEBUG("UPDATELIB::__DSU_update %d  : entered.\n",  __DSU_CurrentVNumber);

  /* hook for triggering garbage collection */
/*   if(__DSU_TRIGGER_FGC)  */
/*     __DSU_FGC(); */

  if (! update_requested)
    {
      return update_performed;
    }
  //OK, an update has been requested.

  if (!patch_pending)
    { // load patch
      handle = load_patch(__DSU_CurrentVNumber, &type_metadata, &def_metadata);

      patch_pending = 1;

#ifdef USE_PATCH_METADATA
      // inform pmeta about our metadata
      new_metadata(type_metadata, def_metadata, __DSU_CurrentVNumber);
#endif /* USE_PATCH_METADATA */
    }

  /* At this point we have the patch ready, but not installed yet
     Check permission
   */

#ifdef USE_PATCH_METADATA
      if (is_allowed_update(capability))
        {
#endif /* USE_PATCH_METADATA */
          __DSU_machndep_get_perf_time(&update_apply_time); 

          DSU_DEBUG("UPDATELIB::__DSU_update %d  : Update allowed. Quiescence reached in %lld usec\n",  
                    __DSU_CurrentVNumber, 
                    __DSU_machndep_usec_of_perf_time(update_apply_time - update_request_time));

          apply_patch(handle, __DSU_CurrentVNumber);

          patch_pending = 0;
          update_requested = 0;
          update_performed = 1;

          __DSU_CurrentVNumber++;

          print_performance_data();

          /* we're now ready for another update */

#ifdef USE_PATCH_METADATA
        }
      else
        DSU_DEBUG("UPDATELIB::__DSU_update %d  : Update not allowed (capabilities say not safe now).\n",  __DSU_CurrentVNumber);
#endif /* USE_PATCH_METADATA */

  return update_performed;
}
#endif // DSU_USE_THREADS
